/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

/**
 * Driver for testing the naming subsystem. Also serves as naming
 * subsystem example code. This program is to HyperCast peer-to-peer
 * networks as "nslookup," "dig," or "host" are to Internet DNS.
 *
 * <code>javac -classpath ../lib/hypercast.jar:../lib/bcprov-jdk14-122.jar NamingShell.java</code>
 * <code>java -classpath ../lib/hypercast.jar:../lib/bcprov-jdk14-122.jar NamingShell [configuration filename]</code>
 *
 * In this shell you can interactively enter naming commands that
 * test/demonstrate the naming system. If no shell runs, the program
 * will participate in the naming system by responding to queries and,
 * depending on the configuration, pushing its name binding to the
 * network.
 *
 * All of the commands of the shell can be seen by entering the "help"
 * command.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @version Spring 2005
 */

import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.PrintStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.FileInputStream;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Attr;
import org.w3c.dom.NodeList;

import org.apache.xpath.XPath;

import javax.xml.transform.TransformerException;

import hypercast.HyperCastConfig;
import hypercast.I_OverlaySocket;
import hypercast.I_OverlayMessage;
import hypercast.I_LogicalAddress;
import hypercast.I_ReceiveCallback;
import hypercast.NotificationHandler;
import hypercast.util.XmlUtil;
import hypercast.HyperCastStatsException;

import hypercast.Naming;

import hypercast.events.NAMING_EVENT;
import hypercast.events.NODE_ISSTABLE;

import java.security.Security;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.KeyFactory;
import java.security.Provider;
import java.security.NoSuchProviderException;

import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;

import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;

import org.bouncycastle.jce.provider.BouncyCastleProvider;


/**
 * This class handles all asynchronous naming events.
 */
class NamingShellNotificationHandler
    extends NotificationHandler 
{
    public void handle_NAMING_EVENT (final NAMING_EVENT event) {
        System.err.println ("NAMING_EVENT handler:\n" + event);
    }
}

/**
 * This class implements the command line interpreter for the naming shell.
 */
public class NamingShell
    implements I_ReceiveCallback
{ 
    // Needed for I_ReceiveCallback
    public void ReceiveCallback (final I_OverlayMessage message)
    {
        final String source = message.getSourceAddress().toString();
        final String payload = new String (message.getPayload(), 0, message.getPayloadLength());
	
        System.out.println ("From \"" + source + "\" received: \"" + payload + "\"");
    }

    private static I_OverlaySocket socket;

    private static CertificateFactory certificateFactory;
    private static KeyFactory keyFactory;

    private static String certificateFormat = "X.509";
    private static String asymmetricEncryptionAlgorithm = "RSA";

    private static int MILLISECONDS_PER_SECOND = 1000;

    public static void main (final String [] args)
    {
        if (Security.getProvider ("BC") == null)
            Security.insertProviderAt (new BouncyCastleProvider(), 2);

        try
        {
            certificateFactory = CertificateFactory.getInstance (certificateFormat, "BC");
            keyFactory = KeyFactory.getInstance (asymmetricEncryptionAlgorithm, "BC");
        }
        catch (final CertificateException ce)
        {
            System.err.println ("Could not instanciate CertificateFactory (" + certificateFormat + ")\n" + ce);
            System.exit (1);
        }
        catch (final NoSuchAlgorithmException nsae)
        {
            System.err.println ("Could not instanciate KeyFactory (" + asymmetricEncryptionAlgorithm + ")\n" + nsae);
            System.exit (1);
        }
        catch (final NoSuchProviderException nspe)
        {
            System.err.println ("Could not instanciate " + certificateFormat + " certificate factory, or " + asymmetricEncryptionAlgorithm + " key factory, provider BC unknown: " + nspe);
            System.exit (1);
        }

        final NamingShell namingShell = new NamingShell();
        final NotificationHandler notificationHandler = new NamingShellNotificationHandler();

        String configurationFilename = "hypercast.xml";

        if (args.length > 1)
            throw new RuntimeException ("usage: NamingShell [configuration filename]");
        else if (args.length > 0)
            configurationFilename = args [0];

        final HyperCastConfig configuration = HyperCastConfig.createConfig (configurationFilename);
	
        socket = configuration.createOverlaySocket (namingShell, notificationHandler);
        socket.joinOverlay();
	
        /*
         * Application blocks here until event NODE_ISSTABLE occurs.
         */
        System.out.print ("Waiting for Naming Shell socket to become stable in the overlay...");
        notificationHandler.waitUntil_NODE_ISSTABLE();
        System.out.println ("stability notification received!");
	
        System.out.println ("Welcome to the Naming Shell");
        shell();
    }

    private static void shell()
    {
        final PrintStream output = System.out;
        try
        {
            final BufferedReader input = new BufferedReader (new InputStreamReader (System.in));
            final String prompt = "> ";
            String myAddress = "";

            if (socket != null)
                myAddress = socket.getLogicalAddress().toString();

            String line;

            /*
             *  This while loop is a bit of a hack. The readLine
             *  method of BufferedReader will return null if the user
             *  presses enter without having typed any
             *  characters. This causes the shell to exit, which is
             *  not what we want.  Unfortunately, with this hack
             *  Ctrl-d to exit does not work, "quit" or "exit" must
             *  typed explicitly.
             */
            boolean keepGoing = true;
          out:
            while (keepGoing)
            {
                for (output.print (myAddress + prompt), line = input.readLine ();
                     line != null && line.length () > 0;
                     output.print (myAddress + prompt), line = input.readLine ())
                {
                    List tokens = tokenize (line);
                    if (! processLine (output, tokens))
                    {
                        keepGoing = false;
                        break out;
                    }

                    if (socket != null)
                        myAddress = socket.getLogicalAddress().toString();
                }
            }
        }
        catch (IOException e)
        {
            System.err.println ("shell IOException");
            System.err.print (e);
        }
        catch (Exception e)
        {
            System.err.println ("shell Exception");
            System.err.print (e);
            e.printStackTrace();
        }
        finally
        {
            System.exit (0);
        }
    }

    private final static  String USAGE_SET          = "set <name> [logical address]; create a binding using a name and optional a logical address (local LA used if none given)";
    private final static  String USAGE_CRYPTO_SET   = "crypto_set <certificate filename> <private key filename> [<name> <logical address>]; create a binding using crypto (optionally specify a name and logical address)";
    private final static  String USAGE_UNSET        = "unset <name> ; unset name of local socket";
    private final static  String USAGE_CRYPTO_UNSET = "crypto_unset <certificate filename> <private key filename> ; unset name of local socket using crypto";
    private final static  String USAGE_TRUSTED_CERT = "trusted_cert <certificate filename>; install a trusted certificate";
    private final static  String USAGE_NAME         = "name [-timeout<secs>] [-block] [-trusted] [-auth] [-age<secs>] [-minhops<count>] [-maxhops<count>] [-responses<count>] [-matchsn] <logical adddress>; name query";
    private final static  String USAGE_LA           = "la [-timeout<secs>] [-block] [-trusted] [-auth] [-age<secs>] [-minhops<count>] [-maxhops<count>] [-responses<count>] [-matchsn] <name>; logical address query";
    private final static  String USAGE_CHANGE_LA    = "changela <new logical address> ; changes logical address of the local socket";
    private final static  String USAGE_CERTS        = "certs ; displays table of certificates";
    private final static  String USAGE_BINDINGS     = "bindings ; displays table of bindings";
    private final static  String USAGE_PENDING      = "pending ; displays table of pendings queries and certificates";

    /**
     * Executes the command encoded in the tokenized list of strings.
     */
    private static boolean processLine (final PrintStream output, final List tokens)
    {
        if (tokens == null)
            throw new IllegalArgumentException ("Node.processLine: tokens is null");
        else if (tokens.isEmpty())
            return true;

        String command = (String) tokens.get (0);

        if (command.equalsIgnoreCase ("quit") || command.equalsIgnoreCase ("exit"))
        {
            return false;
        }
        else if (command.equalsIgnoreCase ("set"))
        {
            if (tokens.size() != 2 && tokens.size() != 3)
            {
                output.println (USAGE_SET);
            }
            else
            {
                final String name = (String) tokens.get(1);

                if (tokens.size() > 2)
                {
                    final String logicalAddressString = (String) tokens.get (2);
                    final I_LogicalAddress logicalAddress = socket.createLogicalAddress (logicalAddressString);
                    socket.setName (name, logicalAddress);
                }
                else
                {
                    socket.setName (name);
                }
            }
        }
        else if (command.equalsIgnoreCase ("unset"))
        {
            if (tokens.size() != 2)
            {
                output.println (USAGE_UNSET);
            }
            else
            {
                final String name = (String)tokens.get(1);
                socket.unsetName (name);
            }
        }
        else if (command.equalsIgnoreCase ("changela"))
        {
            if (tokens.size() != 2)
            {
                output.println (USAGE_CHANGE_LA);
            }
            else
            {
                final String logicalAddressString = (String) tokens.get (1);
                final I_LogicalAddress newLogicalAddress = socket.createLogicalAddress (logicalAddressString);
                socket.setLogicalAddress (newLogicalAddress);
            }
        }
        else if (command.equalsIgnoreCase ("la"))
        {
            String name = null;

            boolean blocking                 = false;
            boolean requireTrusted           = false;
            long    timeout                  = 3; // seconds
            boolean requireSerialNumberMatch = false;
            boolean requireAuthoritative     = false;
            int     maxAge                   = 300; // seconds
            int     maxResponses             = 1;
            short   minHopCount              = (short) 0;
            short   maxHopCount              = Naming.USE_CONFIGURED_QUERY_RADIUS;

            for (int i = 1; i < tokens.size(); ++i)
            {
                final String token = (String) tokens.get (i);

                if (token.equals ("-trusted")) {
                    requireTrusted = true;
                }
                else if (token.startsWith ("-timeout")) {
                    timeout = Long.parseLong (token.substring ("-timeout".length()));
                }
                else if (token.equals ("-matchsn")) {
                    requireSerialNumberMatch = true;
                }
                else if (token.startsWith ("-age")) {
                    maxAge = Integer.parseInt (token.substring ("-age".length()));
                }
                else if (token.startsWith ("-responses")) {
                    maxResponses = Integer.parseInt (token.substring ("-responses".length()));
                }
                else if (token.equals ("-auth")) {
                    requireAuthoritative = true;
                }
                else if (token.equals ("-block")) {
                    blocking = true;
                }
                else if (token.startsWith ("-minhops")) {
                    minHopCount = Short.parseShort (token.substring ("-minhops".length()));
                }
                else if (token.startsWith ("-maxhops")) {
                    maxHopCount = Short.parseShort (token.substring ("-maxhops".length()));
                }
                else if (token.startsWith ("-")) {
                    output.println (USAGE_LA);
                }
                else {
                    name = token;
                }
            }

            if (name == null)
            {
                final I_LogicalAddress logicalAddress = socket.getLogicalAddress();
                output.println ("Result of logical address  query for local socket: " + logicalAddress);
            }
            else if (blocking)
            {
                output.println ("Issuing logical address query for name \"" + name + "\"...");

                final I_LogicalAddress [] logicalAddresses = socket.getLogicalAddressByName (name,
                                                                                             requireTrusted,
                                                                                             timeout * MILLISECONDS_PER_SECOND,
                                                                                             requireSerialNumberMatch,
                                                                                             requireAuthoritative,
                                                                                             maxAge * MILLISECONDS_PER_SECOND,
                                                                                             maxResponses,
                                                                                             minHopCount,
                                                                                             maxHopCount);
                final StringBuffer logicalAddressBuffer = new StringBuffer();
                for (int i = 0; i < logicalAddresses.length; ++i)
                    logicalAddressBuffer.append (logicalAddresses[i] + " ");
                output.println ("Result of logical address query for name \"" + name + "\": " + logicalAddressBuffer);
            }
            else
            {
                output.println ("Issuing logical address query for name \"" + name + "\"");

                socket.getLogicalAddressByNameNonBlocking (name,
                                                           requireTrusted,
                                                           timeout * MILLISECONDS_PER_SECOND,
                                                           requireSerialNumberMatch,
                                                           requireAuthoritative,
                                                           maxAge * MILLISECONDS_PER_SECOND,
                                                           maxResponses,
                                                           minHopCount,
                                                           maxHopCount);
            }
        }
        else if (command.equalsIgnoreCase ("name"))
        {
            String logicalAddressString = null;

            boolean blocking                 = false;
            boolean requireTrusted           = false;
            long    timeout                  = 3; // seconds
            boolean requireSerialNumberMatch = false;
            boolean requireAuthoritative     = false;
            int     maxAge                   = 300; // seconds
            int     maxResponses             = 1;
            short   minHopCount              = (short) 0;
            short   maxHopCount              = Naming.USE_CONFIGURED_QUERY_RADIUS;

            for (int i = 1; i < tokens.size(); ++i)
            {
                final String token = (String) tokens.get (i);

                if (token.equals ("-trusted")) {
                    requireTrusted = true;
                }
                else if (token.startsWith ("-timeout")) {
                    timeout = Long.parseLong (token.substring ("-timeout".length()));
                }
                else if (token.startsWith ("-age")) {
                    maxAge = Integer.parseInt (token.substring ("-age".length()));
                }
                else if (token.startsWith ("-responses")) {
                    maxResponses = Integer.parseInt (token.substring ("-responses".length()));
                }
                else if (token.equals ("-auth")) {
                    requireAuthoritative = true;
                }
                else if (token.equals ("-block")) {
                    blocking = true;
                }
                else if (token.startsWith ("-minhops")) {
                    minHopCount = Short.parseShort (token.substring ("-minhops".length()));
                }
                else if (token.startsWith ("-maxhops")) {
                    maxHopCount = Short.parseShort (token.substring ("-maxhops".length()));
                }
                else if (token.equals ("-matchsn")) {
                    requireSerialNumberMatch = true;
                }
                else if (token.startsWith ("-")) {
                    output.println (USAGE_NAME);
                }
                else {
                    logicalAddressString = token;
                }
            }

            if (logicalAddressString == null)
            {
                final String [] names = socket.getNames(); 
                final StringBuffer nameBuffer = new StringBuffer();

                for (int i = 0; i < names.length; ++i)
                    nameBuffer.append (names[i] + " ");

                output.println ("Result of name query for local socket: " + nameBuffer);
            }
            else
            {
                try
                {
                    final I_LogicalAddress logicalAddress = socket.createLogicalAddress (logicalAddressString);

                    if (blocking)
                    {
                        output.println ("Issuing name query for logical address \"" + logicalAddress.toString() + "\"...");

                        final String [] names = socket.getNames (logicalAddress,
                                                                 requireTrusted,
                                                                 timeout * MILLISECONDS_PER_SECOND,
                                                                 requireSerialNumberMatch,
                                                                 requireAuthoritative,
                                                                 maxAge * MILLISECONDS_PER_SECOND,
                                                                 maxResponses,
                                                                 minHopCount,
                                                                 maxHopCount);

                        final StringBuffer nameBuffer = new StringBuffer();
                        for (int i = 0; i < names.length; ++i)
                            nameBuffer.append (names[i]);

                        output.println ("Result of name query for logical address \"" + logicalAddressString + "\": " + nameBuffer);
                    }
                    else
                    {
                        output.println ("Issuing name query for logical address \"" + logicalAddressString + "\"");
                        socket.getNamesNonBlocking (logicalAddress,
                                                    requireTrusted,
                                                    timeout * MILLISECONDS_PER_SECOND,
                                                    requireSerialNumberMatch,
                                                    requireAuthoritative,
                                                    maxAge * MILLISECONDS_PER_SECOND,
                                                    maxResponses,
                                                    minHopCount,
                                                    maxHopCount);
                    }
                }
                catch (final IllegalArgumentException iae)
                {
                    output.println (logicalAddressString + " is not a valid logical address");
                }
            }
        }
        else if (command.equalsIgnoreCase ("crypto_set"))
        {
            if (tokens.size() != 3 && tokens.size() != 5)
            {
                output.println (USAGE_CRYPTO_SET);
            }
            else
            {
                final String certificateFilename = (String) tokens.get (1);
                final String privateKeyFilename = (String) tokens.get (2);

                final Certificate certificate = readX509Certificate (certificateFilename, certificateFactory);
                final PrivateKey privateKey = readPrivateKey (privateKeyFilename, keyFactory);

                if (tokens.size() == 5)
                {
                    final String name = (String) tokens.get (3);
                    final String logicalAddressString = (String) tokens.get (4);
                    final I_LogicalAddress logicalAddress = socket.createLogicalAddress (logicalAddressString);
                    socket.setName (certificate, privateKey, name, logicalAddress);
                }
                else
                {
                    socket.setName (certificate, privateKey);
                }
            }
        }
        else if (command.equalsIgnoreCase ("crypto_unset"))
        {
            if (tokens.size() != 3)
            {
                output.println (USAGE_CRYPTO_UNSET);
            }
            else
            {
                final String certificateFilename = (String) tokens.get (1);
                final String privateKeyFilename = (String) tokens.get (2);

                final Certificate certificate = readX509Certificate (certificateFilename, certificateFactory);
                final PrivateKey privateKey = readPrivateKey (privateKeyFilename, keyFactory);

                socket.unsetName (certificate, privateKey);
            }
        }
        else if (command.equalsIgnoreCase ("trusted_cert"))
        {
            if (tokens.size() != 2)
            {
                output.println (USAGE_TRUSTED_CERT);
            }
            else
            {
                final String certificateFilename = (String) tokens.get (1);
                final Certificate certificate = readX509Certificate (certificateFilename, certificateFactory);

                socket.installTrustedNamingCertificate (certificate);
            }
        }
        else if (command.equalsIgnoreCase ("bindings"))
        {
            try
            {
                final Document doc = XmlUtil.createDocument();

                final String xPathString = (tokens.size() > 1) ? (String) tokens.get (1) : "/Socket/MessageStore/Naming/BindingTable";

                final Element [] reply = socket.getStats (doc, XmlUtil.createXPath (xPathString));

                for (int s = 0; s < reply.length; ++s)
                    displayNodeRecursive (reply[s], 2, output);
            }
            catch (final HyperCastStatsException hcse)
            {
                output.println ("HyperCastStatsException caught while getting statistics for Binding Table: " + hcse);
            }
        }
        else if (command.equalsIgnoreCase ("certs"))
        {
            if (tokens.size() != 1)
            {
                output.println (USAGE_CERTS);
            }
            else
            {
                try
                {
                    final Document doc = XmlUtil.createDocument();
                    final String xPathString = "/Socket/MessageStore/Naming/CertificateCache";

                    final Element [] reply = socket.getStats (doc, XmlUtil.createXPath (xPathString));

                    for (int s = 0; s < reply.length; ++s)
                        displayNodeRecursive (reply[s], 2, output);
                }
                catch (final HyperCastStatsException hcse)
                {
                    output.println ("HyperCastStatsException caught while getting statistics for Binding Table: " + hcse);
                }
            }
        }
        else if (command.equalsIgnoreCase ("pending"))
        {
            if (tokens.size() != 1)
            {
                output.println (USAGE_PENDING);
            }
            else
            {
                try
                {
                    final Document doc = XmlUtil.createDocument();
                    final String xPathString = "/Socket/MessageStore/Naming/Pending";

                    final Element [] reply = socket.getStats (doc, XmlUtil.createXPath (xPathString));

                    for (int s = 0; s < reply.length; ++s)
                        displayNodeRecursive (reply[s], 2, output);
                }
                catch (final HyperCastStatsException hcse)
                {
                    output.println ("HyperCastStatsException caught while getting statistics for Binding Table: " + hcse);
                }
            }
        }
        else if (command.equalsIgnoreCase ("help"))
        {
            output.println ("Available commands are:");
            output.println ("  " + USAGE_SET);
            output.println ("  " + USAGE_CRYPTO_SET);
            output.println ("  " + USAGE_UNSET);
            output.println ("  " + USAGE_CRYPTO_UNSET);
            output.println ("  " + USAGE_NAME);
            output.println ("  " + USAGE_LA);
            output.println ("  " + USAGE_CHANGE_LA);
            output.println ("  " + USAGE_TRUSTED_CERT);
            output.println ("  " + USAGE_CERTS);
            output.println ("  " + USAGE_BINDINGS);
            output.println ("  " + USAGE_PENDING);
            output.println ("  help");
            output.println ("  quit/exit");
        }
        else
        {
            output.println ("unknown command: " + command);
        }

        return true;
    }

    /**
     * Reads data that encodes an RSA private key from a file and returns an object that is instaciated using the data read.
     */
    private static PrivateKey readPrivateKey (String privateKeyFilename, KeyFactory kf)
    {
        PrivateKey privateKey = null;

        try
        {
            FileInputStream fis = new FileInputStream (privateKeyFilename);
            DataInputStream dis = new DataInputStream (fis);
            byte [] privateKeyData = new byte [dis.available()];
            dis.readFully (privateKeyData);
            dis.close();

            // Instanciate private key object using data from file
            PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec (privateKeyData);
            privateKey = kf.generatePrivate (privateKeySpec);
        }
        catch (final FileNotFoundException fnfe)
        {
            System.err.println (fnfe);
        }
        catch (final IOException ioe)
        {
            System.err.println (ioe);
        }
        catch (final InvalidKeySpecException ikse)
        {
            System.err.println (ikse);
        }

        return privateKey;
    }

    /**
     * Reads data that encodes an X.509 certificate from a file and returns an object that is instaciated using the data read.
     */
    private static X509Certificate readX509Certificate (String certificateFilename, CertificateFactory cf)
    {
        X509Certificate certificate = null;

        try
        {
            // Read certificate data from file

            FileInputStream fis = new FileInputStream (certificateFilename);
            DataInputStream dis = new DataInputStream (fis);
            byte [] certificateData = new byte [dis.available()];
            dis.readFully (certificateData);
            dis.close();

            // Instanciate certificate object using data from file

            ByteArrayInputStream bais = new ByteArrayInputStream (certificateData);
            certificate = (X509Certificate) cf.generateCertificate (bais);
        }
        catch (final FileNotFoundException fnfe)
        {
            System.err.println (fnfe);
        }
        catch (final IOException ioe)
        {
            System.err.println (ioe);
        }
        catch (final CertificateException ce)
        {
            System.err.println (ce);
        }

        return certificate;
    }

    /*
     * Breaks up a line of input into space separated strings, returns
     * those strings in a list. If the line is only spaces, an empty
     * list is returned.
     */
    private static List tokenize (String s)
    {
        final boolean debugging = false;

        List result = new LinkedList();

        boolean inToken = false;
        boolean inQuote = false;
        int b = 0;
        int e = 0;
        int i = 0;

        switch (s.charAt (0))
        {
            case '"': inQuote = true; inToken = true; i = 1; b = 1; break;
            case ' ': inToken = false; break;
            default: inToken = true; b = 0; break;
        }

        for (; i < s.length (); ++i)
        {
            if (s.charAt (i) == '"')
            {
                if (! inQuote)
                {
                    inQuote = true; inToken = true; b = i + 1;
                }
                else
                {
                    e = i;
                    inToken = false;
                    inQuote = false;
                    result.add (s.substring (b, e));
                }
            }
            else if (s.charAt (i) == ' ' && ! inQuote)   // skip whitespace
            {
                if (inToken)
                {
                    e = i;
                    inToken = false;
                    result.add (s.substring (b, e));
                }
            }
            else
            {
                if (! inToken)
                {
                    b = i;
                    inToken = true;
                }
            }
        }

        if (inToken == true)
            result.add (s.substring (b, s.length ()));

        if (debugging)
        {
            System.err.print ("tokens: ");
            for (Iterator iterator = result.iterator(); iterator.hasNext();)
            {
                String token = (String) iterator.next();
                System.err.print ("(" + token + ") ");
            }
            System.err.println();
        }

        return result;
    }

    /**
     * Display a subtree with "node" as the root.  This code was
     * copied from RunControl.GetRCValueCommand - in general is there
     * some better way to display this data?
     *
     * @param node The root of the subtree.
     * @param indentDistance The number of base indent space for displaying statements.
     * @param out The output device.
     */
    public static void displayNodeRecursive (final Node node, final int indentDistance, final PrintStream out)
    {
        if (node == null)
            return;

        String indentString = "";
        for (int i = 0; i < indentDistance; ++i)
        {
            indentString = indentString + " ";
        }
		
        if (node.getNodeType() == Node.TEXT_NODE)
        {
            out.println(indentString + "= " + node.getNodeValue());
        }
        else
        {
            if ("ROOT".equals (node.getNodeName()))
            {
                out.print (indentString + "/");
            }
            else
            {
                out.print (indentString + node.getNodeName());
            }

            if (node instanceof Element)
            {
                Attr attr = ((Element) node).getAttributeNode ("index");
                if (attr != null)
                {
                    out.print ("[" + attr.getNodeValue() + "]");
                }
            }

            out.println();

            final NodeList nodeList = node.getChildNodes();

            for (int i = 0; i < nodeList.getLength(); ++i)
            {
                if (nodeList.item (i) == null)
                {
                    System.err.println ("nodeList.item(" + i + ") is null");
                }
                else
                {
                    displayNodeRecursive (nodeList.item (i), indentDistance + 2, out);
                }
            }
        }
    }
}
