/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;
import hypercast.HyperCastWarningRuntimeException;
import java.io.*;
import java.util.*;
import java.net.*;

/**
 * The cache of buddies in addition to those in the configuration file.  The
 * cache is implemented as a linear structure, addition are to the front.
 *
 * @author HyperCast Team
 * @author Weisheng Si
 * @version 2.0, Nov 11, 2001 
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */
class BuddyCache 
{
    /**
     * The cache of nodes in the Overlay.
     *
     * Note that although Vector is synchronized, the various BuddyCache
     * methods must perform synchronization on this data structure because some
     * BuddyCache methods perform multiple operations on the Vector object,
     * these operations should be atomic.
     */
    private Vector buddyCache = new Vector();

    /** The max number of buddies that this cache may store. */
    private int maxSize;

    /** The name of cache file. */
    private String dynamicCacheFilename;

    /** The prefix of property definition. */
    private String propertyPrefix;

    /**
     * The name of temporary file for saveCache method.
     * It only exists for a short time.
     */
    private static final String TEMPORARY_COPY_FILE = ".temp_buffer_file";

    /**
     * Construct the BuddyCache with its maximum size, cache file name,
     * adapter, and property prefix.
     *
     * @param maxSize the size of the dynamic cache; 0 means dynamic list is not used.
     * @throws IllegalArgumentException if maxSize is less than zero
     * @throws HyperCastWarningRuntimeException if buddy cache file cannot be read
     */
    public BuddyCache (int maxSize, String dynamicCacheFilename, I_UnicastAdapter adapter, String propertyPrefix)
    {
        this.maxSize = maxSize;
        this.dynamicCacheFilename = dynamicCacheFilename;
        this.propertyPrefix = propertyPrefix;

        if (maxSize == 0)
        {
            return;
        }
        else if (maxSize < 0)
        {
            throw new IllegalArgumentException ("Dynamic buddy cache: Negative value given for maxSize: " + maxSize);
        }

        FileInputStream cacheFileInputStream = null;
        try
        {
            cacheFileInputStream = new FileInputStream (dynamicCacheFilename);
        }
        catch (FileNotFoundException fnfe)
        {
            System.out.println ("Dynamic buddy cache: Buddy cache file " + dynamicCacheFilename + " not found; it will be be created");
            return;
        }

        Properties properties = new Properties();
        try
        {
            properties.load (cacheFileInputStream);
        }
        catch (IOException ioe)
        {
            throw new HyperCastWarningRuntimeException ("Dynamic buddy cache: Could not load buddy cache file " + dynamicCacheFilename, ioe);
        }

        /*
         * Read buddies from configuration file. Note that at most
         *  maxSize buddies are read from the file.
         */
        for (int i = 0; i < maxSize ; ++i)
        {
            String buddyString = properties.getProperty (propertyPrefix + ".Buddy" + i);
            
            if (buddyString != null)
            {
                BuddyRecord buddyRecord = new BuddyRecord (adapter, buddyString);
                buddyCache.addElement (buddyRecord);
            }
            else
            {
                break; // no more buddies in file; end loop.
            }
        }
    }

	/**
	 * Return the size of buddy cache.
	 */
    public int size()
    {
        return buddyCache.size();
    }
    
	/**
	 * Return the BuddyRecord object at index i. If i is larger than
	 * or equals to the size of buddy cache, return null.
	 */
    public synchronized BuddyRecord elementAt (int i) 
    {
        BuddyRecord result = null;
        if (i < 0)
        {
            throw new IllegalArgumentException ("index imust be non-negative!");
        }
        else if (i < buddyCache.size())
        {
            result = (BuddyRecord) buddyCache.elementAt (i);
        }
        return result;
    }

    /**
     * Adds a new entry at the beginning of the cache if the entry is not
     * already in the cache.
     */
	public synchronized void addToFront (I_NetworkAddress buddyaddress)
    {
        if (maxSize != 0)
        {
            /*
             * Add only if not already in the cache
             */
			BuddyRecord buddyRecord = new BuddyRecord (buddyaddress);
            int searchResultIndex = buddyCache.indexOf (buddyRecord);
            if (searchResultIndex == -1)
            {
                /*
                 * Ensure that there is enough space to add one BuddyRecord. If
                 * cache is full, evict the BuddyRecord at the end of the cache.
                 */
                int currentCacheSize = buddyCache.size();
				if (currentCacheSize == maxSize)
                {
                    buddyCache.removeElementAt (currentCacheSize - 1);
                }
				buddyCache.insertElementAt (buddyRecord, 0);
            }
        }
    }

	/**
	 * Move the element at index i to the front position oin the buddy cache.
	 */
    public synchronized void moveToFront (int index) 
    {
        if (index > 0
            &&
            index < buddyCache.size())
        {
            doMoveToFront (index);
        }
        else
        {
            throw new HyperCastWarningRuntimeException ("index is out of bounds: " + index);
        }
    }

	/**
	 * Move the element that has specified underlay or physical address to the 
	 * front position in the buddy cache.
	 */
	public synchronized void moveToFront (I_NetworkAddress buddyaddress)
    {
		BuddyRecord buddyRecord = new BuddyRecord (buddyaddress);
        int searchResultIndex = buddyCache.indexOf (buddyRecord);
        if (searchResultIndex != -1)
        {
            doMoveToFront (searchResultIndex);
        }
        else
        {
			throw new HyperCastWarningRuntimeException ("Dynamic buddy cache: Cannot move " + buddyaddress.toString() + " to the front, record not found");
        }
    }

	/**
	 * Execute the movement of element at index i to the front position in the buddy cache.
	 */
    private void doMoveToFront (int index)
    {
        BuddyRecord buddyRecord = (BuddyRecord) buddyCache.elementAt (index);
        System.out.println ("Dynamic buddy cache: Moving " + buddyRecord + " at " + index + " to the front");
        buddyCache.removeElementAt (index);
        buddyCache.insertElementAt (buddyRecord, 0);
    }

	/**
	 * Remove the element that has the specified underlay or physical address.
	 */
	public synchronized void removeRecord (I_NetworkAddress buddyaddress)
    {
        if (maxSize != 0)
        {
			BuddyRecord buddyRecord = new BuddyRecord (buddyaddress);
            int searchResultIndex = buddyCache.indexOf (buddyRecord);
            if (searchResultIndex >= 0)
            {
                //System.out.println ("Dynamic buddy cache: Removing " + buddyaddress.toString() + " at index " + searchResultIndex);
                buddyCache.removeElementAt (searchResultIndex);
            }
		}
    }

	/**
	 * Save the buddy cache into cache file that can be used next time.
	 */
    public synchronized void saveCache()
    {
        createTemporaryFile (TEMPORARY_COPY_FILE);

        /*
         * After temporary file is created, rename it to the cache filename.
         */
        File tempFile = null;
        File cacheFile = null;
        try
        {
            tempFile = new File (TEMPORARY_COPY_FILE);
            cacheFile = new File (dynamicCacheFilename);
            tempFile.renameTo (cacheFile);
        }
        catch (SecurityException se)
        {
            throw new HyperCastWarningRuntimeException ("Could not copy temporary buddy cache \"" + TEMPORARY_COPY_FILE + "\" to \"" + dynamicCacheFilename +"\"", se);
        }
        catch (NullPointerException npe)
        {
            throw new HyperCastWarningRuntimeException ("Null pointer encountered when copying temporary buddy cache: tempFile = " + tempFile + ", cacheFile = " + cacheFile, npe);
        }
    }

    /**
     * The buddylist data, with other existing data in dynamicCacheFilename, are
     * written into a temporary file.
     */
    private void createTemporaryFile (final String temporaryCopyFile)
    {
        BufferedReader inputStream = null;
        PrintWriter outputStream = null;

        try
        {
            inputStream = new BufferedReader (new FileReader (dynamicCacheFilename));
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println ("Dynamic buddy cache: File " + dynamicCacheFilename + " not found.");
            inputStream = null;
        }

        try
        {
            outputStream = new PrintWriter (new BufferedWriter (new FileWriter (temporaryCopyFile)));
        }
        catch (IOException ioe)
        {
           throw new HyperCastFatalRuntimeException ("Dynamic buddy cache: Attempted to create temporary buddy file", ioe);
        }

        /*
         * Write buddy cache contents into temporary file.
         */
        for (int i = 0; i < buddyCache.size(); ++i)
        {
            String buddy = buddyCache.elementAt(i).toString();
			String entry = propertyPrefix + ".Buddy" + i + " = " + buddy;
            outputStream.println (entry);
        }

        try
        {
            String line = null;
            while (inputStream != null
                   &&
                   (line = inputStream.readLine()) != null)
            {
                if (! line.startsWith (propertyPrefix + ".Buddy"))
                {
                    outputStream.println (line);
                }
            }
        }
        catch (IOException ioe)
        {
           throw new HyperCastFatalRuntimeException ("Dynamic buddy cache: Error while writing data to temporary buddy file", ioe);
        }

        try
        {
            if (outputStream != null)
            {
                outputStream.close();
            }

            if (inputStream != null)
            {
                inputStream.close();
            }
        }
        catch (IOException ioe)
        {
            throw new HyperCastFatalRuntimeException ("Dynamic buddy cache: Could not close files for use with buddy list", ioe);
        }
    }
}
