/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.*;

import java.security.*;

import javax.crypto.*;

	/**
	* This class extens class PayloadExtension and defines an encrypted payload
	* extension. An instance of this class is constructed only when security 
	* policy is set to privacy.
	*
	* @author HyperCast Team
	* @version 2005 (version 3.0)
	*/

	public class EncryptedPayloadExtension extends PayloadExtension
	{

		/** encrypted payload byte array  */
		private byte[] EncryptedPayload;
		
		/**
		 * Key vault. It stores information for encryption and decryption operation.
		 */
		KeyVault keyVault;

		/**
		 * Message key for payload decryption/decryption.
		 */
		Key msgKey = null;

		/**
		 * Constructs an EncryptedPayloadExtension header object with all the parameters
		 * 
		 * @param payload		payload byte array
		 * @param isEncrypted	flag indicating the given payload is encrypted or
		 *  plaintext payload. True if payload is encrypted, false otherwise.
		 */
		public EncryptedPayloadExtension(byte[] payload, boolean isEncrypted, KeyVault kv)
		{
			keyVault = kv;
			
			/*
			 * Set encrypted payload byte array and plaintext payload byte array
			 * based on the value of isEncrypted flag. 
			 */
			 
			/*
			 * If isEncrypted flag is true, meaning the given payload is encrypted,
			 * only set encrypted payload. The decryotion is not done immediately.
			 * Instead, when EncryptedPayloadExtension.getPayload() is called,
			 * the encryption is executed if necessary.
			 */
			if (isEncrypted)
			{
				EncryptedPayload = payload;
				Payload = null;
			}
			/*
			 * Otherwise, only plaintext payload. The decryotion is not done 
			 * immediately. When EncryptedPayloadExtension.toByteArray() is called, 
			 * the decryption is executed if necessary.
			 * 
			 */
			else
			{
				Payload = payload;
				EncryptedPayload = null;
			}
		}


		/**
		 * Constructs an Extension from an existing one
		 */
		public EncryptedPayloadExtension(EncryptedPayloadExtension e)
		{
			keyVault = e.keyVault;
			msgKey = e.msgKey;
			Payload = e.Payload;
			EncryptedPayload = e.EncryptedPayload;
		}
		
		/**
		 * Constructs an empty Extension
		 */
		public EncryptedPayloadExtension()
		{
			keyVault = null;
			Payload = null;
			EncryptedPayload = null;
		}

		
		/**
		 * Return a new EncryptedPayloadExtension extension instance by cloning this extension.
		 */
		public Extension Clone()
		{
			return new EncryptedPayloadExtension(this);
		}
		
		/**
		 * Sets the message key which is used to do encryption/decryption of the payload.
		 */
		public void setMessageKey(Key messagekey)
		{
			msgKey = messagekey;
		}

		/**
		 * Converts this object to byte array.
		 */
		public byte[] toByteArray()
		{
			if ((EncryptedPayload == null) && (Payload == null))
			{
				throw new HyperCastWarningRuntimeException ("No payload byte aray is found in payload extension.");
			}
			
			/*
			 * This object has two variables which store plaintext payload and 
			 * encrypted payload seperately. This method returns encrypted
			 * payload byte array which is the payload data carried in the packet. 
			 * When this method is called but encrypted payload has not been 
			 * created, the encryption is done here.
			 */
			if (EncryptedPayload == null)
			{
				encryptPayload();
			}

			return EncryptedPayload;
		}


		/*
		 * This method provides a way for all extensions to get the length of 
		 * the byte array of an extension without actually executing toByteArray().
		 */
		/**
		 * Returns the length of encrypted payload byte array in this extension.
		 */
		public int getSize()
		{
			/*
			 * The length of encrypted payload in this extension is returned. The 
			 * encrypted payload is the payload data actually carried in the packet.
			 */
			return this.toByteArray().length;
		}

		/**
		 * Returns the payload byte array.
		 */
		public byte[] getPayload()
		{
			if ((EncryptedPayload == null) && (Payload == null))
			{
				throw new HyperCastWarningRuntimeException ("No payload byte aray is found in payload extension.");
			}
			
			/*
			 * This method returns plaintext payload byte array. When this 
			 * method is called but plaintext payload has not been created, 
			 * the decryption is done here.
			 */
			if (Payload == null)
			{
				decryptPayload();
			}

			return Payload;
		}

		
		/**
		 * Sets the plaintext payload in this extension to given byte array.
		 */
		public void setPayload(byte[] payload)
		{
			/*
			 * Sets the plaintext payload byte array in
			 * this extension.
			 */
 			Payload = payload;
 			
 			/*
 			 * When new plaintext payload is set in this extension, the encryption
 			 * is not done immediately, so the encrypted payload is set to null.
 			 * When EncryptedPayloadExtension.toByteArray() is called later, the 
 			 * encryption is executed and encrypted payload is returned.
 			 */
			EncryptedPayload = null;
		}

		/*
		 * It is implemented in the way that actual decryption operation is
		 * only executed one time.
		 */
		/**
		 * Decrypts the payload in this message.
		 *
		 * @throws IllegalArgumentException.
		 */
		private void decryptPayload() 
		{
			if ((EncryptedPayload == null) && (Payload == null))
			{
				throw new HyperCastWarningRuntimeException ("No payload byte aray is found in payload extension.");
			}
			
			/*
			 * If this message has an plaintext payload, don't do decryption.
			 */
			if (Payload != null)
			{
				return;
			}

			if (msgKey == null)
			{
				throw new HyperCastWarningRuntimeException ("Try to decrypt an encrypted message without message key.");
			}
	
			/*
			 * Get the decryption algorithm for this message.
			 */
			String confidentialityAlgorithm = keyVault.getConfidentialityAlgorithm();
	
			/*
			 * Decrypt the payload with message key.
			 */
			Cipher decrypt;
	
			try
			{
				decrypt = Cipher.getInstance(confidentialityAlgorithm);
			}
			catch (NoSuchPaddingException ipe)
			{
				throw new IllegalArgumentException("Invalid security parameters");
			}
			catch (NoSuchAlgorithmException nsae)
			{
				throw new IllegalArgumentException("Invalid security parameters");
			}

			try
			{
				decrypt.init(Cipher.DECRYPT_MODE, msgKey);
			}
			catch (InvalidKeyException ike)
			{}
	
			try
			{
				/*
				 * Decrypt the payload.
				 */
				Payload = decrypt.doFinal(EncryptedPayload);
			}
			catch (IllegalStateException e)
			{
				e.printStackTrace();
			}
			catch (IllegalBlockSizeException e)
			{
				//System.out.println("Payload length was:  " + getPayloadWithoutDecryption().length);
				//System.out.println("Expected length was: " + payloadLen);
				//System.out.println("Payload length is " + getPayloadWithoutModification().length + ", it is corrupted.");
				System.out.println("Payload length is " + EncryptedPayload.length + ", it is corrupted.");
				e.printStackTrace();
			}
			catch (BadPaddingException e)
			{
				//System.out.println("Payload length was:  " + getPayloadWithoutDecryption().length);
				//System.out.println("Expected length was: " + payloadLen);
				//System.out.println("Payload length is " + getPayloadWithoutModification().length + ", it is corrupted.");
				System.out.println("Payload length is " + EncryptedPayload.length + ", it is corrupted.");
				e.printStackTrace();
			}
		}//end of decryptPayload


		/*
		 * It is implemented in the way that actual encryption operation is
		 * only executed one time.
		 */
		/**
		 * Encrypts the payload in this message.
		 *
		 * @throws IllegalArgumentException.
		 */
		private void encryptPayload()
		{
			if ((EncryptedPayload == null) && (Payload == null))
			{
				throw new HyperCastWarningRuntimeException ("No payload byte aray is found in payload extension.");
			}
			
			/*
			 * If this message has an encrtpted payload, don't do encryption.
			 */
			if (EncryptedPayload != null)
			{
				return;
			}

			if (msgKey == null)
			{
				throw new HyperCastWarningRuntimeException ("Try to encrypt an plaintext message without message key.");
			}

	
			/*
			 * Get the encryption algorithm for this message.
			 */
			String confidentialityAlgorithm = keyVault.getConfidentialityAlgorithm();
	
			/*
			 * Encrypt the payload with message key.
			 */
			Cipher encrypt;
	
			try
			{
				encrypt = Cipher.getInstance(confidentialityAlgorithm);
			}
			catch (NoSuchPaddingException ipe)
			{
				throw new IllegalArgumentException("Invalid security parameters");
			}
			catch (NoSuchAlgorithmException nsae)
			{
				throw new IllegalArgumentException("Invalid security parameters");
			}
	
			try
			{
				encrypt.init(Cipher.ENCRYPT_MODE, msgKey);
			}
			catch (InvalidKeyException ike)
			{}
	
			/*
			 * Encrypt the payload.
			 */
			try
			{
					EncryptedPayload = encrypt.doFinal(Payload);
			}
			catch (BadPaddingException bpe)
			{
				throw new IllegalArgumentException("Invalid security parameters");
			}
			catch (IllegalBlockSizeException ibse)
			{
				throw new IllegalArgumentException("Invalid security parameters");
			}
	
		}//end of encryptPayload()

		
	}
