/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.*;

/**
 * This class defines a FSM instance pool which is built when the 
 * MessageStore is instantialized. The MessageStore creates instances
 * for each type FSM with the amount specified in the configuration
 * file. These pre-allocated FSM instances are used when a new 
 * FSM message is received by the MessageStore and a new FSM is
 * needed. Using pre-allocated FSMs instead of creating a new FSM 
 * instance on fly for a new FSM message aims at reducing the processing
 * time of each new FSM message. By default, there is no limit on the
 * amount of FSM instances with certain type that can be stored in
 * the pool.
 * 
 * @author HyperCast Team
 * @author Jianping Wang
 * @version Apr. 5, 2005
 */
public class FSMPool extends ObjectPool
{
	/**
	 * Debug flag
	 */
	protected boolean debug = false;

	/**
	 * Constructor.
	 */
	public FSMPool()
	{
		super();
	}

	/**
	 * Add a vector of FSM instances with specified FSM id to the pool.
	 */
	public void addFSMVector(final Vector v, final short fsmid)
	{
		super.addObjVector(v, fsmid);

	}

	/**
	 * Get the vector of FSM instances with specified FSM id in the pool.
	 */
	public Vector getFSMVector(final short fsmid)
	{
		return super.getObjVector(fsmid);
	}
	
	/**
	 * Add a FSM instance with specified FSM id to the pool.
	 */
	public void addFSMInstance(final Object fsmInstance, final short fsmid)
	{
		if (!checkFSMType(fsmInstance, fsmid))
		{
				throw new HyperCastWarningRuntimeException ("Try to add a FSM instance to the vector which stores other type FSM instances.");
		}
		
		super.addObjInstance(fsmInstance, fsmid);
	}

	/**
	 * Return a FSM instances with specified FSM id from the pool.
	 * If no required FSM instance is found, return null.
	 */
	public I_MessageStoreFSM getFSMInstance(final short fsmid)
	{
		Object result = getObjInstance(fsmid);
		
		if ((result != null) && (!checkFSMType(result, fsmid)))
		{
				throw new HyperCastWarningRuntimeException ("Try to get a object instance from the pool which is not a FSM instance.");
		}
		
		
		return (I_MessageStoreFSM) result;
	}
	
	/**
	 * Check the type of a given I_MessageStoreFSM.
	 */
	private boolean checkFSMType(final Object fsmInstance, final short fsmid)
	{
		if ((fsmInstance != null) &&
			(((fsmid == MessageStore.FSM_E2E_ACK) && (fsmInstance instanceof MessageStoreFSM_E2EACK)) ||
			((fsmid == MessageStore.FSM_H2H_ACK) && (fsmInstance instanceof MessageStoreFSM_H2HACK)) ||
			((fsmid == MessageStore.FSM_DUPLICATE_ELIMINATION) && (fsmInstance instanceof MessageStoreFSM_DuplicateElimination)) ||
			((fsmid == MessageStore.FSM_SYNCHRONIZATION) && (fsmInstance instanceof MessageStoreFSM_Sync)) ||
			((fsmid == MessageStore.FSM_INCAST) && (fsmInstance instanceof MessageStoreFSM_Incast)) ||
			((fsmid == MessageStore.FSM_BEST_EFFORT_ORDERING) && (fsmInstance instanceof MessageStoreFSM_Ordering)) ||
			((fsmid == MessageStore.FSM_NAMING) && (fsmInstance instanceof MessageStoreFSM_Naming))))
				return true;
		else
			return false;
	}

	/**
	 * Set the maximum number of a certain type FSM instances that the pool
	 * can store.
	 */
	public void setMaxNumOfFSMInstances(final short fsmid, final int num)
	{
		super.setMaxNumOfObjInstances(fsmid, num);
	}

	/**
	 * Get the maximum number of a certain type FSM instances that the pool
	 * can store.
	 */
	public int getMaxNumOfFSMInstances(final short fsmid)
	{
		return super.getMaxNumOfObjInstances(fsmid);
	}
}
