/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.HC;

import hypercast.*;

/**	
 * This class includes all the functions related with logical address <p>
 * Functionality: <br>
 * It provides the logical address for a hypercube node.
 *  It includes all operations which one might need to do regarding 
 *  a Logical address. This would include, 
 *  1. Converting from Gray to Colex and vice-versa <br>
 *  2. Printing the Logical Address <br>
 *  3. Find parent of a node with respect to a root <br>
 *  4. Finding neighbor information <br>
 * 
 * @author HyperCast Team
 * @author Tyler Beam
 * @author Dongwen Wang
 * @version	2.0, Feb. 23, 2001
 */


public final class HC_LogicalAddress implements I_LogicalAddress {

    /** The index of this node in the hypercube or HC_Node.MaxAddress() if unvalid. */
	private int ColexIndex;
	/** The gray code encoding of <code>ColexIndex</code> */
	private int GrayIndex;

    /** True if <code>bitToOrdered</code> and <code>OrderedToBit</code> have been set. */
	private boolean OrderingDone;

    /** Permutation required to turn a neighborhood sorted by the bit
    * flipped in the Gray index into a neighborhood table sorted by
    * the Colex index.*/
	private int[] BitToOrdered;
    /** Permutation required to turn a neighborhood sorted by Colex
    * index into a neighborhood table sorted by the bit flipped in the
    * Gray index.*/
	private int[] OrderedToBit;


	/**
	 * Constructs HC_LogicalAddress object from an Gray index or Colex index <p>
	 * Algorithm: <br>
	 *  If the second argument is true, the integer is the colex ordering, and
	 *  the gray ordering is trivial to find from the colex ordering, using the
	 *  the formula (colex XOR (colex/2)) <br>
	 *  If the second argument is false, and the integer is a gray index, the
	 *  reverse conversion is a bit more difficult, and is explained in greater
	 *  detail in the tech report <br>
	 *  The SortedNeighborhood variable is set to false <br>
	 * @param	index	An Integer which can represent the nodes gray index or
	 *                  colex index
	 * @param	iscolex	a boolean to indicate if the first argument is
	 *                  colex or a gray index
	 */

	public HC_LogicalAddress(int index, boolean iscolex){
		
		// index cannot have most significant bit set.
		if (index < 0) {
			throw new IllegalArgumentException("HC_LogicalAddress - Argument cannot be negative");
		}
    
		if (iscolex == true){
			ColexIndex = index;
			GrayIndex  = ColexIndex ^ (ColexIndex >>> 1);
		} 
        else 
        {
			GrayIndex = index;
			ColexIndex = convertGrayToColex(GrayIndex);
		}	

		OrderingDone = false;
	}
	/** added by Guangyu May 02, 04 */
	public HC_LogicalAddress(String addr){
		int index=Integer.parseInt(addr);
		
		ColexIndex = index;
		GrayIndex  = ColexIndex ^ (ColexIndex >>> 1);

		OrderingDone = false;
	}

/**
 * Constructs HC_LogicalAddress object from an byte array 
 * using  Colex index Algorithm: <br>
 * @param	byteAddr	a bytearray containing the Colex index.
 * @param	offset		the start position of index in the byte array.
 */
	public HC_LogicalAddress(byte[] byteAddr, int offset){
		

		int index = ByteArrayUtility.toInteger(byteAddr, offset);

		// index cannot have most significant bit set.
		if (index < 0) {
			throw new IllegalArgumentException("HC_LogicalAddress - Argument cannot be negative");
		}
    
		ColexIndex = index;
		GrayIndex  = ColexIndex ^ (ColexIndex >>> 1);

		OrderingDone = false;
	}

	/**
	 * Constructs a new HC_LogicalAddress object from another HC_LogicalAddress
	 * @param	laddr	An object of type HC_LogicalAddress
	 */

	public HC_LogicalAddress(HC_LogicalAddress laddr){
		
		ColexIndex = laddr.getColexIndex();
		GrayIndex  = laddr.getGrayIndex();
		OrderingDone = false;
	}
    

	
	
	/** 
	 * Converts Gray index to Colex index 
	 * @param	index	An integer which is supposed to represent the GrayIndex
	 * @return	the corresponding ColexIndex
	 */
	
	private int convertGrayToColex(int index){
		// Do conversion from Gray to Colex
		boolean EvenNumberOfOnes = true;
		int MSB = 0;
		int ColexIndex = 0;

		// This function is generic and works on 32 bit numbers. Any
		// restriction on the index has top be put by the calling function
		// when it gets the argument from somebody else.

		for (int i = 32; i > 0; i--) {
			if (((index << 1) >>> 1) == index) MSB = 0;
			else MSB = 1;

			if (!EvenNumberOfOnes) { 
				ColexIndex = (ColexIndex << 1) | (1-MSB);
			} else {
				ColexIndex = (ColexIndex << 1) | MSB;
			}

			if (MSB == 1) EvenNumberOfOnes = !EvenNumberOfOnes;
			index = index << 1;
		}

		return ColexIndex;
	}



	/** 
     * Initializes the values of <code>OrderedToBit</code> and
	 * <code>BitToOrdered</code>.  */
	private void orderNeighborhood() {
		if (OrderingDone) 
            return;

		// Now arrange the neighborhood
		BitToOrdered = new int[HC_Node.MaxHypercubeDim];
		OrderedToBit = new int[HC_Node.MaxHypercubeDim];

		int Neighborhood[] = new int[HC_Node.MaxHypercubeDim];

		for (int i = 0; i < Neighborhood.length; i++){
			Neighborhood[i] = convertGrayToColex(GrayIndex ^ (1 << i));
			OrderedToBit[i] = i;
		}

		// Now sort the neighborhood.
		// Quick and dirty bubble sort
		for (int i = 0; i < Neighborhood.length; i++) {
			int min = i;
			for(int j = i + 1; j < Neighborhood.length; j++) {	
				if (Neighborhood[min] > Neighborhood[j]) min = j;
			}

			int temp = Neighborhood[min];
			Neighborhood[min] = Neighborhood[i];
			Neighborhood[i]   = temp;

			temp = OrderedToBit[min];
			OrderedToBit[min] = OrderedToBit[i];
			OrderedToBit[i] = temp;
		}

		// Make the arrays needed
		for(int i = 0; i < Neighborhood.length; i++){
			BitToOrdered[OrderedToBit[i]] = i;
		}

		OrderingDone = true;
	}


		

	/** 
	 * Flip the Nth bit of the gray code. <p>
	 * Algorithm:
	 *  Takes the GrayIndex of the node itself, and flips the bit at which
	 *  the neighbor differs from the node. It then builds the logical
	 *  address of the neighbor using the appropriate constructor.
	 * @param	i	The bit in which the neighbor differs from the node. 
	 *              The nodes are numbered from 0 to 31
	 * @return	HC_LogicalAddress object corresponding to Neigbor
	 * @exception	No formal Exceptions raised; however, it exits
	 *              if the Neighbor index is invalid.
	 */

	private HC_LogicalAddress flipNthBit(int i){
		if ((i >= HC_Node.MaxHypercubeDim) || (i < 0)) 
			throw new IllegalArgumentException("Logical Address: Invalid Index");

		return new HC_LogicalAddress(GrayIndex ^ (1 << i), false);
	}




	/** 
	 * Gets the Nth Neighbor's logical address <p>
	 * Algorithm: <br>
	 *  1. Create the ordering if it is not created, by calling orderNeighborhood <br>
	 *  2. Find the bit in which the required neighbor differs and then XOR <br>
	 *     that bit in GrayIndex, and create a new HC_LogicalAddress.
	 * @param	index	The index of the neighbor, where all neighbors are ordered
	 *                  by the colex number
	 * @return	HC_LogicalAddress object corresponding to Neigbor
	 * @exception	No formal Exceptions raised; however, it ArrayBoundsException
	 *              will be raised by the system if index is out of bounds
	 */

	public HC_LogicalAddress getNthNeighbor(int index){
		if (!OrderingDone) orderNeighborhood();

		return new HC_LogicalAddress(GrayIndex ^ (1<<OrderedToBit[index]), false);
	}




	/** 
	 * Finds the colex index of a neighbor with a specified logical address.
	 * Find the HammingVector by XORing the GrayIndex variable of the argument
	 * and the node itself. If this Vector has only one bit then it is a 
	 * neighbor and we return the index in the neighborhood, else return -1
	 * The index in the neighborhood is found by finding the position of the
	 * flipped bit, and using BitToOrdered to figure out its place in the 
	 * ordering. 
	 * 
	 * @param	laddr	HC_LogicalAddress of a node
	 * @return	-1 if the node is not a neighbor, and the ColexIndex otherwise
	 */

	public int findNeighborIndex(HC_LogicalAddress laddr) {
		int i;
		int HammingVector = GrayIndex ^ laddr.GrayIndex;

		if (HammingVector == 0) {
			return HC_Node.MaxAddress();
		}

		// Shift right till the LSB is a 1
		for(i = 0; (((HammingVector >>> 1) << 1) == HammingVector); i++) 
			HammingVector = HammingVector >>> 1;

		// Only the LSB should be on at this stage and nothing else.
		if (HammingVector != 1) 
			return HC_Node.MaxAddress();

		if (!OrderingDone) 
            orderNeighborhood();

		return BitToOrdered[i];
	}

   


 


	/** 
	 * Gets the Parent of this node <p>
	 * Algorithm: <br>
	 *  1. Compare the ColexIndex of the node with that of the root <br>
	 *  2. If ColexIndex of the node is less than that of the root, flip
	 *     the least bit of difference <br>
	 *  3. If the situation is otherwise, then flip the greatest bit of 
	 *     difference <br>
	 * @param	root	Takes as argument the HC_LogicalAddress object of the node which
	 *				    will be the root of the tree against which you want to find
	 *                  this nodes parent
	 * @return	HC_LogicalAddress of the Parent
	 * @exception	Exits, if the node itself is passed as a root
	 */

	public HC_LogicalAddress parent(HC_LogicalAddress root){
		int i;
		int HammingVector = GrayIndex ^ root.GrayIndex;
		
		
		if (HammingVector == 0) 
            return null;
		   // return new LogicalAddress(HC_Node.MaxAddress(),true);

		if (ColexIndex > root.ColexIndex) {
			// Pick greatest bit of difference
			for(i = 0; HammingVector != 1; i++) {
				HammingVector = HammingVector >>> 1;
			}
		} else {
			// Pick least bit of difference
			for(i = 0; HammingVector == ((HammingVector >>> 1) <<1); i++) {
				HammingVector = HammingVector >>> 1;
			}
		}

		// Flip the chosen bit
		return new HC_LogicalAddress(GrayIndex ^ (1 << i), false);
		
	}



	/**
	 * Gets the children of this node <p>
	 * Algorithm: <br>
	 *  1. Compare the ColexIndex of the node with that of the root <br>
	 *  2. If ColexIndex of the node is less than that of the root, flip
	 *     the least bit of difference <br>
	 *  3. If the situation is otherwise, then flip the greatest bit of 
	 *     difference <br>
	 * @param	root	the HC_LogicalAddress object of the node which
	 *                  will be the root of the tree
	 * @param	hroot	the HC_LogicalAddress object of the HRoot of the Hypercube 
	 * @return	HC_LogicalAddress array of the Children
	 * @exception	Exits, if the node itself is passed as a root
	 */

	public HC_LogicalAddress[] children(HC_LogicalAddress root, HC_LogicalAddress hroot) {

		int hammingVector = GrayIndex ^ root.getGrayIndex();

		int hammingDistance = 0;
		int left = 0;
		int right = 0;

		// calculate the hamming distance and the left and right bounds for the 'middle section'
		// skip this step if this node is the root
		if (hammingVector != 0) {
			int index = 0;
			while ((hammingVector & 1) != 1) {
				hammingVector = hammingVector >>> 1;
				index++;
			}
			right = index;
			hammingDistance++;
			while (hammingVector != 1) {
				hammingVector = hammingVector >>> 1;
				if ((hammingVector & 1) == 1) {
					hammingDistance++;
				}
				index++;
			}
			left = index;
		}

		// calculate the dimension of the hypercube
		int hrootGrayIndex = hroot.getGrayIndex();
		int hypercubeDimension = 0;
		while (hrootGrayIndex != 0) {
			hrootGrayIndex = hrootGrayIndex >>> 1;
			hypercubeDimension++;
		}


		
		
		int childCount = 0;
		int hrootIndex = hroot.getColexIndex();
		int rootIndex = root.getColexIndex();

		// there can be at most hypercubeDimension - hammingDistance children for this parent
		HC_LogicalAddress[] children = new HC_LogicalAddress[hypercubeDimension - hammingDistance];

		if (hammingVector == 0) {
			// this is the root, so add all children
			for (int i = 0; i < hypercubeDimension; i++) {
				HC_LogicalAddress candidate = new HC_LogicalAddress(GrayIndex ^ (1 << i), false);
				int candidateIndex = candidate.getColexIndex();
				if (candidateIndex <= hrootIndex) {
					children[childCount] = candidate;
					childCount++;
				}
			}
		}
		else {
			// this is not the root, so only flip bits that are not in the 'middle section'
			for (int i = 0; i < right; i++) {
				HC_LogicalAddress candidate = new HC_LogicalAddress(GrayIndex ^ (1 << i), false);
				int candidateIndex = candidate.getColexIndex();
				if (candidateIndex < rootIndex) {
					children[childCount] = candidate;
					childCount++;
				}
			}

			for (int i = left + 1; i < hypercubeDimension; i++) {
				HC_LogicalAddress candidate = new HC_LogicalAddress(GrayIndex ^ (1 << i), false);
				int candidateIndex = candidate.getColexIndex();
				if (candidateIndex > rootIndex && candidateIndex <= hrootIndex) {
					children[childCount] = candidate;
					childCount++;
				}
			}
		}

		
		// if too much space was allocated for the array, copy the children array to a new array
		if (childCount == hypercubeDimension - hammingDistance) 
			return children;
		
		HC_LogicalAddress result[] = new HC_LogicalAddress[childCount];
		for (int i = 0; i < childCount; i++) {
			result[i] = children[i];
		}
		return result;
	}



	/** 
	 * Gets the colex index 
	 * @return	ColexIndex of the object
	 */

	public int getColexIndex(){		
		return ColexIndex;
	}



	/** 
	 * Gets the gray index
	 * @return	Gray Index of the object
	 */

	public int getGrayIndex(){
		return GrayIndex;
	}



	/**
	 * Check the equality of two HC_LogicalAddress objects <p>
	 * Algorithm: <br>
	 *  Compares the GrayIndex of both nodes, and the ColexIndex of both.
	 *  If both pairs match, return true. If both pairs don't match return false
	 *  If however, only one of the pairs matches, there is an inconsistency,
	 *  So it exits.
	 * 
	 * @param	hcla	HC_LogicalAddress object against which testing is to be done
	 * @return	true or false depending on result of equality test
	 * @exception	No formal exceptions; however it exits if it finds an inconsistency
	 */
	
	
	public boolean equals(HC_LogicalAddress hcla){		
		if ((hcla.getColexIndex() == ColexIndex) && (hcla.getGrayIndex() == GrayIndex))
			return true;

		if ((hcla.getColexIndex() != ColexIndex) && (hcla.getGrayIndex() != GrayIndex))
			return false;

		throw new IllegalStateException("Logical Address: Inconsistent State!");
	}
	

    /** Implements Object.equals(Object) as required by I_Address. */
	
    public boolean equals(Object o) {
        if (o == null)
            return false;
        else if (o instanceof HC_LogicalAddress)
            return equals((HC_LogicalAddress) o);
		else
            throw new IllegalArgumentException("Call to HC_LogicalAddress.equals() with argument not of type HC_LogicalAddress!");
	}
	
	
	

	
	/**
	 * Compares two HC_LogicalAddress objects <p>
	 * Algorithm: <br>
	 *  Compares the ColexIndex of both nodes
	 * @param	laddr	HC_LogicalAddress object against which testing is to be done
	 * @return	true or false depending on result of compare result
	 */

	public boolean greaterThan(I_Address laddr){
        if (laddr == null || !(laddr instanceof HC_LogicalAddress))
            throw new IllegalArgumentException("Call to HC_LogicalAddress.equals() with argument not of type HC_LogicalAddress!");

		HC_LogicalAddress hcla = (HC_LogicalAddress) laddr;
		if (ColexIndex > hcla.getColexIndex())
			return true;
		else 
            return false;
	}
	
	

	
	/**
	 * Returns the byte array
	 */
	public byte[] toByteArray() {
		byte[] la;
		
		la = ByteArrayUtility.toByteArray(ColexIndex);
		return la;
	}
	
	
	
	/**
	 * Returns the size of the byte array of this HC_LogicalAddress 
	 */
	public int getSize() {
		return 4;
	}
	
	

	/** 
	 * Converts logical address object to String
	 * @return	String which is a printable representation of object contents 
	 *          colexindex (grayindex)
	 */

	public String toString(){
		if (ColexIndex == HC_Node.MaxAddress())
			return("Unknown");
		else
			return (Integer.toString(ColexIndex,10));
	}
}

