/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.net.*;
import java.io.*;

/** This class contains a set of static functions used by
 * UDP_HTTP_ServerAdapter and HTTP_Server for server transactions.
 * These functions translate a message from bytes into a string of
 * hexadecimal characters ('0'-'9' or 'A'-'F'), insert that string
 * into a CGI query, read the response, translate the response
 * back into a byte[] and return it.
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @version 2.0, Feb. 26, 2001 
 */

public class HTTP_ServerUtility 
{

/** Maximum size of a response coming from the server. */
public static final int MAX_SERVER_RESPONSE_SIZE = 8192;

/** Builds a simple query without a message and return the
 * response as a string. */
public static String query(String urlString) throws MalformedURLException, IOException
    {
        URL url = new URL(urlString);
        String result = null;
        HttpURLConnection connection = null;
        try
        {
            connection = (HttpURLConnection) url.openConnection();
            result = firstLineFromStream(connection.getInputStream());
        }
        finally 
        {
            connection.disconnect();
        }
           
        return result;
    }
    

/** Builds a query with a message and return the
 * response as another message. */
public static byte[] query(String server_query_prefix, String overlayID, byte[] input_message) throws MalformedURLException, IOException
    {
        String urlString = server_query_prefix + "?cmd=msg&OverlayID=" + encodeURLString(overlayID) +
            "&msg=" + toHexString(input_message);

        URL url = new URL(urlString);
        String result = null;
        HttpURLConnection connection = null;
        try
        {
            connection = (HttpURLConnection) url.openConnection();
            result = firstLineFromStream(connection.getInputStream());
        }
        finally 
        {
            connection.disconnect();
        }

        return toByteArray(result);
    }

/** Reads the first line from the InputStream and returns it. */
private static String firstLineFromStream(InputStream response_is) throws IOException
    {
        InputStreamReader response_isr = new InputStreamReader(response_is);
        BufferedReader response = new BufferedReader(response_isr, MAX_SERVER_RESPONSE_SIZE);
        
        try 
        {
            String s =  response.readLine();     
            response.close();
            return s;
        }
        finally 
        {
            response.close();
        }
    }
    

/* Converts a byte[] into a hex string.  The lower indexed bytes come first;
 * of each byte, the higher nibble is printed first. 
* e.g. if b[0]= 15, b[1] = 16 then toHexString(b) would return "0F10".
*/
public static String toHexString(byte[] a)
    {
        StringBuffer sb = new StringBuffer();
    
        for (int i = 0; i < a.length; i++) 
        {
            final int high_order_nibble = (a[i] >>> 4) & 0xF;
            if (high_order_nibble < 10)
                sb.append( (char) ('0' + high_order_nibble));
            else if (high_order_nibble >= 10  && high_order_nibble <16)
                sb.append( (char) ('A' + (high_order_nibble - 10)));
            else 
            {
                throw new HyperCastFatalRuntimeException ("ERROR translating high order nibble:" + high_order_nibble);
            }
            


            final int low_order_nibble = a[i] & 0xF;
            if (low_order_nibble < 10)
                sb.append( (char) ('0' + low_order_nibble));
            else if (low_order_nibble >= 10 && low_order_nibble < 16)
                sb.append( (char) ('A' + (low_order_nibble - 10)));
            else 
            {
                throw new HyperCastFatalRuntimeException ("ERROR translating low order nibble:" + low_order_nibble);
            }
        }

        return sb.toString();
    }


/** This is the inverse function of toHexString(). 
 * It takes a string of hexidecimal digits and returns a byte[].
 * The string can end with nothing, a newline character, or a carriage return.
 * Any extraneous characters causes an error message to be printed and
 * null returned.
 */
public static byte[] toByteArray(String s) 
    {
        byte[] buffer = new byte[s.length()/2];

        
        int i;
        for (i = 0; i < s.length(); i += 2) 
        {
            byte output_byte = 0;
            char first_char = s.charAt(i);
            
            if (first_char == '\n' || first_char == '\r')
                break;
            else if (first_char >= 'A' && first_char <= 'F')
                output_byte = (byte) (((byte) (first_char - 'A' + 10)) << 4);
            else if (first_char >= '0' && first_char <= '9')
                output_byte = (byte) (((byte) (first_char - '0')) << 4);
            else  {
                System.err.println("Error at webserver server.  Received msg: \"" + s + "\"");
                return null;
            }
            
            
            char second_char = s.charAt(i+1);

            if (second_char == '\n' || second_char == '\r')
                break;
            else if (second_char >= 'A' && second_char <= 'F')
                output_byte |= ((byte) (second_char - 'A' + 10));
            else if (second_char >= '0' && second_char <= '9')
                output_byte |= ((byte) (second_char - '0'));
            else {
                System.err.println("Error at webserver server.  Received msg: \"" + s + "\"");
                return null;
            }
            
            buffer[i/2] = output_byte;
        }
        
        byte[] output_buffer = new byte[i/2];
        System.arraycopy(buffer, 0, output_buffer, 0, i/2);

        return output_buffer;
    }
    



/*        
            byte[] test_message = new byte[16];
            test_message[0]  = 0x0F;
            test_message[1]  = 0x1E;
            test_message[2]  = 0x2D;
            test_message[3]  = 0x3C;
            test_message[4]  = 0x4B;
            test_message[5]  = 0x5A;
            test_message[6]  = 0x69;
            test_message[7]  = 0x78;
            test_message[8]  = (byte) 0x87;
            test_message[9]  = (byte) 0x96;
            test_message[10] = (byte) 0xA5;
            test_message[11] = (byte) 0xB4;
            test_message[12] = (byte) 0xC3;
            test_message[13] = (byte) 0xD2;
            test_message[14] = (byte) 0xE1;
            test_message[15] = (byte) 0xF0;
*/


/**
 * This method takes a string from a URL (e.g. CGI forms request) and 
 * converts it into a normal string.  This means that '+' gets converted
 * into ' ', and for any '%', the next two characters are treated as 
 * the hex encoding of an ASCII character.
 * WARNING: THIS IS POORLY CODED AND MAY NOT WORK ON NON-ASCII SYSTEMS!
 */
public static String decodeURLString(String s) 
    {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < s.length(); i++) 
        {
            if (s.charAt(i) == '+')
                sb.append(" ");
            else if (s.charAt(i) == '%') 
            {
                if (i+2 >= s.length())
                    throw new IllegalArgumentException("URL String was poorly formatted!");

                char first_char = s.charAt(i+1);
                char second_char = s.charAt(i+2);
                byte output_byte = 0;
            
                if (first_char >= 'A' && first_char <= 'F')
                    output_byte = (byte) (((byte) (first_char - 'A' + 10)) << 4);
                else if (first_char >= '0' && first_char <= '9')
                    output_byte = (byte) (((byte) (first_char - '0')) << 4);
                else  
                    throw new IllegalArgumentException("URL String was poorly formatted!");

                if (second_char >= 'A' && second_char <= 'F')
                    output_byte |= ((byte) (second_char - 'A' + 10));
                else if (second_char >= '0' && second_char <= '9')
                    output_byte |= ((byte) (second_char - '0'));
                else                 
                    throw new IllegalArgumentException("URL String was poorly formatted!");

                // Not good Java:
                sb.append((char) output_byte);
                i += 2;
            }
            else 
            {
                sb.append(s.charAt(i));
            }
        }

        return sb.toString();
    }


/** 
 * This function encodes a string in a URL format.  This means that 
 * all ' ' become '+' and any of the characters "&=?+" are encoded
 * using '%' and the character encoded as the hexidecimal of their
 * ASCII value.
 */
public static String encodeURLString(String s) 
    {
        return URLEncoder.encode (s);
        // I think that using this line of code will make the warning go away:
        //return URLEncoder.encode (s, "UTF-8");
        /*
         * The problem is that URLEncoder.encode (s, "UTF-8") is not supported 
         * in Java 2 Platform SE v1.3.1.
         */

/*        
        StringBuffer sb = new StringBuffer();

        for (int i = 0; i < s.length(); i++) 
        {
            if (s.charAt(i) == ' ')
                sb.append('+');
            else if (s.charAt(i) == '&')
                sb.append("%26");
            else if (s.charAt(i) == '=')
                sb.append("%3D");
            else if (s.charAt(i) == '?')
                sb.append("%3F");
            else if (s.charAt(i) == '+')
                sb.append("%2B");
            else
                sb.append(s.charAt(i));
        }
        
        return sb.toString();
*/
    }
}











