/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;


import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class implements a physical address which works on the 
 * TCP/IP networks. It contains a IP version 4 address and a port
 * number.
 * <p>
 *
 * @author HyperCast Team
 * @author Jianping Wang
 * @version	3.0, Jul. 24, 2004
 */

public class INETv4OnePort implements I_PhysicalAddress, I_UnderlayAddress {


	// Minimum and maximum port numbers. 
	public static final int MIN_PORT = 0;
	public static final int MAX_PORT = 256*256-1;  // = 2^16 - 1

	/**
	 * The size of IP address when carried in the packet.
	 */
	public static final int INETv4_ADDRESS_SIZE = 4;

	/**
	 * The size of port number when carried in the packet.
	 */
	public static final int PORT_NUMBER_SIZE = 2;

	/**
	 * The size of INETv4OnePort type physical address when 
	 * carried in the packet.
	 */
	public static final int ADDRESS_SIZE = INETv4_ADDRESS_SIZE + PORT_NUMBER_SIZE;//=6;

	/**
	 * The IPv4 address of the node. 
	 */
	protected InetAddress Address = null;

	/**
	 * The port number.
	 */
	protected int Port;


	/** 
	 * Constructs a <code>INETv4OnePort</code> object with another INETv4OnePort object.
	 * @param	inetv4oneport	An INETv4OnePort object.
	 */
	public INETv4OnePort(INETv4OnePort inetv4oneport)
	{
		Address = inetv4oneport.Address;
		Port = inetv4oneport.Port;
	}

	/** 
	 * Constructs a <code>INETv4OnePort</code> object with an InetAddress and port number.
	 * This method is called by the constructor of INETv4TwoPorts.
	 * @param	a	An InetAddress.
	 * @param	p	port number.
	 */
	public INETv4OnePort(InetAddress a, int p)
	{
		Port = p;


		if ((a != null) && (!getIPString(a).equals("0.0.0.0"))) {
			Address = a;
			//System.out.println(" (1) INETv4OnePort: InetAddress a = " + getIPString(address) + ",port=" + port);
		}
		else
		{
			try 
			{
			  if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
			  {
				 String LocalIPStr = INETv4OnePort.getLocalIPbyDNS();
				 if (!(LocalIPStr == null))
				 {
					Address = InetAddress.getByName(LocalIPStr);
				 } else {
					throw new HyperCastFatalRuntimeException ("Could not get local IP by DNS query!");
				 }
			  } else {
				 Address = InetAddress.getLocalHost();
			  }
			}
			catch (UnknownHostException uhe) 
			{
			  throw new HyperCastWarningRuntimeException ("InetAddress.getLocalHost() threw an UnknownHostException.");
			}
		}

	}


	/** 
	 * Constructs a <code>INETv4OnePort</code> object with DatagramSocket.
	 * @param	a	A DatagramSocketObject.
	 */
	public INETv4OnePort(DatagramSocket a)
	{
		this(a.getLocalPort());
	}


	/** 
	 * Constructs the <code>INETv4OnePort</code> object which has the specified
	 * port number.
	 * @param	portnum		a port number.
	 */
	 public INETv4OnePort(int portnum) {
		 try 
		 {
			/* Some Unix hosts return 127.0.0.1 as local IP addresses that 
			 * cause problem in the communication between OL Sockets. In this case, 
			 * a DNS inquire command is used to get local IP address.
			 */
			 if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
			 {
				String LocalIPStr = getLocalIPbyDNS();
				if (!(LocalIPStr == null))
				{
				   Address = InetAddress.getByName(LocalIPStr);
				} else {
				   throw new HyperCastFatalRuntimeException ("Could not get local IP address!");
				}
			 } else {
				Address = InetAddress.getLocalHost();
			 }
		 }
		 catch (UnknownHostException uhe) 
		 {
			 throw new HyperCastWarningRuntimeException (uhe);
		 }
        
		 Port = portnum;
	 }

	
	/**
	 *  Constructs an instance of InetAddress from a byte array.
	 */
	static InetAddress createInetAddress(byte[] byteAddress) {
		String stringAddress = null;
		try 
		{
			stringAddress = (byteAddress[0] & 0xff) + "." +
				(byteAddress[1] & 0xff) + "." +
				(byteAddress[2] & 0xff) + "." +
				(byteAddress[3] & 0xff);
			return InetAddress.getByName(stringAddress);
		}
		catch (UnknownHostException uhe)
		{
                    throw new HyperCastFatalRuntimeException ("Unknown Host Exception thrown by new InetAddress on \"" + stringAddress + "\"", uhe);
		}
	}


   /** 
	* Constructs a <code>INETv4OnePort</code> object from an six byte array.
	* 
	* @param	byteAddressAndPort	an Internet address and port stored in a byte array 
	* @exception IllegalArgumentException when p is not in 0 to (2^16)-1
	*/
	public INETv4OnePort(byte[] byteAddressAndPort) {
		if (byteAddressAndPort.length < INETv4OnePort.ADDRESS_SIZE)
			throw new IllegalArgumentException("length of the byte array is too short");

		Address = createInetAddress(byteAddressAndPort);
		Port = (ByteArrayUtility.toShort(byteAddressAndPort, INETv4OnePort.INETv4_ADDRESS_SIZE) & 0xFFFF);

		if (Port < MIN_PORT || Port > MAX_PORT)
			throw new IllegalArgumentException("Port value outside of range.");
	}

	
	/** 
	 * Constructs a <code>INETv4OnePort</code> object from a string. The string may contain the address and port
	 * separated by a '/' or ':', or only contain a port number.
	 * 
	 * @param	addrAndPortString	a string representing the INETv4 address and port (eg. "128.143.71.50/4747" or
	 * 								"128.143.71.50:4747" or "4747")
	 * @exception IllegalArgumentException when port is not in 0 to (2^16)-1
	 * @exception HyperCastWarningRuntimeException
	 */
	 public INETv4OnePort(String addrAndPortString) {
		StringTokenizer token = new StringTokenizer(addrAndPortString, ":/\n");

		int count = token.countTokens();
	
		/* If the string only contains a port number, the InetAddress is created by the system call and
		 * is the default local IP address.
		 */
		//if (slash_location < 0)
		if (count <= 1)
		{
			   try 
			   {
				 if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
				 {
					String LocalIPStr = getLocalIPbyDNS();
					if (!(LocalIPStr == null))
					{
					   Address = InetAddress.getByName(LocalIPStr);
					} else {
					   throw new HyperCastFatalRuntimeException ("Could not get local IP by DNS query!");
					}
				 } else {
					Address = InetAddress.getLocalHost();
				 }
			   }
			   catch (UnknownHostException uhe) 
			   {
				 //System.err.println("INETv4OnePort: InetAddress.getLocalHost() threw an UnknownHostException.");
				 //address = null;
				 throw new HyperCastWarningRuntimeException ("InetAddress.getLocalHost() threw an UnknownHostException.");
			   }
		 }
		/* If the string contains IP address and port number, the InetAddress and port are created from the string.
		 */
		 else
		 {
			   	//String addrString = addrAndPortString.substring(0, slash_location);
			   	String addrString = token.nextToken();
	
			   	try
			   	{
				  	Address = InetAddress.getByName(addrString);
			   	}
			   	catch (UnknownHostException e)
			   	{
				  	//System.err.println("INETv4OnePort: " + addrAndPortString + " is a bad address string.**" + e);
				  	//address = null;
				  	throw new HyperCastWarningRuntimeException (addrAndPortString + " is a bad address string.**" + e);
			   	}
		 }
	
		 //String portString = addrAndPortString.substring(slash_location+1, addrAndPortString.length());
		 String portString = token.nextToken();
		 try 
		 {
			 Port = Integer.parseInt(portString);
		 }
		 catch (NumberFormatException e) 
		 {
			 throw new IllegalArgumentException("Unable to parse  \"" + portString + "\" as a port number.");
		 }
	        
		 if (Port < MIN_PORT || Port > MAX_PORT)
			 throw new IllegalArgumentException("Port value outside of range.");
	 }

	/**
	 * Get string representation of IP address in a InetAddress object.
	 * @param	addr	An InetAddress object. 
	 */
	private String getIPString(InetAddress addr) {
		if (addr == null) return new String("None");
		
		String addrstr = addr.toString();
		int slash_loc = addrstr.indexOf('/');
		return addrstr.substring(slash_loc+1, addrstr.length());
	}


	/** 
	 * Gets INETv4 address of this object.
	 */
	 public InetAddress getInetAddress() {
		 return Address;
	 }
	

	/** 
	 * Gets byte address of this object.
	 * @return	A byte[] as returned by <code>InetAddress.getAddress</code>.
	 */
	 public byte[] getByteAddress() {
		 return Address.getAddress();
	 }

	
	/** 
	 * Gets port number of this INETv4OnePort object.
	 */	
	 public int getPort() {
		 return Port;
	 }


	/** 
	 * Sets port number of this IP_Address object.
	 */	
	 public void setPort(int port) {
		 Port = port;
	 }


	 /**
	  * Returns the byte array size.
	  */
	 public int getSize() {
		 return INETv4OnePort.ADDRESS_SIZE;
	 }

	
	/** 
	 * Converts the Address/Port pair to a 6 bytes array. The format is
	  * |-------4------|----2----|
	  * |      IP      |   port  |
	  * |              |         |
	  * |--------------|---------|
	 * @return	a 6 byte array where the first 4 bytes holds the address, 
	 * the next 2 bytes holds the port.
	 */	
	 public byte[] toByteArray() {
		 byte[] byteAddressAndPort = new byte[INETv4OnePort.ADDRESS_SIZE];
		
		 byte[] byteAddress = Address.getAddress();
		 byte[] bytePort = ByteArrayUtility.toByteArray((short) Port);
			
		 System.arraycopy(byteAddress, 0, byteAddressAndPort, 0, INETv4_ADDRESS_SIZE);
		 System.arraycopy(bytePort, 0, byteAddressAndPort, INETv4_ADDRESS_SIZE, PORT_NUMBER_SIZE);

		 return byteAddressAndPort;
	 }

	
	/** 
	 * Test if two INETv4OnePort addresses are identical.
	 * 
	 * @param	obj 	an object (should be of type <code>INETv4OnePort</code>.)
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>INETv4OnePort</code>.
	 */
	 public boolean equals(Object obj) {
		 if (obj == null)
		 {
			 return false;
		 }
		 else if (!(obj instanceof INETv4OnePort))  
		 {
			 throw new IllegalArgumentException("The passed object is not of type INETv4OnePort!");
		 }
		 else 
		 {
			INETv4OnePort ipAddr = (INETv4OnePort) obj;

			 //Using the method InetAddress.equals does not seem to work
			 byte[] remoteByteAddress = ipAddr.getByteAddress();
			 byte[] localByteAddress = Address.getAddress();

			 return ((remoteByteAddress[0]==localByteAddress[0]) && (remoteByteAddress[1]==localByteAddress[1]) && 
					 (remoteByteAddress[2]==localByteAddress[2]) && (remoteByteAddress[3]==localByteAddress[3]) && 
					 (Port == ipAddr.getPort()));

		 }
	 }
	

	/** 
	 * Test if two INETv4OnePort addresses are identical.
	 * 
	 * @param	obj 	an object (should be of type <code>INETv4OnePort</code>.)
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>INETv4OnePort</code>.
	 */
	 public boolean addressCheck(Object obj) {
	 	return this.equals(obj);
	 }

	
	/** 
	 * Compares two INETv4OnePort objects. 
	 * 
	 * @param	addr	an object of type <code>I_Address</code>
	 * @return	true if (this > addr), otherwise false 
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>INETv4OnePort</code>.
	 */
	 public boolean greaterThan(I_Address addr) {
		 if (addr == null || !(addr instanceof INETv4OnePort)) 
			 throw new IllegalArgumentException("The I_Address instance passed to INETv4OnePort.greaterThan is not of type INETv4OnePort!");
        
		INETv4OnePort ipAddr = (INETv4OnePort) addr;

		 byte[] localByteAddress = Address.getAddress();
		 byte[] remoteByteAddress = ipAddr.getByteAddress();

		 for (int i = 0; i < 4; i++) 
		 {
			 if (localByteAddress[i] == remoteByteAddress[i]) 
				 continue;
			 else
				 return ((localByteAddress[i]&0xff) > (remoteByteAddress[i]&0xff));
		 }

		 return (Port > ipAddr.getPort());
	 }
	
	
	/** 
	 * Converts the IP address object to string format "xxx.xxx.xxx.xxx/port" (e.g. "128.143.71.50/4747")
	 */
	 public String toString() {
		 byte[] byteAddress = Address.getAddress();
		 return ((byteAddress[0]&0xff) + "." + (byteAddress[1]&0xff) +"."+
			 (byteAddress[2]&0xff) + "." + (byteAddress[3]&0xff) + ":" + Port);
	 }
	
	
	 /**
	  * Returns a hashcode of this object
	  * It is important that the argument is of type <code>Object</code> so that 
	  * it can be used in a <code>Hashtable</code>.
	  */
	 public int hashCode() {
		 byte[] byteAddress = Address.getAddress();

		 int i = Port;
		 i ^= (byteAddress[0] << 0);
		 i ^= (byteAddress[1] << 8);
		 i ^= (byteAddress[2] << 16);
		 i ^= (byteAddress[3] << 24);
		 return i;
	 }
	

	/*
	 * This method is platform specific. But it solves a very common problem on Unix
	 * platforms and is very useful in the experiments. It may be deleted when found
	 * better solution.
	 */
	/**
	 * Obtain local IP address by a DNS query and return local IP address in String form.
	 */
	 public static String getLocalIPbyDNS() {
		try
		{
		   String LocalIP = InetAddress.getLocalHost().getHostAddress();
		   String LocalHostname = InetAddress.getLocalHost().getHostName();
			  //System.out.println("The local IP address is tested : " + LocalIP);
	
	
				boolean bDebug=false;
				String sOs=System.getProperty("os.name");
	
				Process p=null;
				Process q=null; //for ifconfig command
	
				BufferedReader r=null;
				String sLine;
				String Local_IP_DNS = null;
	
				if (sOs.equalsIgnoreCase("Linux") || sOs.equalsIgnoreCase("SunOS")) {
						//System.out.println("The local Host Name is tested : " + LocalHostname);
	
						try {
								p=Runtime.getRuntime().exec(new String[] {"host", LocalHostname});
						} catch (IOException e) {
								System.out.println(e);
						}
						r=new BufferedReader(new InputStreamReader(new BufferedInputStream(p.getInputStream())));
						try {
								int sig = 0;
								while ((sLine=r.readLine())!=null) {
									if (sLine.trim().startsWith(LocalHostname))
									{
										//System.out.println("The result is : " + sLine);
										//System.out.println("The local HostName is : " + LocalHostname);
										Local_IP_DNS = sLine.substring(sLine.indexOf("address")+7).trim(); 
										//System.out.println("The result DNS IP is : *****" + Local_IP_DNS + "****");
										return Local_IP_DNS;
									} else if (sLine.trim().startsWith("Name"))
									{   
										sig++;
									} else if ((sig == 1) && (sLine.trim().startsWith("Address")))
									{   
										sig++;
										Local_IP_DNS = sLine.substring(sLine.indexOf("Address")+8).trim(); 
										//System.out.println("The result DNS IP is : *****" + Local_IP_DNS + "****");
										return Local_IP_DNS;
									}
								}
						} catch (IOException e) {
								System.out.println(e);
						}
	
						//System.out.println("The DNS query is failed");
	
						try {
								q=Runtime.getRuntime().exec(new String[] {"/sbin/ifconfig"});
						} catch (IOException e) {
								System.out.println(e);
						}
						r=new BufferedReader(new InputStreamReader(new BufferedInputStream(q.getInputStream())));
						try {
								System.out.println("Start ifconfig try");
								while ((sLine=r.readLine())!=null) {
									if (sLine.trim().startsWith("inet"))
									{
										int start_ind = sLine.trim().indexOf("inet");
										int end_ind = sLine.trim().indexOf("Bcast");
										Local_IP_DNS = sLine.trim().substring(start_ind+10, end_ind-1).trim();
										//System.out.println("The result DNS IP is : *****" + Local_IP_DNS + "****");
										return Local_IP_DNS;
									}
								}
						} catch (IOException e) {
								System.out.println(e);
						}
						//System.out.println("The ifconfig command is failed");
	
				}
	        
		} catch (UnknownHostException ee) { 
		   System.err.println("Unknown Host :" + ee);
		}
	        
		return null;
	 }

}
