/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This class implements a physical address which works on the 
 * TCP/IP networks and contains a IP version 4 address and two
 * port numbers.
 * <p>
 *
 * @author HyperCast Team
 * @author Jianping Wang
 * @version	3.0, Jul. 24, 2004
 */

public class INETv4TwoPorts implements I_PhysicalAddress {

	/**
	 * The size of INETv4TwoPorts type physical address when 
	 * carried in the packet.
	 */
	public static final int ADDRESS_SIZE = INETv4OnePort.INETv4_ADDRESS_SIZE + 2 * INETv4OnePort.PORT_NUMBER_SIZE; //=8

	/**
	 * The IPv4 address of the node. 
	 */
	protected InetAddress Address = null;

	/**
	 * The port number used in the Node Adapter.
	 */
	protected int NodeAdptPort;

	/**
	 *  The port number used in the Socket Adapter.
	 */
	protected int ScktAdptPort = 0;


	/** 
	 * Constructs a INETv4TwoPorts object with another INETv4TwoPorts object.
	 * @param	a	An INETv4TwoPorts object.
	 */
	public INETv4TwoPorts(INETv4TwoPorts a)
	{
		NodeAdptPort = a.getNodeAdptPort();
		ScktAdptPort = a.getScktAdptPort();
		Address = a.getInetAddress();
	}

	/** 
	 * Constructs a INETv4TwoPorts object with a string representing 
	 * address and two port numbers.
	 * @param	addrAndPortsString	An string representing an address
	 * 			and two port numbers.
	 */
	public INETv4TwoPorts(String addrAndPortsString)
	{
		/*
		 * String addrAndPortsString has format "IP address:port1:port2", so
		 * it has three fields after seperated by ":".
		 */
		String[] paStr = INETv4TwoPorts.getPartsFromAddrAndPortsString(addrAndPortsString);

		/*
		 * Create InetAddress.
		 */			
		if (!paStr[0].equals("")) {
			try
			{
				Address = InetAddress.getByName(paStr[0]);
			}
			catch (UnknownHostException e)
			{
				Address = null;
			   	throw new HyperCastWarningRuntimeException ("Cannot create InetAddress from " + paStr[0] + " : " + e);
			}
		}
		else
		{
			try 
			{
			  if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
			  {
				 String LocalIPStr = INETv4OnePort.getLocalIPbyDNS();
				 if (!(LocalIPStr == null))
				 {
					Address = InetAddress.getByName(LocalIPStr);
				 } else {
					throw new HyperCastFatalRuntimeException ("Could not get local IP by DNS query!");
				 }
			  } else {
				 Address = InetAddress.getLocalHost();
			  }
			}
			catch (UnknownHostException uhe) 
			{
			  throw new HyperCastWarningRuntimeException ("InetAddress.getLocalHost() threw an UnknownHostException.");
			}
		}
		

		/*
		 * Obtain the first port number.
		 */
		if (!paStr[1].equals("")) {
			try 
			{
				NodeAdptPort = Integer.parseInt(paStr[1]);
			  if ((NodeAdptPort < 0) && (NodeAdptPort != -1))
			  {
				throw new IllegalArgumentException("" + paStr[1] + " is not a available value for port number.");
			  }
			}
			catch (NumberFormatException e) 
			{
			  throw new IllegalArgumentException("Unable to parse  \"" + paStr[1] + "\" as a port number.");
			}
		}
		else
		{
			NodeAdptPort = 0;
		}
		
		/*
		 * Obtain the second port number.
		 */
		if (!paStr[2].equals("")) {
			try 
			{
				ScktAdptPort = Integer.parseInt(paStr[2]);
				if ((ScktAdptPort < 0) && (ScktAdptPort != -1))
				{
				  throw new IllegalArgumentException("" + paStr[2] + " is not a available value for port number.");
				}
			}
			catch (NumberFormatException e) 
			{
			  throw new IllegalArgumentException("Unable to parse  \"" + paStr[2] + "\" as a port number.");
			}
		}
		else
		{
			ScktAdptPort = 0;
		}


	}

	/** 
	 * Constructs a INETv4AndOnePort object from an InetAddress instance 
	 * and two port numbers.
	 * 
	 * @param	a 	an instance of InetAddress
	 * @param	nodeport	port number for node adapter
	 * @param	scktport	port number for socket adapter
	 */
	public INETv4TwoPorts(InetAddress a, int nodeport, int scktport) {
		NodeAdptPort = nodeport;
		ScktAdptPort = scktport;

		if ((a != null) && (!this.getIPString(a).equals("0.0.0.0"))) {
			Address = a;
			//System.out.println("INETv4TwoPorts: Address=" + getIPString(Address) + ",nodeport=" + nodeport+", scktport="+scktport);
		}
		else
		{
			try 
			{
			  if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
			  {
				 String LocalIPStr = INETv4OnePort.getLocalIPbyDNS();
				 if (!(LocalIPStr == null))
				 {
					Address = InetAddress.getByName(LocalIPStr);
				 } else {
					throw new HyperCastFatalRuntimeException ("Could not get local IP by DNS query!");
				 }
			  } else {
				 Address = InetAddress.getLocalHost();
			  }
			}
			catch (UnknownHostException uhe) 
			{
			  throw new HyperCastWarningRuntimeException ("InetAddress.getLocalHost() threw an UnknownHostException.");
			}
		}

		//System.out.println("INETv4TwoPorts: Address=" + getIPString(Address) + ",nodeport=" + nodeport+", scktport="+scktport);

	}

	
	/** 
	 * Constructs a INETv4TwoPorts object with two port numbers. 
	 * The IP addeess is obtained from local host.
	 * @param	nodeport	port number for Node Adapter.
	 * @param	scktport	port number for Socket Adapter.
	 */
	public INETv4TwoPorts(int nodeport,  int scktport)
	{
		NodeAdptPort = nodeport;
		ScktAdptPort = scktport;
		
		try 
		{
		  if (InetAddress.getLocalHost().getHostAddress().equalsIgnoreCase("127.0.0.1"))
		  {
			 String LocalIPStr = INETv4OnePort.getLocalIPbyDNS();
			 if (!(LocalIPStr == null))
			 {
				Address = InetAddress.getByName(LocalIPStr);
			 } else {
				throw new HyperCastFatalRuntimeException ("Could not get local IP by DNS query!");
			 }
		  } else {
			 Address = InetAddress.getLocalHost();
		  }
		}
		catch (UnknownHostException uhe) 
		{
		  throw new HyperCastWarningRuntimeException ("InetAddress.getLocalHost() threw an UnknownHostException.");
		}
		
	}


	/** 
	 * Constructs a INETv4TwoPorts object from a byte array.
	 * @param	byteAddrPort	a byte array storing IP 
	 * 			address	and two port numbers.
	 */
	public INETv4TwoPorts(byte[] byteAddrPort)
	{
		if (byteAddrPort.length < INETv4TwoPorts.ADDRESS_SIZE)
			throw new IllegalArgumentException("length of the byte array is too short");

		Address = INETv4OnePort.createInetAddress(byteAddrPort);
		NodeAdptPort = (ByteArrayUtility.toShort(byteAddrPort, INETv4OnePort.INETv4_ADDRESS_SIZE) & 0xFFFF);

		if (NodeAdptPort < INETv4OnePort.MIN_PORT || NodeAdptPort > INETv4OnePort.MAX_PORT)
			throw new IllegalArgumentException("Port value outside of range.");


		ScktAdptPort = (ByteArrayUtility.toShort(byteAddrPort, INETv4OnePort.ADDRESS_SIZE) & 0xFFFF);

		if (ScktAdptPort < INETv4OnePort.MIN_PORT || ScktAdptPort > INETv4OnePort.MAX_PORT)
			throw new IllegalArgumentException("The value of second port is out of range.");

	}

	/**
	 * Get different parts from a string which carries address and ports information.
	 * The address and ports string may have formats like "128.143.71.29:9800:9801", 
	 * "0.0.0.0:0:0". It also may be in the format of "9800", i.e., only consists of 
	 * a number. In this case, it is considered as the port number for Node Adapter. 
	 * An example of this case is that when we start a DT server, the port number is 
	 * specified in the commandline and is used to initialize the socket in the Node 
	 * Adapter.
	 * 
	 * @param	addrAndPortString	the string of address and ports.
	 */
	public static String[] getPartsFromAddrAndPortsString(String addrAndPortString)
	{

		/*
		 * String array paStr is used to return the result. paStr[0] 
		 * is the IP address, psStr[1] is the port number for node
		 * adapter and paStr[2] is the port number for socket adapter.
		 */
		String[] paStr = new String[3];
		for (int i=0; i<3; i++) {
			paStr[i] = "";
		}

		String remainedStr = addrAndPortString.trim();
		int comaLoc = remainedStr.indexOf (":");
		int index = 0;

		/*
		 * If address string only consists of a number. 
		 */
		if (comaLoc < 0)
		{
			paStr[1] = remainedStr;
		}
		/*
		 * Otherwise, the address and ports string is 
		 */
		else
		{
			while (comaLoc >=0)
			{
				//System.out.println ("remainedStr="+remainedStr+", comaLoc="+comaLoc);
				paStr[index]=remainedStr.substring (0, comaLoc);
				remainedStr = remainedStr.substring (comaLoc+1);
				comaLoc = remainedStr.indexOf (":");
				index++;
			
				if (index == paStr.length)
				{
					if (comaLoc < 0)
					{
						break;
					}
					else
					{
						throw new HyperCastWarningRuntimeException (addrAndPortString + " is a bad address string.");
					}
				}
			}

			if (index < paStr.length)
			{
				paStr[index]=remainedStr.substring (0);
			}
		}
		
		return paStr;
	}

	/**
	 * Get string representation of IP address in a InetAddress object.
	 * @param	addr	An InetAddress object. 
	 */
	private String getIPString(InetAddress addr) {
		if (addr == null) return new String("None");
		
		String addrstr = addr.toString();
		int slash_loc = addrstr.indexOf('/');
		
		//System.out.println("Address string is "+addrstr+", IP string is ***"+addrstr.substring(slash_loc+1, addrstr.length())+"***");
		
		return addrstr.substring(slash_loc+1, addrstr.length());
	}

	/** 
	 * Gets INETv4 address of this object.
	 */
	 public InetAddress getInetAddress() {
		 return Address;
	 }

	/**
	 * Return the first port number in ths physical address.
	 */
	public int getNodeAdptPort()
	{
		return NodeAdptPort;
	}


	/**
	 * Return the second port number in ths physical address.
	 */
	public int getScktAdptPort()
	{
		return ScktAdptPort;
	}


	/**
	 * Returns the byte array size
	 */
    public int getSize() {
	   return INETv4TwoPorts.ADDRESS_SIZE;
    }

	/** 
	 * Converts the Address/Port1/Port2 to a 8 bytes array. The format is
	 * |--------4-------|------2------|------2------|
	 * |       IP       |NodeAdpt port|ScktAdpt port|
	 * |----------------|-------------|-------------|
	 * @return	a 8 byte array where the first 4 bytes holds the address, 
	 *  the next 2 holds the Node Adapter port and the last 2 bytes holds
	 * the Socket Adapter port.
	 */	
	 public byte[] toByteArray() {
		byte[] byteAddressAndTwoPorts = new byte[INETv4TwoPorts.ADDRESS_SIZE];
		
		byte[] byteAddress = Address.getAddress();
		byte[] nodeAdptPort = ByteArrayUtility.toByteArray((short) NodeAdptPort);
		byte[] scktAdptPort = ByteArrayUtility.toByteArray((short) ScktAdptPort);

			
		System.arraycopy(byteAddress, 0, byteAddressAndTwoPorts, 0, 
						INETv4OnePort.INETv4_ADDRESS_SIZE);
		System.arraycopy(nodeAdptPort, 0, byteAddressAndTwoPorts, 
						INETv4OnePort.INETv4_ADDRESS_SIZE, 
						INETv4OnePort.PORT_NUMBER_SIZE);
		System.arraycopy(scktAdptPort, 0, byteAddressAndTwoPorts, 
						INETv4OnePort.INETv4_ADDRESS_SIZE+INETv4OnePort.PORT_NUMBER_SIZE, 
						INETv4OnePort.PORT_NUMBER_SIZE);

		 return byteAddressAndTwoPorts;
	 }


	/** 
	 * Test if two INETv4TwoPorts address are identical.
	 * 
	 * @param	obj an object (should be of type <code>INETv4TwoPorts</code>.)
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>INETv4TwoPorts</code>.
	 */
	public boolean equals(Object obj)
	{
		if (obj == null)
		{
			return false;
		}
		else if ((!(obj instanceof INETv4TwoPorts)) && (!(obj instanceof INETv4OnePort))) 
		{
			throw new IllegalArgumentException("The passed object is not of INETv4OnePort or type INETv4TwoPorts!");
		}
		/*
		 * If obj is an instance of INETv4TwoPorts, compare IP address and two ports.
		 */
		else if (obj instanceof INETv4TwoPorts) 
		{
			INETv4TwoPorts inetv4Obj = (INETv4TwoPorts) obj;

			 /*
			  * Using the method InetAddress.equals(InetAddress ia) to compare two
			  * InetAddress objects does not work.
			  */
			 byte[] remoteByteAddress = inetv4Obj.getInetAddress().getAddress();
			 byte[] localByteAddress = Address.getAddress();

			 return ((remoteByteAddress[0]==localByteAddress[0]) && (remoteByteAddress[1]==localByteAddress[1]) && 
					 (remoteByteAddress[2]==localByteAddress[2]) && (remoteByteAddress[3]==localByteAddress[3]) && 
					 (NodeAdptPort == inetv4Obj.getNodeAdptPort()) &&
					 (ScktAdptPort == inetv4Obj.getScktAdptPort()));
		}
		/*
		 * If obj is an instance of INETv4OnePort, (for example, obj is a underlay address),
		 * then only compare IP address and the first port.
		 */
		else
		{
			INETv4OnePort inetv4Obj = (INETv4OnePort) obj;

			 byte[] remoteByteAddress = inetv4Obj.getInetAddress().getAddress();
			 byte[] localByteAddress = Address.getAddress();

			 return ((remoteByteAddress[0]==localByteAddress[0]) && (remoteByteAddress[1]==localByteAddress[1]) && 
					 (remoteByteAddress[2]==localByteAddress[2]) && (remoteByteAddress[3]==localByteAddress[3]) && 
					 (NodeAdptPort == inetv4Obj.getPort()));
		}
	}



	/** 
	 * Test if the node adapter related part of two INETv4TwoPorts address are identical.
	 * 
	 * @param	obj an object (should be of type <code>INETv4TwoPorts</code>.)
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>INETv4TwoPorts</code>.
	 */
	public boolean addressCheck(Object obj)
	{
		if (obj == null)
		{
			return false;
		}
		else if ((!(obj instanceof INETv4TwoPorts)) && (!(obj instanceof INETv4OnePort))) 
		{
			throw new IllegalArgumentException("The passed object is not of INETv4OnePort or type INETv4TwoPorts!");
		}
		/*
		 * If obj is an instance of INETv4TwoPorts, compare IP address and two ports.
		 */
		else if (obj instanceof INETv4TwoPorts) 
		{
			INETv4TwoPorts inetv4Obj = (INETv4TwoPorts) obj;

			 /*
			  * Using the method InetAddress.equals(InetAddress ia) to compare two
			  * InetAddress objects does not work.
			  */
			 byte[] remoteByteAddress = inetv4Obj.getInetAddress().getAddress();
			 byte[] localByteAddress = Address.getAddress();

			 return ((remoteByteAddress[0]==localByteAddress[0]) && (remoteByteAddress[1]==localByteAddress[1]) && 
					 (remoteByteAddress[2]==localByteAddress[2]) && (remoteByteAddress[3]==localByteAddress[3]) && 
					 (NodeAdptPort == inetv4Obj.getNodeAdptPort()));
		}
		/*
		 * If obj is an instance of INETv4OnePort, (for example, obj is a underlay address),
		 * then only compare IP address and the first port.
		 */
		else
		{
			INETv4OnePort inetv4Obj = (INETv4OnePort) obj;

			 byte[] remoteByteAddress = inetv4Obj.getInetAddress().getAddress();
			 byte[] localByteAddress = Address.getAddress();

			 return ((remoteByteAddress[0]==localByteAddress[0]) && (remoteByteAddress[1]==localByteAddress[1]) && 
					 (remoteByteAddress[2]==localByteAddress[2]) && (remoteByteAddress[3]==localByteAddress[3]) && 
					 (NodeAdptPort == inetv4Obj.getPort()));
		}
	}
 

 

	/** 
	 * Compares two INETv4TwoPorts address objects. 
	 * 
	 * @param	addr	an object of type <code>I_Address</code>
	 * @return	true if (this > addr), otherwise false 
	 * @exception IllegalArgumentException if <code>obj</code> is not of type <code>INETv4TwoPorts</code>.
	 */
	 public boolean greaterThan(I_Address addr) {
		if (addr == null || !(addr instanceof INETv4TwoPorts)) 
			 throw new IllegalArgumentException("The passed I_Address instance is not of type INETv4TwoPorts!");
        
		INETv4TwoPorts remoteInetv4Addr = (INETv4TwoPorts) addr;

		byte[] localByteAddress = Address.getAddress();
		byte[] remoteByteAddress = remoteInetv4Addr.getInetAddress().getAddress();

		for (int i = 0; i < 4; i++) 
		{
			if (localByteAddress[i] == remoteByteAddress[i]) 
			   continue;
			else
			   return ((localByteAddress[i]&0xff) > (remoteByteAddress[i]&0xff));
		}

		if (NodeAdptPort == remoteInetv4Addr.getNodeAdptPort()) 
		   return (ScktAdptPort > remoteInetv4Addr.getScktAdptPort());
		else
		   return (NodeAdptPort > remoteInetv4Addr.getNodeAdptPort());
	 }
	
	
	/** 
	 * Converts the IP address object to string format "xxx.xxx.xxx.xxx/port1/port2" (e.g. "128.143.71.50/4747/5001")
	 */
	 public String toString() {
		byte[] byteAddress = Address.getAddress();
		return ((byteAddress[0]&0xff) + "." + (byteAddress[1]&0xff) +"."+
			(byteAddress[2]&0xff) + "." + (byteAddress[3]&0xff) + ":" + 
			 NodeAdptPort+":"+ScktAdptPort);
	 }

	 /**
	  * Returns a hashcode of this object
	  * It is important that the argument is of type <code>Object</code> so that 
	  * it can be used in a <code>Hashtable</code>.
	  */
	 public int hashCode() {
		 byte[] byteAddress = Address.getAddress();

		 int i = NodeAdptPort ^ ScktAdptPort;
		 i ^= (byteAddress[0] << 0);
		 i ^= (byteAddress[1] << 8);
		 i ^= (byteAddress[2] << 16);
		 i ^= (byteAddress[3] << 24);
		 return i;
	 }
}
