/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import hypercast.HyperCastConfig;
import hypercast.I_LogicalAddress;
import hypercast.HyperCastStatsException;
import hypercast.StatsProcessor;
import hypercast.StatsElement;
import hypercast.SimpleStats;

import hypercast.HyperCastConfigException;
import hypercast.util.XmlUtil;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xml.sax.SAXException;

import org.apache.xpath.XPath;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Arrays;

import java.util.Random;

/** 
 * This class maintains a list of ActiveStatsAccessors that have contacted it.
 * The status of each contact is kept in an PortalListEntry object.
 *
 * @see PortalListEntry
 * @see PortalAdvertiser
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January, 2005
 * @author Greg Mattes, January, 2005
 * @author Michael Nahas
 */
/* package */ class PortalListManager
    implements I_MCRaw_Callback
{
    private static final boolean debug = false;

    private static final Random random = new Random();

    private static final String PROP_NAME_TIME_ADVERTISE = "TimeAdvertise";
    private static final String PROP_NAME_PORTAL_TIMEOUT = "PortalTimeout";

    private static final String PROP_NAME_PORTAL_LIST    = "PortalList";
	
    /** The time interval, in milliseconds, between AdvertiseMonitor messages that are sent. */
    private long advertisePeriod; // ms

    /**
     * The time, in milliseconds, required to wait until an ActiveStatsAccessor is considered to be dead.
     * @see PortalListEntry#isActive 
     */
    private long portalTimeout; // ms

    /** Socket used to send and receive messages (e.g. send AdvertiseMonitor and receive AdvertisePortal messages). */
    private MCSocket mcSocket;
	
    /** Contains all of the portals, both dead and active. */
    private Vector portalList = new Vector(); // PortalListEntries

    /** Maps a Portal logical address to a Portal index integer. */
    private Hashtable portalLogicalAddressToIndex = new Hashtable(); // String -> Integer (String is I_LogicalAddress.toString)

    /** config object */
    private HyperCastConfig config;
	
    /**
     * Configuration prefix selector
     * @see Monitor#MONITOR_CONFIG_PREFIX
     */
    private String propertiesPrefix;
	
    /* 
     * In silent mode, Monitor does not announce itself; also portal
     * entries are not timed-out. // TODO: is this verified?
     */
    private boolean silentMode = false;

    /** Object/thread that periodically sends MonitorAdvertise messages to the Portal - only sends when not in silent mode. */
    private MonitorAdvertiser advertiser;

    /**
     * @param config HyperCast configuration object
     * @param mcSocket The monitor/control socket that will be used to monitor portals
     * @param propertiesPrefix TODO should this ever be anything other than Monitor?
     */
    /* package */ PortalListManager (final HyperCastConfig config, final MCSocket mcSocket, final String propertiesPrefix)
    {
        this.config = config;
        this.propertiesPrefix = propertiesPrefix;
		
		this.advertisePeriod = config.getLongAttribute
				(XmlUtil.createXPath(propertiesPrefix + PROP_NAME_TIME_ADVERTISE));

		this.portalTimeout = config.getLongAttribute
				(XmlUtil.createXPath(propertiesPrefix + PROP_NAME_PORTAL_TIMEOUT));

        this.mcSocket = mcSocket;
        
        this.advertiser = new MonitorAdvertiser();
        new Thread (this.advertiser, "Periodic AdvertiseMonitor Message Sender").start();

        clearPortalList();
    }

    /**
     * Remove all Portals, read configuration file for static Portals
     * and add them to the Portal list/table.
     */
    /* package */ synchronized void clearPortalList()
    {
        portalList.clear();
        portalLogicalAddressToIndex.clear();

        for (int i = 0; true; ++i)
        {
			String portalAddressString = null;
			try
			{
				portalAddressString = config.getTextAttribute
						(XmlUtil.createXPath(propertiesPrefix + PROP_NAME_PORTAL_LIST + "/PortalAddress["+(i+1)+"]"));
			}
			catch (HyperCastConfigException e)
			{
				/*
				 * The attribute doesn't exist.
				 */
				break;
			}

            if (portalAddressString == null)
            {
                break;
            }

            config.log.println ("NOTICE: PortalListManager adding <" + portalAddressString + "> into static portal list.");

            final I_LogicalAddress portalAddress = mcSocket.createLogicalAddress (portalAddressString);

            if (! portalAddressString.equals (portalAddress.toString()))
            {
                System.err.println ("*** PortalListManager.initPortalList portal address string do not match: \"" + portalAddressString + "\" \"" + portalAddress.toString() + "\" ***");
            }
			
            addPortal (portalAddress, true);
        }
    }
    /**
     * Compact the portal list
     *
     * Dangerous read the comment at this reference:
     *
     * @see I_Monitor#compactPortalList
     */
    public synchronized void compactPortalList(){
		int j=0;
		long now=System.currentTimeMillis();
		while(j<portalList.size()){
			PortalListEntry pe=(PortalListEntry)portalList.elementAt(j);
			if(!pe.isActive()){
				portalList.removeElementAt(j);
				continue;
			}
			j++;
		}		
	}
	
    /**
     * Adds a Portal to the list of known Portals.
     *
     * @param portalAddress Logical address of Portal.
     * @param isStatic Set to <code>true</code> if Portal comes from configuration file
     */
    private synchronized void addPortal (final I_LogicalAddress portalAddress, final boolean isStatic)
    {
        final int portalIndex = portalList.size();
        portalLogicalAddressToIndex.put (portalAddress.toString(), new Integer (portalIndex));
        final PortalListEntry portalListEntry = new PortalListEntry (portalAddress, portalIndex, isStatic);
        portalList.addElement (portalListEntry);
    }

    /* package */ int [] getActivePortalIndices()
    {
        final int [] activePortalIndices;
        int activePortalCount = 0;

        synchronized (this)
        {
            final int portalCount = portalList.size();
            activePortalIndices = new int [portalCount];

            final long now = System.currentTimeMillis(); // freeze time

            for (int i = 0; i < portalCount; ++i)
            {
                final PortalListEntry portalListEntry = (PortalListEntry) portalList.elementAt (i);
                if (portalListEntry.isActive (now))
                {
                    activePortalIndices [activePortalCount] = i;
                    ++activePortalCount;
                }
            }
        }

        final int [] result = new int [activePortalCount];
        System.arraycopy (activePortalIndices, 0, result, 0, activePortalCount);

        return result;
    }
    
    /* package */ int [] getInActivePortalIndices()
    {
        final int [] inActivePortalIndices;
        int inActivePortalCount = 0;

        synchronized (this)
        {
            final int portalCount = portalList.size();
            inActivePortalIndices = new int [portalCount];

            final long now = System.currentTimeMillis(); // freeze time

            for (int i = 0; i < portalCount; ++i)
            {
                final PortalListEntry portalListEntry = (PortalListEntry) portalList.elementAt (i);
                if (!portalListEntry.isActive (now))
                {
                	inActivePortalIndices [inActivePortalCount] = i;
                    ++inActivePortalCount;
                }
            }
        }

        final int [] result = new int [inActivePortalCount];
        System.arraycopy (inActivePortalIndices, 0, result, 0, inActivePortalCount);

        return result;
    }

    /* package */ synchronized int [] getAllPortalIndices()
    {
        final int portalCount = portalList.size();
        final int [] portalIndices = new int [portalCount];

        for (int i = 0; i < portalCount; ++i)
        {
            portalIndices[i] = i;
        }
        return portalIndices;
    }

    /**
     * Get a list of all of the ActiveStatsAccessors that have contacted this object.  
     * <b>Note: This includes both dead and active contacts.</b> 
     * Clones vector, since internal one can be modified by multiple threads.
     */
    public synchronized int getPortalCount()
    {
        return portalList.size();
    }

    /**
     * Retrieves the Portal having the specified index.
     *
     * @param portalIndex Index to use when doing lookup for Portal
     *
     * @throws NoSuchPortalException
     */
    public synchronized PortalListEntry getPortal (final int portalIndex)
        throws NoSuchPortalException
    {
        try
        {
            return (PortalListEntry) portalList.elementAt (portalIndex);
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            throw new NoSuchPortalException ("No Portal known with index: " + portalIndex, aioobe);
        }
    }

    /**
     * @return index of <code>Portal</code> with given logical address.
     *
     * @see #indexOfPortal(String)
     *
     * @throws NoSuchPortalException
     */
    /* package */ int indexOfPortal (final I_LogicalAddress logicalAddress)
        throws NoSuchPortalException
    {
        // TODO: there exists a map in this class from portal la to index - use it!
        return indexOfPortal (logicalAddress.toString());
    }

    /**
     * @return index of <code>Portal</code> with given logical address string.
     *
     * @throws NoSuchPortalException
     */
    /* package */ synchronized int indexOfPortal (final String logicalAddressString)
        throws NoSuchPortalException
    {
        final int PORTAL_NOT_FOUND = -1;
        int result = PORTAL_NOT_FOUND;
        for (int i = 0; i < portalList.size(); i++)
        {
            final PortalListEntry portalListEntry = (PortalListEntry) portalList.elementAt (i);

            final String portalAddressString = portalListEntry.getPortalAddress().toString();

            if (portalAddressString.equals (logicalAddressString)) // NOTE: Relying on String equality always...
            {
                result = i;
            }
	}

        if (result == PORTAL_NOT_FOUND)
            throw new NoSuchPortalException ("No Portal is known having logical address: " + logicalAddressString);

        return result;
    }

    /* package */ boolean getSilentMode()
    {
        return silentMode;
    }

    public synchronized void setSilentOn()
    {
        this.silentMode = true;
    }

    public synchronized void setSilentOff()
    {
        this.silentMode = false;
    }

    public void signalOpenMonitor()
    {
        advertiser.signalOpenMonitor();
    }

    public void signalCloseMonitor()
    {
        advertiser.signalCloseMonitor();
    }

    /** This class is responsible for sending out the periodic AdvertiseMonitor messages. */
    class MonitorAdvertiser
        implements Runnable
    {
        private static final boolean debug = false;

        /**
         * The <code>MonitorlAdvertiser</code> will send advertisement
         * messages only when this value is <code>true</code>.
         */
        private boolean sendAdvertisements = false;

        /** A cached copy of the AdvertiseMonitor message that gets sent to every active ActiveStatsAccessor. */
        private MCAdvertiseMessage advertiseMonitorMessage;
	
        public MonitorAdvertiser ()
        {
            this.advertiseMonitorMessage = mcSocket.createAdvertiseMessage (MCRawMessage.ADVERTISE_MONITOR);
        }

        /*
         * This is called from the Monitor when the Monitor.openMonitor() method is
         * invoked. It signals the advertiser thread to begin advertising.
         *
         * @see Monitor#openMonitor()
         */
        public synchronized void signalOpenMonitor()
        {
            if (MonitorAdvertiser.debug)
                config.err.println ("Monitor Advertiser: open monitor signaled");

            sendAdvertisements = true;
            notifyAll();
        }

        /*
         * This is called from the Monitor when the Monitor.closeMonitor() method is
         * invoked. It signals the advertiser thread to stop advertising.
         *
         * @see Monitor#closeMonitor()
         */
        public synchronized void signalCloseMonitor()
        {
            if (MonitorAdvertiser.debug)
                config.err.println ("Monitor Advertiser: close monitor signaled");

            sendAdvertisements = false;
            notifyAll();
        }

        /**
         * @see PortalListManager#receiveMCRawMessage
         */
        public synchronized void run()
        {
            while (true)
            {
                // If this thread is running but the socket has not
                // started there is nothing that this thread can do!

                while (! sendAdvertisements
                       ||
                       ! mcSocket.isOpen())
                {
                    if (PortalListManager.debug)
                        config.err.println ("Portal List Manager: waiting for monitor to be opened");

                    try { wait(); } catch (final InterruptedException ignored) { }

                    // A delay for some random amount of time (uniform
                    // between (0,advertisePeriod)) to stagger
                    // advertisements of many portals that are started
                    // simultaneously.

                    final long delayForStaggering = Math.abs(random.nextLong()) % advertisePeriod;

                    // Wait at least few seconds to make sure that the
                    // adapters have started. When the sematics of
                    // OL_Socket.joinOverlay have been changed so that the
                    // caller can be assured that the adapters are
                    // prepared to receive calls to sendUnicastMessage,
                    // this extra delay can go away.

                    final long delayForAdapterStart = 2000L;

                    final long delay = Math.max (delayForAdapterStart, delayForStaggering);

                    try { wait (delay); } catch (final InterruptedException ignored) { }
                }

                /*
                 * While in Silent Mode, no AdvertiseMonitor is sent
                 */
                if (! silentMode)
                {
                    if (PortalListManager.debug)
                        System.err.println ("PortalListManager: Broadcasting AdvertiseMonitor");

                    // TODO: "note, not added to transcript" <-- not clear what this comment means
                    mcSocket.sendToAll (advertiseMonitorMessage); // non-blocking
                }
                else
                {
                    /*
                     * In silent mode, no advertisement message
                     * is sent by portal (sic should be Monitor
                     * not Portal, right? TODO) but none of the
                     * entries will be timed-out.  So we need to
                     * touch each active entry periodically.
                     *
                     * I think that this means that we have to
                     * freshen the portal entries implicitly when
                     * no monitor advertisements go out. There
                     * needs to be a better description of
                     * "silent mode" to understand this.
                     */

                    // TODO: should it be enforced that time advertise << portal timeout?

                    synchronized (PortalListManager.this)
                    {
                        for (int i = 0; i < portalList.size(); i++)
                        {
                            final PortalListEntry portalListEntry = (PortalListEntry) portalList.elementAt (i);
                            if (portalListEntry.isActive())
                            {
                                portalListEntry.updateMostRecentContactTime();
                            }
                        }
                    }
                }

                try { wait (advertisePeriod); } catch (final InterruptedException ignored) { }
            }
        }
    }

    /**
     * @see hypercast.MonitorAndControl.I_ReceiveCallback#receiveMessage(hypercast.MonitorAndControl.MonMessage)
     */
    public void receiveMCRawMessage (final MCRawMessage message)
    {
        if (MCSocket.debug)
            System.err.println ("PortalListManager: Received MCRawMessage: " + message.getType());

        switch (message.getTypeNum())
        {
            case MCRawMessage.ADVERTISE_PORTAL:
                handleAdvertisePortal (message);
                break;

            case MCRawMessage.ADVERTISE_MONITOR:
                // We're hearing our own broadcast - that's ok; ignore.
                break;

            default:
            config.log.println ("PortalListManager.receiveMCRawMessage(): not an Advertise message; ignoring: " + message.getType());
        }
    }

    /**
     * @param advertisePortalMessageRaw Message that is assumed to be of type AdvertisePortal, verified by caller
     * @see #receiveMCRawMessage(MCRawMessage)
     */
    private void handleAdvertisePortal (final MCRawMessage advertisePortalMessageRaw)
    {
        final MCAdvertiseMessage advertisePortalMessage = (MCAdvertiseMessage) advertisePortalMessageRaw;

        /*
         * When in silent mode, there should not be any AdvertisePortal message
         *
         * TODO: what if Monitor is in Silent Mode but Portals are not?
         */
        if (silentMode)
        {
            config.log.println ("PortalListManager received \"AdvertisePortal\" message while in Silent Mode.");
            // TODO: OK to continue with this method after this point?
        }

        final I_LogicalAddress portalAddress = advertisePortalMessage.getSender();

        if (portalAddress == null)
        {
            config.log.println ("WARNING: Received AdvertisePortal message with bad Sender Address: \"" + portalAddress  + "\"");
            return;
        }

        if (PortalListManager.debug)
            System.err.println ("PortalListManager: Received AdvertisePortal from Portal: " + portalAddress);
        
        /*
         * If Portal is contained in the list of known Portals, update the time at which it was last heard.
         * If Portal has not been seen previously, add it to the list of known Portals.
         */
        synchronized (this)
        {
            final Integer portalIndex = (Integer) portalLogicalAddressToIndex.get (portalAddress.toString());
            if (portalIndex != null)
            {
                PortalListEntry portalListEntry = (PortalListEntry) portalList.elementAt (portalIndex.intValue());
                final long timeSinceMostRecentContact = portalListEntry.getTimeSinceMostRecentContact();

                if (PortalListManager.debug)
                    config.err.println ("Reacquired Portal at: " + portalAddress + " (last contact was " + (timeSinceMostRecentContact / 1000L) + " seconds ago)");

                portalListEntry.updateMostRecentContactTime();
             
            }
            else
            {
                if (PortalListManager.debug)
                    config.log.println ("Acquired new Portal at: " + portalAddress);

                addPortal (portalAddress, false);
            }
        }
    }

    /** 
     * Holds the information about one Portal including the time at
     * which it was last heard from as well as whether or not it was
     * specified in the configuration file. This is the in-memory data
     * structure that represents remote Portals.
     *
     * @see PortalListManager
     * @see PortalAdvertiser
     * 
     * @author Michael Nahas
     * @version	2.0, Feb. 28, 2001
     */
    /* package */ class PortalListEntry
        //implements Comparable
    {
        private int portalIndex;

        /** Data socket used by StatsAccessor. */
        private I_LogicalAddress portalAddress;

        /** The last time a message was received from the ActiveStatsAccessor. */
        private long mostRecentContactTime;

        

        /** Some Portals can be specified in the configuration file, in this case they are considered to be static */
        private boolean isStatic;

        /**
         * It records the time of creation.  
         */
        public PortalListEntry (final I_LogicalAddress portalAddress, final int portalIndex, final boolean isStatic)
        {
            this.portalIndex = portalIndex;
            this.isStatic = isStatic;
            this.portalAddress = portalAddress;
            this.mostRecentContactTime = isStatic ? 0 : System.currentTimeMillis();
        }

        public long getMostRecentContactTime()
        {
            return mostRecentContactTime;
        }

        public int getPortalIndex()
        {
            return portalIndex;
        }

        /** Get the address of the StatsAccessors's MCSocket */
        public I_LogicalAddress getPortalAddress()
        {
            return portalAddress;
        }

        /**
         * @see #isActive(long)
         */
        public boolean isActive() 
        {
            final long now = System.currentTimeMillis();
            return isActive (now);
        }
    
        /**
         * @param currentTime The current time in milliseconds.
         *
         * TODO: Just because a Portal appears in the config file we *assume* that it is active??? maybe until timeout
         */
        public boolean isActive (final long currentTime)
        {
            final long timeSinceMostRecentContact = getTimeSinceMostRecentContact (currentTime);
            final boolean timeoutOccured = timeSinceMostRecentContact > portalTimeout;

            if (PortalListManager.debug
                &&
                timeoutOccured)
            {
                final String eventString = isStatic ? "Ignoring static" : "Detected non-static";
                System.err.println ("PortalListManager: " + eventString+ " Portal timeout for Portal at: " + portalAddress +
                                    "; most recent Portal advertisement was " + (timeSinceMostRecentContact / 1000L) +
                                    " seconds ago; Portal timeout is: " + (portalTimeout / 1000L) + " seconds.");
            }

            return isStatic ? true : ! timeoutOccured;
        }

        /**
         * Updates the time field when an AdvertisePortal message is received for
         * this entry from a neighboring <code>ActiveStatsAccessor</code>. TODO: is this verified?
         */
        public void updateMostRecentContactTime() 
        {
            mostRecentContactTime = System.currentTimeMillis();
        }
        
        /**
         * Returns the amount of time (in milliseconds) since the most
         * recent AdvertisePortal message was received. A return value
         * of zero means that this Portal was configured statically
         * (in the configuration file) but no AdvertisePortal message
         * has ever been received from it.
         */
        public long getTimeSinceMostRecentContact()
        {
            return getTimeSinceMostRecentContact (System.currentTimeMillis());
        }

        public long getTimeSinceMostRecentContact (final long currentTime)
        {
            return (mostRecentContactTime == 0) ? 0 : currentTime - mostRecentContactTime;
        }

        public String toString()
        {
            final long timeSinceMostRecentContact = getTimeSinceMostRecentContact();

            return
                "<Portal - Index: " + portalIndex + ", " +
                "Address: " + portalAddress + ", " +
                (isStatic ? "    " : "Not ") + "Static, " +
                (isActive() ? "    " : "Not ") + "Active, " +
                "Contact: " + (timeSinceMostRecentContact / 1000L) + "s ago>";
        }
    }
}
