/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import hypercast.util.XmlUtil;

/**
 * A stats entry for a neghbor
 *
 * @author HyperCast Team
 * @author Guangyu Dong
 */

public class NeighborhoodStats implements I_Stats
{
	
	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	protected StatsProcessor statsPro;
	
	/**
	 * Address pair of this neighbor
	 */
	public I_AddressPair addressPair;
	
	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	protected String statisticsName = null;

	
	/**
	 * Constructor.
	 */
	public NeighborhoodStats(I_AddressPair ap)
	{
		addressPair=ap;
		
		InitStatisticsStructure();
	}

	/**
	 * Constructor.
	 */
	public NeighborhoodStats(I_AddressPair ap, String statsName)
	{
		addressPair=ap;
		
		statisticsName = statsName;
		
		InitStatisticsStructure();
	}
	
	/**
	 * Initialize the statistics processor instance contained in this object.
	 */
	protected void InitStatisticsStructure()
	{

		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, false/* writable flag */);
		
		statsPro.addStatsElement("LogicalAddress"/* name */, new LogicalAddressOperator()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("PhysicalAddress"/* name */, new PhysicalAddressOperator()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);
   }


   /**
    * Set the value of this entry by another entry
    * @param entry
    */
   public void set(NeighborhoodStats entry){
   		this.addressPair=entry.addressPair;
   }

   /**
	* Check if the given object (a address piar instance) matchs the
	* addres pair in this object.
	* 
	* @param obj	an instance of I_AddressPair.
	*/
   public boolean equals(Object obj)
   {
   		return addressPair.getLogicalAddress().equals(((I_AddressPair) obj).getLogicalAddress());
   }


	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getStatsResult(doc, xpath);
	}
	
	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	
	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	
	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}

  /** @see I_Stats#getStatsName() */
  public String getStatsName ()
  {
	  return statisticsName;
  }

  /** @see I_Stats#setStatsName(String) */
  public void setStatsName (String name)
  {
	  statisticsName = name;
  }


  /**
   * This class manages simple statistic "LogicalAddress".
   */
  class LogicalAddressOperator extends SimpleStats
  {
	  /**
	   * Gets the value of a simple statistic managed by the implementation of this class.
	   * 
	   * @see SimpleStats#getStats()
	   */
	  protected String getStats()
	  {
		   if(addressPair==null)
			   return null;
		  	  
		   I_LogicalAddress LA = addressPair.getLogicalAddress();
		   if (LA == null)
		   {
			   return null;
		   }
		   else
		   {
			   return LA.toString();
		   }
	  }
	
	  /**
	   * Sets the value of a simple statistic managed by the implementation of this class.
	   * 
	   * @see SimpleStats#setStats(String newvalue)
	   */
	  protected String setStats(String newValue) throws HyperCastStatsException
	  {
	   throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
	  }

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
  }

  /**
   * This class manages simple statistic "PhysicalAddress".
   */
  class PhysicalAddressOperator extends SimpleStats
  {
	  /**
	   * Gets the value of a simple statistic managed by the implementation of this class.
	   * 
	   * @see SimpleStats#getStats()
	   */
	  protected String getStats()
	  {
		   if(addressPair==null)
			   return null;
		  	  
		   I_PhysicalAddress PA = addressPair.getPhysicalAddress();
		   if (PA == null)
		   {
			   return null;
		   }
		   else
		   {
			   return PA.toString();
		   }
	  }
	
	  /**
	   * Sets the value of a simple statistic managed by the implementation of this class.
	   * 
	   * @see SimpleStats#setStats(String newvalue)
	   */
	  protected String setStats(String newValue) throws HyperCastStatsException
	  {
		  throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
	  }

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
  }
	
}
