/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.Random;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.xpath.XPath;

import hypercast.util.XmlUtil;

import java.security.cert.Certificate;
import java.security.PrivateKey;

/** 
 * This class implements all the functions of an overlay socket. It contains
 * an overlay node that maintains the overlay topology. A forward engine is
 * runnig inside this class for sending and receiving application messages.
 * This class also provides interfaces for applications to handling data
 * sending and receiving.
 *
 * @author HyperCast Team
 * @author Dongwen Wang
 * @author Jianping Wang
 * @author Guimin Zhang
 * @author Guangyu Dong
 * @author Greg Mattes
 * @version 3.0, October, 2004 
 */
public class OL_Socket
	implements I_OverlaySocket, I_AdapterCallback, I_ReceiveCallback
{
    /** Conditional compilation flag used to include or exclude debugging output. */
    private static final boolean debug = false;

    /**
     * Conditional compilation flag used to include or exclude
     * debugging output specifically regarding the initialization of
     * the overlay socket.
     */
    private final boolean debugInitializationMessages = false;

    /** Overly property object for configuration */
    protected HyperCastConfig config;

    /** Overlay Node */
    protected I_Node node;

    /** Overlay Socket Adapter */
    protected I_UnicastAdapter socketAdapter; 
    
    /** Key vault */
    private KeyVault keyvaultObj = null;

    /** Overlay receive buffer */
    protected MessageBuffer receiveBuffer; 

    /** messagestore for formatted message and stream */
    /* package */ MessageStore messagestore = null; 

    /* A random number generator used to create message and stream ids */
    private Random random = new Random (System.currentTimeMillis());

    /**
     * A notification handler for an overlay socket is initialized
     * when the socket is created. The reference is given to the node
     * of the socket as well as the message store (if the message
     * store is configured). The reference is also needed in the
     * stream manager for the socket. Since the stream manager is
     * instanciated on-demand (not at construction time), the
     * reference to the notification handler is kept in the socket so
     * that the socket can pass it to the stream manager.
     */
    private NotificationHandler notificationHandler = null;
 
    /** Default overlay message hop limit */
    protected short hopLimit;

    /** Overlay Call Back interface */
    protected I_ReceiveCallback callback = null;

    /**
     * <code>true</code> after a call to <code>joinOverlay()</code>,
     * until a call to <code>leaveOverlay()</code>.
     *
     * <code>false</code> after a call to <code>leaveOverlay()</code>,
     * until a call to <code>joinOverlay()</code>.
     */
    private boolean running = false;

    /**
     * Interception Callback interface
     *
     * If interceptionCallback is non-null, all unicast messages that
     * would be forwarded by this overlay socket are first passed to
     * the the interceptionCallback.ReceiveCallback method. The
     * ReceiveCallback method may return a message or null. If it
     * returns a message, that message is forwarded.
     */
    protected I_InterceptionCallback interceptionCallback = null;
    
    /**
     * The naming proxy object for this socket; all naming APIs will be
     * handled by this proxy. This object is not the naming cache
     * itself - that is a message store stream based finite state
     * machine.
     *
     * @see Naming
     * @see MessageStoreFSM_Naming
     */
    private Naming naming = null;

    /**
     *  Maximum size of route record header
     */
    private int routeRecordSize;

    /*
     * Singleton stream manager object
     * @see getStreamManager()
     */
    private StreamManager streamManager = null;

    /**
     * A StatsProcessor instance which handles the statistics in this object.
     */
    private StatsProcessor statsProcessor;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;

    /**
     * Create an overlay socket with a HyperCastConfig object and a Callback 
     * interface.
     *
     * @param config HyperCastConfig object which is used for configuration.
     * @param callback I_ReceiveCallback interface for the application.
     *
     * @see #OL_Socket(HyperCastConfig, I_ReceiveCallback, NotificationHandler, I_InterceptionCallback)
     */
    public OL_Socket (final HyperCastConfig config, final I_ReceiveCallback callback)
    {
        this (config, callback, null, null);
    }

    /**
     * Create an overlay socket with a HyperCastConfig object, a callback
     * object and a notification handler.
     * @param config HyperCastConfig object which is used for configuration.
     * @param callback I_ReceiveCallback interface for the application.
     * @param notificationHandler a notification handler.
     *
     * @see #OL_Socket(HyperCastConfig, I_ReceiveCallback, NotificationHandler, I_InterceptionCallback)
     */
    public OL_Socket (final HyperCastConfig config,
                      final I_ReceiveCallback      callback,
                      final NotificationHandler    notificationHandler)
    {
        this (config, callback, notificationHandler, null);
    }

    /**
     * Create an overlay socket with a HyperCastConfig object, a callback
     * object and a notification handler.
     * @param config HyperCastConfig object which is used for configuration.
     * @param callback I_ReceiveCallback interface for the application.
     * @param notificationHandler_parameter a notification handler.
     * @param interceptionCallback a interception object.
     */
    public OL_Socket (final HyperCastConfig config,
                      final I_ReceiveCallback      callback,
                      final NotificationHandler    notificationHandler_parameter,
                      final I_InterceptionCallback interceptionCallback)
    {
        /*
         * Validate and store parameters (notificationHandler is set
         * for appropriate subcomponents as they are created).
         */
        if (config == null)
            throw new IllegalArgumentException ("HyperCastConfig was null when constructing OL_Socket.");

        this.config = config;

        if (debugInitializationMessages)
            this.config.log.println ("Creating Overlay Socket...");

        this.callback = (callback == null) ? this : callback;
        this.interceptionCallback = interceptionCallback;
        
        this.notificationHandler = (notificationHandler_parameter == null) ? new NotificationHandler() : notificationHandler_parameter;

        if (debugInitializationMessages)
            this.config.log.println ("Reading Socket Configuration...");

        /*
         * Read relevant configuration properties.
         */
        final short configHopLimit = (short) config.getPositiveIntAttribute (XmlUtil.createXPath("/Public/HopLimit"));

        setHopLimit (configHopLimit);
        
        final int receiveBufferSize = config.getPositiveIntAttribute (XmlUtil.createXPath("/Public/ReceiveBuffer/ReceiveBufferSize"));

        final int socketTimeout = config.getNonNegativeIntAttribute (XmlUtil.createXPath("/Public/ReadTimeout"));

        this.receiveBuffer = new MessageBuffer (config, receiveBufferSize, socketTimeout);

        this.routeRecordSize = config.getNonNegativeIntAttribute (XmlUtil.createXPath("/Public/RouteRecordSize"));

        /*
         * Instantiate message store.
         */
        if (debugInitializationMessages)
            this.config.log.println ("Creating Message Store...");

        this.messagestore = new MessageStore (this, config);
        this.messagestore.setNotificationHandler (notificationHandler);

        final AdapterFactory adapterFactory = new AdapterFactory (debugInitializationMessages);

        /*
         * Instantiate required subcomponents.
         */
        if (debugInitializationMessages)
            this.config.log.println ("Creating Socket Adapter...");

        this.socketAdapter = adapterFactory.createAdapter (config, null, "SocketAdapter", null);
        this.socketAdapter.setCallback (this);

        if (debugInitializationMessages)
            this.config.log.println ("Creating Node Adapter...");

        I_UnicastAdapter nodeAdapter = adapterFactory.createAdapter (config, socketAdapter, "NodeAdapter", null);
        /*
         * If security policy is not plaintext, i.e., integrity or privacy, the node adapter passed
         * to the node is a <code>SecurityProcessor</code> instance. In this case, the adapter created 
         * above is passed to the constructor of the class <code>SecurityProcessor</code> and kept by the 
         * created <code>SecurityProcessor</code> instance. 
         * 
         * See class <code>SecurityProcessor</code> for more information about the relationship between 
         * the overlay node, security processor and normal node adapter.
         */	
		if (! config.getTextAttribute (XmlUtil.createXPath("/Public/Security/SecurityLevel")).equals ("plaintext"))
        {
            keyvaultObj = new KeyVault (config);

            /*
             * Create a security processor with the keyvaultObj and nAdapt. It will be passed to overlay
             * node as the node adapter.
             */
            nodeAdapter = (I_UnicastAdapter) new SecurityProcessor (config, keyvaultObj, nodeAdapter);
            if(notificationHandler==null){
            	notificationHandler = (SecurityProcessor)nodeAdapter;
            }else{
            	notificationHandler.addNotificationHandler((SecurityProcessor)nodeAdapter);
            }
        }

        if (debugInitializationMessages)
            this.config.log.println ("Creating Node...");

        final NodeFactory nodeFactory = new NodeFactory();
        this.node = nodeFactory.create_newNode (config, nodeAdapter);
        this.node.setNotificationHandler (notificationHandler);

        /*
         * Naming requires the presence of the Message Store and the
         * Node. This means that the naming object must be
         * instantiated after the node is instantiated, and only if
         * the message store is instantiated.
         */

        final String activateNamingString = config.getTextAttribute (XmlUtil.createXPath ("/Public/NetworkService/Naming/StartService"));
        final boolean activateNaming = config.booleanFromYesNoString (activateNamingString);
        if (activateNaming)
            this.naming = new Naming (this);

        initStatisticsStructure();
    }

    /*
     * Each overlay socket may have at most one StreamManager so a singleton pattern is used.
     * @return singleton stream manager object, or null if message store is not configured.
     */
    public synchronized StreamManager getStreamManager()
    {
        if (messagestore == null)
            throw new HyperCastWarningRuntimeException ("Stream manager can only be created with a message store");

        if (streamManager == null)
        {
            if (debug)
                System.out.println ("create streamManager...");

            streamManager = new StreamManager (this);
            MessageStore.setFSMCallback (MessageStore.FSM_BEST_EFFORT_ORDERING, streamManager);
        }

        return streamManager;
    }

    /**
     * TODO: document me!
     */
    public void ReceiveCallback (final I_OverlayMessage message)
    {
        try 
        {
            receiveBuffer.blockingWrite ((OL_Message) message); // blind cast
        }
        catch (InterruptedException ignored)
        {
        }
    }   
    

    /**
     * <p>Implements the operation of joining an overlay network. The
     * call is asynchronous in the sense that upon return to the
     * caller, the socket has *not necessarily* completed the join
     * operation of the particular overlay protocol (or protocols in
     * the multi-node case NB multi-node is not implemented as of Fall
     * 2004). If an application is interested in knowing when the join
     * operation is complete it must use the notification mechanism to
     * receive such an event.</p>
     *
     * <p>This call is responsible for starting all threads needed by
     * the socket that were not started at socket construction
     * time. These threads should be stopped in reverse order by
     * <code>leaveOverlay()</code>.</p>
     *
     * @see #leaveOverlay
     */
    public synchronized void joinOverlay()
    {
        if (running == false)
        {
            node.joinOverlay();

            socketAdapter.Start();
            socketAdapter.startTimer();

            if (messagestore != null)
            {
                messagestore.start();
            }

            if (naming != null) {
                naming.joinOverlay();
            }

            running = true;
        }
    }

    /**
     * Leaves a multicast group, stopping all threads started by
     * <code>joinOverlay()</code> in reverse order.
     *
     * @see #joinOverlay
     */
    public synchronized void leaveOverlay()
    {
        if (running == true)
        {
            running = false;

            if (naming != null) {
                naming.leaveOverlay();
            }

            if (messagestore != null)
            {
                messagestore.stop();
            }

            socketAdapter.stopTimer();
            socketAdapter.Stop();

            node.leaveOverlay();
        }
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.setName()</code>. Delegates call to an
     * instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#setName(String)
     */
    public void setName (final String name)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        naming.setName (name);
    }

    public void setName (final String name, final I_LogicalAddress logicalAddress)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        naming.setName (name, logicalAddress);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.setName()</code>. Delegates call to an
     * instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#setName(Certificate, PrivateKey)
     */
    public void setName (final Certificate certificate, final PrivateKey privateKey)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        naming.setName (certificate, privateKey);
    }

    public void setName (final Certificate certificate, final PrivateKey privateKey, final String name, final I_LogicalAddress logicalAddress)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        naming.setName (certificate, privateKey, name, logicalAddress);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.unsetName()</code>. Delegates call to an
     * instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#unsetName(String)
     */
    public void unsetName (final String name)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        naming.unsetName (name);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.unsetName()</code>. Delegates call to an
     * instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#unsetName(Certificate, PrivateKey)
     */
    public void unsetName (final Certificate certificate, final PrivateKey privateKey)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        naming.unsetName (certificate, privateKey);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getName()</code>. Delegates call to an
     * instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getNames()
     */
    public String [] getNames()
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getNames();
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getName(I_LogicalAddress)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getNames(I_LogicalAddress)
     */
    public String [] getNames (final I_LogicalAddress logicalAddress)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getNames (logicalAddress);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getName(I_LogicalAddress,boolean,long,boolean,boolean,int,int,short)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getNames(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     */
    public String [] getNames (final I_LogicalAddress peer,
                               final boolean          requireTrusted,
                               final long             timeout,
                               final boolean          requireSerialNumberMatch,
                               final boolean          requireAuthoritative,
                               final int              maxAge,
                               final int              maxResponses,
                               final short            maxHopCount,
                               final short            minHopCount)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getNames (peer, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getNameNonBlocking(I_LogicalAddress)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getNamesNonBlocking(I_LogicalAddress)
     */
    public int getNamesNonBlocking (final I_LogicalAddress address)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getNamesNonBlocking (address);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getNameNonBlocking(I_LogicalAddress)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getNamesNonBlocking(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getNamesNonBlocking (final I_LogicalAddress peer,
                                    final boolean          requireTrusted,
                                    final long             timeout,
                                    final boolean          requireSerialNumberMatch,
                                    final boolean          requireAuthoritative,
                                    final int              maxAge,
                                    final int              maxResponses,
                                    final short            minHopCount,
                                    final short            maxHopCount)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getNamesNonBlocking (peer, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getLogicalAddress(String)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getLogicalAddressByName(String)
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String name)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getLogicalAddressByName (name);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getLogicalAddress(String)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getLogicalAddressByName(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String  name,
                                                        final boolean requireTrusted,
                                                        final long    timeout,
                                                        final boolean requireSerialNumberMatch,
                                                        final boolean requireAuthoritative,
                                                        final int     maxAge,
                                                        final int     maxResponses,
                                                        final short   minHopCount,
                                                        final short   maxHopCount)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getLogicalAddressByName (name, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getLogicalAddressNonBlocking(String)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getLogicalAddressByNameNonBlocking(String)
     */
    public int getLogicalAddressByNameNonBlocking (final String name)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getLogicalAddressByNameNonBlocking (name);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.getLogicalAddressNonBlocking(String)</code>. Delegates
     * call to an instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#getLogicalAddressByNameNonBlocking(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getLogicalAddressByNameNonBlocking (final String name,
                                                   final boolean requireTrusted,
                                                   final long    timeout,
                                                   final boolean requireSerialNumberMatch,
                                                   final boolean requireAuthoritative,
                                                   final int     maxAge,
                                                   final int     maxResponses,
                                                   final short   minHopCount,
                                                   final short   maxHopCount)
    {
        if (naming == null)
            throw new UnsupportedOperationException ("Naming is not configured");

        return naming.getLogicalAddressByNameNonBlocking (name, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
    }

    /**
     * <p>Implementation of
     * <code>I_OverlaySocket.installTrustedNamingCertificate()</code>. Delegates call to an
     * instance of <code>Naming</code>.</p>
     *
     * @throws UnsupportedOperationException if naming is not configured
     * @see Naming
     * @see I_OverlaySocket#installTrustedNamingCertificate(Certificate)
     */
    public void installTrustedNamingCertificate (final Certificate certificate)
    {
        naming.installTrustedNamingCertificate (certificate);
    }

    /**
     * @return the notification handler
     */
    protected NotificationHandler getNotificationHandler()
    {
        return notificationHandler;
    }

  
    /**
     * @deprecated The length parameter is redundant.
     */
    public I_OverlayMessage createMessage (final byte [] payload, final int length)
    {
        return createMessage (payload);
    }

    /**
     * @deprecated The length parameter is redundant.
     */
    public I_OverlayMessage createMessage (final byte [] payload, final int length, final int servicenumber)
    {
        return createMessage (payload, (short) servicenumber);
    }

    /**
     * @deprecated The length parameter is redundant.
     */
    public I_OverlayMessage createMessage (final byte [] payload, final int length, final byte [] streamid)
    {
        return createMessage (payload, streamid);
    }

    /**
     * Same as calling <code>createMessage(byte [])</code> with parameter of <code>null</code>.
     *
     * @see #createMessage(byte [])
     */
    public I_OverlayMessage createMessage()
    {
        final I_AddressPair myAddressPair = node.getAddressPair();
        final I_LogicalAddress myLogicalAddress = myAddressPair.getLogicalAddress();

        final OL_Message message = new OL_Message (this.hopLimit, myLogicalAddress, keyvaultObj);

        return message;
    }

    /**
     * Creates an overlay message with an application payload
     * extension.
     * 
     * @param payload The payload of the message.
     */
    public I_OverlayMessage createMessage (final byte [] payload)
    {
        final OL_Message message = (OL_Message) createMessage();

        if (payload != null && payload.length > 0)
        {
            final Extension extension = Extension.createExtension (Extension.RAW_DATA, payload, null, keyvaultObj, false);
            message.addExtension (extension);
        }

		if (debug)
			System.out.println("create normal OL_Message.");

        return message;
    }

    /**
     * Create an OL_Message containing an extension that contains an
     * appliction payload extension and a message store extension.
     *
     * @param payload The data of the application program
     * @param serviceNumber The service (FSM) that will handle the application data message
     *
     * @throws HyperCastWarningRuntimeException if message store is not active or service type is not supported
     */
    public I_OverlayMessage createMessage (final byte [] payload, final short serviceNumber)
    {
        if (messagestore == null)
            throw new HyperCastWarningRuntimeException ("Cannot create this type of message without a message store");
            
        if (! MessageStore.isValidServiceNumber (serviceNumber))
            throw new HyperCastWarningRuntimeException ("Service type " + serviceNumber + " is not supported in Hypercast");

        final OL_Message message = (OL_Message) createMessage (payload);

        if (debug)
            System.out.println("Create a message with service ID "+serviceNumber);

        final FSM_Extension extension = messagestore.getMessageServiceExtension (serviceNumber);
        message.addExtension (extension);

        return message;
    }  

    /**
     * Creates a message with specified stream id.
     *
     * @return a new overlay message with an FSM extension and a payload extension. This message will be processed by a message store.
     *
     * @param payload the payload of the new message, it will be stored in a payload extension
     * @param streamid the stream id that identifies the service that the new message will be associated with (for (de)multiplexing), this value must be 4 bytes in length
     *
     * @throws HyperCastWarningRuntimeException if there is no message store configured, this is not a configuration exception because the programmer choose to create this message despite the configuration which may be correct or incorrect.
     */
    public I_OverlayMessage createMessage (final byte [] payload, final byte [] streamid)
    {
        if (messagestore == null)
            throw new HyperCastWarningRuntimeException ("Cannot create this type of message without a message store.");

        if (streamid == null
            ||
            streamid.length != 4)
        {
            throw new HyperCastWarningRuntimeException ("Stream ID must be 4 bytes in length: " + streamid);
        }

        final OL_Message message = (OL_Message) createMessage (payload);
        final FSM_Extension extension = messagestore.getStreamServiceExtension (streamid);
        message.addExtension (extension);

        return message;
    }

    /**
     * Creates an I_LogicalAddress from a bytes array
     */
    public I_LogicalAddress createLogicalAddress (final byte [] encodedLogicalAddress, final int offset)
    {
        return node.createLogicalAddress (encodedLogicalAddress, offset);
    }

    /** Creates a logical address object from a String.
     */
    public I_LogicalAddress createLogicalAddress (final String logicalAddressString)
    {
        return node.createLogicalAddress (logicalAddressString);
    }

    /**
     * Sets the logical address to specified one.
     */
    public void setLogicalAddress (final I_LogicalAddress logicalAddress)
    {
        node.setLogicalAddress (logicalAddress);
    }

    /**
     * Returns the node's logical address
     */
    public I_LogicalAddress getLogicalAddress()
    {
        final I_AddressPair addressPair = node.getAddressPair();
        return addressPair.getLogicalAddress();
    }

    /**
     * Passes the message to the application via the callback. This
     * serves as the central point to pass messages to the
     * application, subclasses can implement this method to do
     * subclass specific operations, e.g. decrypt/encrypted payload
     * before delivering to the application.
     */
    public void appmsgArrived (final I_OverlayMessage message, final I_ReceiveCallback callback)
    {
        callback.ReceiveCallback ((OL_Message) message); // blind cast
    }
    
    /**
     * Returns the unique identifier for this node. It is actually the node physical
     * address byte array.  
     */
    public byte [] getUniqueIdentifier()
    {
        final I_AddressPair addressPair = node.getAddressPair();
        final byte [] uniqueid = addressPair.getPhysicalAddress().toByteArray();

        return uniqueid;
    }

    /**
     * Returns a message identifier. 
     */
    public short generateMsgIdentifier()
    {
        return (short) random.nextInt();
    }

    /**
     * Returns a stream identifier. 
     */
    public short generateStreamIdentifier()
    {
        return (short) random.nextInt();
    }

    /**
     * Returns this node's parent with respect to a tree starting at "root".
     */
    public I_LogicalAddress getParent (final I_LogicalAddress root)
    {
        final I_AddressPair [] parent = node.getParent (root);

        // Node may have no parent
        return (parent == null || parent.length != 1) ? null : parent[0].getLogicalAddress();
    }

    /**
     * TODO: document me!
     */
    private I_LogicalAddress [] extractLogicalAddressArray (final I_AddressPair [] addressPairs)
    {
        I_LogicalAddress [] logicalAddresses = null;
        
        if (addressPairs != null)
        {
            logicalAddresses = new I_LogicalAddress [addressPairs.length];
            for (int i = 0; i < addressPairs.length; i++)
            {
                logicalAddresses[i] = addressPairs[i].getLogicalAddress();
            }
        }
        return logicalAddresses;
    }

    /**
     * Returns an array consisting of the logical addresses of the
     * children of this node.
     */
    public I_LogicalAddress [] getChildren (final I_LogicalAddress root)
    {
        final I_AddressPair [] addressPairs = node.getChildren (root);
        return extractLogicalAddressArray (addressPairs);
    }

    /**
     * Returns an array consisting of the logical addresses of this node's neighbors.
     */
    public I_LogicalAddress [] getNeighbors()
    {
        final I_AddressPair [] addressPairs = node.getAllNeighbors();
        return extractLogicalAddressArray (addressPairs);
    }


    /**
     * Extracts a route record from an overlay message. If no route record is found,
     * a new empty route record is created and returned.
     *
     * TODO: should most likely be a method of the OL_Message class
     */
    private RouteRecord getRouteRecord (final OL_Message message)
    {
        final RouteRecord routeRecord = (RouteRecord) message.getFirstExtensionByType (Extension.ROUTE_RECORD);
        return (routeRecord != null) ? routeRecord : new RouteRecord();
    }

    /**
     * Updates the route record with local logical address, possibly
     * creating a new route record if non existed.
     *
     * TODO: should most likely be a method of the OL_Message class
     */
    private void updateRouteRecord (final OL_Message message)
    {
        final RouteRecord routeRecord = getRouteRecord (message);
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        routeRecord.insertLogicalAddress (myLogicalAddress, routeRecordSize);

        /*
         * Writes the updated (or new) route record back to message.
         */
        message.setFirstExtensionByType (Extension.ROUTE_RECORD, routeRecord);
    }

    /**
     * <p>This method implements the heart of the "Forwarding Engine."
     * The bulk of the remaining portions of the forwarding engine can
     * be found in the methods with the prefix "forward"</p>
     *
     * <p>Validates the received message, delivers it to either an
     * application or a buffering store (e.g. message store), and
     * decides whether or not to forward it to other nodes depending
     * on the delivery mechanism of the message and the configuration
     * of any buffering stores. This method is called from the socket
     * adapter thread.</p>
     *
     * <p>This method is executed for its side effects only; there is
     * no return value. Before calling one of the "forward" methods,
     * this method can return, effectively dropping the message</p>.
     *
     * @param msg The message that has arrived from the adapter.
     *
     * @throws HyperCastWarningRuntimeException
     * @throws HyperCastFatalRuntimeException
     */
    public void messageArrivedFromAdapter (final I_Message msg)
    {
        if (! (msg instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("Passed object of type other than OL_Message!");

        OL_Message message = (OL_Message) msg;

        if (debug)
            System.out.println("OL_Socket: messageArrivedFromAdapter: Received message from "+((OL_Message) msg).getSourceAddress().toString());

        String msgInfo =
            "OL_Socket: messageArrivedFromAdapter: The received message: " +
            "src = " + message.getSourceAddress() + ", " +
            "dst = " + message.getDestinationAddress() + ", " +
            "prevhop = " + message.getPreviousHopAddress();
		
        if (debug)
            System.out.println (msgInfo);
        
        /*
         * The route record that is attached to a message records the
         * logical addresses of nodes that have previously forwarded
         * the message. If the current node appears in this route
         * record it means that the message can be dropped since it
         * has already been processed by this node.
         */
        final I_AddressPair addressPair = node.getAddressPair();
        final I_LogicalAddress myLogicalAddress = addressPair.getLogicalAddress();
        
        final RouteRecord routeRecord = getRouteRecord (message);
        final boolean haveSeenThisMessageBefore = routeRecord.contains (myLogicalAddress);

        if (haveSeenThisMessageBefore)
        {
            String warningMessage =
                "Dropping message, route record indicates that this node has already seen this message: " +
                "hop limit = " + (message.getHopLimit() + 1) + ", " +
                "src = " + message.getSourceAddress();
            config.log.println (warningMessage);
            return; // drop message
        }

        /*
         * The hop limit field is interpreted to mean the maximum number of
         * hops that a message is permitted to make in the overlay before it is
         * dropped. This can also be described as the maximum number of times
         * that a message is permitted to be forwarded. It is useful to think
         * in terms of forwarding rather than dropping because there is no code
         * that explicitly drops a message due to the value of its hop limit
         * field, rather the hop limit field is used to decide if the
         * message can be forwarded.
         *
         * The hop limit field of a message is decremented here because when
         * a message arrives from the adapter the message is considered to have
         * completed one hop.
         *
         * A check is done to ensure that the contents of the hop limit
         * field are valid.
         */
        short hopLimit = 0; // initialize to mean "cannot forward," avoids compiler error
        try
        {
            hopLimit = message.decrementHopLimit();
        }
        catch (IllegalArgumentException iae)
        {
            String warningMessage =
                "Dropping message due to invalid hop limit: " +
                "hop limit = " + (message.getHopLimit() + 1) + ", " +
                "src = " + message.getSourceAddress();
            config.log.println (warningMessage);
            return; // drop message
        }

        /*
         * After the hop limit field has been decremented a decision can be
         * made about whether or not it is possible to forward this
         * message. Note that simply because this node *can* forward this
         * message based on the hop limit field does not mean that the message
         * *will* be forwarded, e.g. perhaps the message is destined for this
         * node so it is delivered locally, the message store will assume the
         * responsibility of forwarding, or the interception callback will
         * cancel forwarding.
         */
        boolean isPossibleToForwardThisMessage = hopLimit != 0;

        /*
         * Validates the delivery mode of incoming messages. In
         * HyperCast the delivery mode field is a four bit
         * quantity that allows for sixteen unique settings, although
         * on four are defined. Only three of the four defined
         * settings are used, the fourth is called "invalid." No
         * message passed to this method should ever use invalid. The
         * invalid setting used for new messages that have not been
         * assigned a valid delivery mode (i.e. a message has this
         * valid after construction but before being passed to a
         * "send" method). The detecting a message that has this field
         * set to unknown is exceptional.
         *
         * In addition to having the delivery mode set to invalid
         * explicitly, all unused values of delivery mode are also
         * exceptional.
         */
        byte deliveryMode = message.getDeliveryMode();
        switch (deliveryMode)
        {
            case OL_Message.DELIVERY_MODE_MULTICAST: /* fall-through */
            case OL_Message.DELIVERY_MODE_FLOOD:     /* fall-through */
            case OL_Message.DELIVERY_MODE_UNICAST:
                /* valid modes: OK */
                break;

            case OL_Message.DELIVERY_MODE_INVALID:
            default:
            {
                String warningMessage =
                    "Dropping message with unknown delivery mode: " +
                    "src = " + message.getSourceAddress() + ", " +
                    "dst = " + "null" + ", " +
                    "prevhop = " + message.getPreviousHopAddress();
                throw new HyperCastWarningRuntimeException (warningMessage);
            }
        }

        /*
         * Decide if the sending node of the previous hop is a valid
         * sender. The message is dropped if the sender is not valid.
         * 
         * This is useful because some protocols (e.g. the Pastry distributed
         * hash table (DHT) protocol) do not have the ability to make a local
         * decision about which neighbors are children in an arbitrary multicast
         * tree, however, they do have the ability to decide if a sending node
         * is their parent.
         *
         * When doing multicast, these protocols must send messages to all
         * neighbors assuming that all neighbors are children. When a
         * neighboring node receives a message it calls
         * <code>prevhopCheck</code> to decide if the sending node is its
         * parent.
         *
         * TODO: it is not clear if we should do this check on flood.
         */
        if (deliveryMode != OL_Message.DELIVERY_MODE_FLOOD)
        {
            final I_LogicalAddress sourceAddress = message.getSourceAddress();
            final I_LogicalAddress destinationAddress = message.getDestinationAddress();
            final I_LogicalAddress previousHopAddress = message.getPreviousHopAddress();
            if (! node.previousHopCheck (sourceAddress, destinationAddress, previousHopAddress))
            {
                if (debug)
                    System.out.println("The delivery mode of droped message is "+deliveryMode);

                String logMessage =
                    "Dropping message from invalid sender: " +
                    "src = " + message.getSourceAddress() + ", " +
                    "dst = " + message.getDestinationAddress() + ", " +
                    "prevhop = " + message.getPreviousHopAddress();
                config.log.println (logMessage);
                return; // drop message
            }
        }

        /*
         * NB: Some messages do not have application data, but do have message
         * store extensions. Such messages are message store control messages.
         */
        final boolean messageHasApplicationData = message.getFirstExtensionByType (Extension.RAW_DATA) != null;
        final boolean messageHasMessageStoreExtension = hasMessageStoreExtension (message);

        /*
         * All messages must have at least one extension other than a route
         * record.
         */
        if (! messageHasApplicationData
            &&
            ! messageHasMessageStoreExtension)
        {
            String logMessage =
                "Dropping message, it does not have any extension (other than possibly a route record): " +
                "src = " + message.getSourceAddress() + ", " +
                "dst = " + message.getDestinationAddress() + ", " +
                "prevhop = " + message.getPreviousHopAddress();
            config.log.println (logMessage);
            return; // drop message
        }

        /*
         * At this point in the method, the message is considered to be
         * COMPLETELY VALIDATED.
         */

        /*
         * The application may have registered an "interception
         * callback" because it is interested in seeing, and possibly
         * modifying, all messages that are received at the local
         * node. The interception callback has the ability to decide
         * whether or not a message should be processed further. If it
         * decides that a message should be processed further, it will
         * return the message, otherwise it will return null. Note
         * that all messages are passed to the interception handler,
         * regardless of whether the messages are destined for this
         * node or this node is simply an intermediate node.
         */
        if (interceptionCallback != null)
        {
            final OL_Message result = interceptionCallback.InterceptionCallback (message);
            if (result == null)
            {
                final String logMessage =
                    "Dropping message, interception callback signaled a drop: " +
                    "src = " + message.getSourceAddress() + ", " +
                    "dst = " + message.getDestinationAddress() + ", " +
                    "prevhop = " + message.getPreviousHopAddress();
                config.log.println (logMessage);
                return; // drop message;
            }
            else
            {
                message = result;
            }
        }

        /*
         * If there is a message store active, the message is received
         * by it rather than by an application. The message store is
         * always responsible for passing the message to an
         * application. The message store is *sometimes* responsible
         * for forwarding messages depending on which state machine is
         * being used. Since the forwarding engine cannot know which
         * state machine is being used, the message store informs the
         * forwarding engine whether or not it will forward a message
         * by the return value of the receiveMessage() method. If the
         * return value is true, that means that the message store
         * will be responsible for forwarding the message.
         *
         * NB: At this point in the method, no decision has been made
         * regarding whether the message will be received by the local
         * node. This means that the message store will receive all
         * messages, even those that are not destined for this
         * node. For example, if this node is an intermediate node on
         * the path from the source to the destination of a unicast
         * message. In such a case, it is the responsibility of the
         * message store to compare the destination logical address
         * and the local logical address to determine if it should
         * process the message.
         *
         * The value of the boolean variable
         * messageReceivedByMessageStore indicates whether the message
         * store has received the message.
         */

        boolean messageReceivedByMessageStore = false;
        if (messageHasMessageStoreExtension)
        {
            if (debug)
                System.out.println("OL_Socket:MessageArrivedFromAdapter: call messagestore.receiveMessage (message).");

			/*
			 * Create a copy of the original message and send the copy to MessageStore.
			 */
			OL_Message clonedMsg = new OL_Message(message);

            boolean messageStoreWillDoForwarding = messagestore.receiveMessage (clonedMsg);
            
            messageReceivedByMessageStore = true;
            if (messageStoreWillDoForwarding)
            {
                isPossibleToForwardThisMessage = false;
            }
        }

        /*
         * If the message should be received by the application, the
         * flag applicationShouldReceiveMessage is set to true and the
         * message will be received after any forwarding occurs *only*
         * if the message store has not already received the
         * message. The reason for setting a flag rather than simply
         * sending the message directly to the application is that the
         * receive code is identical in all three delivery mode cases;
         * it should not be duplicated. The value of this flag must be
         * initialized to something, either true or false, false is
         * chosen; this means that in order for a message to be
         * received, the flag must be changed explicitly.
         */
        boolean applicationShouldReceiveMessage = false; // avoids compiler warning

        switch (deliveryMode)
        {
            /*
             * A multicast message continues to propagate through the overlay
             * by being forwarded to children in a multicast tree defined by
             * the overlay topology. Each node in the multicast tree
             * unconditionally receives the message.
             */
            case OL_Message.DELIVERY_MODE_MULTICAST:
            {
                if (isPossibleToForwardThisMessage)
                {
                    if (debug)
                        System.out.println("OL_Socket:messageArrivedFromAdapter: forward to children.");

                    forwardToChildren (message);
                }

                applicationShouldReceiveMessage = ! messageReceivedByMessageStore;
            }
            break;

            /*
             * A unicast message continues to be routed through the
             * network until one of two things occurs:
             *
             * (1) Its destination logical address matches the logical
             * address of the current node, in which case the message
             * is received by the local node.
             *
             * (2) It has exceeded its maximum hop count, or
             * "hop limit" value, before reaching its
             * destination. This means that it can no longer be
             * forwarded to other nodes in an effort to reach its
             * destination - the message must be dropped.
             *
             * Thus, if it has not reached its destination and it has
             * not exceeded its maximum hop count, it will be
             * forwarded to another node.
             */
            case OL_Message.DELIVERY_MODE_UNICAST:
            {
                I_LogicalAddress destinationLogicalAddress = message.getDestinationAddress();
                boolean messageHasReachedDestination = myLogicalAddress.equals (destinationLogicalAddress);
                if (! messageHasReachedDestination
                    &&
                    isPossibleToForwardThisMessage)
                {
                    if (debug)
                        System.out.println("OL_Socket:messageArrivedFromAdapter: forward to parent.");

                    forwardToParent (message);
                }
                else if (messageHasReachedDestination)
                {
                    applicationShouldReceiveMessage = ! messageReceivedByMessageStore;
                }
                else
                {
                    // message is dropped silently; exceeded hop limit
                }
            }
            break;

            /*
             * A message that is in the overlay with flood mode
             * delivery has no destination or route, it is simply
             * forwarded to all neighboring nodes (save the node that
             * sent it) at each hop until its hop limit is
             * exceeded. Messages are received unconditionally at each
             * hop in a flood.
             *
             * NB: If the message store is active, it is responsible for
             * continuing to forward messages in the flood.
             */
            case OL_Message.DELIVERY_MODE_FLOOD:
            {
                if (isPossibleToForwardThisMessage)
                {
                    if (debug)
                        System.out.println("OL_Socket:messageArrivedFromAdapter: forward flood.");

                    forwardFlood (message);
                }

                applicationShouldReceiveMessage = ! messageReceivedByMessageStore;
            }
            break;

            case OL_Message.DELIVERY_MODE_INVALID: /* should never happen due to previous validation */
                throw new HyperCastFatalRuntimeException ("Previous check for unknown delivery mode failed");

            default:  /* should never happen due to previous validation */
                throw new HyperCastFatalRuntimeException ("Message with undefined delivery mode detected: " + deliveryMode);
        }

        if (applicationShouldReceiveMessage)
        {
            if (debug)
                System.out.println("OL_Socket:messageArrivedFromAdapter: pass the received message to application.");

            appmsgArrived (message, callback);
        }
    }

    /**
     * Sends a multicast overlay message to all nodes in the multicast
     * group from this socket; the sender receives a copy as well. The
     * message may traverse multiple hops.
     *
     * @param message The overlay application message that is sent.
     * 
     * @throws HyperCastFatalRuntimeException when the message is not an OL_Message
     */
    public void sendToAll (final I_OverlayMessage message)
    {
        if (! (message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("Passed object of type other than OL_Message.");

        final OL_Message applicationMessage = (OL_Message) message;
        applicationMessage.setDeliveryMode (OL_Message.DELIVERY_MODE_MULTICAST);

        // Logical address could have changed since message creation.
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        applicationMessage.setSourceAddress (myLogicalAddress);
        applicationMessage.setPreviousHopAddress (myLogicalAddress);

        forwardToChildren (applicationMessage);
        final boolean deliveredToMessageStore = checkForMessageStoreSend (applicationMessage);

        if (debug)
            System.out.println("OL_Socket: sendToAll: deliveredToMessageStore is "+deliveredToMessageStore);

        /*
         * With this type of send, the message should be delivered to
         * the local application as well. The only exception to this
         * is when a message store is active. In that case, the
         * message store is responsible for delivering the message to
         * the local application.
         */
        if (! deliveredToMessageStore)
        {
            appmsgArrived (applicationMessage, callback);
        }
    }

    /**
     * Sends message to all the children nodes from this socket. 
     * (Send message to children using Unicast. It is single hop)
     *
     * @param message the overlay application message to be sent out
     * @param root the logical address of the data source root
     *
     * @throws HyperCastFatalRuntimeException
     */
    public void sendToChildren (final I_OverlayMessage message, final I_LogicalAddress root)
    {
        if (!(message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendToChildren: Passed object of type other than OL_Message.");

        final OL_Message appmsg = (OL_Message) message;
        appmsg.setDeliveryMode (OL_Message.DELIVERY_MODE_UNICAST);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();

        // Logical address could have changed since message creation.
        appmsg.setSourceAddress (myLogicalAddress);
        appmsg.setPreviousHopAddress (myLogicalAddress);

        appmsg.setHopLimit ((short) 1);

        final I_AddressPair [] children = node.getChildren(root);
        if (children == null)
            throw new HyperCastFatalRuntimeException ("Children array is null");

        for (int i = 0; i < children.length; i++)
        {
            final I_AddressPair child = children[i];

            if (child == null) // HC protocol can return a null child entry
                continue;

            final I_PhysicalAddress destinationPhysicalAddress = child.getPhysicalAddress();
            if (destinationPhysicalAddress == null)
                throw new HyperCastFatalRuntimeException ("Physical Address cannot be null in a child entry");

            final I_LogicalAddress destinationLogicalAddress = child.getLogicalAddress();
            if (destinationLogicalAddress == null)
                throw new HyperCastFatalRuntimeException ("destination Logical Address must never be null");
            appmsg.setDestinationAddress (destinationLogicalAddress);

            socketAdapter.sendUnicastMessage (destinationPhysicalAddress, appmsg);
        }

        checkForMessageStoreSend (appmsg);
    }

    /**
     * Sends message to all the neighbors. Single hop, Multicast
     *
     * @param message the overlay application message to be sent out
     * 
     * @throws HyperCastFatalRuntimeException
     */
    public void sendToAllNeighbors (final I_OverlayMessage message)
    {
        if (! (message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendToAllNeighbors: Passed object of type other than OL_Message.");

        final OL_Message appmsg = (OL_Message) message;

        appmsg.setDeliveryMode (OL_Message.DELIVERY_MODE_MULTICAST);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();

        // Logical address could have changed since message creation.
        appmsg.setSourceAddress (myLogicalAddress);
        appmsg.setPreviousHopAddress (myLogicalAddress);

        appmsg.setHopLimit ((short) 1);

        final I_AddressPair [] neighbors = node.getAllNeighbors();
        if (neighbors == null)
            throw new HyperCastFatalRuntimeException ("Neighbor array is null");

        for (int i = 0; i < neighbors.length; i++)
        {
            final I_AddressPair neighbor = neighbors[i];

            if (neighbor == null) // HC protocol can return a null neighbor entry
                continue;

            final I_PhysicalAddress destinationPhysicalAddress = neighbor.getPhysicalAddress();
            if (destinationPhysicalAddress == null)
                throw new HyperCastFatalRuntimeException ("Physical Address cannot be null in a neighbor entry");

            socketAdapter.sendUnicastMessage (destinationPhysicalAddress, appmsg);
        }

        checkForMessageStoreSend (appmsg);
    }

    /**
     * Sends a message to parent. Single hop
     *
     * @param message the overlay application message to be sent out
     * @param root logical address of the data source root
     *
     * @throws HyperCastFatalRuntimeException
     */
    public void sendToParent (final I_OverlayMessage message, final I_LogicalAddress root)
    {
        if (! (message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendToParent: Passed object of type other than OL_Message.");

        final OL_Message appmsg = (OL_Message) message;
        appmsg.setDeliveryMode (OL_Message.DELIVERY_MODE_UNICAST);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();

        // Logical address could have changed since message creation.
        appmsg.setSourceAddress (myLogicalAddress);
        appmsg.setPreviousHopAddress (myLogicalAddress);

        appmsg.setHopLimit ((short)1);

        final I_AddressPair parents [] = node.getParent (root);
        if (parents == null)
            throw new HyperCastFatalRuntimeException ("Parent array is null");

        if (parents.length == 1)
        {
            I_AddressPair parent = parents[0];
            I_LogicalAddress parentLogicalAddress = parent.getLogicalAddress();
            appmsg.setDestinationAddress (parentLogicalAddress);
        
            I_PhysicalAddress parentPhysicalAddress = parent.getPhysicalAddress();
            if (parentPhysicalAddress == null)
                throw new HyperCastFatalRuntimeException ("Physical Address cannot be null in a parent entry");

            socketAdapter.sendUnicastMessage (parentPhysicalAddress, appmsg);
            checkForMessageStoreSend (appmsg);
        }
        else
        {
            // TODO: What to do if many or zero parents exist?
            // should this be an error? warning? nothing? no parent? what does this mean?
            // there should be a comment/documentation explaining the reasoning for the semantics.
        }
    }

    /**
     * Sends a message to a specified node. Multi hops.
     * @param message the overlay application message to be sent out
     * @param destinationLogicalAddress the logical address of destination node 
     *
     * @throws HyperCastFatalRuntimeException
     */
    public void sendToNode (final I_OverlayMessage message, final I_LogicalAddress destinationLogicalAddress)
    {
        if (! (message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendToNode: Passed object of type other than OL_Message.");

        final OL_Message appmsg = (OL_Message) message;

        appmsg.setDeliveryMode (OL_Message.DELIVERY_MODE_UNICAST);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();

        // Logical address could have changed since message creation.
        appmsg.setSourceAddress (myLogicalAddress);
        appmsg.setPreviousHopAddress (myLogicalAddress);
        appmsg.setDestinationAddress (destinationLogicalAddress);

        if (debug)
        {
            String infoMessage =
                "OL_Socket: sendToNode: send unicast message:" +
                "src = " + appmsg.getSourceAddress() + ", " +
                "dst = " + appmsg.getDestinationAddress() + ", " +
                "prevhop = " + appmsg.getPreviousHopAddress();
	
            System.out.println(infoMessage);
        }

        forwardToParent (appmsg);
        checkForMessageStoreSend (appmsg);
    }

    /**
     * Floods a message to all nodes. 
     *
     * @param message the overlay application message to be sent out
     *
     * @throws HyperCastFatalRuntimeException
     */
    public void sendFlood (I_OverlayMessage message)
    {
        if (! (message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendFlood: Passed object of type other than OL_Message.");

        final OL_Message appmsg = (OL_Message) message;

        appmsg.setDeliveryMode (OL_Message.DELIVERY_MODE_FLOOD);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();

        // Logical address could have changed since message creation.
        appmsg.setSourceAddress (myLogicalAddress);
        appmsg.setPreviousHopAddress(myLogicalAddress);

        forwardFlood (appmsg);
        checkForMessageStoreSend (appmsg);
    }

    /**
     * Send Incast message to root node. 
     *
     * @param message the overlay application message to be sent out
     *
     * @throws HyperCastFatalRuntimeException
     */
    public void sendIncast (final I_OverlayMessage message,
                            final byte []          messageid,
                            final I_LogicalAddress root)
    {
        if (messagestore == null)
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendIncast: the overlay socket is not configured to support formatted message ");

        if (! (message instanceof OL_Message))
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendIncast: Passed object of type other than OL_Message.");

        final OL_Message appmsg = (OL_Message) message;

        if (appmsg.getServiceIdentifier()!= MessageStore.FSM_INCAST)
            throw new HyperCastFatalRuntimeException ("OL_Socket.sendIncast: The service number has to be set as INCAST to use this method.");

        appmsg.setDeliveryMode (OL_Message.DELIVERY_MODE_UNICAST);
 
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
 
        // Logical address could have changed since message creation.
        appmsg.setSourceAddress (myLogicalAddress);
        appmsg.setPreviousHopAddress (myLogicalAddress);
 
        appmsg.setHopLimit ((short) 1);

        appmsg.setDestinationAddress (root);
        checkForMessageStoreSend (appmsg);
    }

    /**
     * TODO: document me!
     */
    protected boolean hasMessageStoreExtension (final OL_Message appmsg)
    {
        return
            messagestore != null
            &&
            appmsg.getFirstExtensionByType (Extension.FSM) != null;
    }

    /**
     * TODO: document me!
     */
    private boolean checkForMessageStoreSend (final OL_Message appmsg)
    {
        boolean deliveredToMessageStore = false;
        if ((hasMessageStoreExtension (appmsg)) && (appmsg.getFirstExtensionByType (Extension.RAW_DATA) != null))
        {
            if (debug)
                System.out.println("OL_Socket: checkForMessageStoreSend: call messagestore.sendMessage (appmsg).");

            messagestore.sendMessage (appmsg);
            deliveredToMessageStore = true;
        }
        return deliveredToMessageStore;
    }

    /**
     * Receives an overlay application message from this socket. 
     * This call will block until there a message returns.
     * @return overlay application message
     *
     * This method should be called *ONLY* when the user has NOT specified a callback to the constructor
     */
    public I_OverlayMessage receive()
    {
        if (callback != this)
            throw new HyperCastFatalRuntimeException ("Called receive() in OL_Socket when a user Callback had been specified.");

        try 
        {
            return (OL_Message) receiveBuffer.read();
        }
        catch (InterruptedException ie)
        {
            throw new HyperCastWarningRuntimeException ("OverlaySocket.receive() was interrupted.", ie);
        }
    }

    /**
     * Enable/disable SO_TIMEOUT with the specified timeout, in milliseconds. 
     * With this option set to a non-zero timeout, a call to receive() 
     * for this OL_Socket will block for only this amount of time. 
     * If the timeout expires, a java.io.InterruptedIOException is raised, 
     * though the OL_Socket is still valid. 
     * The option must be enabled prior to entering the blocking operation 
     * to have effect. The timeout must be > 0. 
     * A timeout of zero is interpreted as an infinite timeout.
     * @param timeout the specified timeout in milliseconds.
     */
    public void setSoTimeout (final int timeout)
    {
        try
        {
            receiveBuffer.setSoTimeout (timeout);
        }
        catch (IllegalArgumentException iae)
        {
            throw new HyperCastWarningRuntimeException (iae);
        }
    }

    /**
     * Retrive setting for SO_TIMEOUT. 
     * 0 returns implies that the option is disabled (i.e., timeout of infinity).
     * @return the setting for SO_TIMEOUT
     */
    public int getSoTimeout()
    {
        return receiveBuffer.getSoTimeout();
    }

    /**
     * Set the default hop limit for multicast packets sent out on
     * this socket.  The hop limit sets the IP hop limit for
     * DatagramPackets sent to a MulticastGroup, which specifies how
     * many "hops" that the packet will be forwarded on the network
     * before it expires.  The hop limit is an unsigned 16-bit quantity
     */
    public void setHopLimit (final short hopLimit)
    {
        if (hopLimit == 0)
        {
            IllegalArgumentException iae = new IllegalArgumentException ("Default hop limit cannot be zero - no messages would ever be transmitted!");
            throw new HyperCastWarningRuntimeException (iae);
        }
        this.hopLimit = hopLimit;
    }

    /**
     * Get the default hop limit for multicast packets sent out on the OL_Socket. 
     *
     * @return the default hop limit value
     */
    public short getHopLimit()
    {
        return hopLimit;
    }

    /**
     *  This method is called from the timer thread of the Socket
     *  adapter. This timer is used by the message store.
     */
    public void timerExpired (final Object time_id)
    {
        if (! (time_id instanceof Socket_Timer_ID))
            throw new HyperCastFatalRuntimeException ("Timer Id not supported at the socket adapter!");

        final Socket_Timer_ID id = (Socket_Timer_ID) time_id;
         
        switch (id.getType())
        {
            case Socket_Timer_ID.typeOfMessageStore:
            {
                if (messagestore==null) throw new HyperCastFatalRuntimeException ("The OL_Socket is not configured to use Message Store!");
                messagestore.timerExpired (id.getEvent());
            }
            break;
            
            case Socket_Timer_ID.typeOfStreamManager:
            {
                if (streamManager==null) throw new HyperCastFatalRuntimeException ("The OL_Socket is not configured to use Stream Manager!");
                streamManager.timerExpired (id.getEvent());
            }
            break;

            default:
                throw new HyperCastWarningRuntimeException ("Undefined timer processor");
        }
    }

    /**
     * Creates an I_Message from a byte array.
     *
     * @param recvbuff          byte array of data
     * @param validBytesStart   the offset from where to create the message
     * @param validBytesEnd     the total valid bytes length in this array
     *
     * @return an I_Message object or null if the array contains no complete I_Message
     */
    public I_Message restoreMessage (final byte [] recvbuff, final int [] validBytesStart, final int validBytesEnd)
    {
        return OL_Message.restoreMessage (recvbuff, validBytesStart, validBytesEnd, node, keyvaultObj);
    }


    /**
     * Sends the message to all children of this node.<p>
     * Root address is included in appmsg header.
     *
     * @param appmsg OL_Message object
     *
     * @throws HyperCastFatalRuntimeException If the parameter of the
     * method cannot be validated as a unicast message or it any method
     * call returns a null reference.
     */
    protected void forwardToChildren (final OL_Message appmsg)
    {
        if (appmsg.getDeliveryMode() != OL_Message.DELIVERY_MODE_MULTICAST)
            throw new HyperCastFatalRuntimeException ("Delivery mode must be MULTICAST, not: " + appmsg.getDeliveryMode());

        final I_LogicalAddress root = appmsg.getSourceAddress();

        if (root == null)
            throw new HyperCastFatalRuntimeException ("source Logical Address must never be null");

        final RouteRecord routeRecord = getRouteRecord (appmsg);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        appmsg.setPreviousHopAddress (myLogicalAddress);

        final I_AddressPair [] children = node.getChildren (root);

        /*
         * TODO: In the system how do you say no children? what should
         * getChildren return if there are no chlidren? null or zero
         * length list?
         */
        if (children != null && children.length > 0)
        {
            for (int i = 0; i < children.length; i++)
            {
                final I_AddressPair child = children[i];

                if (child == null) // HC protocol can return a null child entry
                    continue;

                final I_PhysicalAddress destinationPhysicalAddress = child.getPhysicalAddress();
                if (destinationPhysicalAddress == null)
                    throw new HyperCastFatalRuntimeException ("Physical Address cannot be null in a children entry");

                final I_LogicalAddress destinationLogicalAddress = child.getLogicalAddress();
                if (destinationLogicalAddress == null)
                    throw new HyperCastFatalRuntimeException ("destination Logical Address must never be null");

                //don't send message to previous hops
                if (! routeRecord.contains (destinationLogicalAddress))
                {
                    updateRouteRecord (appmsg);

                    if (debug)
                        System.out.println("send Unicast message to "+child.toString()+", next hop is "+destinationPhysicalAddress.toString());

                    socketAdapter.sendUnicastMessage (destinationPhysicalAddress, appmsg);
                }
            }
        }
        else
        {
            //added by Guangyu to support Spanning tree routing
            //to be completed;
            //I_AddressPair[] neighbors=node.getAllNeighbors(root);
            I_AddressPair[] neighbors=node.getAllNeighbors();
            //I_LogicalAddress prev = appmsg.getSenderAddress();
        }
    }

    /**
     * Forwards the message to the node's parents in the spanning tree used for routing. <p>
     * Root address is included in the message header.
     *
     * @param appmsg an OL_Message object
     *
     * @throws HyperCastFatalRuntimeException If the parameter of the
     * method cannot be validated as a unicast message or it any method
     * call returns a null reference.
     */
    protected void forwardToParent (final OL_Message appmsg)
    {
        if (appmsg.getDeliveryMode() != OL_Message.DELIVERY_MODE_UNICAST)
            throw new HyperCastFatalRuntimeException ("Delivery mode must be UNICAST, not: " + appmsg.getDeliveryMode());

        final I_LogicalAddress root = appmsg.getDestinationAddress();

        if (root == null)
            throw new HyperCastFatalRuntimeException ("OL_Socket: forwardToParent() all UNICAST messages MUST have a destination!");

        final RouteRecord routeRecord = getRouteRecord (appmsg);
        
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        appmsg.setPreviousHopAddress (myLogicalAddress);

        final I_AddressPair [] parents = node.getParent (root);

        if (parents == null)
            throw new HyperCastFatalRuntimeException ("Parent array is null");

        for (int i = 0; i < parents.length; i++)
        {
            final I_AddressPair parent = parents[i];

            if (parent == null) // HC protocol can return a null parent entry
                continue;

            final I_PhysicalAddress destinationPhysicalAddress = parent.getPhysicalAddress();                    
            if (parent.getPhysicalAddress() == null)
                throw new HyperCastFatalRuntimeException ("Physical Address cannot be null in a neighborhood entry");

            final I_LogicalAddress destinationLogicalAddress = parent.getLogicalAddress();
            if (destinationLogicalAddress == null)
                throw new HyperCastFatalRuntimeException ("destination Logical Address must never be null");

            // don't send it back to previous hops
            if (! routeRecord.contains (destinationLogicalAddress))
            {
                updateRouteRecord (appmsg);

                if (debug)
                    System.out.println("send Unicast message to "+parent.toString()+", next hop is "+destinationPhysicalAddress.toString()+"deliveryMode="+appmsg.getDeliveryMode());

                socketAdapter.sendUnicastMessage (destinationPhysicalAddress, appmsg);
            }
        }
    }

    /**
     * Forwards the flood message to all neighbors 
     * except the one that this message comes from.
     *
     * @param appmsg an OL_Message object
     *
     * @throws HyperCastFatalRuntimeException If the parameter of the
     * method cannot be validated as a flood message or it any method
     * call returns a null reference.
     */
    protected void forwardFlood (final OL_Message appmsg)
    {
        if (appmsg.getDeliveryMode() != OL_Message.DELIVERY_MODE_FLOOD)
            throw new HyperCastFatalRuntimeException ("Delivery mode must be FLOOD, not: " + appmsg.getDeliveryMode());

        final I_LogicalAddress lasthop = appmsg.getPreviousHopAddress();
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        appmsg.setPreviousHopAddress (myLogicalAddress);

        final I_AddressPair [] neighbors = node.getAllNeighbors();
        if (neighbors == null)
            throw new HyperCastFatalRuntimeException ("Neighbor list is null");

        for (int i = 0; i < neighbors.length; i++)
        {
            /*
             *  HC protocol returns null neighbor entries
             */
            if (neighbors[i] != null)
            {
                final I_LogicalAddress neighborLogicalAddress = neighbors[i].getLogicalAddress();
                if (neighborLogicalAddress == null)
                    throw new HyperCastFatalRuntimeException ("Logical Address cannot be null in a neighborhood entry");

                if (! neighborLogicalAddress.equals (lasthop))
                {
                    final I_PhysicalAddress destinationPhysicalAddress = neighbors[i].getPhysicalAddress();
                    if (destinationPhysicalAddress == null)
                        throw new HyperCastFatalRuntimeException ("Physical Address cannot be null in a neighborhood entry");

                    socketAdapter.sendUnicastMessage (destinationPhysicalAddress, appmsg);
                }
            }
        }
    }

	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
    protected void initStatisticsStructure()
    {
    	/*
    	 * Get the statistics name defined in the configuration file for this I_Stats instance.
    	 */
		statisticsName = config.getTextAttribute (XmlUtil.createXPath("/Public/StatName"));

        /*
         * Initial StatsProcessor instance.
         */
        statsProcessor = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);

        statsProcessor.addStatsElement ("Running"/* name */, new olsocketRunning()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

        statsProcessor.addStatsElement ("Node", node, 1, 1);

        statsProcessor.addStatsElement ("SocketAdapter", socketAdapter, 1, 1);

        statsProcessor.addStatsElement ("RecvBuf", receiveBuffer, 1, 1);

        statsProcessor.addStatsElement ("Config", config, 1, 1);

        statsProcessor.addStatsElement ("MessageStore", messagestore, 1, 1);
    }

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

    /** @see I_Stats#getStats */
    public Element[] getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getStatsResult (doc, xpath);
    }

    /** @see I_Stats#setStats */
    public Element[] setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);
    }

    /** @see I_Stats#getReadSchema */
    public Element[] getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }

    /** @see I_Stats#getWriteSchema */
    public Element[] getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }
    
    
    /**
     * This class manages simple statistic "Running".
     */
    class olsocketRunning extends SimpleStats
    {
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			synchronized (this)
			{
				return ""+ running;
			}
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newvalue)
		{
			final boolean newRunningValue = Boolean.valueOf (newvalue).booleanValue();

			synchronized (this)
			{
				if (running == false
					&&
					newRunningValue == true)
				{
					joinOverlay();
				}
				else if (running == true
						 &&
						 newRunningValue == false)
				{
					leaveOverlay();
				}
				return ""+ running;
			}
		}
		

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		

		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
				
    }
    
}
