/*************************************************************************

"FreePastry" Peer-to-Peer Application Development Substrate

Copyright 2002, Rice University. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither  the name  of Rice  University (RICE) nor  the names  of its
contributors may be  used to endorse or promote  products derived from
this software without specific prior written permission.

This software is provided by RICE and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall RICE or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

********************************************************************************/

package hypercast.Pastry.rice.pastry.hc;

import hypercast.Pastry.rice.pastry.*;
import hypercast.Pastry.rice.pastry.hc.*;
import hypercast.Pastry.rice.pastry.leafset.*;
import hypercast.Pastry.rice.pastry.routing.*;
import hypercast.Pastry.rice.pastry.messaging.*;
import hypercast.Pastry.rice.pastry.join.*;
import hypercast.Pastry.rice.pastry.client.*;

import hypercast.Pastry.HyperCastTimer;
import hypercast.Pastry.HyperCastNodeAdapter;
import hypercast.Pastry.HyperCastScheduledMessage;
import hypercast.Pastry.HyperCastPhysicalAddress;

import java.util.*;
import java.net.*;

/**
 * @author Weisheng Si
 * @version 2.0, Nov 11, 2001 
 *
 * @author Jianping Wang
 * @version 3.0, Apr. 28, 2004 
 *
 * @author Greg Mattes
 * @version May 2004
 */
public class HCPastryNode
    extends PastryNode
{
    private final static boolean debug = false;

    // Period (in seconds) at which the leafset and routeset maintenance tasks, respectively, are invoked.
    // 0 means never.
    protected int leafSetMaintFreq, routeSetMaintFreq;

    private HyperCastNodeAdapter adapter;
    private HCNodeHandlePool handlepool;

    // timer that supports scheduled messages
    private HyperCastTimer hcTimer;

    // join retransmission stuff
    private ScheduledMessage joinEvent;
    
    public HCPastryNode (final NodeId id, final HyperCastTimer hcTimer, final HyperCastNodeAdapter adapter)
    {
        super (id);
        this.hcTimer = hcTimer;
        this.adapter = adapter;
        this.handlepool = null;
    }

    public void setHCElements (final HCNodeHandlePool handlepool, final int leafSetMaintFreq, final int routeSetMaintFreq)
    {
        this.handlepool = handlepool;
        this.leafSetMaintFreq = leafSetMaintFreq;
        this.routeSetMaintFreq = routeSetMaintFreq;
    }

    public long getCurrentTime() {
        return adapter.getCurrentTime();
    }

    /**
     * Sends an InitiateJoin message to itself.
     *
     * @param bootstrap Node handle to bootstrap with.
     */
    public final void initiateJoin (final NodeHandle bootstrap)
    {
        // schedule (re-)transmission of the join message, every 5s
	if (bootstrap != null)
            joinEvent = scheduleMsg (new InitiateJoin (bootstrap), 0, 5000);
	else
	    setReady(); // no bootstrap node, so ready immediately
    }

    /**
     * Method which kills a PastryNode (used only for testing).
     */
    public void kill() {
	// cancel all scheduled messages
	hcTimer.cancel();
    }

   /**
     * Schedule the specified message to be sent to the local node after a specified delay.
     * Useful to provide timeouts.
     *
     * @param msg a message that will be delivered to the local node after the specified delay
     * @param delay time in milliseconds before message is to be delivered
     * @return the scheduled event object; can be used to cancel the message
     */
    public ScheduledMessage scheduleMsg(Message msg, long delay) {
        final HyperCastScheduledMessage hsm = new HyperCastScheduledMessage(this, msg, delay);
        hcTimer.schedule(hsm);
        return hsm;
    }

    /**
     * Schedule the specified message for repeated fixed-delay delivery to the local node,  
     * beginning after the specified delay. Subsequent executions take place at approximately regular 
     * intervals separated by the specified period. Useful to initiate periodic tasks.
     *
     * @param msg a message that will be delivered to the local node after the specified delay
     * @param delay time in milliseconds before message is to be delivered
     * @param period time in milliseconds between successive message deliveries
     * @return the scheduled event object; can be used to cancel the message 
     */
    public ScheduledMessage scheduleMsg(Message msg, long delay, long period) {
        final HyperCastScheduledMessage hsm = new HyperCastScheduledMessage(this, msg, delay, period);
        hcTimer.schedule(hsm);
	return hsm;
    }

    /**
     * Schedule the specified message for repeated fixed-rate delivery to the local node,  
     * beginning after the specified delay. Subsequent executions take place at approximately regular 
     * intervals, separated by the specified period.
     *
     * @param msg a message that will be delivered to the local node after the specified delay
     * @param delay time in milliseconds before  message is to be delivered
     * @param period time in milliseconds between successive message deliveries
     * @return the scheduled event object; can be used to cancel the message 
     */
    public ScheduledMessage scheduleMsgAtFixedRate (Message msg, long delay, long period) {
        final HyperCastScheduledMessage hsm = new HyperCastScheduledMessage (this, msg, delay, period);
        hcTimer.schedule (hsm);
	return hsm;
    }

    public HCNodeHandlePool getNodeHandlePool() { return handlepool; }

    /**
     * Called after the node is initialized.
     * @param bootstrap The node which this node should boot off of.
     */
    public void doneNode (final NodeHandle bootstrap)
    {
        if (debug)
            System.out.println ("PASTRY DEBUG: HCPastryNode.doneNode(): calling initiateJoin...");

        // schedule the leafset maintenance event
	if (leafSetMaintFreq > 0)
	    scheduleMsgAtFixedRate (new InitiateLeafSetMaintenance(), leafSetMaintFreq * 1000, leafSetMaintFreq * 1000);

        // schedule the routeset maintenance event
	if (routeSetMaintFreq > 0)
	    scheduleMsgAtFixedRate (new InitiateRouteSetMaintenance(), routeSetMaintFreq * 1000, routeSetMaintFreq * 1000);

        initiateJoin (bootstrap);
    }

    public final void leaveOverlay()
    {
        hcTimer.clearTimer (new HyperCastScheduledMessage (this, new InitiateLeafSetMaintenance(), 0));
        hcTimer.clearTimer (new HyperCastScheduledMessage (this, new InitiateRouteSetMaintenance(), 0));
    }

    /**
     * Called from PastryNode when the join succeeds.
     */
    public final void nodeIsReady()
    {
        if (debug)
            System.err.println (myNodeId + " " + hcTimer.getCurrentTime() + ": PASTRY DEBUG: NODE IS READY: CANCELLING JOIN RETRANSMISSIONS!!");

        // cancel join retransmissions
	if (joinEvent != null)
            joinEvent.cancel();

        adapter.nodeIsReady();
    }

    public void ping (final HyperCastPhysicalAddress dstPhysical, final NodeId dstLogical) {
        adapter.ping (dstPhysical, dstLogical);
    }

    /**
     *  Uses HyperCast to send a message over the network
     */
    public void sendMessage (final HyperCastPhysicalAddress dstPhysical,
                             final NodeId                   dstLogical,
                             final HyperCastPhysicalAddress srcPhysical,
                             final NodeId                   srcLogical,
                             final Message                  msg)
    {
        if (debug)
            System.err.println ("PASTRY DEBUG: Sending message from PastryNode.sendMessage()");

        adapter.sendUnicastMessage (dstPhysical, dstLogical, srcPhysical, srcLogical, msg);
    }
}
