/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.SPT;

import hypercast.*;
import hypercast.events.*;
import hypercast.util.XmlUtil;

import java.util.*;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.apache.xpath.*;


/** This class keeps track of the Neighborhood table and LogicalAddress
 * of a node in the Spanning Tree Network.  
 * <p>
 * The neighbors are kept unsorted in a Vector. The neighbors can be accessed
 * either through their index in the Vector, or by the neighbor's Physical Address.
 * For each neighbor, SPT_Neighborhood stores an AddressPair and the time
 * the most recent message was received from that neighbor.
 * <p>
 * SPT_Neighborhood maintains a property on the neighbors.  This property is:
 * the node's Logical Address
 *
 * For any problems, contact Jorg Liebeherr at jorg@cs.virginia.edu
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version	3.0, May 25, 2003
 */

final class SPT_Neighborhood
{
	/** Source of constants & error/log message display */
	private HyperCastConfig config;

	/** Node adapter (used for times and Physical Address). */
	private I_UnicastAdapter adapter;
	
	/** Used for parsing logical address*/
	private SPT_Node node;

	/** The physical and logical address of this node.
	 *  Since this object might be refered by several different place,
	 *  when the logical address need to be changed, try to use SPT_LogicalAddress.changeLogicalAddress()
	 *  instead of making another new address.
	 */
	private SPT_AddressPair myAddressPair;

	/** NeighborhooEntries of the node's neighbors.  This is
	 unordered.  @see SPT_NeighborhoodEntry */
	Hashtable neighbors_table;
	
	/** store all the adjacent nodes */
	Hashtable adjacency_table;
	
	/** Root History Table */
	SPT_RootHistory root_history;

	/** parent entry */
	private SPT_NeighborhoodEntry parentEntry;	
	
	/**
	 * Whether the neighborhood table is fixed.
	 * If it's fixed, neither a new message nor timeout would change this
	 * neighborhood table.
	 * 
	 * For experimental purpose only.
	 */
	boolean NeighborhoodFixed=false;
	
	/** Topology Policy */
	SPT_TopologyPolicy Topo_Policy;
	
	final static String PROP_NAME_NEIGHBOR_TIMEOUT="NeighborTimeout";
	final static String PROP_NAME_ADJACENCY_TIMEOUT="AdjacencyTimeout";
	final static String PROP_NAME_PRESET="Preset";
	final static String PROP_NAME_PRESET_NEIGHBORS="PresetNeighbors";
	
	long neighborTimeout;
	long adjacencyTimeout;
	
	//private NeighborhoodStatsTable neighborhoodStats;

	/**
	 * The reference to the StatsProcessor instance which handles the statistics in this object.
	 */
	protected StatsProcessor statsPro;
	
	/** preset neighbors */
	private boolean Preset=false;

	/** Constructor.
	  */
	public SPT_Neighborhood(HyperCastConfig c, String prefix, SPT_Node nd, I_UnicastAdapter ua, SPT_AddressPair ap, StatsProcessor sPro) {
		config = c;
		adapter = ua;
		node = nd;
		//neighborhoodStats=ns;
		statsPro = sPro;
		
        //A adapter knows the type of physical address, but it doesn't store aphysical address.
		//myAddressPair = new SPT_AddressPair(adapter.getPhysicalAddress(), la);
		myAddressPair = ap;

		parentEntry = null;

		neighbors_table = new Hashtable();
		adjacency_table = new Hashtable();
		root_history=new SPT_RootHistory(config, prefix, ua);
		
		Topo_Policy = new SPT_TopologyPolicy(c, prefix, ap.getSPTLogicalAddress(), ua);
		
		//prefix+=".";
		
		//neighborTimeout=config.getLongProperty(prefix+PROP_NAME_NEIGHBOR_TIMEOUT, SPT_Node.BEACON_TIME_DEFAULT*5);
		neighborTimeout=config.getLongAttribute
				(XmlUtil.createXPath(prefix+PROP_NAME_NEIGHBOR_TIMEOUT));

		//adjacencyTimeout=config.getLongProperty(prefix+PROP_NAME_ADJACENCY_TIMEOUT, SPT_Node.BEACON_TIME_DEFAULT*20);
		adjacencyTimeout=config.getLongAttribute
				(XmlUtil.createXPath(prefix+PROP_NAME_ADJACENCY_TIMEOUT));

		/** for experiment purpose */
		//String preset = c.getStringProperty(prefix+PROP_NAME_PRESET, "false");
		String preset = "false";
		try
		{
			preset = c.getTextAttribute(XmlUtil.createXPath(prefix+PROP_NAME_PRESET));
		}
		catch (HyperCastConfigException e)
		{
			/*
			 * The attribute doesn't exist.
			 */
			preset = "false";
		}
			

		/** preset neighbors */
		if(preset.equals("true")){
		
/*
			String preset_parent = c.getStringProperty(prefix+PROP_NAME_PRESET_NEIGHBORS+".Parent", "");
			String preset_children = c.getStringProperty(prefix+PROP_NAME_PRESET_NEIGHBORS+".Children", "");
			String preset_root = c.getStringProperty(prefix+PROP_NAME_PRESET_NEIGHBORS+".Root", preset_parent);
			int preset_cost = c.getIntProperty(prefix+PROP_NAME_PRESET_NEIGHBORS+".Cost", 1);
*/
			String preset_parent = "";
			String preset_children = "";
			String preset_root = "";
			int preset_cost = 1;
			try
			{
				preset_parent = c.getTextAttribute(XmlUtil.createXPath(prefix+prefix+PROP_NAME_PRESET_NEIGHBORS+"/Parent"));
			}
			catch (HyperCastConfigException e)
			{
				preset_parent = "";
			}
			try
			{
				preset_children = c.getTextAttribute(XmlUtil.createXPath(prefix+prefix+PROP_NAME_PRESET_NEIGHBORS+"/Children"));
			}
			catch (HyperCastConfigException e)
			{
				preset_children = "";
			}
			try
			{
				preset_root = c.getTextAttribute(XmlUtil.createXPath(prefix+prefix+PROP_NAME_PRESET_NEIGHBORS+"/Root"));
			}
			catch (HyperCastConfigException e)
			{
				preset_root = preset_parent;
			}
			try
			{
				preset_cost = c.getIntAttribute(XmlUtil.createXPath(prefix+prefix+PROP_NAME_PRESET_NEIGHBORS+"/Cost"));
			}
			catch (HyperCastConfigException e)
			{
				preset_cost = 1;
			}
			
			Preset=true;
			readPresetNeighbors(preset_parent, preset_children, preset_root, preset_cost);
			
			String fixNeighborFlag = "false";
			try
			{
				fixNeighborFlag = c.getTextAttribute(XmlUtil.createXPath(prefix+PROP_NAME_PRESET+"/FixNeighbor"));
			}
			catch (HyperCastConfigException e)
			{
				fixNeighborFlag = "flase";
			}
			
			//if(c.getStringProperty(prefix+PROP_NAME_PRESET+".FixNeighbor", "false").equals("true"))
			if(fixNeighborFlag.equals("true"))
			{
				c.log.println("Neighborhood Fixed");
				this.NeighborhoodFixed=true;
			}
		}
		
	}
	
	
	
	private void readPresetNeighbors(String parent, String children, String root, int cost) {
		
		
		SPT_AddressPair parentAP = null;
		SPT_AddressPair childrenAP = null;
		SPT_LogicalAddress rootLA = (SPT_LogicalAddress) node.createLogicalAddress(root);

		long now = adapter.getCurrentTime();

		/* read parent */
		if (parent != null && !parent.equals(""))
			parentAP = parseAddressPair(parent);

		if (parentAP != null) {
			// add new parent
			
			System.out.println("add "+parentAP+" as parent root="+rootLA+" cost="+cost);
			
			SPT_Message beacon = 
				new SPT_Message(
						I_Node.PROTOSUB_MULTICAST,
						SPT_Message.Hello,
						config.getOverlayHash(),
						parentAP,
						rootLA,
						rootLA,
						cost-1,
						adapter.getCurrentTime(),
						null,
						(short)I_PathMetric.FULL_PATHMETRIC_VALUE
					);
			
			SPT_NeighborhoodEntry ne = this.updateAdjacentNode(beacon);
			this.updateNeighbor(beacon);
			
			ne.fixed=true;
			
			Topo_Policy.NextParent=parentAP.getSPTLogicalAddress();
			
			parentEntry=ne;

		}

		/* read children */
		if (children!=null && !children.equals("")){
			StringTokenizer tokens = new StringTokenizer(children, ";");
			while(tokens.hasMoreTokens()){
				String child=tokens.nextToken();
				childrenAP=parseAddressPair(child);
				
				System.out.println("add "+childrenAP+" as children");
				
				SPT_Message beacon = 
					new SPT_Message(
							I_Node.PROTOSUB_MULTICAST,
							SPT_Message.Hello,
							config.getOverlayHash(),
							childrenAP,
							rootLA,
							this.getMyLogicalAddress(),
							cost+1,
							adapter.getCurrentTime(),
							null,
							(short)I_PathMetric.FULL_PATHMETRIC_VALUE
						);
				
				SPT_NeighborhoodEntry ne = this.updateAdjacentNode(beacon);
				
				ne.fixed=true;
			
				this.updateNeighbor(beacon);
			}
		}

	}
	/** split address pair string to LA and PA */
	private SPT_AddressPair parseAddressPair(String addr) {
		String la = "", pa = "";
		int index = addr.indexOf("@");
		if (index >= 0) {
			la = addr.substring(0, index);
			pa = addr.substring(index + 1, addr.length());
		} else
			la = addr;

		return new SPT_AddressPair(
			adapter.createPhysicalAddress(pa),
			(SPT_LogicalAddress) node.createLogicalAddress(la));

	}

	/**
	 * get the entry in neighbors table by LA
	 */
	public SPT_NeighborhoodEntry getNeighborEntry(SPT_LogicalAddress la){
		return (SPT_NeighborhoodEntry)neighbors_table.get(la);
	}
	/**
	 * get the entry in neighbors table by Address Pair
	 */
	public SPT_NeighborhoodEntry getNeighborEntry(SPT_AddressPair ap){
		SPT_NeighborhoodEntry ne=this.getNeighborEntry(ap.getSPTLogicalAddress());
		if(ne.getAddressPair().equals(ap)){
			return ne;
		}
		else return null;
	}
	
	/**
	 * update a node to adjacency table
	 */
	private SPT_NeighborhoodEntry updateAdjacentNode(SPT_Message beacon){
		SPT_NeighborhoodEntry entry=this.getAdjacencyEntry(beacon.getSrcLA());
		if(entry!=null){
			entry.update(beacon);
			
		}else{
			//add a entry
			LinkQuality link=Topo_Policy.createLinkQuality();
			I_PathMetric pm=Topo_Policy.createPathMetric(beacon, link);
			
			entry=new SPT_NeighborhoodEntry(beacon, link, pm, adapter, neighborTimeout, adjacencyTimeout, false);
			
			adjacency_table.put(beacon.getSrcLA(), entry);
		}
		
		return entry;
	}
	/**
	 * add this entry into neighbor table
	 */
	private void addNeighbor(SPT_NeighborhoodEntry entry){
		if(!neighbors_table.containsKey(entry.getLogicalAddress())){
			neighbors_table.put(entry.getLogicalAddress(), entry);

			//if(neighborhoodStats!=null)
			//	neighborhoodStats.addNeighbor(entry.getAddressPair());
			
			if(statsPro!=null)
				this.addNeighborToStatsProcessor(entry.getAddressPair());
			
		}
	}
	

	/**
	 * Create a StatsElement instance for a neighbor and add it to the StatsProcessor object.
	 */
	private void addNeighborToStatsProcessor(I_AddressPair ap)
	{
		 statsPro.addStatsElement("NeighborTable", new NeighborhoodStats(ap), 0/* minoccurs */, -1/* maxoccurs */);
	}
	
	/**
	 * update a node to neighbor table
	 */
	private SPT_NeighborhoodEntry updateNeighbor(SPT_Message beacon){
		SPT_NeighborhoodEntry entry=this.getAdjacencyEntry(beacon.getSrcLA());
		if(entry!=null){
			entry.update(beacon);
			
		}else{
			//add a entry
			LinkQuality link=Topo_Policy.createLinkQuality();
			I_PathMetric pm=Topo_Policy.createPathMetric(beacon, link);
			
			entry=new SPT_NeighborhoodEntry(beacon, link, pm, adapter, neighborTimeout, adjacencyTimeout, false);
			
			adjacency_table.put(beacon.getSrcLA(), entry);
			
			
		}
		
		if(!neighbors_table.containsKey(beacon.getSrcLA())){
			
			addNeighbor(entry);
			
		}
		
		return entry;
	}
	
	/**
	 * get the entry in adjacency table by LA
	 */
	public SPT_NeighborhoodEntry getAdjacencyEntry(SPT_LogicalAddress la){
		return (SPT_NeighborhoodEntry)adjacency_table.get(la);
	}
	/**
	 * get the entry in adjacency table by Address Pair
	 */
	public SPT_NeighborhoodEntry getAdjacencyEntry(SPT_AddressPair ap){
		SPT_NeighborhoodEntry ne=this.getAdjacencyEntry(ap.getSPTLogicalAddress());
		if(ne.getAddressPair().equals(ap)){
			return ne;
		}
		else return null;
	}
	/**
	 * get the adjacency table put into the beacon message
	 */
	public Hashtable getMsgAdjacencyTable(){
		/** check the age of entries */

		Hashtable sl = new Hashtable();

		for (Enumeration e = adjacency_table.elements(); e.hasMoreElements();) {
			SPT_NeighborhoodEntry se = (SPT_NeighborhoodEntry) e.nextElement();
			
			
			if (!se.isExpiredAdjacentNode() && se.getLinkQuality().getLinkQuality()>=se.getLinkQuality().ReliableThreshold) {
				
				byte sender_byte=(byte)se.getLinkQuality().getPingNum();
				if(containsNeighbor(se.getLogicalAddress()))sender_byte= (byte)(sender_byte | (byte)0x80);
				sl.put("" + se.getLogicalAddress(), new Byte(sender_byte));

			}
		}

		return sl;
	}

	/** 
	 * @return  address pair of this node 
	 */
	public SPT_AddressPair getMyAddressPair() {
		return myAddressPair;
	}
	

	/** 
	 * @return logical address of this node
	 */
	public SPT_LogicalAddress getMyLogicalAddress() {
		return myAddressPair.getSPTLogicalAddress();
	}

	/** 
	 * @return physical address of this node
	 */
	public I_PhysicalAddress getMyPhysicalAddress() {
		return myAddressPair.getPhysicalAddress();
	}
	
	
	/** Returns true if there exists a neighbor with a LogicalAddress equal to the parameter. 
	 * @param la		logical address of neighbor
	 * @return	whether this neighbor exists
	 */
	public boolean containsNeighbor(SPT_LogicalAddress la) {
		return this.getNeighborEntry(la)!=null;
	}
	
	/** Returns true if there exists an adjacent node with a LogicalAddress equal to the parameter. 
	 * @param la		logical address of node
	 * @return	whether this adjacent node exists
	 */
	public boolean containsAdjacentNode(SPT_LogicalAddress la) {
		return this.getAdjacencyEntry(la)!=null;
	}
	
	/** Returns true if there exists a neighbor with a specified address pair. 
	 * @param ap		address pair
	 * @return	whether this neighbor exists
	 */
	public boolean containsNeighbor(SPT_AddressPair ap) {
		return this.getNeighborEntry(ap)!=null;
	}
	
	/** Returns true if there exists an adjacent with a specified address pair. 
	 * @param ap		address pair
	 * @return	whether this adjacent exists
	 */
	public boolean containsAdjacentNode(SPT_AddressPair ap) {
		return this.getAdjacencyEntry(ap)!=null;
	}
	
	/** general methods */
	
	public boolean isRoot() {
		return parentEntry == null;
	}
	/** Returns true if this node is the parent of the neighbor <code>node</code> 
	 * in the spanning tree 
	 * @see #getParentLogicalAddress()
	 */
	public boolean isParentOf(SPT_NeighborhoodEntry nodeEntry) {
		if (nodeEntry.getParentLogicalAddress() == null)
			return false;

		return nodeEntry.getLogicalAddress().equals(this.getMyLogicalAddress());

	}

	/**
	 * Judge if a neighbor is this node's parent
	 * @param nodeEntry	the neighbor's entry
	 * @return	whether the neighbor is parent
	 */
	public boolean isChildOf(SPT_NeighborhoodEntry nodeEntry) {
		if (parentEntry == null)
			return false;
		return nodeEntry.getLogicalAddress().equals(this.getParentLogicalAddress());

	}
	
	/**
	 *  @return the number of neighbors of this node 
	 */
	public final int getNumOfNeighbors() {
		return neighbors_table.size();
	}
	/** @return the number of neighbors
	 */
	public final int getNumOfAdjacentNodes() {
		return adjacency_table.size();
	}

	
	/**
	 * When the node is root, timestamp is generated using an
	 * incremental sequence number, otherwise it's set to the
	 * timestamp got from the parent.
	 * 
	 * @return	the next timestamp to put in the beacon message
	 */
	public long getTimeStamp() {
		if (parentEntry == null) {
			return adapter.getCurrentTime();
		} else {
			return parentEntry.getTimeStamp();
		}
	}

	/**
	 * @return	the current path cost to root
	 */
	public int getCost() {
		if (parentEntry == null)
			return 0;
		else
			return parentEntry.getCost() + 1;

	}
	/**
	 * @return	parent's logical address, or null if no parent
	 */
	public SPT_LogicalAddress getParentLogicalAddress() {
		if (parentEntry == null)
			return null;

		return parentEntry.getLogicalAddress();
	}
	/**
	 * @return  root's logical address
	 */
	public SPT_LogicalAddress getRootLogicalAddress() {
		if (parentEntry == null)
			return getMyLogicalAddress();

		return parentEntry.getRootLogicalAddress();
	}
	
	/**
	 * @return	address pair of parent, or null if no parent
	 */
	public SPT_AddressPair getParentAddressPair() {
		if (parentEntry == null)
			return null;

		return parentEntry.getAddressPair();
	}
	
	/**
	 * get the path metric of this node
	 */
	public int getPathMetric(){
		if (parentEntry == null)
			return Topo_Policy.createPathMetric(null, null).getPathMetric();
		
		return parentEntry.getPathMetric();
	}

	
	/**
	 * @return	all spanning tree children
	 */
	public SPT_AddressPair[] getChildren() {
		Vector children = new Vector(neighbors_table.size());
		
		for(Enumeration e=neighbors_table.elements();e.hasMoreElements();){
			SPT_NeighborhoodEntry entry=(SPT_NeighborhoodEntry)e.nextElement();
			if(isParentOf(entry)){
				children.addElement(entry.getAddressPair());
			}
		}

		SPT_AddressPair[] array = new SPT_AddressPair[children.size()];
		children.copyInto(array);

		return array;
	}
	/**
	 * @return	all neighbors
	 */
	public SPT_AddressPair[] getAllNeighbors() {
		SPT_AddressPair[] array = new SPT_AddressPair[neighbors_table.size()];
		int i=0;
		for(Enumeration e=neighbors_table.elements();e.hasMoreElements();){
			SPT_NeighborhoodEntry entry=(SPT_NeighborhoodEntry)e.nextElement();
			array[i++]=(entry.getAddressPair());
			
		}
		return array;
	}
	
	private SPT_NeighborhoodEntry getBackupParent(long now) {
		return getBetterParent(null, now);
	}
	private SPT_NeighborhoodEntry getBetterParent(SPT_NeighborhoodEntry parent, long now) {
		
		SPT_NeighborhoodEntry bestNode=null;
		
		for(Enumeration e= adjacency_table.elements();e.hasMoreElements();){
			SPT_NeighborhoodEntry entry=(SPT_NeighborhoodEntry)e.nextElement();
			
			if(entry.isExpiredNeighbor())
				continue;
			
			/* TODO: Determine if the following commented statement is still relevant
			if(node.root_history.isTooOld(m2.getRootId(), now, node.MSG_MAX_AGE))
				continue;
			*/
			
			if(!Topo_Policy.isBetterThanCurrentParent(entry, parent)){
				continue;
			}
			
			if(entry.getLinkQuality().isReliable())
			{
				if(bestNode==null){
					bestNode=entry;
				}else{
					if(Topo_Policy.isBetterThan(entry, bestNode)){
						bestNode=entry;
					}
				}
			}
		}
		
		return bestNode;
	}
	
	/**
	 * Change current parent to a backup parent
	 */
	private void changeToBackupParent() {
		/*## may need further work to refine the backup parent table*/
		if(this.NeighborhoodFixed)return ;
		
		parentEntry = getBackupParent(adapter.getCurrentTime());
		if(parentEntry!=null && !this.containsNeighbor(parentEntry.getLogicalAddress())){
			addNeighbor(parentEntry);

			//if(neighborhoodStats!=null)
			//	neighborhoodStats.addNeighbor(parentEntry.getAddressPair());

			
			if(statsPro!=null)
				this.addNeighborToStatsProcessor(parentEntry.getAddressPair());
		}

		//for debug
		/*
		config.log.println(
			"Node "+this.getMyLogicalAddress()+" - Root: " + getRootId() + " Parent: " + getParentId() + " Cost=" + getCost());
			*/

	}

	/** end of parent table methods */

	/** neighbors update routines */
	/**
	 * update the neighbor table by a beacon message
	 * @return 		true if the neighbor table has changed
	 */
	public boolean update(SPT_Message beacon){
		
		/* if neighborhood is fixed, do nothing */
		if(NeighborhoodFixed)return false;
		
		boolean changed=false;
		
		/* is it from myself? */
		if (myAddressPair.getPhysicalAddress().equals(beacon.getSrcPA())) {
			return false;
		}
		
		SPT_LogicalAddress neLA=beacon.getSrcLA();
		
		// if a neighbor's LA have changed - remove and reinsert.
		if (containsAdjacentNode(neLA) && !containsAdjacentNode(beacon.getSrc())) {
			if(containsNeighbor(neLA))
				changed=true;
			
			removeAdjacentNode(neLA);
			//            System.out.println("Was neighbor - changed LA - removed.");
		}
		
		// guarantee unique id
		// may need some work to be able to adjust id automatically
		if (myAddressPair.getSPTLogicalAddress().equals(neLA)) {
			config.err.println("Node with duplicate ID detected.");
			return false;
		}
		
		//System.out.println(""+this.getMyLogicalAddress()+" from "+beacon.getSrcLA()+" ts="+beacon.getTimeStamp());
		
		SPT_NeighborhoodEntry entry = this.updateAdjacentNode(beacon);
		SPT_NeighborhoodEntry old_parent_entry=parentEntry;
		

		if(this.containsNeighbor(entry.getLogicalAddress())){
			if(entry.fixed)return false;

			//if(!entry.getNeighbor().equals(n))return false;
			if (this.isChildOf(entry) && !beacon.getRoot().equals(this.getRootLogicalAddress())) {
				config.log.println("Node "+this.getMyLogicalAddress()+" - Root changed to " + beacon.getRoot() + " Cost=" + (beacon.getCost() + 1));
								
			}
		}
		
		if(!beacon.getRoot().equals(this.getMyLogicalAddress())){
			root_history.updateRootHistory(beacon.getRoot(), beacon.getTimeStamp());
		}
		if (root_history.isTooOld(getRootLogicalAddress())) {
			//root is too old
			//delete the parent
			setParent(null);
			changed=true;
		}
		if (root_history.isTooOld(beacon.getRoot())) {
			return false; //message too old
		}
		
		/* used to judge if any change happens */
		int old_root = this.getRootLogicalAddress().getId();
		int old_cost = getCost();
		
		
		/* update the root and parent */
		updateParent(entry);

		if (entry != parentEntry && (!getMyLogicalAddress().equals(beacon.getParent()))) {
			/* this is not a neighbor */
			if(this.containsNeighbor(beacon.getSrcLA())){
				removeNeighbor(entry.getLogicalAddress());
				return true;
			}
		}else{
			/* this is a neighbor */
			if(!this.containsNeighbor(beacon.getSrcLA())){
				changed=true;
			}
			this.updateNeighbor(beacon);
			
		}
		
		if(old_parent_entry!=null && old_parent_entry!=parentEntry && !old_parent_entry.getLogicalAddress().equals(entry.getLogicalAddress())){
			//old parent removed
			removeNeighbor(old_parent_entry.getLogicalAddress());
			return true;
		}

		if (old_root != this.getRootLogicalAddress().getId() || old_cost != getCost() || old_parent_entry!=parentEntry)
			return true;
		else
			return changed;
	}

	/**
	 * clear neighbors
	 *
	 */
	public void clear() {
		
		if(this.NeighborhoodFixed)return;
		
		if(Preset)return;
		
		neighbors_table.clear();
		adjacency_table.clear();
		root_history.clear();
		
		//if(neighborhoodStats!=null)
		//	neighborhoodStats.clear();

		if (statsPro != null)
			statsPro.removeAllStatsElement("NeighborTable");

		parentEntry = null;
		
	}
	
	/**
	 * clear timeout entries
	 */
	public boolean clearExpired(){
		int old_size=neighbors_table.size();
		
		for(Enumeration e=adjacency_table.elements();e.hasMoreElements();){
			SPT_NeighborhoodEntry entry=(SPT_NeighborhoodEntry)e.nextElement();
			if(entry.isExpiredNeighbor()){
				removeNeighbor(entry.getLogicalAddress());
				
			}if(entry.isExpiredAdjacentNode())
				removeAdjacentNode(entry.getLogicalAddress());
		}
		
		if(neighbors_table.size()!=old_size)
			return true;
		else 
			return false;
	}
	/**
	 * remove adjacent node
	 */
	public void removeAdjacentNode(SPT_LogicalAddress la){
		removeNeighbor(la);
		adjacency_table.remove(la);
	}
	/**
	 * Remove a neighbor
	 * @param la		la of the neighbor
	 */
	public void removeNeighbor(SPT_LogicalAddress la) {
		
		if(this.NeighborhoodFixed)return;
		
		/* check if it is to delete parent */
		SPT_NeighborhoodEntry ne = getNeighborEntry(la);
		boolean removeParent = false;

		if(ne==null || ne.fixed)return;

		neighbors_table.remove(la);

		//if(neighborhoodStats!=null)
		//	neighborhoodStats.removeNeighbor(ne.getAddressPair());

		if(statsPro!=null)
			statsPro.removeStatsElement(ne.getAddressPair());
       
		//TODO: Determine if the following commented statement is still relevant
		//node.sender_list.setNeighbor(ne.getNeighborId(), false);
                
        if (ne != null) {
        			
   			if (ne == parentEntry) {
        		/* change to a backup parent */
   				removeParent = true;
   				changeToBackupParent();
   			}

   		}

		if (!removeParent){
			//System.out.println("Children=" + getChildren().length);
			
		}

	}

	/**
	 * Set the parent to a neighbor
	 * @param ne	the neighbor entry
	 */
	public void setParent(SPT_NeighborhoodEntry ne) {
		
		if(this.NeighborhoodFixed)return;
		
		if (parentEntry == ne)
			return;

		if(parentEntry!=null && parentEntry.fixed)
			return;
		
		if(parentEntry!=null){
			removeNeighbor(parentEntry.getLogicalAddress());
		}
			
		
		parentEntry = ne;
		if(ne!=null)
			addNeighbor(ne);

		config.log.println(
			"Node " + this.getMyLogicalAddress()+ " - Root: " + this.getRootLogicalAddress() + " Parent: " + this.getParentLogicalAddress() + " Cost=" + getCost());
		
		
		
		
	}

	
	/**
	 * Update the root/parent information with a neighbor entry.
	 * Check if the neighbor can be a better parent.
	 * @param ne	the neighbor entry
	 */
	private boolean updateParent(SPT_NeighborhoodEntry ne) {

		if (this.NeighborhoodFixed)
			return false;
			
		if (parentEntry!=null && parentEntry.fixed)return false;
		
		//to see if the parent has change the root setting
		//so that it becomes a worse root than this node itself
		if (this.getRootLogicalAddress().greaterThan(this.getMyLogicalAddress())) {
			//the root set to this node
			parentEntry = null;
		}

		//to check if the neighbor's parent is this node
		if (ne.getParentLogicalAddress()!=null && ne.getParentLogicalAddress().equals(this.getMyLogicalAddress()))
			return false;
		
		
		/* if this neighbor is better than the current parent to be a parent
		 * set it as the parent
		 */
		if (Topo_Policy.isBetterThanCurrentParent(ne, parentEntry)) {
			setParent(ne);
			return true;
		}
		
		
		
		/* parent not changed */
		return false;

	}



}


/** 
 * Class containing fields for an entry in the neighbor table. 
 * All of the information in an neighborhood entry is copied from
 * the beacon message received from that neighbor
 * @see SPT_Neighborhood
 */

final class SPT_NeighborhoodEntry {
	/** the latest beacon received */
	private SPT_Message lastBeacon;

	/** This is the time the last message was received from the neighbor. */
	private long lastContactTime;
	
	/** The time for a neighbor to expire */
	private long neighborTimeout;
	
	/** the time for an adjacent node to expire */
	private long adjacencyTimeout;
	
	/** the TimeStamp of the previous beacon */
	long prevMsgTimeStamp=0;
	
	/** for experiment purpose */
	protected boolean fixed = false;
	
	/** for getting time */
	private I_UnicastAdapter adapter;
	
	/** link quality measurement object */
	private LinkQuality link;
	
	/** Path Metric */
	private I_PathMetric pathMetric;
	
	/**
	 * Constructor
	  */
	SPT_NeighborhoodEntry(
		SPT_Message msg, LinkQuality lnk, I_PathMetric pm, I_UnicastAdapter adpt, long nTimeout, long aTimeout, boolean fx)
	{
		lastBeacon=msg;
		
		link=lnk;
		
		pathMetric=pm;

		lastContactTime = adpt.getCurrentTime();
		
		adapter=adpt;
		
		neighborTimeout=nTimeout;
		
		adjacencyTimeout=aTimeout;
		
		fixed=fx;
	}
	
	public LinkQuality getLinkQuality(){
		return link;
	}
	
	/**
	 * check if it's a expried neighbor
	 */
	public boolean isExpiredNeighbor(){
		if(fixed)return false;
		
		return this.lastContactTime+this.neighborTimeout<adapter.getCurrentTime();
	}
	/**
	 * check if it's a expired adjacent node
	 */
	public boolean isExpiredAdjacentNode(){
		if(fixed)return false;
		
		return this.lastContactTime+this.adjacencyTimeout<adapter.getCurrentTime();
	}
	
	/**
	 * get the path metric value
	 */
	public int getPathMetric(){
		return pathMetric.getPathMetric();
	}
	/**
	 * Update this entry
	 * @param msg	the beacon mesage
	 */
	public void update(
			SPT_Message msg
			) 
	{
		if(fixed)return;
		
		if(msg!=null){
			if(lastBeacon!=null){
				if(prevMsgTimeStamp<msg.getTimeStamp())
					prevMsgTimeStamp=lastBeacon.getTimeStamp();
				
			}
			if(msg!=lastBeacon){
				link.beaconReceived(msg);
				pathMetric.beaconReceived(msg);
			}
			
			lastBeacon=msg;
			
		}

		lastContactTime = adapter.getCurrentTime();

	}
	/**
	 * @return the address pair of this neighbor
	 */
	public SPT_AddressPair getAddressPair() {
		return lastBeacon.getSelf();
	}
	/**
	 * @return	the address of this neighbor
	 */
	public SPT_LogicalAddress getLogicalAddress() {
		return lastBeacon.getSrcLA();
	}
	/**
	 * @return	the address of root
	 */
	public SPT_LogicalAddress getRootLogicalAddress() {
		return lastBeacon.getRoot();
	}
	/**
	 * @return	the Id of parent, or -1 if this neighbor has no parent
	 */
	public SPT_LogicalAddress getParentLogicalAddress() {
		return lastBeacon.getParent();
	}
	/**
	 * @return	the path cost
	 */
	public int getCost() {
		return lastBeacon.getCost();
	}
	
	/**
	 * @return	last contact time
	 */
	long getLastContactTime() {
		return lastContactTime;
	}
	/**
	 * @return	timestamp
	 */
	public long getTimeStamp() {
		if(fixed)return adapter.getCurrentTime();
		else return lastBeacon.getTimeStamp();
	}
	
	public SPT_Message getLastBeacon(){
		return lastBeacon;
	}
	
	public String toString() {
		return this.getAddressPair().toString();
	}
	

	/**
	 * @return Returns the prevMsgTimeStamp.
	 */
	public long getPrevMsgTimeStamp() {
		return prevMsgTimeStamp;
	}
}


/** 
 * Used to keep track of the last changing time of the timestamp 
 * from every root id
 * @see SPT_RootHistory
 */
class SPT_RootHistoryEntry {
	/** the Id of root */
	SPT_LogicalAddress rootLA;
	/** the timestamp coming from the root */
	long time_stamp;
	/** the last time the timestamp changed */
	long lastChangeTime;
	
	/** for getting time */
	I_UnicastAdapter adapter;
	
	/** timeout */
	long timeout;
	
	/**
	 * Constructor
	 * @param Id	root Id
	 * @param ts	timestamp from the root
	 */
	public SPT_RootHistoryEntry(SPT_LogicalAddress Id, long ts, long to, I_UnicastAdapter adpt) {
		
		adapter=adpt;
		timeout=to;
		
		rootLA = Id;
		time_stamp = ts;
		lastChangeTime = adapter.getCurrentTime();
	}
	/**
	 * Update the entry
	 * @param ts	timestamp
	 */
	public void update(long ts) {
		if (ts > time_stamp) {
			lastChangeTime = adapter.getCurrentTime();
			time_stamp = ts;

		}
	}
	/**
	 * @return	last changeing time
	 */
	public long getLastChange() {
		return lastChangeTime;
	}
	/**
	 * @return	root's Id
	 */
	public SPT_LogicalAddress getRootLogicalAddress() {
		return rootLA;
	}
	/**
	 * Check if this entry expired
	 * @return			if it has expired
	 */
	public boolean isExpired(){
		return (lastChangeTime + timeout) < adapter.getCurrentTime();
	}
}
/** 
 * This is a class recording the newest timestamp coming from every
 * roots ever appeared in a incoming beacon message. It's used to 
 * solve count-to-infinity problem
 * 
 * @author Guangyu Dong, Jan 13, 2004
 */
class SPT_RootHistory {
	/** the history table*/
	Hashtable history;

	static final String PROP_NAME_RH_TIMEOUT="RootHistoryTimeout";
	static final String PROP_NAME_RH_MAXAGE="MsgMaxAge";
	
	long rootHistoryTimeout;
	long msgMaxAge;
	
	/** for getting time */
	I_UnicastAdapter adapter;
	
	/**
	 * Constructor
	 *
	 */
	public SPT_RootHistory(HyperCastConfig c, String prefix, I_UnicastAdapter adpt) {
		adapter=adpt;
		history = new Hashtable();
		
		//prefix+=".";
		
		//rootHistoryTimeout=c.getLongProperty(prefix+PROP_NAME_RH_TIMEOUT, 20*SPT_Node.BEACON_TIME_DEFAULT);
		rootHistoryTimeout=c.getLongAttribute
				(XmlUtil.createXPath(prefix+PROP_NAME_RH_TIMEOUT));

		//msgMaxAge=c.getLongProperty(prefix+PROP_NAME_RH_MAXAGE, 5*SPT_Node.BEACON_TIME_DEFAULT);
		msgMaxAge=c.getLongAttribute
				(XmlUtil.createXPath(prefix+PROP_NAME_RH_MAXAGE));
		
	}
	/**
	 * Clear table
	 *
	 */
	public void clear(){
		history.clear();
	}
	/**
	 * update root history entry
	 * @param rootId		root's Id
	 * @param time_stamp	timestamp coming from root
	 */
	
	public void updateRootHistory(SPT_LogicalAddress rootId, long time_stamp) {
		SPT_RootHistoryEntry re = (SPT_RootHistoryEntry) history.get(rootId);
		if (re == null)
			history.put(rootId, new SPT_RootHistoryEntry(rootId, time_stamp, rootHistoryTimeout, adapter));
		else
			re.update(time_stamp);
	}
	/*
	public boolean updateRootHistory(int rootId, long time_stamp, long now) {
		SPT_RootHistoryEntry re = (SPT_RootHistoryEntry) history.get("" + rootId);
		if (re == null){
			history.put("" + rootId, new SPT_RootHistoryEntry(rootId, time_stamp, now));
			return true;
		}else{
			if(re.time_stamp<time_stamp){
				re.update(time_stamp, now);
				return true;
			}else
				return false;
		}
	}
	*/
	/**
	 * Check if the timestamp coming from a root is too old to use
	 * @param rootId	root's Id
	 */
	public boolean isTooOld(SPT_LogicalAddress rootId) {
		SPT_RootHistoryEntry re = (SPT_RootHistoryEntry) history.get(rootId);
		if (re == null)
			return false;
		else
			return (re.getLastChange() + this.msgMaxAge) < adapter.getCurrentTime();
			
	}
	/**
	 * Clear expiring entry
	 * @param now		current time
	 * @param timeout	timeout time
	 */
	public void clearExpired(long now, long timeout) {
		for (Enumeration rt = history.elements(); rt.hasMoreElements();) {
			SPT_RootHistoryEntry re = (SPT_RootHistoryEntry) rt.nextElement();

			if (re.isExpired()) {
				history.remove(re.getRootLogicalAddress());
			}
		}
	}
}
