/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import hypercast.events.*;
import hypercast.util.XmlUtil;

/**
 * This class provides methods to allow the application to join a streaming group, leave a streaming group, 
   send/receive stream within the group. 
 * 
 * @author HyperCast Team
 * @author Guimin Zhang
 * @version 3.0, Mar. 8, 2004 
 */

public class StreamManager implements I_ReceiveCallback
{
	/** The overlay socket*/
	protected OL_Socket socket;
	
	/** The inputstream table*/
	protected Hashtable inputtable;
	
	/** The outputstream table*/
	protected Hashtable outputtable;	
	
	/**The outputstream sending period */
	private long sendingPeriod;
	 
	/**the config object */
	protected HyperCastConfig config;  
	
	/** the object used to synchronize the thread */
	private Object guard = new Object();
	
	/** Notification handler. When new stream arrives, it is notified to the application.*/
	protected NotificationHandler notificationHandler = null;
	
	public StreamManager (OL_Socket socket)
	{
		/*
		 * Intialization
		 */
		this.socket = socket;
        this.notificationHandler = socket.getNotificationHandler();
		this.config = socket.config;
		this.inputtable = new Hashtable();
		this.outputtable = new Hashtable();
		//sendingPeriod = config.getPositiveLongProperty ("OutputStreamSendingPeriod");
		sendingPeriod = config.getPositiveLongAttribute (XmlUtil.createXPath("/Public/OutputStreamSendingPeriod"));
		
		/*
		 * set the periodically checking timer
		 */
		socket.socketAdapter.setTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfStreamManager, new Integer(1)),sendingPeriod);
		 
	 }
	
    /**
     * Sets the notification handler that is used to pass events from
     * the message store to an application.
     */
    public void setNotificationHandler (NotificationHandler notificationHandler)
    {
        this.notificationHandler = notificationHandler;
    }

	/**
	 *  The method called each time a message arrives at the overlay socket. */
	public void ReceiveCallback(I_OverlayMessage message)
	{
		/*
		 * check if the message is in valid format
		 */
		if (!(message instanceof OL_Message))
		{
			throw new IllegalArgumentException("OL_Socket.processMessage() passed object of type other than OL_Messagee!");
		}

	
		/*
		 * extract the payload
		 */	
		OL_Message msg = (OL_Message) message;
		byte[] payload = message.getPayload();
		/*
		 * extract the formatted message header
		 */
		FSM_Extension e = (FSM_Extension)msg.getFirstExtensionByType(Extension.FSM);
		if (e == null)
		{
			throw new HyperCastWarningRuntimeException ("No Stream extension found");
		}
		/*
		 * obtain the stream id
		 */
		byte[] uniqueID = e.getUniqueID();	
		/*
		 * use the stream id as key, obtain the inputstream  
		 */
		HCastInputStream ps = (HCastInputStream)inputtable.get(new String(uniqueID));
        
		/*
		 * a new stream.
		 */
		if (ps == null)
		{

//TODO: remove debug statement if no longer needed
			//System.out.println("StreamManager: messageArrived: create new HCastInputStream with unique ID "+(new String(uniqueID)));

			/*
			 * create a new inputstream
			 */
			ps = new HCastInputStream(this,uniqueID);		
			inputtable.put((new String(uniqueID)),ps);
			/*
			 * notify the other waiting thread
			 */
			 synchronized(guard)
			 {
		 		guard.notify();
			 }
			 
			/*
			 * notify the application
			 */
			NEWSTREAM_ARRIVED_EVENT ne = new NEWSTREAM_ARRIVED_EVENT(System.currentTimeMillis(),uniqueID);
			if(notificationHandler != null)
			{
				notificationHandler.eventOccurred(ne);
			}
            
		}
		/*
		 * write the payload in the inputstream
		 */
		ps.nonblockingWrite(payload,0,payload.length);
	}
	
	
	/**
	 * get the HCastOutputSteam to write bytes to with the specified stream id (integer)
	 * @param sid in string format
	 * @return HCastOutputstream
	 */	  		
	public HCastOutputStream getOutputStream(int sid)
	{
		/*
		 * generate the hash in the stream id format
		 */
		byte[] id = ByteArrayUtility.toByteArray(sid); 
		HCastOutputStream out = (HCastOutputStream)outputtable.get(new String(id));

		if(out==null)
		{
			 out = new HCastOutputStream(this,id);	
			 outputtable.put(String.valueOf(id),out);
		}
		return  out;
	}

	
	/**
	 * get the HCastOutputSteam to write bytes to with the specified stream id (integer)
	 * @param sid in byte array format
	 * @return HCastOutputstream
	 */	  		
/*
	public HCastOutputStream getOutputStream(byte[] sid)
	{
		HCastOutputStream out = (HCastOutputStream)outputtable.get(new String(sid));

		if(out==null)
		{
			 out = new HCastOutputStream(this,sid);	
			 outputtable.put(String.valueOf(sid),out);
		}
		return  out;
	}
*/	
	
	/**
	 * get the HCastOutputSteam to write bytes to with no specified stream id
	 * @return HCastOutputstream
	 */	  		
	public HCastOutputStream getOutputStream() 
	{
		/*
		 * create a random number as the stream id
		 */
		int sid =(new Random (System.currentTimeMillis ())).nextInt();
		
		return getOutputStream(sid);
	}

	
 	/**
 	 * get the available inputstream with specified stream id
 	 * @param sid streamid in stream format
 	 * @return Inputstream
 	 */
	
	public HCastInputStream acceptInputStream(int sid)
	{
		/*
		 * generate the hash in the stream id format
		 */
		byte[] id = ByteArrayUtility.toByteArray(sid); 
		
		HCastInputStream input;
		while((input = (HCastInputStream)(inputtable.get(new String(id))))==null)
		{
			synchronized(guard)
			{
				try
				{
					guard.wait();

//TODO: The following should be logged and/or displayed in debug mode
					//System.out.println("StreamManager:acceptInputStream: obtained new HCastInputStream");
				}
				catch (InterruptedException ie) 
				{
				}

			}
		}

		return input;
	}

	/**
	 * get the available inputstream with specified stream id
	 * @param sid streamid in byte array format
	 * @return Inputstream
	 */
/*	
	public HCastInputStream acceptInputStream(byte[] sid)
	{
		HCastInputStream input;
		while((input = (HCastInputStream)(inputtable.get(new String(sid))))==null)
		{
			synchronized(guard)
			{
				try
				{
					guard.wait();

				}
				catch (InterruptedException ie) 
				{
				}

			}
		}

		return input;
	}
*/

	/**
 	 * get the available inputstream with no specified stream id
 	 * iterate through the hashtable
	 * find and return the first stream that has not been assigned to the application
	 * if no stream or all stream have been assigned, block until new stream arrives
 	 * @return HCastInputstream
 	 */
	
	public HCastInputStream acceptInputStream()
	{
		HCastInputStream is = null;
		/*
		 * wait until the input stream number is not zero
		 */
		int size = inputtable.size();
		while((size = inputtable.size())==0)
		{
			synchronized(guard)
			{
				try
				{
					guard.wait();
				}
				catch (InterruptedException ie) 
				{
				}

			}
		}
		
		/*
		 * find the available stream and return to the application
		 */
		Iterator it = inputtable.keySet().iterator();

		boolean noStreamAvailable = true;
		
		while(it.hasNext())
		{
			String sid = (String)it.next();
			is = (HCastInputStream) inputtable.get(sid);
			if(!is.isAssignedToApplication())
			{
				is.setAssignedFlag(true);
				noStreamAvailable = false;
				return is;
			}
		}			
  		/*
  		 * if no available stream, block here until new stream arrives
  		 */
		while(noStreamAvailable)
		{
			synchronized(guard)
			{
				try
				{
					guard.wait();
				}
				catch (InterruptedException ie) 
				{
				}

			}
			while(it.hasNext())
			{
				String sid = (String)it.next();
				is = (HCastInputStream) inputtable.get(sid);
				if(!is.isAssignedToApplication())
				{
					is.setAssignedFlag(true);
					noStreamAvailable = false;
					return is;
				}
			}			
		}
		
		return is;

	}
	/** The time to periodically check the sending status of each stream */
	public void timerExpired (Object time_id)
	{
		Enumeration e = outputtable.elements();
		while (e.hasMoreElements()) 
		{
			HCastOutputStream so = (HCastOutputStream) e.nextElement();
               
			// if no sending in timeout time...
			if (!so.hasSent()) 
			{
			 try
			 {
			   // flush data.
			   so.flush();
			  }catch (IOException ee){}
			   so.reset();
			}
			
		}	
		/*
		 * set the timer
		 */
		socket.socketAdapter.setTimer(new Socket_Timer_ID(Socket_Timer_ID.typeOfStreamManager, new Integer(1)),sendingPeriod);
		
	}

}



 
 
