/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.adapters;


import java.io.*;
import java.util.*;



/** 
 * This class provides synchronized access for a thread to a TimerEventQueue.
 * It also provides for waiting until an event. 
 *
 * NOTE: Calls to this object use delays, while the times in the TimerEventQueue
 * are absolute!!!
 *
 * @author HyperCast Team
 * @author Dongwen Wang
 * @author Michael Nahas
 * @version	2.0, Feb. 28, 2001
 */

public class Timer {
	/** The queue of events. */
	TimerEventQueue TimerEventQueue;

	/**
	 * Constructs a Timer object which contains a TimerEventQueue.
	 */
	
	public Timer() {
		TimerEventQueue = new TimerEventQueue();
	}	

	
	/**
	 * Adds a TimerEvent to the TimerEventQueue.
	 * @param	timeID		the id of a time event
	 * @param	delay		the time value in ms
	 */

	public synchronized void setTimer(Object timeID, long delay) {
        TimerEventQueue.remove(timeID);

        long time = System.currentTimeMillis() + delay;
        TimerEvent te = new TimerEvent(timeID, time);
		
        if (TimerEventQueue.isEmpty() ||
            TimerEventQueue.getMinEventTime() > time) 
            notify();
        
        TimerEventQueue.insert(te);						

    }


	/** Return delay until a timer goes off.  */
	public synchronized long getTimer(Object timeID) {
        TimerEvent te = TimerEventQueue.getEvent(timeID);

        if (te == null)
            return -1;

        long t = te.getTime();
        long delay = t - System.currentTimeMillis();
        if (delay < 0)
            delay = 0;
		return delay;
	}

	
	
	/**
	 * Removes a TimerEvent from the TimerEventQueue.
	 * @param	timeID		the id of the time event to be removed
	 */
	
	public synchronized void clearTimer(Object timeID) {
        TimerEvent te = TimerEventQueue.remove(timeID);

        if (te != null && !TimerEventQueue.isEmpty() && te.getTime() <= TimerEventQueue.getMinEventTime())
            notify();
    }

	
	
	/**
	 * Returns after a certain amount of time. 
	 */
	
	public synchronized Object sleepUntilNextEvent() throws InterruptedException {
        Object resultID;
        
        while(true) 
        {
            long sleepTime = 10000;
            if (!TimerEventQueue.isEmpty())
                sleepTime = TimerEventQueue.getMinEventTime() - System.currentTimeMillis();

            if (sleepTime > 0) 
            {
                wait(sleepTime);
            }

            if (!TimerEventQueue.isEmpty() && TimerEventQueue.getMinEventTime() <= System.currentTimeMillis()) 
            {
                TimerEvent te = TimerEventQueue.removeTop();
                resultID = te.getTimeID();
                break;
            }
        }
        
        return resultID;
    }


// TEST

public static void main(String args[])
    {
        Timer t = new Timer();
        TimerTestObject to = new TimerTestObject(t);
        Thread th = new Thread(to);
        th.start();

        try { Thread.sleep(1000); } catch (Exception e) {}
        System.out.println("Insert to infinite sleeping thread.");
        System.out.println("Correct response: 0:200");        
        to.resetStartTime();
        t.setTimer(new Integer(0), 200);


        try { Thread.sleep(1000); } catch (Exception e) {}
        System.out.println();
        System.out.println("Insert of two timers first follows second.");
        System.out.println("Correct response: 1:200 0:400");        
        to.resetStartTime();
        t.setTimer(new Integer(0), 400);
        t.setTimer(new Integer(1), 200);


        try { Thread.sleep(1000); } catch (Exception e) {}
        System.out.println();
        System.out.println("Insert of two timers second follows first.");
        System.out.println("Correct response: 0:200 1:400");        
        to.resetStartTime();
        t.setTimer(new Integer(0), 200);
        t.setTimer(new Integer(1), 400);

        try { Thread.sleep(1000); } catch (Exception e) {}
        System.out.println();
        System.out.println("Insert of two timers with same expiration time.");
        System.out.println("Correct response: 0:200 1:200 or 1:200 0:200");        
        to.resetStartTime();
        t.setTimer(new Integer(0), 200);
        t.setTimer(new Integer(1), 200);

        try { Thread.sleep(1000); } catch (Exception e) {}
        System.out.println();
        System.out.println("Insert of two timers, clearing second.");
        System.out.println("Correct response: 0:200");        
        to.resetStartTime();
        t.setTimer(new Integer(0), 200);
        t.setTimer(new Integer(1), 400);
        t.clearTimer(new Integer(1));

        try { Thread.sleep(1000); } catch (Exception e) {}
        System.out.println();
        System.out.println("Insert of two timers, clearing first.");
        System.out.println("Correct response: 1:400");        
        to.resetStartTime();
        t.setTimer(new Integer(0), 200);
        t.setTimer(new Integer(1), 400);
        t.clearTimer(new Integer(0));

        try { Thread.sleep(1000); } catch (Exception e) {}
        to.die();
    }

}	
	

/** Simple object used to test Timer object. */
class TimerTestObject implements Runnable
{
    /** Timer being tested.*/
    private Timer t;

    /** Time at which the current test started.  */
    private long startTime;

    /** True if the function die() has been called. */
    private boolean dying;

    /** Thread performing test. */
    private Thread myThread;
        
    TimerTestObject(Timer timer) 
    {
        t = timer;
        dying = false;
    }

    /** resets start time to current time. */
    void resetStartTime() 
    {
        startTime = System.currentTimeMillis();
    }

    /** Kills <code>myThread</code> indirectly. */
    void die() 
    {
        dying = true;
        myThread.interrupt();
    }
    
    /** Waits until next event and then prints out the ID and time of the event. */
public void run() 
        {
            myThread = Thread.currentThread();
            while (!dying) 
            {
                try 
                {
                   Integer j = (Integer) t.sleepUntilNextEvent();
                    //System.out.println("Index = " + j + "    at displacement: " + (System.currentTimeMillis() - startTime));
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }
        
        
    }
    



