/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.NOTIFICATION_EVENT;
import hypercast.events.NAMING_EVENT;

import hypercast.util.XmlUtil;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Vector;
import java.util.GregorianCalendar;
import java.util.Date;

import java.io.ByteArrayInputStream;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.DataInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.lang.ref.SoftReference;
import java.lang.ref.ReferenceQueue;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Provider;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.Signature;
import java.security.SignatureException;

import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;

import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;

import javax.crypto.Cipher;
import javax.crypto.NoSuchPaddingException;
import javax.security.auth.x500.X500Principal;

import org.apache.xpath.XPath;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

/**
 * <p>This class implements a finite state machine that provides a naming
 * service. In particular, it implements a "getLogicalAddressByName"
 * service in an logical overlay network.</p>
 *
 * <p>This class contains:
 * <ul>
 *   <li>A variety of miscellaneous serices that connect the Naming Service to the other parts of the HyperCast system, in particular, the Message Store.</li>
 *   <li>Various tables that are used to hold naming and trust information.</li>
 *   <li>A hierarchy of message types that are used by the Naming Service protocol.</li>
 *   <li>A hierarchy of classes that implement various timer tasks used by the Naming Service</li>
 * </ul>
 * 
 * @author HyperCast Team
 * @author Greg Mattes
 * @version Spring 2005
 */
public class MessageStoreFSM_Naming
    extends I_MessageStoreFSM
    implements I_Stats
{
    private static final boolean masterDebugSwitch = false;

    private static final boolean debugCertificates                   = true && masterDebugSwitch;
    private static final boolean debugMessages                       = true && masterDebugSwitch;
    private static final boolean debugTimers                         = true && masterDebugSwitch;
    private static final boolean debugSend                           = true && masterDebugSwitch;
    private static final boolean debugPendingQueryTable              = true && masterDebugSwitch;
    private static final boolean debugPeerBindingCache               = true && masterDebugSwitch;
    private static final boolean debugPendingCertificateRequestTable = true && masterDebugSwitch;
    private static final boolean debugBindingsMessage                = true && masterDebugSwitch;
    private static final boolean debugPropagateBindingsMessage       = true && masterDebugSwitch;
    private static final boolean debugBindingQueryMessageControlData = true && masterDebugSwitch;
    private static final boolean debugQueryMessage                   = true && masterDebugSwitch;
    private static final boolean debugCertificateMessages            = true && masterDebugSwitch;
    private static final boolean debugPendingCertificateTable        = true && masterDebugSwitch;
    private static final boolean debugPendingUntrustedQueryTable     = true && masterDebugSwitch;
    private static final boolean debugUserBindingMessage             = true && masterDebugSwitch;

    private final boolean tracingQueries;
    private final int tracingModulo;
    private final boolean traceToStderr;

    /** A table of pending name queries. */
    private final PendingNameQueryTable pendingNameQueryTable = new PendingNameQueryTable();

    /** A table of pending logical address queries. */
    private final PendingLogicalAddressQueryTable pendingLogicalAddressQueryTable = new PendingLogicalAddressQueryTable();

    /** A table of trust chains that are in the process of being built. */
    private final PendingCertificateRequestTable pendingCertificateRequestTable = new PendingCertificateRequestTable();

    /** A table of bindings that satisfy pending queries, but are not yet trusted. */
    private final PendingUntrustedQueryTable pendingUntrustedQueryTable = new PendingUntrustedQueryTable();

    /** A table of names that are used to create NamingBinding objects for the local logical address. */
    private final LocalBindingTable localBindingTable = new LocalBindingTable();

    /** A table of bindings received from peers via push. */
    private final PeerBindingCache peerBindingCachePush;

    /** A table of bindings received from peers via pull. */
    private final PeerBindingCache peerBindingCachePull;

    /** A count of the number of bindings that are cached. */
    private int cachedBindingsCount = 0;

    /** A table of all known certificates. This data structure is always needed to store trust anchors even if not certs received from others are not cached */
    private final CertificateCache certificateCache;

    /** Reference to the OL_Socket that is using this Naming FSM; set in constructor. */
    private final OL_Socket socket;

    /** Reference to the MessageStore that is using this Naming FSM; set in constructor. */
    private final MessageStore messageStore;

    /** Reference to the NotificationHandler that will be used by Naming FSM; set in constructor. */
    private final NotificationHandler notificationHandler;


    /**
     * The security/crypto object that will be to create and verify digital signatures of NamingBinding objects. Mutual
     * exclusive access to this object must be ensured.
     */
    private final Signature digitalSigner;

    /** A statistic, incremented each time a binding is found in the peer binding cache by a query message. */
    private long cacheHits = 0;

    /** A statistic, incremented each time a binding is found in the peer binding cache by a query message. */
    private long queryMessagesProcessed = 0;

    /** A statistic, incremented each time a binding is forced out of the peer cache. */
    private long peerBindingCacheEvictions = 0;

    /** EXPERIMENTS a stat that tells how many pushes remain: used to fill caches initially in experiments */
    private long experimentPushes = 0;

    /** The name by which this particular FSM will be known */
    private String statisticsName;

    /** The object that handle access to the statsitics of this FSM. */
    private final StatsProcessor statsProcessor;

    /** A counter that tracks the number of times that the logical address of the socket using this FSM has changed. */
    private int logicalAddressChangeCounter = 0;

    /* Control message types */
    protected static final byte SET_NAME              = (byte) 0x01;
    protected static final byte UNSET_NAME            = (byte) 0x02;
    protected static final byte INSTALL_CERTIFICATE   = (byte) 0x03;
    protected static final byte LOGICAL_ADDRESS_QUERY = (byte) 0x04;
    protected static final byte NAME_QUERY            = (byte) 0x05;
    protected static final byte LOCAL_NAME_QUERY      = (byte) 0x06;
    protected static final byte PUSH_BINDINGS         = (byte) 0x07;
    protected static final byte PULL_BINDINGS         = (byte) 0x08;
    protected static final byte INVALIDATE_BINDINGS   = (byte) 0x09;
    protected static final byte CERTIFICATE_REQUEST   = (byte) 0x0B; // NB skipping 0x0A to avoid (potential) bug in MessageStore, search for QUERY_SYNC_ALL to find the issue
    protected static final byte CERTIFICATE_RESPONSE  = (byte) 0x0C;

    /* These are types passed by the application that command the FSM to preform an operation. */
    protected static final byte PUSH_OPERATION        = (byte) 0x0E;
    protected static final byte INVALIDATE_OPERATION  = (byte) 0x0F;

    // flag indicating a locally generated message: cannot rely on testing equality of local logical address and source address since logical addresses can change
    protected static final byte DUMMY_BIT          = (byte) (1 << 6);
    
    /** Constants used when binding is received and queries are pending */
    private final static int REJECT_NOT_PENDING       = 0;
    private final static int REJECT_NOT_TRUSTED       = 1;
    private final static int REJECT_NOT_AUTHORITATIVE = 2;
    private final static int REJECT_TOO_OLD           = 3;
    private final static int ACCEPTED                 = 4;

    /** Distinguished value for a deadline that means: no deadline. Used for queries. */
    private final static int WAIT_INDEFINITELY = 0;

    //
    // !NOTE! These constants are need in the Naming API to encode dummy messages. They are *copied* there (not referenced
    // from here because the API is not supposed to be able to access this portion of the system). THE COPIES OF THESE
    // CONSTANTS MUST BE KEPT IN SYNC.
    //

    // byte the identifies the type of message
    private static final int MESSAGE_TYPE_LENGTH        = 1;

    // of local interest only
    private static final int TRUSTED_LENGTH             = 1;
    private static final int DEADLINE_LENGTH            = 8;
    private static final int SERIAL_NUMBER_MATCH_LENGTH = 1;

    // sent on wire (depending on message type)
    private static final int AUTHORITATIVE_LENGTH       = 1;
    private static final int MAXIMUM_AGE_LENGTH         = 4;
    private static final int MAX_RESPONSES_LENGTH       = 4;
    private static final int MINIMUM_HOP_COUNT_LENGTH   = 2;
    private static final int MAXIMUM_HOP_COUNT_LENGTH   = 2;
    /* package */ static final int SERIAL_NUMBER_LENGTH       = 4;

    // end of fixed length fields

    // sent on wire (depending on message type)
    private static final int STRING_LENGTH_BYTES        = 2;
    private static final int CERTIFICATE_LENGTH_BYTES   = 2;

    // of local interest only
    private static final int PRIVATE_KEY_LENGTH_BYTES   = 2;

    // length of fields that are the prefix of query messages sent from peer to peer
    private static final int PEER_QUERY_PREFIX_LENGTH =
    AUTHORITATIVE_LENGTH     +
    MAXIMUM_AGE_LENGTH       +
    MAX_RESPONSES_LENGTH     +
    MINIMUM_HOP_COUNT_LENGTH +
    SERIAL_NUMBER_LENGTH;

    // length of fields that are the prefix of query messages sent from API to FSM
    private static final int LOCAL_QUERY_PREFIX_LENGTH =
    MESSAGE_TYPE_LENGTH        +
    TRUSTED_LENGTH             +
    DEADLINE_LENGTH            +
    SERIAL_NUMBER_MATCH_LENGTH +
    MAXIMUM_HOP_COUNT_LENGTH   +
    PEER_QUERY_PREFIX_LENGTH;

    /** Used in a dummy message to signal that the configured query radius should not be overridden. */
    private static final short USE_CONFIGURED_QUERY_RADIUS = (short) 0;

    /** Created as needed when local bindings are periodically pushed to peers. */
    private PeriodicPushTimerTask periodicPushTimerTask;

    /** The name of the class that provides crypto algorithm implementations. */
    public static final String securityProviderClassName = "org.bouncycastle.jce.provider.BouncyCastleProvider";

    /** The short name of the provider that has crypto algorithm implementations. */
    public static final String securityProviderShortName = "BC";

    private static final Random random = new Random();

    // TODO(EXPERIMENTS)
    final java.text.SimpleDateFormat trace_sdf = new java.text.SimpleDateFormat();

    // TODO(EXPERIMENTS)
    private static final java.util.Timer timer = new java.util.Timer();

    // TODO(EXPERIMENTS): used in experiment that assumes each socket has a single name
    private String localName;

    /**
     * Class that prepends socket logical address to all log message to facilitate ease in debugging in distributed
     * environments.
     */
    class LogicalAddressPrefixPrintWriter
    {
        private PrintWriter pw;

        public LogicalAddressPrefixPrintWriter (final PrintWriter pw) {
            this.pw = pw;
        }

        public void print (final Object object) {
            pw.print (getLogicalAddress() + ": " + object);
            pw.flush();
        }

        public void println (final Object object) {
            print (object + "\n");
        }

        public void println() {
            print ("\n");
        }

        public void close() {
            try { pw.flush(); pw.close(); } catch (Exception ignored) { }
        }
    }

    /** Log used to show informational messages generated by the Naming FSM */
    private LogicalAddressPrefixPrintWriter log;

    //private static final int MILLISECONDS_PER_SECOND = 1000;
    private static final double MILLISECONDS_PER_SECOND = 1000.0;

    /** Object that describes the properties of this type of FSM. */
    private static FiniteStateMachineMetaOperations metaOps;

    /**
     * Called by the Message Store to acquire the meta-ops for this FSM.
     */
    protected static FiniteStateMachineMetaOperations registerMetaOperations () {
        if (metaOps == null)
            metaOps = new NamingMetaOperations();

        return metaOps;
    }

    /**
     * <p>This object is used to separate the policy (i.e. socket
     * configuration) from the mechanism in the naming service.</p>
     *
     * @see NamingPolicy
     */
    private final NamingPolicy policy;

    private HyperCastConfig configuration;

    /**
     * Creates a naming service finite state machine with the
     * specified parameters.
     *
     * @param socket overlay socket associated with this instance of the naming service
     * @param messageStore message store within which this instance of the naming service runs
     * @param config overlay socket config that contains a reference to the log used by this naming service
     */
    public MessageStoreFSM_Naming (final OL_Socket socket, final MessageStore messageStore, final HyperCastConfig config)
    {
        // TODO(EXPERIMENTS)
        trace_sdf.applyPattern ("hh:mm:ss.SSS");

        this.socket = socket;
        this.messageStore = messageStore;
        this.configuration = config;

        this.log = new LogicalAddressPrefixPrintWriter (config.log);

        this.notificationHandler = socket.getNotificationHandler();

        new Thread (new LogicalAddressChangeListener(), "Logical Address Change Listener").start();

        this.policy = NamingPolicy.getInstance (config);

        this.statisticsName = policy.getStatName();

        peerBindingCachePush = new PeerBindingCache (policy.getPeerBindingCachePushMaximumSize());
        peerBindingCachePull = new PeerBindingCache (policy.getPeerBindingCachePullMaximumSize());
        peerBindingCachePush.setOtherCache (peerBindingCachePull);
        peerBindingCachePull.setOtherCache (peerBindingCachePush);

        certificateCache = new CertificateCache (policy.getWriteCertificatesToBackingStore());

        final String digitalSignatureAlgorithm = policy.getDigitalSignatureAlgorithm();

        if (! digitalSignatureAlgorithm.equalsIgnoreCase ("none"))
        {
            if (policy.getPreallocateCrypto())
                CertificateCache.initStaticCryptoFactories();

            CertificateCache.ensureSecurityProviderInstalled (securityProviderShortName, securityProviderClassName);

            try {
                this.digitalSigner = Signature.getInstance (digitalSignatureAlgorithm, securityProviderShortName);
            } catch (final NoSuchProviderException nspe) {
                throw new HyperCastFatalRuntimeException ("Cannot instanciate naming FSM, error when instanciating digital Signature object: " + nspe);
            } catch (final NoSuchAlgorithmException nsae) {
                throw new HyperCastFatalRuntimeException ("Cannot instanciate naming FSM, error when instanciating digital Signature object: " + nsae);
            }
        }
        else
        {
            this.digitalSigner = null;
        }

        this.statsProcessor = new StatsProcessor (this, true, true);

        statsProcessor.addStatsElement ("BindingTable", new BindingTableStat(), 1, 1);
        statsProcessor.addStatsElement ("CertificateCache", new CertificateCacheStat(), 1, 1);
        statsProcessor.addStatsElement ("Pending", new PendingStat(), 1, 1);
        statsProcessor.addStatsElement ("Experiment", new ExperimentStat(), 1, 1);

//        statsProcessor.addStatsElement ("Debug", new DebugNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("SocketsInNetwork", new SocketsInNetworkNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("MaximumResponseBindingAge", new MaximumResponseBindingAgeNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("MaximumResponsesPerQuery", new MaximumResponsesPerQueryNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("BindingTimeout", new BindingTimeoutNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PeerBindingCachePushMaximumSize", new PeerBindingCachePushSizeNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PeerBindingCachePullMaximumSize", new PeerBindingCachePullSizeNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PushPeriod", new PushPeriodNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("BlockingAPITimeout", new BlockingAPITimeoutNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("CertificateRequestTimeout", new CertificateRequestTimeoutNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PushRadius", new PushRadiusNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("InvalidateRadius", new InvalidateRadiusNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("QueryRadius", new QueryRadiusNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("MinimumQueryHopCount", new MinimumQueryHopCountNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("DigitalSignatureAlgorithm", new DigitalSignatureAlgorithmNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("StatName", new StatNameNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("TrustedResponsesOnly", new TrustedResponsesOnlyNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("AuthoritativeResponsesOnly", new AuthoritativeResponsesOnlyNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PushOnSetName", new PushOnSetNameNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PushOnJoinOverlay", new PushOnJoinOverlayNamingStat(), 1, 1);
        statsProcessor.addStatsElement ("PushOnChangeOfLogicalAddress", new PushOnChangeOfLogicalAddressNamingStat(), 1, 1);

        // TODO(EXPERIMENTS)
        statsProcessor.addStatsElement ("ExperimentPushes", new ExperimentPushesStat(), 1, 1);

        statsProcessor.addStatsElement ("BindingCacheHits", new CacheHitsStat(), 1, 1);
        statsProcessor.addStatsElement ("QueryMessagesProcessed", new QueryMessagesProcessedStat(), 1, 1);
        statsProcessor.addStatsElement ("BindingCacheEvictions", new PeerBindingCacheEvictionsStat(), 1, 1);
        statsProcessor.addStatsElement ("BindingCachePushSize", new PeerBindingCachePushSizeStat(), 1, 1);
        statsProcessor.addStatsElement ("BindingCachePullSize", new PeerBindingCachePullSizeStat(), 1, 1);
        statsProcessor.addStatsElement ("BindingCacheSize", new PeerBindingCacheSizeStat(), 1, 1);

        final String traceToStderrString = configuration.getTextAttribute (XmlUtil.createXPath ("/Public/NetworkService/Naming/TraceToStderr"));

        traceToStderr = configuration.booleanFromYesNoString (traceToStderrString);

        tracingQueries = policy.getTracingQueries();
        tracingModulo = policy.getQueryTracingModulo();
    }

    abstract class NamingStat extends SimpleStats
    {
        protected final String statType;

        public NamingStat (final String statType) { this.statType = statType; }

        protected String setStats (final String newValue) throws HyperCastStatsException {
            throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
        }

        public final Element [] getReadSchema (final Document doc, final XPath xpath) throws HyperCastStatsException {
            return XmlUtil.createSchemaElement (doc, statisticsName, statType, null, null);
        }

        public final Element [] getWriteSchema(final Document doc, final XPath xpath) throws HyperCastStatsException {
            return XmlUtil.createSchemaElement (doc, statisticsName, statType, null, null);
        }
    }

    final class BindingTableStat extends NamingStat {
        public BindingTableStat() { super ("xsd:String"); }
        protected String getStats() {
            synchronized (MessageStoreFSM_Naming.this) {
                return localBindingTable + "\nPush:\n" + peerBindingCachePush + "\nPull:\n" + peerBindingCachePull;
            }
        }
    }

    final class CacheHitsStat extends NamingStat {
        public CacheHitsStat() { super ("xsd:Integer"); }
        protected String getStats() {
            return Long.toString (cacheHits);
        }
    }

    final class ExperimentPushesStat extends NamingStat {
        public ExperimentPushesStat() { super ("xsd:Integer"); }
        protected String getStats() {
            return Long.toString (experimentPushes);
        }
    }

    final class PeerBindingCacheEvictionsStat extends NamingStat {
        public PeerBindingCacheEvictionsStat() { super ("xsd:Integer"); }
        protected String getStats() {
            return Long.toString (peerBindingCacheEvictions);
        }
    }

    final class PeerBindingCachePushSizeStat extends NamingStat {
        public PeerBindingCachePushSizeStat() { super ("xsd:Integer"); }
        protected String getStats() {
            synchronized (MessageStoreFSM_Naming.this) {
                return Long.toString (peerBindingCachePush.size());
            }
        }
    }

    final class PeerBindingCachePullSizeStat extends NamingStat {
        public PeerBindingCachePullSizeStat() { super ("xsd:Integer"); }
        protected String getStats() {
            synchronized (MessageStoreFSM_Naming.this) {
                return Long.toString (peerBindingCachePull.size());
            }
        }
    }

    final class PeerBindingCacheSizeStat extends NamingStat {
        public PeerBindingCacheSizeStat() { super ("xsd:Integer"); }
        protected String getStats() {
            return Long.toString (cachedBindingsCount);
        }
    }

    final class QueryMessagesProcessedStat extends NamingStat {
        public QueryMessagesProcessedStat() { super ("xsd:Integer"); }
        protected String getStats() {
            return Long.toString (queryMessagesProcessed);
        }
    }

    final class ExperimentStat extends NamingStat {
        public ExperimentStat() { super ("xsd:String"); }
        protected String getStats() { return experimentRunning ? "Experiment running" : "Experiment not running"; }
        protected String setStats (final String newValue) {
            if (newValue.equalsIgnoreCase ("start"))
                startExperiment (false);
            else if (newValue.equalsIgnoreCase ("stop"))
                stopExperiment (false);
            else if (newValue.equalsIgnoreCase ("pause"))
                pauseExperiment();
            else if (newValue.equalsIgnoreCase ("unpause"))
                unpauseExperiment();
            else if (newValue.equalsIgnoreCase ("startPush"))
                startExperiment (true);
            else if (newValue.equalsIgnoreCase ("stopPush"))
                stopExperiment (true);
            return newValue;
        }
    }

    final class CertificateCacheStat extends NamingStat {
        public CertificateCacheStat() { super ("xsd:String"); }
        protected String getStats() { return certificateCache.toString(); }
    }

    final class PendingStat extends NamingStat {
        public PendingStat() { super ("xsd:String"); }
        protected String getStats() {
            return
                pendingLogicalAddressQueryTable + "\n\n" +
                pendingNameQueryTable           + "\n\n" +
                pendingUntrustedQueryTable      + "\n\n" +
                pendingCertificateRequestTable;
        }
    }

    final class SocketsInNetworkNamingStat extends NamingStat {
        public SocketsInNetworkNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getEXPERIMENTsocketsInNetwork()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setEXPERIMENTsocketsInNetwork (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class MaximumResponseBindingAgeNamingStat extends NamingStat {
        public MaximumResponseBindingAgeNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getMaximumResponseBindingAge()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setMaximumResponseBindingAge (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class MaximumResponsesPerQueryNamingStat extends NamingStat {
        public MaximumResponsesPerQueryNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getMaximumResponsesPerQuery()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setMaximumResponsesPerQuery (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class PeerBindingCachePushSizeNamingStat extends NamingStat {
        public PeerBindingCachePushSizeNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getPeerBindingCachePushMaximumSize()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setPeerBindingCachePushMaximumSize (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class PeerBindingCachePullSizeNamingStat extends NamingStat {
        public PeerBindingCachePullSizeNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getPeerBindingCachePullMaximumSize()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setPeerBindingCachePullMaximumSize (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class PushPeriodNamingStat extends NamingStat {
        public PushPeriodNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getPushPeriod()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setPushPeriod (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class BindingTimeoutNamingStat extends NamingStat {
        public BindingTimeoutNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Integer.toString (policy.getBindingTimeout()); }
        protected String setStats (final String newValue) {
            try {
                int nv = Integer.parseInt (newValue);
                policy.setBindingTimeout (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class BlockingAPITimeoutNamingStat extends NamingStat {
        public BlockingAPITimeoutNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Long.toString (policy.getBlockingAPITimeout()); }
        protected String setStats (final String newValue) {
            try {
                long nv = Long.parseLong (newValue);
                policy.setBlockingAPITimeout (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class CertificateRequestTimeoutNamingStat extends NamingStat {
        public CertificateRequestTimeoutNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Long.toString (policy.getCertificateRequestTimeout()); }
        protected String setStats (final String newValue) {
            try {
                long nv = Long.parseLong (newValue);
                policy.setCertificateRequestTimeout (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class PushRadiusNamingStat extends NamingStat {
        public PushRadiusNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Short.toString (policy.getPushRadius()); }
        protected String setStats (final String newValue) {
            try {
                short nv = Short.parseShort (newValue);
                policy.setPushRadius (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class InvalidateRadiusNamingStat extends NamingStat {
        public InvalidateRadiusNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Short.toString (policy.getInvalidateRadius()); }
        protected String setStats (final String newValue) {
            try {
                short nv = Short.parseShort (newValue);
                policy.setInvalidateRadius (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class QueryRadiusNamingStat extends NamingStat {
        public QueryRadiusNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Short.toString (policy.getQueryRadius()); }
        protected String setStats (final String newValue) {
            try {
                short nv = Short.parseShort (newValue);
                policy.setQueryRadius (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class MinimumQueryHopCountNamingStat extends NamingStat {
        public MinimumQueryHopCountNamingStat() { super ("xsd:integer"); }
        protected String getStats() { return Short.toString (policy.getMinimumQueryHopCount()); }
        protected String setStats (final String newValue) {
            try {
                short nv = Short.parseShort (newValue);
                policy.setMinimumQueryHopCount (nv);
                return newValue;
            } catch (final NumberFormatException nfe) {
                return null;
            }
        }
    }

    final class DigitalSignatureAlgorithmNamingStat extends NamingStat {
        public DigitalSignatureAlgorithmNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.getDigitalSignatureAlgorithm(); }
        protected String setStats (final String newValue) { policy.setDigitalSignatureAlgorithm (newValue); return newValue;}
    }

    final class StatNameNamingStat extends NamingStat {
        public StatNameNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.getStatName(); }
        protected String setStats (final String newValue) { policy.setStatName (newValue); return newValue;}
    }

/*
this is no longer used, specific debug constants are now used
    final class DebugNamingStat extends NamingStat {
        public DebugNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.getDebug() ? "Yes" : "No"; }
        protected String setStats (final String newValue) { policy.setDebug (configuration.booleanFromYesNoString (newValue)); return newValue;}
    }
*/

    final class TrustedResponsesOnlyNamingStat extends NamingStat {
        public TrustedResponsesOnlyNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.getTrustedResponsesOnly() ? "Yes" : "No"; }
        protected String setStats (final String newValue) { policy.setTrustedResponsesOnly (configuration.booleanFromYesNoString (newValue)); return newValue;}
    }

    final class AuthoritativeResponsesOnlyNamingStat extends NamingStat {
        public AuthoritativeResponsesOnlyNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.getAuthoritativeResponsesOnly() ? "Yes" : "No"; }
        protected String setStats (final String newValue) { policy.setAuthoritativeResponsesOnly (configuration.booleanFromYesNoString (newValue)); return newValue;}
    }

    final class PushOnSetNameNamingStat extends NamingStat {
        public PushOnSetNameNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.doPushOnSetName() ? "Yes" : "No"; }
        protected String setStats (final String newValue) { policy.setPushOnSetName (configuration.booleanFromYesNoString (newValue)); return newValue;}
    }

    final class PushOnJoinOverlayNamingStat extends NamingStat {
        public PushOnJoinOverlayNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.doPushOnJoinOverlay() ? "Yes" : "No"; }
        protected String setStats (final String newValue) { policy.setPushOnJoinOverlay (configuration.booleanFromYesNoString (newValue)); return newValue;}
    }

    final class PushOnChangeOfLogicalAddressNamingStat extends NamingStat {
        public PushOnChangeOfLogicalAddressNamingStat() { super ("xsd:string"); }
        protected String getStats() { return policy.doPushOnChangeOfLogicalAddress() ? "Yes" : "No"; }
        protected String setStats (final String newValue) { policy.setPushOnChangeOfLogicalAddress (configuration.booleanFromYesNoString (newValue)); return newValue;}
    }

    /**
     * Implements the algorithm that waits for change of logical address and records the number of changes.
     */
    private class LogicalAddressChangeListener
        implements Runnable
    {
        /*
         * Each time the logical address of the socket changes, update the corresponding counter and push binding
         * information if configured to do so in response to this event.
         */
        public void run()
        {
            while (true)
            {
                notificationHandler.waitUntil_NODE_LOGICALADDRESSCHANGED();

                //log.println ("Detected change of logical address");

                ++logicalAddressChangeCounter;
                
                final boolean pushOnChangeOfLogicalAddress = policy.doPushOnChangeOfLogicalAddress();
                if (pushOnChangeOfLogicalAddress)
                    pushOperation();
            }
        }
    }

    /**
     * This finite state machine takes no action when the neighborhood
     * changes.
     */
    public void changingNeighbor() {
    }

    /**
     * Return the ID of this FSM.
     */
    public short getFSMID() {
        return MessageStore.FSM_NAMING;
    }

    /*
     * This finite state machine uses application data messages as "dummy" messages. They must be converted to naming
     * service control messages before processing.
     *
     * Basically, any message that this FSM receives will be sent to the process() method; payload messages will be
     * converted to control messages first.
     */
    public void newADMsg (final OL_Message message) { newApplicationDataMessage (message); }
    public void newApplicationDataMessage (final OL_Message message) { process (convertPayloadExtensionToControlExtension (message)); }

    public void updateADMsg (final OL_Message message) { updateApplicationDataMessage (message); }
    public void updateApplicationDataMessage (final OL_Message message) { process (convertPayloadExtensionToControlExtension (message)); }

    public void newACMsg (final OL_Message message) { newApplicationControlMessage (message); }
    public void newApplicationControlMessage (final OL_Message message) { process (message); }

    public void updateACMsg (final OL_Message message) { updateApplicationControlMessage (message); }
    public void updateApplicationControlMessage (final OL_Message message) { process (message); }

    private int message_count = 0;

    /**
     * <p>Processes a naming message, the message is assume to be a control message. Any message received by the naming
     * FSM that was an application message is a dummy message which should have been converted to a control
     * message by the caller of this method.</p>
     *
     * @see #convertPayloadExtensionToControlExtension(OL_Message)
     *
     * <p>I believe that it can be assumed that this is called from a context that guarentees mutual exclusive access,
     * i.e. tables can be accessed safely without a lock.</p>
     *
     * Serves as "factory" create method for NamingMessages - this create method is implemented here rather than as a
     * static method of the NamingMessage class due to the details of Java inner classes.
     *
     * @param olMessage The overlay message containing the naming service control message that is being processed.
     */
    private synchronized void process (final OL_Message olMessage)
    {
        final FSM_Extension extension = (FSM_Extension) olMessage.getFirstExtensionByType (Extension.FSM);

        if (extension == null)
            throw new HyperCastWarningRuntimeException ("Message does not have a finite state machine extension.");

        final boolean locallyGeneratedByAPI = testAndUnsetDummyBit (extension);
        final byte [] controlData = extension.getControlData();

        final NamingMessage message;

        final byte type = extension.getType();
        switch (type)
        {
            case SET_NAME:              message = new SetBindingMessage                             (controlData);                        break;
            case UNSET_NAME:            message = new UnsetBindingMessage                           (controlData);                        break;
            case INSTALL_CERTIFICATE:   message = new InstallTrustedCertificateMessage              (controlData);                        break;

            case LOGICAL_ADDRESS_QUERY: message = new LogicalAddressQueryMessage         (olMessage, controlData, locallyGeneratedByAPI); break;
            case NAME_QUERY:            message = new NonLocalNameQueryMessage           (olMessage, controlData, locallyGeneratedByAPI); break;

            case LOCAL_NAME_QUERY:      message = new LocalNameQueryMessage              (olMessage, controlData);                        break;
            case PUSH_BINDINGS:         message = new PushBindingsMessage                (olMessage, controlData);                        break;
            case PULL_BINDINGS:         message = new PullBindingsMessage                (olMessage, controlData);                        break;
            case INVALIDATE_BINDINGS:   message = new InvalidateBindingsMessage          (olMessage, controlData);                        break;
            case CERTIFICATE_RESPONSE:  message = new CertificateResponseMessage         (olMessage, controlData);                        break;
            case CERTIFICATE_REQUEST:
            {
                final byte deliveryMode = olMessage.getDeliveryMode();
                switch (deliveryMode)
                {
                    case OL_Message.DELIVERY_MODE_MULTICAST:
                        //case OL_Message.DELIVERY_MODE_FLOOD:
                        message = new CertificateRequestMessageBroadcast (olMessage, controlData);
                        break;

                    case OL_Message.DELIVERY_MODE_UNICAST:
                        message = new CertificateRequestMessageUnicast (olMessage, controlData);
                        break;

                    default: throw new HyperCastFatalRuntimeException ("Invalid delivery mode encounted in Naming Message: " + type);
                }
            }
            break;

            case PUSH_OPERATION:       pushOperation();       return;
            case INVALIDATE_OPERATION: invalidateOperation(); return;

            default: throw new HyperCastWarningRuntimeException ("Naming FSM: Message received with unknown control type: " + type);
        }

        if (message != null)
            process (message);
    }

    /**
     * Calls the particular process method of each message based on type. This method is called by timer tasks as well when
     * they generate a message that is to be processed. Basically this method exists so that the lock for the FSM can be
     * acquired before processing a message or a timer event to ensure mutual exclusivity.
     */
    private synchronized void process (final NamingMessage message)
    {
        if (debugMessages)
            log.println (message);

        try {
            message.process();
        }
        catch (Exception e) {
            System.err.println ("Exception while processing message: " + e);
            e.printStackTrace (System.err);
        }
    }

    /**
     * Determine if this is a locally generated message extension, i.e. "dummy" message, then make the dummy bit zero
     * (i.e. turn it off) so that it does not interfere when switching on message type in <code>process()</code>.
     *
     * @see #process(OL_Message)
     */
    private boolean testAndUnsetDummyBit (final FSM_Extension extension)
    {
        byte type = extension.getType();
        final boolean test = (type & DUMMY_BIT) != 0;
        type &= ~DUMMY_BIT; // unset
        extension.setType (type); // unset in extension
        return test;
    }

    /**
     * @return a count of the number of times that the logical address has changed.
     */
    public int getLogicalAddressChangeCounter() {
        return logicalAddressChangeCounter;
    }
    
    /**
     * Delegated method call to get local logical address.
     */
    public I_LogicalAddress getLogicalAddress() {
        return socket.getLogicalAddress();
    }

    /**
     * No longer used, is has been decided that multicasting is the preferred way to send a broadcast message for this FSM.
     */
    /*
    private void sendFlood (final OL_Message olMessage)
    {
        if (debugSend)
            log.println ("NamingFSM forwarding message via flooding.");

        olMessage.setDeliveryMode (OL_Message.DELIVERY_MODE_FLOOD);
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        olMessage.setSourceAddress (myLogicalAddress);
        olMessage.setPreviousHopAddress (myLogicalAddress);

        if (olMessage.getHopLimit() != 0)
            socket.forwardFlood (olMessage);
        else {
            if (false) {
                log.println ("Attempt to send a message with a zero length hop limit");
                Thread.dumpStack();
            }
        }
    }
    */

    /**
     * See the comment in sendToAll(OL_Message) for an explanation of why OL_Socket.forwardToChildren() is not called
     * directly.
     *
     * Send a broadcast message. Does not reset the source address. Drop message if it has reached its hop limit.
     */
    private void forwardToChildren (final OL_Message olMessage)
    {
        if (olMessage.getHopLimit() != 0)
        {
            //socket.forwardFlood (olMessage);
            socket.forwardToChildren (olMessage);
        }
        else
        {
            //log.println ("Naming Service FSM: Attempt to send a broadcast message with a zero length hop limit - dropping");
            // drop message
        }
    }

    /**
     * This method is an adapter that compensates for the interface provided by <code>OL_Socket</code>. Ideally,
     * <code>OL_Socket.sendToAll()</code> would be called directly by the Naming FSM, but it cannot be called from here
     * since it would send this message back to the message store. <code>OL_Socket.sendToAll()</code> calls
     * <code>checkForMessageStoreSend()</code> which would always return true for any naming message. Instead, the code
     * for <code>OL_Socket.sendToAll()</code> must be duplicated here.
     *
     * Maybe this should be implemented by the message store so that all FSMs can take advantage of it.
     *
     * @param olMessage The message containing a naming FSM control extension that will be forwarded to neighbors.
     *
     * @see OL_Socket#sendToAll(I_OverlayMessage)
     * @see OL_Socket#forwardToChildren(OL_Message)
     */
    private void sendToAll (final OL_Message olMessage)
    {
        if (debugSend)
            log.println ("NamingFSM forwarding message via multicast.");

        olMessage.setDeliveryMode (OL_Message.DELIVERY_MODE_MULTICAST);
        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        olMessage.setSourceAddress (myLogicalAddress);
        olMessage.setPreviousHopAddress (myLogicalAddress);

        forwardToChildren (olMessage);
    }

    /**
     * See the comment in sendToAll(OL_Message) for an explanation of why OL_Socket.forwardToParent() is not called
     * directly.
     *
     * Send a unicast message. Does not reset the source address. Drop message if it has reached its hop limit.
     */
    private void forwardToParent (final OL_Message olMessage)
    {
        if (olMessage.getHopLimit() != 0)
            socket.forwardToParent (olMessage);
        else {
            //log.println ("Naming Service FSM: Attempt to forward a unicast message with a zero length hop limit - dropping");
            // drop message
        }
    }

    /**
     * See the comment in sendToAll(OL_Message) for an explanation of why OL_Socket.sendToAll() is not called
     * directly.
     *
     * Send a unicast message. Use this logical address as the source address.
     */
    private void sendToNode (final OL_Message olMessage, final I_LogicalAddress destinationAddress)
    {
        olMessage.setDeliveryMode (OL_Message.DELIVERY_MODE_UNICAST);

        final I_LogicalAddress myLogicalAddress = getLogicalAddress();
        olMessage.setSourceAddress (myLogicalAddress);
        olMessage.setPreviousHopAddress (myLogicalAddress);
        olMessage.setDestinationAddress (destinationAddress);

        forwardToParent (olMessage);
    }

    /*
     * @see #sendToNode(OL_Message,I_LogicalAddress)
     */
    private void sendToNode (final OL_Message olMessage, final String logicalAddressString) {
        sendToNode (olMessage, socket.createLogicalAddress (logicalAddressString));
    }

    /**
     * Returns the short name of a full inner class name, that is the name after the dollar sign character. This is
     * useful in many toString methods in this FSM.
     */
    private String shortNameOfClass (final String fullName) {
        return fullName.substring (fullName.indexOf ('$') + 1);
    }

    /**
     * Uses a dummy message with data in an application payload
     * extension to create a proper naming control message with
     * information in the control data portion of the FSM extension.
     *
     * Only the first character of the payload array is interpretted
     * to determine the type of naming message.
     *
     * @return the naming control message.
     */
    private OL_Message convertPayloadExtensionToControlExtension (final OL_Message message)
    {
        final PayloadExtension payloadExtension = (PayloadExtension) message.getFirstExtensionByType (Extension.RAW_DATA);
        final byte [] payload = payloadExtension.getPayload();

        final short hopLimit = message.getHopLimit();

        final byte typePrefix = payload[0];
        final int TYPE_PREFIX_LENGTH = 1;

        byte type = DUMMY_BIT;

        switch (typePrefix)
        {
            case Naming.MESSAGE_TYPE_SET_NAME:              type |= SET_NAME;              break;
            case Naming.MESSAGE_TYPE_SET_CERT:              type |= SET_NAME;              break;
            case Naming.MESSAGE_TYPE_SET_NAME_DELEGATE:     type |= SET_NAME;              break;
            case Naming.MESSAGE_TYPE_SET_CERT_DELEGATE:     type |= SET_NAME;              break;
            case Naming.MESSAGE_TYPE_UNSET_NAME:            type |= UNSET_NAME;            break;
            case Naming.MESSAGE_TYPE_UNSET_CERT:            type |= UNSET_NAME;            break;
            case Naming.MESSAGE_TYPE_NAME_QUERY:            type |= NAME_QUERY;            break;
            case Naming.MESSAGE_TYPE_LOCAL_NAME_QUERY:      type |= LOCAL_NAME_QUERY;      break;
            case Naming.MESSAGE_TYPE_LOGICAL_ADDRESS_QUERY: type |= LOGICAL_ADDRESS_QUERY; break;
            case Naming.MESSAGE_TYPE_INSTALL_CERT:          type |= INSTALL_CERTIFICATE;   break;
            case Naming.MESSAGE_TYPE_PUSH:                  type |= PUSH_OPERATION;        break;
            case Naming.MESSAGE_TYPE_INVALIDATE:            type |= INVALIDATE_OPERATION;  break;

            default:
                throw new HyperCastFatalRuntimeException ("Unknown type encoding in naming channel: " + typePrefix);
        }

        final short fsmid = MessageStore.FSM_NAMING;
        final byte [] streamID = FSM_Extension.NAMING_STREAM_ID;

        final FSM_Extension controlExtension = new FSM_Extension (fsmid, type, streamID, payload);

        // if I could simply replace the FSM extension in existing message could reuse message memory instead of creating a
        // new OL_Message object OL_Message.removeAllExtensions() ?

        final OL_Message controlMessage = (OL_Message) socket.createMessage();
        controlMessage.setHopLimit (hopLimit);
        controlMessage.addExtension (controlExtension);

        return controlMessage;
    }

    /**
     * Sends a Push Message that contains all local bindings. Does not send anything if there are no local bindings.
     */
    public void pushOperation()
    {
        final NamingBinding [] localBindings = localBindingTable.getBindings();

        if (localBindings.length > 0)
            new PushBindingsMessage (localBindings).send();
    }

    /**
     * Sends a Push Message that contains one binding. Does not send anything if there is no local binding that matches the specified name.
     * @param name The name that should be pushed, if it is in the local binding table
     */
    public void pushOperation (final String name)
    {
        final NamingBinding localBinding = localBindingTable.get (name);

        if (localBinding != null)
            new PushBindingsMessage (new NamingBinding [] { localBinding }).send();
    }

    /**
     * Sends an Invalidate Message that contains all local bindings. Does not send anything if there are no local bindings.
     */
    public void invalidateOperation()
    {
        final BindingInvalidation [] invalidations = localBindingTable.getInvalidations();

        if (invalidations.length > 0)
            new InvalidateBindingsMessage (invalidations).send();
    }

    /**
     * Sends a Push Message that contains all local bindings. Does not send anything if there are no local bindings.
     */
    public void invalidateOperation (final String name)
    {
        final BindingInvalidation invalidation = localBindingTable.getInvalidation (name);

        if (invalidation != null)
            new InvalidateBindingsMessage (new NamingBinding [] { invalidation }).send();
    }

    /**
     * @param obj The object that is to be associated with this timer event, available when event occurs
     * @param delay milliseconds until timer expires
     * @deprecated
     */
    public void setTimer (final Object obj, final long delay) {
        messageStore.setTimer (this, obj, delay);
    }

    /**
     * Method that is called when a timer expires
     */
    public void timerExpired (final Object object)
    {
        if (object instanceof NamingTimerTask)
        {
            try
            {
                final NamingTimerTask timerTask = (NamingTimerTask) object;
                timerTask.run();
            }
            catch (final ClassCastException cce)
            {
                log.println ("Unknown type in timer expired: " + cce);
            }
            catch (Exception e)
            {
                log.println ("NamingFSM: Timer expired exception:");
                e.printStackTrace (System.err);
            }
        }
    }

    /**
     * Parent of all naming timer task objects. This class defines a run method that is executed when a timer event
     * occurs.
     */
    private abstract class NamingTimerTask
    {
        public abstract void run();
    }

    // TODO(EXPERIMENTS)
    private final class MetaRandomLAQueryTimerTask
        extends java.util.TimerTask
    {
        private final RandomLAQueryTimerTask randomLAQueryTimerTask;
        public MetaRandomLAQueryTimerTask (final RandomLAQueryTimerTask randomLAQueryTimerTask) {
            this.randomLAQueryTimerTask = randomLAQueryTimerTask;
        }

        public void run() {
            randomLAQueryTimerTask.run();
        }
    }

    /* *
     *  TODO(EXPERIMENTS)
     *
     * Query rate controller
     *
     * us means microseconds, ms means milliseconds
     */
    private final class RandomLAQueryTimerTask
       extends NamingTimerTask
      //    extends java.util.TimerTask
    {
        private boolean debugRateController = false;

        private final double period;       // average period in (us), each query has a uniform random amount of time from the previous
        private final int socketsInNetworkCount;


        // the time in ms at which the previous timer actually expired, 0 means that no timer has been set yet.
        // reset for a pause
        private long previousMillis = 0;

        private long configuredDelay; // the time that *should* occur between when the previous query and the current query (us)

        private final boolean doNamingQueries; // if false, regular overlay messages will be sent instead of Naming Service queries.
        private final OL_Message helloMessage; // non-Naming Service message

        private int queriesSent = 0;
        private long experimentStartTime; // ms

        // this value was found empirically
        private final int javaClockResolution = 30000; // us

        // limits the amount of messages that can be sent in one burst
        private double maxSend;

        // absolute limit of queries in single burst
        private final int sendLimit;
        private /*final*/ int maximumQuerySerialNumber;

        // The default query string, used in networks where this is only one name (linear experiments)
        private final String allZeroQueryString = zeroPaddedString (0, 4);

        public RandomLAQueryTimerTask (final double period, final int socketsInNetworkCount)
        {
            this.period = period * 1000.0; // ms -> us
            if (debugRateController)
                experimentOut.println (getLogicalAddress() + ": " + "period (us): " + this.period);
            this.socketsInNetworkCount = socketsInNetworkCount;
            this.helloMessage = (OL_Message) socket.createMessage ("-Hello Naming!-".getBytes());
            this.doNamingQueries = policy.getEXPERIMENTdoNamingQueries();
            this.maxSend = 1.0;
            this.sendLimit = 3000;
//            this.maximumQuerySerialNumber = policy.getEXPERIMENTQueryCount() / socketsInNetworkCount;

            this.maximumQuerySerialNumber =
                policy.getEXPERIMENTQueryCount() / socketsInNetworkCount;

            final long leftover = policy.getEXPERIMENTQueryCount() % socketsInNetworkCount;
            final double odds = ((double) leftover) / socketsInNetworkCount;
            if (random.nextDouble() < odds)
                ++maximumQuerySerialNumber;

            if (debugRateController)
                experimentOut.println (getLogicalAddress() + ": " + "maximumQuerySerialNumber: " + this.maximumQuerySerialNumber);
        }
        
        private boolean paused = false;
        private boolean triedToRunWhilePaused = false;
        private long pauseTime = 0;

        private long totalConfiguredTime = 0; // us

        public synchronized void pause() {
            log.println ("pausing after sending " + queriesSent + " queries");
            //           log.println (pendingLogicalAddressQueryTable);
            paused = true;
            previousMillis = 0;
            pauseTime = System.currentTimeMillis();
        }

        public synchronized void unpause(){
            pauseTime = System.currentTimeMillis() - pauseTime;
            log.println ("unpausing after " + (pauseTime / 1000.0) + " seconds. So far, have sent " + queriesSent + " queries");
//            log.println (pendingLogicalAddressQueryTable);
            paused = false;
            if (triedToRunWhilePaused)
                run();
            triedToRunWhilePaused = false;
//            debugRateController = true;
        }

        public synchronized void run()
        {
            if (randomLAQueryTimerTask != null
                &&
                randomLAQueryTimerTask == this)
            {
                if (paused) {
                    triedToRunWhilePaused = true;
                    return;
                }

                if (maximumQuerySerialNumber == 0) {
                    randomLAQueryTimerTask = null;

/*
                    synchronized (experimentOutLock)
                    {
                        experimentOut.println (getLogicalAddress() + ": " +
                                               "FINAL " +
                                               " queriesSent: " + queriesSent);
                    }
*/

                    return;
                }

                final boolean isFirstQuery;

                final long now = System.currentTimeMillis();

                final long delayError; // us

                if (previousMillis == 0)
                {
                    if (pauseTime == 0) {
                        experimentStartTime = now;
                        if (debugRateController)
                            experimentOut.println (getLogicalAddress() + ": " + "experiment starting at: " + now);
                    }

                    delayError = 0;

                    isFirstQuery = true;
                    previousMillis = now;
                }
                else
                {
                    // We're not dealing with a real-time clock here. If we say wait X milliseconds, we might actually
                    // wait Y milliseconds where Y > X (can X > Y?). The difference between Y and X is the amount of
                    // time that we have to "catch up"
                    
                    final long actualDelay = now - previousMillis;
//                    delayError = actualDelay * 1000 - configuredDelay;

                    delayError = (now * 1000) - (experimentStartTime * 1000 + totalConfiguredTime);

                    if (debugRateController)
                        experimentOut.println (getLogicalAddress() + ": " +
                                               "configuredDelay (us): " + configuredDelay +
                                               " (s): " + (configuredDelay / 1000.0 / 1000) +
                                               " actualDelay (ms): " + actualDelay +
                                               " delayError (us): " + delayError);

                    isFirstQuery = false;
                    previousMillis = now;
                }

                int queryCount = isFirstQuery ?
                    ((socketsInNetworkCount > 1) ? 0 : ((int) (javaClockResolution / period)))
                    : 0;

                if (socketsInNetworkCount == 1)
                    configuredDelay = 20000;


//                configuredDelay = Math.max (configuredDelay, 20000);

                if (queryCount == 0 && socketsInNetworkCount == 1)
                {
                    long experimentElapsed = (now - experimentStartTime - pauseTime) * 1000; // us
                    if (experimentElapsed < 0 && debugRateController)
                        experimentOut.println (getLogicalAddress() + ": " +
                                               "experimentElapsed (us): " + experimentElapsed +
                                               " now: " + now +
                                               " experimentStartTime: " + experimentStartTime +
                                               " pauseTime: " + pauseTime);
                        
                    long intervalCompleted = (long) (queriesSent * period);
                    long intervalToDo = experimentElapsed - intervalCompleted;
                    queryCount = (int) (intervalToDo / period);

                    if (debugRateController)
                        experimentOut.println (getLogicalAddress() + ": " +
                                               "experimentElapsed (us): " + experimentElapsed +
                                               " intervalCompleted (us): " + intervalCompleted +
                                               " intervalToDo (us): " + intervalToDo +
                                               " delta (us): " + ((now - previousMillis) * 1000) +
                                               " queriesSent: " + queriesSent +
                                               " queryCount: " + queryCount);
                }
                else if (socketsInNetworkCount > 1)
                {
                    if (delayError < 0) // timer went off too soon
                        configuredDelay = -delayError;
                    else
                    {
                        queryCount = 1;
                        long delayError2 = delayError; // us
                        long nextDelay = Math.abs (random.nextLong() % ((long) (period * 2))); // us
//                        if (debugRateController)
//                            experimentOut.println (getLogicalAddress() + ": " + "nextDelay (us):" + nextDelay);
                        
                        totalConfiguredTime += nextDelay;
                        while (nextDelay < delayError2)
                        {
                            ++queryCount;
                            delayError2 -= nextDelay;
                            nextDelay = Math.abs (random.nextLong() % ((long) (period * 2)));
                            totalConfiguredTime += nextDelay;
                        }
                        configuredDelay = nextDelay - delayError2;
                    }

//                    configuredDelay = (long) (0.5 * configuredDelay);

                    final long experimentElapsed = now - experimentStartTime - pauseTime; // ms

                    if (debugRateController)
                        experimentOut.println (getLogicalAddress() + ": " +
//                                               ((queriesSent + queryCount >= maximumQuerySerialNumber) ? "FINAL " : "") +
                                               "experimentElapsed (ms): " + experimentElapsed +
                                               " qps: " + (queriesSent / (experimentElapsed / 1000.0)) +
//                                               " delta (us): " + ((now - previousMillis) * 1000) +
                                               " configured delay (s): " + (configuredDelay / 1000.0 / 1000) +
                                               " queriesSent: " + queriesSent +
                                               " queryCount: " + queryCount);
                }

                // went off too soon!
/*
                if (delayError < 0)
                {
                    configuredDelay = -delayError;
                }
                else
                {
                    final long previousFrac = delayError % ((long) period);
                    
                    //experimentOut.println ("queryCount: " + queryCount + " previousFrac: " + previousFrac);

                    //configuredDelay = Math.abs (random.nextLong()) % ((long) (period * 2)) - delayError; // can be negative! that's ok!
                    //configuredDelay = ((long) period) - delayError; // can be negative! that's ok!

                    configuredDelay = ((long) period) - previousFrac; // can be negative! that's ok!
                }
*/

                configuredDelay = Math.max (1000, configuredDelay); // if we want to wait, we should wait

//                setTimer (this, configuredDelay / 1000);
                timer.schedule (new MetaRandomLAQueryTimerTask (this), configuredDelay / 1000);

                if (doNamingQueries)
                {
                    final long sendStart = System.currentTimeMillis();

                    final long deadline = policy.getBlockingAPITimeout() + sendStart;

                    // default query string; this value is used in Linear networks in which there is only one name
                    String nextQueryString = allZeroQueryString;

//                    queryCount = Math.min ((int) maxSend, queryCount);
                    queryCount = Math.min (sendLimit, queryCount);

                    if (debugRateController)
                        experimentOut.println (getLogicalAddress() + ": " + "queryCount is: " + queryCount);

                    final int queryCountSave = queryCount;

                    while (queryCount-- > 0)
                    {
                        if (socketsInNetworkCount > 1) // GRID of nodes; randomly select from all names
                        {
                            // ensures that local name is never queried (is this wise? should I query for myself?)
                            do {
                                nextQueryString = zeroPaddedString (Math.abs (random.nextInt()) % socketsInNetworkCount, 4);
                            } while (localName != null && localName.equals (nextQueryString));
                        }

                        final LogicalAddressQueryMessage logicalAddressQueryMessage = new LogicalAddressQueryMessage (getNextQuerySerialNumber(), nextQueryString, deadline);

                        process (logicalAddressQueryMessage);

                        ++queriesSent;

                        if (queriesSent >= maximumQuerySerialNumber)
//                            && socketsInNetworkCount == 1)
                        {
                            if (debugRateController)
                                experimentOut.println (getLogicalAddress() + ": " + "setting randomLAQueryTimerTask = null");

                            randomLAQueryTimerTask = null;
/*

                            final long experimentElapsed = System.currentTimeMillis() - experimentStartTime - pauseTime; // ms

                            String laString = "";
                            try { laString = getLogicalAddress().toString() + ": "; }
                            catch (NullPointerException ignored) {laString = "";}
                            synchronized (experimentOutLock)
                            {
                                experimentOut.println (laString +
                                                       "FINAL " +
                                                       " qps: " + (queriesSent / (experimentElapsed / 1000.0)) +
                                                       " queriesSent: " + queriesSent);
                            }
*/

                            break;
                        }

                    }
                    final long sendFinish = System.currentTimeMillis();

                    if (debugRateController)
                        experimentOut.println (getLogicalAddress() + ": " + "sent " + (queryCountSave - (queryCount + 1)) + " queries in " + (sendFinish - sendStart) + " ms");
                }
                else
                {
                    sendToAll (helloMessage); // non-naming broadcast traffic
                }
//                maxSend *= 1.05;
            }
        }

        private String zeroPaddedString (int origint, final int length)
        {
            if (origint < 0) { origint = -origint; }
            final String orig = Integer.toString (origint);
            final int numOfLeadingZeros = length - orig.length();
            if (numOfLeadingZeros < 0) { throw new RuntimeException ("origint: " + origint + " length: " + length + " orig.length(): " + orig.length()); }
            final StringBuffer stringBuffer = new StringBuffer (length);
            for (int j = 0; j < length; ++j) { stringBuffer.append ('0'); }
            for (int j = numOfLeadingZeros; j < length; ++j) { stringBuffer.setCharAt (j, orig.charAt (j - numOfLeadingZeros)); }
            return stringBuffer.toString();
        }
    }

    /**
     * Timer task that handles time outs of entries in a pending query table. This type of timer task is used to handle
     * both pending name queries and pending logical address queries.
     */
    private final class PendingQueriesTimerTask
    // TODO(EXPERIMENTS): timer
    //extends java.util.TimerTask
        extends NamingTimerTask
    {
        final String query;
        final int serialNumber;
        final PendingQueryTable pendingQuerytable;
        
        public PendingQueriesTimerTask (final String query, final int serialNumber, final PendingQueryTable pendingQuerytable)
        {
            this.query = query;
            this.serialNumber = serialNumber;
            this.pendingQuerytable = pendingQuerytable;
        }

        public void run()
        {
            if (debugTimers)
                log.println ("*** PendingQueriesTimerTask - query (" + query + ") ***");

            pendingQuerytable.remove (query, serialNumber);
        }

        public boolean equals (final Object object)
        {
            final boolean result;
            if (! (object instanceof PendingQueriesTimerTask))
                 result = false;
            else
            {
                final PendingQueriesTimerTask other = (PendingQueriesTimerTask) object;
                result =
                    this.pendingQuerytable == other.pendingQuerytable
                    &&
                    this.query.equals (other.query)
                    &&
                    this.serialNumber == other.serialNumber;
            }
            return result;
        }
    }

    /**
     * Timer task that handles time outs of certificate request messages.
     */
    private abstract class CertificateRequestTimerTask
    // TODO(EXPERIMENTS): timer
    //extends java.util.TimerTask
        extends NamingTimerTask
    {
        protected final String pendingCertificateName;
        protected final int serialNumber;

        public CertificateRequestTimerTask (final int serialNumber, final String pendingCertificateName) {
            this.pendingCertificateName = pendingCertificateName;
            this.serialNumber = serialNumber;
        }

        protected void sendNextMessage() { }

        protected void doPendingTimeout() { }

        public final void run()
        {
            if (debugTimers)
                log.println ("*** " + shortNameOfClass (getClass().getName()) + " - (" + pendingCertificateName + ") ***");

            final boolean continueSearch = pendingCertificateRequestTable.timeout (serialNumber, pendingCertificateName);
            if (continueSearch)
                sendNextMessage();
            else
                doPendingTimeout();
        }
    }

    /**
     * Timer task that handles time outs of unicast certificate request messages. A broadcast message is sent if the
     * unicast message times out.
     */
    private final class CertificateRequestUnicastTimerTask
        extends CertificateRequestTimerTask
    {
        public CertificateRequestUnicastTimerTask (final int serialNumber, final String pendingCertificateName) {
            super (serialNumber, pendingCertificateName);
        }

        protected void sendNextMessage() {
            new CertificateRequestMessageBroadcast (serialNumber, pendingCertificateName).send();
        }
    }

    /**
     * Timer task that handles time outs of broadcast certificate request messages.
     */
    private final class CertificateRequestBroadcastTimerTask
        extends CertificateRequestTimerTask
    {
        public CertificateRequestBroadcastTimerTask (final int serialNumber, final String pendingCertificateName) {
            super (serialNumber, pendingCertificateName);
        }

        protected void doPendingTimeout() {
            pendingUntrustedQueryTable.remove (serialNumber, pendingCertificateName);
        }
    }

    /**
     * Timer task that handles time outs of entries in the peer bindings cache.
     */
    private final class PeerBindingsTimerTask
        extends NamingTimerTask
    {
        private long period;
        private PeerBindingCache peerBindingCache;

        public PeerBindingsTimerTask (final long period, final PeerBindingCache peerBindingCache) {
            this.period = period;
            this.peerBindingCache = peerBindingCache;
        }

        public void run()
        {
            if (debugTimers)
                log.println ("*** PeerBindingsTimerTask - period (" + (period / MILLISECONDS_PER_SECOND) + " seconds) ***");

            synchronized (MessageStoreFSM_Naming.this) {
                peerBindingCache.removeExpiredEntries();
            }

            setTimer (this, period);
        }
    }

    /**
     * Timer task that handles time outs of periodic push intervals.
     */
    private final class PeriodicPushTimerTask
        extends NamingTimerTask
    {
        private long pushPeriod;

        public PeriodicPushTimerTask (final long pushPeriod) {
            this.pushPeriod = pushPeriod;
        }

        public void run()
        {
            if (debugTimers)
                log.println ("*** PeriodicPushTimerTask - period (" + (pushPeriod / MILLISECONDS_PER_SECOND) + " seconds) ***");

            pushOperation();

            // TODO(EXPERIMENTS)
/*
            if (experimentPushes > 0)
                --experimentPushes;

            if (experimentPushes == 0)
                periodicPushTimerTask = null;
*/

            if (periodicPushTimerTask != null)
                setTimer (this, pushPeriod);
        }
    }

    /**
     * Method that is called when a timer expires - no longer used as timers have been generalized to use objects
     * instead of ints.
     */
    public void timerExpired (final int timerIndex)
    {
        throw new HyperCastWarningRuntimeException ("Old style timer expired call");
    }

    /**
     * A table that handles pending queries of either logical addresses or names. This class defines the structure that
     * is common to a table of pending logical address queries or name queries; child classes are defined for each
     * particular type of query. This means that there must be two tables in the system, one for pending logical address
     * queries, and another for pending name queries.
     *
     * Pending logical address queries and pending name queries are kept in separate tables because any single binding that
     * is received could fail to satisfy a pending logical address query in one way, and fail to satisfy a pending name query
     * in a different way. For example, if there is a pending name query for logical address Y that requires authoritative
     * responses and a pending logical address query for name X that requires trusted responses, the receipt of an
     * unauthoritative, untrusted binding of <X,Y> will fail in two ways. The caller that passes a binding to this table to
     * ask, "does this binding satisfy a query, and if not, why not?" would have to receive two answers: not authoritative
     * and not trusted. Returning two answers is not simple, so two separate tables are used.
     *
     * Currently, entries in this table are for queries that originate at the local socket only. There are no "proxy"
     * queries, that is, queries that are being made on behalf of a peer. Such proxy queries would be useful in culling
     * unwanted or redundant query responses at the cost of added state across the network. Note that in the presence
     * of proxy queries a single binding could fail to satisfy many logical address queries or many name queries (or
     * both) requiring that multiple types of failure be returned to the caller.
     *
     * Note that the caller only takes action if the binding is accepted by some query or if it is untrusted (and a
     * trusted response is needed). In some sense only two bits need to be returned to the caller. Each type of failure
     * could be a bit in a bit vector.
     */
    private abstract class PendingQueryTable
    {
        private final class Entry
        {
            /** Unique query serial number. Name and LA queries use same S/N space. Cert requests/replys use a different S/N space. */
            private Integer serialNumberInteger;

            /** true if a satisfying response must be trusted. false means that trusted or untrusted responses can be satisfactory. */
            private final boolean requiresTrusted;

            /** true if a satisfying response must be authoritative. false means that authoritative or non-authoritative responses can be satisfactory. */
            private final boolean requiresAuthoritative;

            /** Interpreted as the difference between timestamp in binding and "now," i.e. received (in particular not how old it was when it was sent) */
            private final int maxAge;

            /** When this number of responses is received the query will no longer be pending, decremented on each satisfying receipt. Responses remaining. */
            private int responsesRemaining;

            /** Absolute time (since epoch) in ms when this query will no longer be valid */
            private final long deadline;

            /** Absolute time (since epoch) in ms when this query began */
            private final long startTime;

            /** @see PendingQueryTable#put(String, boolean, boolean, int, int, long, int) */
            public Entry (final Integer serialNumberInteger,
                          final boolean requiresTrusted,
                          final boolean requiresAuthoritative,
                          final int     maxAge,
                          final int     maxResponses,
                          final long    deadline,
                          final long    startTime)
            {
                this.serialNumberInteger = serialNumberInteger;
                this.requiresTrusted = requiresTrusted;
                this.requiresAuthoritative = requiresAuthoritative;
                this.maxAge = maxAge;
                this.responsesRemaining = maxResponses;
                this.deadline = deadline;
                this.startTime = startTime;
            }

            public Integer getSerialNumber() {
                return serialNumberInteger;
            }

            public String toString()
            {
                final long now = System.currentTimeMillis();

                final String deadlineString = ((deadline == WAIT_INDEFINITELY) ? "indefinitely many" : Double.toString ((deadline - now)/MILLISECONDS_PER_SECOND));
                
                return
                    "serialNumber"          + " (" + serialNumberInteger                           + "), "          +
                    "requiresTrusted"       + " (" + requiresTrusted                               + "), "          +
                    "requiresAuthoritative" + " (" + requiresAuthoritative                         + "), "          +
                    "maxAge"                + " (" + (maxAge / MILLISECONDS_PER_SECOND)            + " secs old), " +
                    "responsesRemaining"    + " (" + responsesRemaining                            + "), "          +
                    "deadline"              + " (" + deadlineString                                + " secs left) " +
                    "startTime"             + " (" + ((now - startTime) / MILLISECONDS_PER_SECOND) + " secs ago)";
            }
        }

        /** The table of pending queries. */
        protected Map pendingQueries = new HashMap(); // String (query) -> HashMap (Integer (Serial Number) -> Entry)

        /** Number of queries in the table. */
        private int size;

        /**
         * Adds a pending query to the table.
         *
         * @param queryString Either name or logical address string
         * @param requiresTrusted If true only trusted bindings will be accepted
         * @param requiresAuthoritative If true only bindings received from the authoritative peer will be accepted
         * @param maxAge The maximum age in milliseconds of an acceptable query. Calculated from the time of the query
         * for already cached bindings; calculated from time of receipt for bindings received after the query is made,
         * i.e. relative time, ms. 0 is magic meaning "don't care about age."
         * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
         * @param deadline Absolute time (since epoch) in milliseconds before this query will be ended; 0 means indefinitely.
         * @param serialNumber The unique serial number of this query, name and logical address queries share the same serial
         * number space since response messages are not specific to type of query: a Pull message is a Pull message; there is
         * no logical address Pull message, etc.
         */
        public synchronized void put (final String  queryString,
                                      final boolean requiresTrusted,
                                      final boolean requiresAuthoritative,
                                      final int     maxAge,
                                      final int     maxResponses,
                                      final long    deadline,
                                      final int     serialNumber)
        {
            final long now = System.currentTimeMillis();
            final Integer serialNumberInteger = new Integer (serialNumber);

            if (deadline == WAIT_INDEFINITELY
                ||
                deadline > now)
            {
//                final Entry queryEntry = new Entry (serialNumberInteger, requiresTrusted, requiresAuthoritative, maxAge, maxResponses, deadline, now);

//                if (debugPendingQueryTable)
//                    log.println ("PendingQueryTable.put(): \"" + queryString + "\" -> " + queryEntry + " S/N: " + serialNumber);

                ++size;

                Map serialNumberMap = (Map) pendingQueries.get (queryString);
                if (serialNumberMap == null)
                {
                    serialNumberMap = new HashMap();
                    pendingQueries.put (queryString, serialNumberMap);
                }

                final Entry queryEntry = new Entry (serialNumberInteger, requiresTrusted, requiresAuthoritative, maxAge, maxResponses, deadline, System.currentTimeMillis());

                serialNumberMap.put (serialNumberInteger, queryEntry);

                final long timeout = deadline - now;
                if (timeout > 0)
                    setTimer (new PendingQueriesTimerTask (queryString, serialNumber, this), timeout);

/*
  // TODO(EXPERIMENTS): timer
                final long timeout = deadline - now;
                if (timeout > 0)
                    timer.schedule (new PendingQueriesTimerTask (queryString, serialNumber, this), timeout);
*/
            }
        }

        // TODO(EXPERIMENTS): used for formatting in log
        private final java.text.SimpleDateFormat sdf;

        // TODO(EXPERIMENTS)
        public PendingQueryTable()
        {
            sdf = new java.text.SimpleDateFormat();
            sdf.applyPattern ("hh:mm:ss.SSS");
        }

        /**
         * @see PendingQueryTable#remove(String,Integer)
         */
        public synchronized void remove (final String queryString, final int serialNumber) {
            remove (queryString, new Integer (serialNumber));
        }

        /**
         * Called to remove a pending query from the table, due to either timeout or receipt of satisfactory number of queries.
         */
        public synchronized void remove (final String queryString, final Integer serialNumberInteger)
        {
            if (debugPendingQueryTable)
                log.println ("PendingQueryTable.remove(): queryString: " + queryString + " serialNumber: " + serialNumberInteger);

            final Map serialNumberMap = (Map) pendingQueries.get (queryString);
            if (serialNumberMap != null)
            {
                // TODO(EXPERIMENTS): when an entry is removed from this table, log it
                if (experimentRunning)
                {
                    final Entry entry = (Entry) serialNumberMap.get (serialNumberInteger);
                    if (entry != null)
                    {
                        final long endTime = (entry.responsesRemaining == 0) ? System.currentTimeMillis() : 0;
                        final String oneQuery =
                            serialNumberInteger + ":" +
                            (localName == null ? "NONE" : localName)           + ":" +
                            queryString         + ":" +
                            entry.startTime     + ":" +
                            endTime;
                        //final String oneQuery = localName + ":" + queryString + ":" + sdf.format (new Date (entry.startTime)) + ":" + ((endTime == 0) ? 0.0 : ((endTime - entry.startTime) / MILLISECONDS_PER_SECOND));

                        if (debugPendingQueryTable)
                            log.println (oneQuery);

                        //if (experimentOut != null && randomLAQueryTimerTask != null)
                        synchronized (experimentOutLock)
                        {
                            if (experimentOut != null)
                                experimentOut.println (oneQuery);
                        }
                    }
                }

                if (serialNumberMap.remove (serialNumberInteger) != null)
                {
                    --size;
                    if (size < 0)
                        throw new HyperCastFatalRuntimeException ("Logic error in Naming FSM: pending query table size: " + size);
                }

                if (serialNumberMap.size() == 0)
                    pendingQueries.remove (serialNumberMap);
            }
        }

        /** Returns either Name or Logical Address depending on the type of table. */
        protected abstract String getTableKey (final NamingBinding binding);

        public synchronized int size() {
            return size;
        }

        /**
         * Given the name and logical address of a binding that was pending the arrival of a certificate to complete a
         * trust chain, attempt to satisfy pending queries and store satisfying bindings in the bindingVector
         *
         * @param serialNumber If NO_SERIAL_NUMBER then any pending query may be satisfied by <code>binding</code>. If a
         * particular serial number, only a query with that serial number may be satisfied by <code>binding</code>.
         * @param binding A binding that is being considered for satisifying one or more pending queries.
         * @param satisfiedQueries A Map: (Integer (serial number) -> Vector (NamingBinding)) that contains queries that are
         * at least partially satisified.
         * @param untrustedQueries A Map: (Integer (serial number) -> Vector (NamingBinding)) that contains queries that
         * would be satisified by this binding, if only the binding were trusted.
         */
        public synchronized void attemptToSatisfyQueries (final int serialNumber, final NamingBinding binding, final Map satisfiedQueries, final Map untrustedQueries)
        {
            if (debugPendingQueryTable)
                log.println ("PendingQueryTable: S/N " + serialNumber + ", considering: " + binding);

            if (binding != null)
            {
                final Integer serialNumberInteger = new Integer (serialNumber);

                final String key = getTableKey (binding);
                final Map serialNumberMap = (Map) pendingQueries.get (key);

                if (serialNumberMap == null)
                {
                    if (debugPendingQueryTable)
                        log.println ("PendingQueryTable: Serial number map is null");

                    return;
                }

                // depending on the serialNumber value we either look at an entire set of pending queries or only one pending query.
                if (serialNumber == NO_SERIAL_NUMBER)
                {
                    final Set serialNumberSet = new HashSet (serialNumberMap.keySet());
                    final Iterator serialNumberSetIterator = serialNumberSet.iterator();
                    while (serialNumberSetIterator.hasNext())
                    {
                        final Integer serialNumberIntegerFromSet = (Integer) serialNumberSetIterator.next();
                        final Entry entry = (Entry) serialNumberMap.get (serialNumberIntegerFromSet);
                        if (entry != null)
                            tryAcceptEntry (entry, binding, satisfiedQueries, untrustedQueries);
                    }
                }
                else
                {
                    final Entry entry = (Entry) serialNumberMap.get (serialNumberInteger);
                    if (entry != null)
                        tryAcceptEntry (entry, binding, satisfiedQueries, untrustedQueries);
                }
            }
        }

        /**
         * This method is a "front-end" to the <code>accept()</code> method. It deals with the various return values, adding
         * information to relevant tables as needed.
         *
         * @param entry A pending query
         * @param binding a binding that might satisfy the query described by <code>entry</code>.
         * @param satisfiedQueries A Map that will have <code>entry</code> added to it if <code>binding</code> satisfies
         * <code>entry</code>.
         * @param untrustedQueries A Map that will have <code>entry</code> added to it if <code>binding</code> satisfies
         * <code>entry</code> but <code>binding</code> is not trusted and <code>entry</code> requires a trusted response.
         */
        private void tryAcceptEntry (final Entry         entry,
                                     final NamingBinding binding,
                                     final Map           satisfiedQueries,
                                     final Map           untrustedQueries)
        {
            switch (accept (entry, binding))
            {
                case REJECT_NOT_PENDING:       if (debugPendingQueryTable) log.println ("PendingQueryTable: S/N (" + entry.getSerialNumber() + ") REJECT_NOT_PENDING"       + " -> " + binding); break;
                case REJECT_NOT_AUTHORITATIVE: if (debugPendingQueryTable) log.println ("PendingQueryTable: S/N (" + entry.getSerialNumber() + ") REJECT_NOT_AUTHORITATIVE" + " -> " + binding); break;
                case REJECT_TOO_OLD:           if (debugPendingQueryTable) log.println ("PendingQueryTable: S/N (" + entry.getSerialNumber() + ") REJECT_TOO_OLD"           + " -> " + binding); break;

                case REJECT_NOT_TRUSTED:       if (debugPendingQueryTable) log.println ("PendingQueryTable: S/N (" + entry.getSerialNumber() + ") REJECT_NOT_TRUSTED"       + " -> " + binding);
                    Vector untrustedBindings = (Vector) untrustedQueries.get (entry.getSerialNumber());
                    if (untrustedBindings == null)
                    {
                        untrustedBindings = new Vector();
                        untrustedQueries.put (entry.getSerialNumber(), untrustedBindings);
                    }
                    untrustedBindings.add (binding);
                    break;

                case ACCEPTED:                 if (debugPendingQueryTable) log.println ("PendingQueryTable: S/N (" + entry.getSerialNumber() + ") ACCEPTED"                 + " -> " + binding);
                    Vector bindings = (Vector) satisfiedQueries.get (entry.getSerialNumber());
                    if (bindings == null)
                    {
                        bindings = new Vector();
                        satisfiedQueries.put (entry.getSerialNumber(), bindings);
                    }
                    bindings.add (binding);
                    break;

                default:
                    throw new HyperCastFatalRuntimeException ("Unknown return value from pendingQueryTable.accept()");
            }
        }

        /**
         * If accepted, decrements response count, possibly removing from pending table if count remaining is
         * zero. Think of this as a conditional remove operation.
         *
         * Note: this method is effectively private to this class - constants could be moved here if Java supported static members of inner classes
         */
        private synchronized int accept (final Entry entry, final NamingBinding binding)
        {
            final int result;

            if (debugPendingQueryTable)
                log.println ("PendingQueryTable.accept(): Entry for <" + binding.getName() + " <-> " + binding.getLogicalAddressString() + " (" + binding.getSignerName() + ")> " +
                             "(key: \"" + getTableKey (binding) + "\"):\n" + "    " + ((entry == null) ? ("No entry for " + getTableKey (binding)) : entry.toString())/* + " - table:\n" + this*/);

            if (entry != null)
            {
                final long now = System.currentTimeMillis();

                if (entry.requiresAuthoritative
                    &&
                    ! binding.isAuthoritative())
                {
                    result = REJECT_NOT_AUTHORITATIVE;
                }
                else if (entry.maxAge > 0
                         &&
                         binding.getAge (now) > entry.maxAge)
                {
                    result = REJECT_TOO_OLD;
                }
                else if ((entry.deadline != WAIT_INDEFINITELY && now > entry.deadline)
                         ||
                         entry.responsesRemaining <= 0)
                {
                    if (debugPendingQueryTable)
                        log.println ("REJECT_NOT_PENDING arm: entry.deadline: " + entry.deadline + " now: " + now +
                                     " entry.maxResponses: " + entry.responsesRemaining);

                    result = REJECT_NOT_PENDING;
                }
                else if (entry.requiresTrusted
                         &&
                         ! binding.isTrusted (digitalSigner, certificateCache.get (binding.getSignerName())))
                {
                    result = REJECT_NOT_TRUSTED;
                }
                else
                {
/*
                    if (! policy.getBindingsStickyTrust())
                        binding.clearTrusted();
*/

                    result = ACCEPTED;

                    if (--entry.responsesRemaining <= 0)
                    {
                        if (debugPendingQueryTable)
                            log.println ("PendingQuery ACCEPT: " + binding);

                        remove (getTableKey (binding), entry.getSerialNumber());
                    }
                }
            }
            else
                result = REJECT_NOT_PENDING;

            return result;
        }

        /**
         * This method allows the type of query to be factored out of an algorithm that is shared by name and logical address
         * queries.
         *
         * @see #attemptToSatisfyQueries(int, NamingBinding, Map, Map)
         */
        protected abstract boolean queryIsLogicalAddress();

        /**
         * Returns a string that contains all of the data of this object.
         */
        public synchronized String toString()
        {
            final StringBuffer buf = new StringBuffer ("Pending Query Table:\n\n");
            final long now = System.currentTimeMillis();

            final Iterator iter = pendingQueries.entrySet().iterator();
            while (iter.hasNext())
            {
                final Map.Entry mapEntry = (Map.Entry) iter.next();

                final String pendingQuery = (String) mapEntry.getKey();
                final Map serialNumberMap = (Map) mapEntry.getValue();
                final Iterator iter2 = serialNumberMap.entrySet().iterator();
                while (iter2.hasNext())
                {
                    final Map.Entry mapEntry2 = (Map.Entry) iter2.next();

                    final Integer serialNumberInteger = (Integer) mapEntry2.getKey();
                    final Entry entry = (Entry) mapEntry2.getValue();
                    buf.append (pendingQuery + " -> " + serialNumberInteger + " -> " + entry + "\n");
                }
            }

            return buf.toString();
        }
    }

    /**
     * The Pending Name Query Table. All algorithms and data are common to pending name and pending logical address
     * table except the knowledge of whether to use the name or logical address of a binding when doing lookups - that
     * information is captured here.
     */
    private final class PendingNameQueryTable
        extends PendingQueryTable
    {
        protected String getTableKey (final NamingBinding binding) {
            return binding.getLogicalAddressString();
        }

        protected boolean queryIsLogicalAddress() {
            return true;
        }
    }

    /**
     * The Pending Logical Address Query Table.
     *
     * @see PendingNameQueryTable
     */
    private final class PendingLogicalAddressQueryTable
        extends PendingQueryTable
    {
        protected String getTableKey (final NamingBinding binding) {
            return binding.getName();
        }

        protected boolean queryIsLogicalAddress() {
            return false;
        }
    }

    /**
     * This table records the status of the process of establishing chains of trust for certificates. This process
     * begins when a certificate cannot be verified because the certificate of the issuer (or next CA) is not
     * present. Messages are sent to retrieve the certificate of the issuer. If the issuer cannot be verified this
     * process continues until the entire chain can be verified.
     *
     * The entries of this table are indexed by the names of the certificate that are currently requested. These names
     * map to the certificates that could not be verified initially.
     *
     * For example, say there is a binding of ("Foo" <-> 123) that cannot be verified. The chain of trust from Foo to the
     * Trust Anchor Boz is Foo -> Bar -> Baz -> Boz. Assume that the certificate for the Trust Anchor Boz is already
     * installed, thus the certificates for Foo, Bar, and Baz must be acquired. Initially, this table would have the entry:
     * Foo -> Unicast, { Foo } meaning that a unicast message has been sent for Foo and that the names of the certificates at
     * the beginning of trust chains for which Foo is needed is { Foo }. If the unicast message for Foo times-out, the entry
     * will change to: Foo -> Broadcast, { Foo }. When the certificate for Foo is received the entry will change to: Bar ->
     * Broadcast, { Foo } meaning that certificate chains starting with names { Foo } are currently waiting for the receipt
     * of the certificate for Bar. When the certificate for Bar is received, the table entry will look like this: Baz ->
     * Broadcast, { Foo } and when Baz is acquired the chain of trust can be established because the Trust Anchor Boz is
     * already installed.
     *
     * If unicast request fails, do a broadcast request, if that fails: timeout
     */
    private final class PendingCertificateRequestTable
    {
        /**
         * An entry of this table is a set of certificates from which build-trust-chain operations began. All
         * certificates in the set are actually building the same chain, though they began at different points - they
         * are all waiting for the same certificate that is the next link in the trust chain.
         */
        private class Entry        
        {
            private Set startingCertificateNames = new HashSet();

            /** false means pendingBroadcast */
            private boolean pendingUnicast;

            public Entry (final String name, final boolean pendingUnicast)
            {
                this.pendingUnicast = pendingUnicast;
                startingCertificateNames.add (name);
            }

            public void add (final String startingCert) {
                startingCertificateNames.add (startingCert);
            }
        }

        /** Implements the table of pending certificate request messages. */
        private Map pendingCerts = new HashMap(); // String (Pending Certificate Name) -> (Map: Integer (serial number) -> Entry)

        /**
         * Called when no response is received for a certificate request. If the request was unicast, next try a
         * broadcast; if the request was broadcast terminate build chain of trust operation.
         *
         * @return true if search for certificate should be continued (via broadcast)
         */
        public synchronized boolean timeout (final int serialNumber, final String pendingCertName)
        {
            final Map serialNumberMap = (Map) pendingCerts.get (pendingCertName);
            if (serialNumberMap != null)
            {
                final Integer serialNumberInteger = new Integer (serialNumber);
                final Entry entry = (Entry) serialNumberMap.get (serialNumberInteger);
                
                if (entry != null)
                {
                    if (entry.pendingUnicast) {
                        entry.pendingUnicast = false;
                        return true;
                    }
                    else
                    {
                        serialNumberMap.remove (serialNumberInteger); // broadcast for next link in chain went unanswered; cannot build chain.
                        if (serialNumberMap.size() == 0)
                            pendingCerts.remove (serialNumberMap);
                    }
                }
            }
            return false;
        }

        /**
         * for each entry of pending cert table:
         * (1) either you need another cert in the chain (still pending)
         * (2) the chain is trusted, note where chain started (no longer pending - trusted)
         * (3) bad chain - untrusted, self-signed cert (no longer pending - untrusted)
         *
         * This method considers certs individuals rather than in sets.
         *
         * @param serialNumber The S/N appearing in a certificate response message. Can be NO_SERIAL_NUMBER in which case any
         * number of trust chains may be extended.
         * @param receivedCertificateName The name of the certificate that was received.
         * @param noLongerPendingCerts Empty set that will be populated by this method to contains names of certificates
         * that have their trust chains established.
         *
         * @return name of next cert needed in chain or null if chain is complete
         */
        public synchronized String update (final int serialNumber, final String receivedCertificateName, final Set noLongerPendingCerts)
        {
            if (debugPendingCertificateTable)
            {
                log.println ("PendingCertificateTable.update(): S/N (" + serialNumber + ") cert name: \"" + receivedCertificateName + "\"");
                final Iterator iterdebug = noLongerPendingCerts.iterator();
                final StringBuffer debugBuf = new StringBuffer();
                debugBuf.append ("    noLongerPendingCerts: ");
                while (iterdebug.hasNext())
                {
                    final String noLongerPendingCert = (String) iterdebug.next();
                    debugBuf.append (noLongerPendingCert + " ");
                }
                log.println (debugBuf.toString());
            }

            String nextCertName = null; // NB: *not* necessarily the issuer of receivedCertificateName!

            final Map serialNumberMap = (Map) pendingCerts.get (receivedCertificateName);

            if (serialNumberMap != null)
            {
                final Integer serialNumberInteger = new Integer (serialNumber);
                final Entry entry = (Entry) serialNumberMap.remove (serialNumberInteger);

                if (serialNumberMap.size() == 0)
                    pendingCerts.remove (serialNumberMap);

                if (entry != null)
                {
                    if (debugPendingCertificateTable)
                        log.println ("PendingCertificateTable.update(): pending entry found for certificate \"" + receivedCertificateName + "\" S/N " + serialNumberInteger);

                    final Iterator iter = entry.startingCertificateNames.iterator();

                    while (iter.hasNext())
                    {
                        final String startCertName = (String) iter.next();
                        final CertificateCache.Entry startCert = certificateCache.get (startCertName);

                        if (startCert == null) {
                            noLongerPendingCerts.add (startCertName);
                            continue; //  certificate is not in cache
                        }
/*
                        if ((policy.getCertificateCaching() && startCert.isTrusted())
                            ||
                            startCert.isTrusted (receivedCertificateName)) // must bound chain by receivedCertificateName if cannot used cached certs
*/
                        if (startCert.isTrusted())
                        {
                            if (debugPendingCertificateTable)
                                log.println ("PendingCertificateTable.update(): trust chain established for certificate \"" + startCertName + "\" S/N " + serialNumberInteger);

                            noLongerPendingCerts.add (startCertName);
                        }
                        else
                        {
                            nextCertName = policy.getCertificateCaching()
                                ? startCert.getNextNeededIssuerInChain()
                                : certificateCache.get (receivedCertificateName).getIssuerCommonName();

                            if (nextCertName != null
                                &&
                                ! nextCertName.equals (receivedCertificateName))
                            {
                                if (debugPendingCertificateTable)
                                    log.println ("PendingCertificateTable.update(): trust chain not yet established for certificate \"" + receivedCertificateName
                                                 + "\" - next needed certificate is \"" + nextCertName + "\" S/N " + serialNumberInteger);;

                                put (serialNumber, nextCertName, startCertName, false);
                            }
                            else // encountered untrusted self-signed cert
                            {
                                noLongerPendingCerts.add (startCertName);
                            }
                        }
                    }
                }
                else
                {
                    // serial number map entry was null - I wasn't waiting on this cert
                }
            }
            else
            {
                // no entry was found for the specified serial number - I wasn't waiting on this cert
            }

            return nextCertName;
        }

        /**
         * Adds an entry to the table.
         */
        public synchronized boolean put (final int serialNumber, final String pendingCertName, final String startCertName, final boolean pendingUnicast)
        {
            if (pendingCertName == null)
                throw new HyperCastFatalRuntimeException ("null passed as parameter for pendingCertName", new NullPointerException());

            if (startCertName == null)
                throw new HyperCastFatalRuntimeException ("null passed as parameter for startCertName", new NullPointerException());

            final Integer serialNumberInteger = new Integer (serialNumber);
            
            final boolean createdNewEntry;
            Map serialNumberMap = (Map) pendingCerts.get (pendingCertName);
            if (serialNumberMap == null)
            {
                serialNumberMap = new HashMap();
                pendingCerts.put (pendingCertName, serialNumberMap);
            }

            final Entry existingEntry = (Entry) serialNumberMap.get (serialNumberInteger);
            if (existingEntry != null)
            {
                existingEntry.add (startCertName);
                createdNewEntry = false;
            }
            else
            {
                final Entry newEntry = new Entry (startCertName, pendingUnicast);
                serialNumberMap.put (serialNumberInteger, newEntry);
                createdNewEntry = true;
            }

            if (debugPendingCertificateTable)
                log.println ("PendingCertificateTable.put(): \n" + this);

            return createdNewEntry;
        }

        /**
         * Returns a string that has information about the contents of this table.
         */
        public String toString()
        {
            final StringBuffer buf = new StringBuffer ("Pending Certificate Request Table:\n\n");

            final Iterator iter = pendingCerts.entrySet().iterator();
            while (iter.hasNext())
            {
                final Map.Entry mapEntry = (Map.Entry) iter.next();

                final String pendingCertName = (String) mapEntry.getKey();
                final Map serialNumberMap = (Map) mapEntry.getValue();
                final Iterator serialNumberMapIter = serialNumberMap.entrySet().iterator();
                while (serialNumberMapIter.hasNext())
                {
                    final Map.Entry serialNumberMapEntry = (Map.Entry) serialNumberMapIter.next();
                    
                    final Integer serialNumberInteger = (Integer) serialNumberMapEntry.getKey();
                    final Entry entry = (Entry) serialNumberMapEntry.getValue();

                    final Set startingCerts = entry.startingCertificateNames;
                
                    final Iterator certsIter = startingCerts.iterator();
                    while (certsIter.hasNext())
                    {
                        final String startCert = (String) certsIter.next();
                        buf.append (pendingCertName + " -> " + serialNumberInteger + " -> " + startCert + " (" + (entry.pendingUnicast ? "unicast" : "broadcast") +")\n");
                    }
                }
            }

            return buf.toString();
        }
    }

    /**
     * When a query cannot be satisfied due to a break in a chain of trust, a certificate request is made to acquire the
     * next missing certificate in the chain. This table records the names of the missing certificates and maps them to
     * the queries that need them. When a missing certificate is acquired, this table is consulted to see which queries
     * should be restarted.
     *
     * This table provides a link between the certificate distribution system and the naming binding exchange system.
     *
     * As a first approximation, the mapping of this table can be thought of as:
     *
     * Pending Chain of Trust -> Pending Query -> { Set of Untrusted Query Responses }
     *
     * The "Pending Chain of Trust" and "Pending Query" portions each have a serial number associated with them which is
     * useful for experiments - by having serial numbers, individual messages that carry the same information can be treated
     * independently. Thus, the actual picture is:
     *
     * Pending Chain of Trust -> Trust S/N -> Pending Query -> Query S/N -> { Set of Untrusted Query Responses }
     *
     * In cases where serial numbers of messages are not being matched the Query S/N will still have a query S/N, however the
     * trust S/N field will always be 0 (i.e. NO_SERIAL_NUMBER).
     */
    private final class PendingUntrustedQueryTable
    {
        /** String (certificate name) -> (Map : Integer (trust S/N) -> (Map : String (name) -> (Map : Integer (query S/N) -> Set : String (logical address)))) */
        private Map pendingLogicalAddressQueries = new HashMap();

        /** String (certificate name) -> (Map : Integer (trust S/N) -> (Map : String (logical address) -> (Map : Integer (query S/N) -> Set : String (name)))) */
        private Map pendingNameQueries = new HashMap();

        /**
         * Adds the name and logical address of an untrusted naming binding that satisfied a pending logical address
         * query. The certificate named is needed to establish trust.
         */
        public void putLogicalAddressQuery (final int trustSerialNumber, final int querySerialNumber, final String certificateName, final String name, final String logicalAddressString)
        {
            if (debugPendingUntrustedQueryTable)
                log.println ("BEGIN PendingUntrustedQueryTable.putLogicalAddressQuery (" +
                             "certificate \"" + certificateName + "\" S/N (" + trustSerialNumber + ")," +
                             " LA \"" + logicalAddressString + "\" S/N (" + querySerialNumber + ")" + " -> name \"" + name + "\")\n" + this);

            Map trustSerialNumberMap = (Map) pendingLogicalAddressQueries.get (certificateName);

            if (trustSerialNumberMap == null) {
                trustSerialNumberMap = new HashMap();
                pendingLogicalAddressQueries.put (certificateName, trustSerialNumberMap);
            }

            final Integer trustSerialNumberInteger = new Integer (trustSerialNumber);
            Map nameMap = (Map) trustSerialNumberMap.get (trustSerialNumberInteger);

            if (nameMap == null) {
                nameMap = new HashMap();
                trustSerialNumberMap.put (trustSerialNumberInteger, nameMap);
            }

            Map querySerialNumberMap = (Map) nameMap.get (name);

            if (querySerialNumberMap == null) {
                querySerialNumberMap = new HashMap();
                nameMap.put (name, querySerialNumberMap);
            }

            final Integer querySerialNumberInteger = new Integer (querySerialNumber);
            Set logicalAddressSet = (Set) querySerialNumberMap.get (querySerialNumberInteger);

            if (logicalAddressSet == null) {
                logicalAddressSet = new HashSet();
                querySerialNumberMap.put (querySerialNumberInteger, logicalAddressSet);
            }
            
            logicalAddressSet.add (logicalAddressString);

            if (debugPendingUntrustedQueryTable)
                log.println ("DONE PendingUntrustedQueryTable.putLogicalAddressQuery (" +
                             "certificate \"" + certificateName + "\" S/N (" + trustSerialNumberInteger + ")," +
                             " LA \"" + logicalAddressString + "\" S/N (" + querySerialNumber + ")" + " -> name \"" + name + "\")\n" + this);
        }

        /**
         * Adds the logical address and name of an untrusted naming binding that satisfied a pending name query. The
         * certificate named is needed to establish trust.
         */
        public void putNameQuery (final int trustSerialNumber, final int querySerialNumber, final String certificateName, final String logicalAddressString, final String name)
        {
            if (debugPendingUntrustedQueryTable)
                log.println ("BEGIN PendingUntrustedQueryTable.putNameQuery (" +
                             "certificate \"" + certificateName + "\" S/N (" + trustSerialNumber + ")," +
                             " name \"" + name + "\" S/N (" + querySerialNumber + ")" + " -> LA \"" + logicalAddressString + "\")\n" + this);

            Map trustSerialNumberMap = (Map) pendingNameQueries.get (certificateName);

            if (trustSerialNumberMap == null) {
                trustSerialNumberMap = new HashMap();
                pendingNameQueries.put (certificateName, trustSerialNumberMap);
            }

            final Integer trustSerialNumberInteger = new Integer (trustSerialNumber);
            Map logicalAddressMap = (Map) trustSerialNumberMap.get (trustSerialNumberInteger);

            if (logicalAddressMap == null) {
                logicalAddressMap = new HashMap();
                trustSerialNumberMap.put (trustSerialNumberInteger, logicalAddressMap);
            }

            Map querySerialNumberMap = (Map) logicalAddressMap.get (logicalAddressString);

            if (querySerialNumberMap == null) {
                querySerialNumberMap = new HashMap();
                logicalAddressMap.put (logicalAddressString, querySerialNumberMap);
            }

            final Integer querySerialNumberInteger = new Integer (querySerialNumber);
            Set nameSet = (Set) querySerialNumberMap.get (querySerialNumberInteger);

            if (nameSet == null) {
                nameSet = new HashSet();
                querySerialNumberMap.put (querySerialNumberInteger, nameSet);
            }
            
            nameSet.add (name);

            if (debugPendingUntrustedQueryTable)
                log.println ("DONE PendingUntrustedQueryTable.putNameQuery (" +
                             "certificate \"" + certificateName + "\" S/N (" + trustSerialNumberInteger + ")," +
                             " name \"" + name + "\" S/N (" + querySerialNumber + ")" + " -> LA \"" + logicalAddressString + "\")\n" + this);
        }

        /**
         * Given a set of newly trusted certificates (newly trusted because a certificate was just received that allows
         * the trust chains for these certificates to be completed) determines if there are any bindings that could have
         * satisfied queries, but did not because they were not trusted. If such bindings exist, use them to attempt to
         * satisfy pending queries.
         *
         * @return vector of mappings, mappings map a string to a set of strings. For example: [ {Foo -> {123, 456}},
         * {345 -> {Bar, Baz}} ] is vector of length two, each element of the vector is a mapping, each mapping maps a
         * string to a set of strings. In this example the pending bindings <Foo, 123>, <Foo, 456>, <Bar, 345>, and
         * <Baz, 345> are represented. There was a pending name query for 345 and a pending logical address query for
         * Foo.
         */
/*
        public synchronized void getNonPendingBindings (final Set    noLongerPendingCerts,
                                                        final Vector nonPendingBindingsOfLogicalAddressQueries,
                                                        final Vector nonPendingBindingsOfNameQueries)
        {
            if (noLongerPendingCerts != null)
            {
                if (debugPendingUntrustedQueryTable) {
                    final int count = noLongerPendingCerts.size();
                    log.println ("PendingUntrustedQueryTable: " + count + " cert" + ((count > 0) ? "s" : "" )+ " no longer pending\n" + this);
                }

                //
                // Loop through the set of starting points of cert chains
                //

                final Iterator noLongerPendingCertsIterator = noLongerPendingCerts.iterator();
                while (noLongerPendingCertsIterator.hasNext())
                {
                    final String nonPendingCertName = (String) noLongerPendingCertsIterator.next();
                    if (debugPendingUntrustedQueryTable)
                        log.println ("PendingUntrustedQueryTable: Non-pending cert: " + nonPendingCertName);

                    //
                    // Inspect tables of untrusted bindings that were waiting for a chain to be established from this certificate starting point.
                    //

                    final Map nameToLAs = (Map) pendingLogicalAddressQueries.remove (nonPendingCertName);
                    if (nameToLAs != null)
                        nonPendingBindingsOfLogicalAddressQueries.add (nameToLAs);

                    final Map laToNames = (Map) pendingNameQueries.remove (nonPendingCertName);
                    if (laToNames != null)
                        nonPendingBindingsOfNameQueries.add (laToNames);
                }
            }

            if (debugPendingUntrustedQueryTable)
                log.println ("PendingUntrustedQueryTable: finished removing non pending binding mappings\n" + this);
        }
*/

        /**
         * Searches table to find all logical address queries that are no longer pending due to receipt of certificate and
         * completion of trust chain.
         *
         * @param noLongerPendingCerts a set of certificates that are no longer pending because trust chains have been established.
         * @param trustSerialNumber The trust serial number of a particular chain or NO_SERIAL_NUMBER if many chains are acceptable.
         * @return Map : String (LA query) -> (Map : Integer (query serial number) -> (Set : String (name))) (can return empty Map; never null)
         */
        public Map getNonPendingLogicalAddressQueries (final int trustSerialNumber, final Set noLongerPendingCerts)
        {
            final Map result = new HashMap();

            final Integer trustSerialNumberInteger = new Integer (trustSerialNumber);

            final Iterator certsIterator = noLongerPendingCerts.iterator();
            while (certsIterator.hasNext())
            {
                final String certificateName = (String) certsIterator.next();

                if (debugPendingUntrustedQueryTable)
                    log.println ("PendingUntrustedQueryTable.getNonPendingLogicalAddressQueries:" +
                                 " certificate \"" + certificateName + "\" S/N (" + trustSerialNumberInteger + ")");

                final Map trustSerialNumberMap = (Map) pendingLogicalAddressQueries.get (certificateName);
                if (trustSerialNumberMap != null)
                {
                    final Map nameMap = (Map) trustSerialNumberMap.get (trustSerialNumberInteger);
                    if (nameMap != null)
                    {
                        final Iterator nameMapIterator = nameMap.entrySet().iterator();
                        while (nameMapIterator.hasNext())
                        {
                            final Map.Entry nameMapEntry = (Map.Entry) nameMapIterator.next();

                            final String name = (String) nameMapEntry.getKey();
                            final Map querySerialNumberMap = (Map) nameMapEntry.getValue();

                            Map resultSerialNumberMap = (Map) result.get (name);
                            if (resultSerialNumberMap == null) {
                                resultSerialNumberMap = new HashMap();
                                result.put (name, resultSerialNumberMap);
                            }

                            resultSerialNumberMap.putAll (querySerialNumberMap);

                            if (debugPendingUntrustedQueryTable)
                                log.println ("PendingUntrustedQueryTable.getNonPendingLogicalAddressQueries:" +
                                             " resultSerialNumberMap for " + name + " has " + querySerialNumberMap.size() + " entries.");
                        }

                        trustSerialNumberMap.remove (trustSerialNumberInteger);
                        if (trustSerialNumberMap.size() == 0)
                            pendingLogicalAddressQueries.remove (certificateName);
                    }
                }
            }

            return result;
        }

        /**
         * Searches table to find all name queries that are no longer pending due to receipt of certificate and completion of
         * trust chain.
         *
         * @param noLongerPendingCerts a set of certificates that are no longer pending because trust chains have been established.
         * @param trustSerialNumber The trust serial number of a particular chain or NO_SERIAL_NUMBER if many chains are acceptable.
         * @return Map : String (name query) -> (Map : Integer (query serial number) -> (Set : String (LA))) (can return empty Map; never null)
         */
        public Map getNonPendingNameQueries (final int trustSerialNumber, final Set noLongerPendingCerts)
        {
            final Map result = new HashMap();

            final Integer trustSerialNumberInteger = new Integer (trustSerialNumber);

            final Iterator certsIterator = noLongerPendingCerts.iterator();
            while (certsIterator.hasNext())
            {
                final String certificateName = (String) certsIterator.next();

                final Map trustSerialNumberMap = (Map) pendingNameQueries.get (certificateName);
                if (trustSerialNumberMap != null)
                {
                    final Map logicalAddressMap = (Map) trustSerialNumberMap.get (trustSerialNumberInteger);
                    if (logicalAddressMap != null)
                    {
                        final Iterator logicalAddressMapIterator = logicalAddressMap.entrySet().iterator();
                        while (logicalAddressMapIterator.hasNext())
                        {
                            final Map.Entry logicalAddressMapEntry = (Map.Entry) logicalAddressMapIterator.next();

                            final String logicalAddressString = (String) logicalAddressMapEntry.getKey();
                            final Map querySerialNumberMap = (Map) logicalAddressMapEntry.getValue();

                            Map resultSerialNumberMap = (Map) result.get (logicalAddressString);
                            if (resultSerialNumberMap == null) {
                                resultSerialNumberMap = new HashMap();
                                result.put (logicalAddressString, resultSerialNumberMap);
                            }

                            resultSerialNumberMap.putAll (querySerialNumberMap);
                        }

                        trustSerialNumberMap.remove (trustSerialNumberInteger);
                        if (trustSerialNumberMap.size() == 0)
                            pendingNameQueries.remove (certificateName);
                    }
                }
            }

            return result;
        }

        /**
         * Called when a certificate broadcast times out.
         */
        public void remove (final int trustSerialNumber, final String certificateName)
        {
            final Integer trustSerialNumberInteger = new Integer (trustSerialNumber);

            if (debugPendingUntrustedQueryTable)
                log.println ("PendingUntrustedQueryTable: timeout removing: \"" + certificateName + "\" S/N " + trustSerialNumberInteger + "\n" + this);

            {
                final Map laSerialNumberMap = (Map) pendingLogicalAddressQueries.get (certificateName);
                if (laSerialNumberMap != null)
                {
                    laSerialNumberMap.remove (trustSerialNumberInteger);
                    if (laSerialNumberMap.size() == 0)
                        pendingLogicalAddressQueries.remove (certificateName);
                }
            }

            {
                final Map nameSerialNumberMap = (Map) pendingNameQueries.get (certificateName);
                if (nameSerialNumberMap != null)
                {
                    nameSerialNumberMap.remove (trustSerialNumberInteger);
                    if (nameSerialNumberMap.size() == 0)
                        pendingNameQueries.remove (certificateName);
                }
            }

            if (debugPendingUntrustedQueryTable)
                log.println ("PendingUntrustedQueryTable: timeout removed: \"" + certificateName + "\" S/N " + trustSerialNumberInteger + "\n" + this);
        }

        /**
         * Used by toString() to generate a string with the data in this table.
         */
        private void generateTable (final StringBuffer buf, final Map mapping)
        {
            final Iterator certificateIterator = mapping.entrySet().iterator();
            while (certificateIterator.hasNext())
            {
                final Map.Entry certificateEntry = (Map.Entry) certificateIterator.next();

                final String certificateName = (String) certificateEntry.getKey();
                final Map trustSerialNumberMap = (Map) certificateEntry.getValue();

                final Iterator trustSerialNumberMapIterator = trustSerialNumberMap.entrySet().iterator();
                while (trustSerialNumberMapIterator.hasNext())
                {
                    final Map.Entry trustSerialNumberEntry = (Map.Entry) trustSerialNumberMapIterator.next();

                    final Integer trustSerialNumberInteger = (Integer) trustSerialNumberEntry.getKey();
                    final Map queryMap = (Map) trustSerialNumberEntry.getValue();

                    final Iterator queryMapIterator = queryMap.entrySet().iterator();
                    while (queryMapIterator.hasNext())
                    {
                        final Map.Entry queryMapEntry = (Map.Entry) queryMapIterator.next();

                        final String queryString = (String) queryMapEntry.getKey();
                        final Map querySerialNumberMap = (Map) queryMapEntry.getValue();

                        final Iterator querySerialNumberMapIterator = querySerialNumberMap.entrySet().iterator();
                        while (querySerialNumberMapIterator.hasNext())
                        {
                            final Map.Entry querySerialNumberMapEntry = (Map.Entry) querySerialNumberMapIterator.next();

                            final Integer querySerialNumberInteger = (Integer) querySerialNumberMapEntry.getKey();
                            final Set responseSet = (Set) querySerialNumberMapEntry.getValue();

                            buf.append (certificateName + " (" + trustSerialNumberInteger + ") -> " + queryString + " (" + querySerialNumberInteger + ") -> {");

                            final Iterator responseSetIterator = responseSet.iterator();
                            boolean first = true;
                            while (responseSetIterator.hasNext())
                            {
                                final String response = (String) responseSetIterator.next();
                                
                                buf.append ((first ? "" : "; ") + response);
                                first = false;
                            }
                            buf.append ("}\n");
                        }
                    }
                }
            }
        }

        /**
         * Creates and returns a string that has all of the entries of this table.
         */
        public String toString()
        {
            final StringBuffer buf = new StringBuffer();

            buf.append ("Pending Untrusted Query Table:\n\n");

            generateTable (buf, pendingLogicalAddressQueries);
            generateTable (buf, pendingNameQueries);

            return buf.toString();
        }
    }

    /**
     * This table contains the binding information that has been set by the local socket. Indeed, this table can be
     * thought of as a binding factory because when it is asked to return bindings it will update each binding with a
     * current timestamp and signature (all other statistics will be updated as well, e.g. logical address change
     * count), only the name is fixed.
     *
     * Note that aliasing can occur between this table and the peer cache. For example, if this socket has name X
     * recently moved to location Y, and another socket that also has name X was recently in location Y, the binding
     * <X,Y> could appear in each table.
     */
    private final class LocalBindingTable
    {
        /** Map that implements this table, mapping names to binding objects. */
        private Map nameToLocalBinding = new HashMap(); // String (name) -> NamingBinding

        /**
         * Returns an array containing all local bindings with updated information.
         */
        public synchronized NamingBinding [] getBindings()
        {
            final Iterator iterator = nameToLocalBinding.entrySet().iterator();
            final Vector result = new Vector();

            final String laString = getLogicalAddress().toString();
            final long now = System.currentTimeMillis();
            final int laChanges = getLogicalAddressChangeCounter();

            while (iterator.hasNext())
            {
                final NamingBinding binding = (NamingBinding) ((Map.Entry) iterator.next()).getValue();

                updateBinding (binding, laString, now, laChanges);
                result.add (binding);
            }

            return (NamingBinding []) result.toArray (new NamingBinding [0]);
        }

        /**
         * Returns an array containing invalidations for all local bindings.
         */
        public synchronized BindingInvalidation [] getInvalidations()
        {
            final Iterator iterator = nameToLocalBinding.entrySet().iterator();
            final Vector result = new Vector();

            final String laString = getLogicalAddress().toString();
            final long now = System.currentTimeMillis();

            while (iterator.hasNext())
            {
                final NamingBinding binding = (NamingBinding) ((Map.Entry) iterator.next()).getValue();
                final BindingInvalidation invalidation = new BindingInvalidation (binding.getName(), laString, now);
                result.add (invalidation);
            }

            return (BindingInvalidation []) result.toArray (new BindingInvalidation [0]);
        }

        /**
         * Returns an invalidate binding for the specified name if the name exists in the table.
         */
        public synchronized BindingInvalidation getInvalidation (final String name)
        {
            BindingInvalidation invalidation = null;
            final NamingBinding binding = (NamingBinding) nameToLocalBinding.get (name);

            if (binding != null)
            {
                final String laString = getLogicalAddress().toString();
                final long now = System.currentTimeMillis();

                invalidation = new BindingInvalidation (name, laString, now);
            }

            return invalidation;
        }

        private void updateBinding (final NamingBinding binding, final String laString, final long now, final int laChanges)
        {
            binding.setLogicalAddressString (laString);
            binding.setTimestamp (now);
            binding.setLogicalAddressChangeCounter (laChanges);
        }

        /**
         * Puts a binding into this table.
         */
        public synchronized void put (final NamingBinding binding) {
            nameToLocalBinding.put (binding.getName(), binding);
        }

        /**
         * Removes one binding from the this table, if a matching binding is found in the table.
         */
        public synchronized void remove (final NamingBinding binding) {
            nameToLocalBinding.remove (binding.getName());
        }

        /**
         * Gets a single binding from this table if one exists with the given name. Null is returned if no matching
         * binding exists.
         */
        public synchronized NamingBinding get (final String name)
        {
            final NamingBinding binding = (NamingBinding) nameToLocalBinding.get (name);

            if (binding != null)
            {
                final String laString = getLogicalAddress().toString();
                final long now = System.currentTimeMillis();
                final int laChanges = getLogicalAddressChangeCounter();

                updateBinding (binding, laString, now, laChanges);
            }

            return binding;
        }

        /**
         * Returns a count of the number of local bindings.
         */
        public synchronized int size() {
            return nameToLocalBinding.size();
        }

        /**
         * Returns a string that represents the bindings in this table.
         */
        public synchronized String toString()
        {
            final StringBuffer buf = new StringBuffer();

            buf.append ("Local Name Binding Table:\n\n");

            final String laString = getLogicalAddress().toString();
            final long now = System.currentTimeMillis();
            final int laChanges = getLogicalAddressChangeCounter();

            final Iterator iterator = nameToLocalBinding.entrySet().iterator();
            while (iterator.hasNext())
            {
                Map.Entry entry = (Map.Entry) iterator.next();
                final String key = (String) entry.getKey();
                final NamingBinding binding = (NamingBinding) entry.getValue();

                updateBinding (binding, laString, now, laChanges);

                buf.append (key + " -> " + binding + "\n");
            }

            return buf.toString();
        }
    }

    /**
     * A table that holds all bindings received from peers. This table is implemented by two maps that are kept
     * synchronized. One maps from names to logical addresses and the other maps from logical addresses to names.
     *
     * Entries are removed in only two ways: timeout (with timer and
     * timeoutQueue) or receipt of explicit invalidate message.
     *
     * The bindings in this table are not stored on backing store; they are stored in volatile memory only.
     */
    private final class PeerBindingCache
    {
        /** A timer that periodical sweeps the peer table for stale entries. */
        private PeerBindingsTimerTask peerBindingsTimerTask = null;

        /**
         * Queue that contains links to binding objects and manages their timeouts along with the PeerBindingsTimerTask.
         */
        private LRUTimeoutQueue timeoutQueue = new LRUTimeoutQueue (policy.getBindingTimeout());

        /** The current number of mappings in the cache. */
        private int size;

        /**
         * The total number of possible mappings in the cache. The LRU mapping is removed if a new mapping is added when
         * there are maxSize mappings already in the cache.
         */
        private final int maxSize;

        /*
         * These two tables implement the mappings for the binding table. There are two tables so that the bindings can
         * be indexed using either name or logical address.
         */
        private Map nameToLogicalAddressMap = new HashMap(); // String (name) -> HashMap (String (LA)   -> NamingBinding)
        private Map logicalAddressToNameMap = new HashMap(); // String (LA)   -> HashMap (String (name) -> NamingBinding)

        /** Creates a peer bindings cache with the give maximum size (binding count). */
        public PeerBindingCache (final int maxSize)
        {
            this.maxSize = maxSize;

            if (debugPeerBindingCache)
                System.err.println ("Peer Binding Cache <init>: " + maxSize);
        }

        /**
         * Returns the number of bindings held in the cache.
         */
        public synchronized int size() {
            return size;
        }
       
        /**
         * @see #getNameMap(String,boolean)
         */
        private Map getNameMap (final String logicalAddressString) {
            return getNameMap (logicalAddressString, true);
        }

        /**
         * Returns the map that maps the given logical address to
         * names. If no such map is in the table and the
         * createIfNotPresent parameter is true, a new empty map will
         * be created.
         */
        private Map getNameMap (final String logicalAddressString, final boolean createIfNotPresent)
        {
            Map nameMap = (Map) logicalAddressToNameMap.get (logicalAddressString);

            if (nameMap == null
                &&
                createIfNotPresent)
            {
                nameMap = new HashMap();
                logicalAddressToNameMap.put (logicalAddressString, nameMap);
            }

            return nameMap;
        }

        /**
         * @see #getLogicalAddressMap(String,boolean)
         */
        private Map getLogicalAddressMap (final String name)
        {
            return getLogicalAddressMap (name, true);
        }

        /**
         * Returns the map that maps the given name to logical
         * addresses. If no such map is in the table and the
         * createIfNotPresent parameter is true, a new empty map will
         * be created.
         */
        private Map getLogicalAddressMap (final String name, final boolean createIfNotPresent)
        {
            Map logicalAddressMap = (Map) nameToLogicalAddressMap.get (name);

            if (logicalAddressMap == null
                &&
                createIfNotPresent)
            {
                logicalAddressMap = new HashMap();
                nameToLogicalAddressMap.put (name, logicalAddressMap);
            }

            return logicalAddressMap;
        }

        /**
         * Get one binding.
         */
        public synchronized NamingBinding get (final String name, final String logicalAddressString)
        {
            final Map logicalAddressMap = (Map) nameToLogicalAddressMap.get (name);
            return (logicalAddressMap != null) ? (NamingBinding) logicalAddressMap.get (logicalAddressString) : null;
        }

        /**
         * Return an array that has one entry for each binding found in the values of the given map.
         */
        private NamingBinding [] bindingMapToArray (final Map map)
        {
            final Vector results = new Vector();

            if (map != null)
            {
                final Iterator iter = map.entrySet().iterator();

                while (iter.hasNext())
                {
                    Map.Entry entry = (Map.Entry) iter.next();
                    final NamingBinding binding = (NamingBinding) entry.getValue();
                    timeoutQueue.append ((I_LRUObject) binding);
                    results.add (binding);
                }
            }

            return (NamingBinding []) results.toArray (new NamingBinding [0]);
        }

        /**
         * Queries the table by name.
         */
        public synchronized NamingBinding [] getByName (final String name) {
            return bindingMapToArray (getLogicalAddressMap (name, false));
        }

        /**
         * Queries the table by logical address.
         */
        public synchronized NamingBinding [] getByLogicalAddress (final String logicalAddressString) {
            return bindingMapToArray (getNameMap (logicalAddressString, false));
        }

        /**
         * Puts many bindings into the table.
         */
        public synchronized void put (final NamingBinding [] bindings) {
            for (int i = 0; i < bindings.length; ++i)
                put (bindings[i]);
        }

        /**
         * Puts one binding into the peer binding cache. No unconditional crypto here. Crypto ops are considered to be
         * expensive. Only do them when bindings are looked-up in this table, not when bindings are added. Always store the
         * most recently received unverified bindings and the most recently received unverified invalidation. Do not
         * overwrite verified information until it expires.
         *
         * @param newBinding A binding received from a peer that is added to the cache of peer bindings
         */
        public synchronized void put (final NamingBinding newBinding)
        {
            if (maxSize <= 0)
                return;

            if (peerBindingsTimerTask == null
                &&
                policy.getBindingTimeout() > 0)
            {
                peerBindingsTimerTask = new PeerBindingsTimerTask (policy.getBindingTimeout(), this);
                setTimer (peerBindingsTimerTask, policy.getBindingTimeout());
            }

            final String name = newBinding.getName();
            final String logicalAddressString = newBinding.getLogicalAddressString();
            
            if (debugPeerBindingCache)
                log.println ("PeerBindingCache.put: " + newBinding + this);

            final NamingBinding existingBinding = get (name, logicalAddressString);

            final NamingBinding existingBindingOther = other.get (name, logicalAddressString);
            if (existingBindingOther == null)
                ++cachedBindingsCount;

            if (existingBinding == null)
            {
                // Remove LRU binding to make room for new binding
                NamingBinding lruBinding;
                while (maxSize > 0
                       &&
                       size >= maxSize
                       &&
                       (lruBinding = (NamingBinding) timeoutQueue.removeLRU()) != null)
                {
                    ++peerBindingCacheEvictions;
                    remove (lruBinding, true);
                }

                ++size;
            }
            else
                timeoutQueue.remove ((I_LRUObject) existingBinding);

            getNameMap (logicalAddressString).put (name, newBinding);
            getLogicalAddressMap (name).put (logicalAddressString, newBinding);

            timeoutQueue.append ((I_LRUObject) newBinding);

            if (debugPeerBindingCache)
                log.println ("PeerBindingCache.put: DONE" + newBinding + this);

            if (existingBinding == null)
            {
                synchronized (experimentOutLock)
                {
                    if (experimentOut != null)
                        experimentOut.println ((localName == null ? "NONE" : localName) + ":P:" + System.currentTimeMillis() + ":" + newBinding.getName());
                }
            }
        }

        /**
         * Removes many bindings from the table. Called from contexts where invalidations are received from peers.
         */
        public synchronized void remove (final NamingBinding [] bindings) {
            for (int i = 0; i < bindings.length; ++i)
                remove (bindings[i]);
        }

        public synchronized void remove (final NamingBinding [] bindings, final boolean force) {
            for (int i = 0; i < bindings.length; ++i)
                remove (bindings[i], force);
        }

        public synchronized void remove (final BindingInvalidation [] bindings) {
            for (int i = 0; i < bindings.length; ++i)
                remove (bindings[i]);
        }

        /**
         * @see #remove(NamingBinding,boolean)
         */
        public synchronized void remove (final NamingBinding invalidationBinding) {
            remove (invalidationBinding, false);
        }

        /**
         * Removes one binding from the table. Called from contexts where a timer has expired.
         *
         * @param force If true, the invalidationBinding need not be trusted in order for a binding to be removed
         */
        public synchronized void remove (final NamingBinding invalidationBinding, final boolean force)
        {
            final String name = invalidationBinding.getName();
            final String logicalAddressString = invalidationBinding.getLogicalAddressString();

            if (debugPeerBindingCache)
                log.println ("PeerBindingCache.remove: name(" + name + ") <-> logical address(" + logicalAddressString + ")" );

            final NamingBinding binding = get (name, logicalAddressString);

            if (binding != null)
            {
                if (debugPeerBindingCache)
                    log.println ("PeerBindingCache.remove: Binding found: name(" + name + ") <-> logical address(" + logicalAddressString + ")" );

                if (force || invalidationBinding.isTrusted (digitalSigner, certificateCache.get (name)))
                {
                    removeFromDoubleMapping (nameToLogicalAddressMap, name,                 logicalAddressString);
                    removeFromDoubleMapping (logicalAddressToNameMap, logicalAddressString, name);

                    --size;

                    final NamingBinding existingBindingOther = other.get (name, logicalAddressString);
                    if (existingBindingOther == null)
                        --cachedBindingsCount;

                    timeoutQueue.remove ((I_LRUObject) binding);
                    
                    synchronized (experimentOutLock)
                    {
                        if (experimentOut != null)
                            experimentOut.println ((localName == null ? "NONE" : localName) + ":R:" + System.currentTimeMillis() + ":" + binding.getName());
                    }

                    if (debugPeerBindingCache)
                    {
                        int realSize = nameToLogicalAddressMap.size();
                        if (size != realSize) {
                            log.println ("Size inconsistency REMOVE - size: " + size + " realSize: " + realSize);
                            Thread.dumpStack();
                        }
                    }
                }
            }
            else
                if (debugPeerBindingCache)
                    log.println ("PeerBindingCache.remove: NO SUCH BINDING FOUND: name(" + name + ") <-> logical address(" + logicalAddressString + ")" );
        }

        /**
         * Helper/utility for: remove one binding
         */
        private void removeFromDoubleMapping (final Map firstMapping, final String firstKey, final String secondKey)
        {
            final Map secondMapping = (Map) firstMapping.get (firstKey);

            if (secondMapping != null)
            {
                secondMapping.remove (secondKey);

                if (secondMapping.isEmpty())
                    firstMapping.remove (firstKey);
            }
        }

        /**
         * Helper for <code>toString()</code>
         * @see #toString()
         */
        private boolean doubleMappingContainsKeys (final Map firstMapping, final String firstKey, final String secondKey) {
            final Map secondMapping = (Map) firstMapping.get (firstKey);
            return (secondMapping != null) ? secondMapping.containsKey (secondKey) : false;
        }

        /**
         * Helper for <code>toString()</code>
         * @see #toString()
         */
        private void generateTable (final StringBuffer buf, final Map forwardMapping, final Map reverseMapping, final boolean nameFirst)
        {
            final Iterator firstIterator = forwardMapping.entrySet().iterator();
            
            while (firstIterator.hasNext())
            {
                final Map secondMapping = (Map) ((Map.Entry) firstIterator.next()).getValue();

                final Iterator secondIterator = secondMapping.entrySet().iterator();
                while (secondIterator.hasNext())
                {
                    final NamingBinding binding = (NamingBinding) ((Map.Entry) secondIterator.next()).getValue();

                    final String name = binding.getName();
                    final String logicalAddressString = binding.getLogicalAddressString();

                    buf.append (binding);

                    if (reverseMapping != null)
                    {
                        final boolean reverseMappingExists =
                            (nameFirst)
                            ? doubleMappingContainsKeys (reverseMapping, logicalAddressString, name)
                            : doubleMappingContainsKeys (reverseMapping, name, logicalAddressString);

                        if (! reverseMappingExists)
                            buf.append (" <<<NO REVERSE MAPPING>>>");
                    }

                    buf.append ("\n");
                }
            }
        }

        public synchronized void removeExpiredEntries()
        {
            final long now = System.currentTimeMillis();
            NamingBinding binding;

            if (debugPeerBindingCache)
                log.println ("PeerBindingsCache removeExpiredEntries, now: " + now + " queue: " + timeoutQueue);

            while ((binding = (NamingBinding) timeoutQueue.remove (now)) != null)
            {
                if (debugPeerBindingCache)
                    log.println ("PeerBindingsCache removeExpiredEntries, removing: " + binding);
                
                remove (binding, true);
            }
        }


        // If this is the pull cache, point to the push cache, if this is the push cache, point to the pull cache
        private PeerBindingCache other;
        public synchronized void setOtherCache (final PeerBindingCache other) {
            this.other = other;
        }

        public synchronized String toString()
        {
            final StringBuffer buf = new StringBuffer ("Peer Name Binding Cache:\n\n");

            generateTable (buf, nameToLogicalAddressMap, logicalAddressToNameMap, true);

            // This is redundant, but useful for debugging to ensure that binding tables are in sync
            //generateTable (buf, logicalAddressToNameMap, nameToLogicalAddressMap, false);

            return buf.toString();
        }
    }

    // TODO(EXPERIMENTS):
    private int querySerialNumber = 1;
    private synchronized int getNextQuerySerialNumber()
    {
        final int result = querySerialNumber;

        ++querySerialNumber; // increment could overflow: that's ok

        // 0 is a distinguished value meaning "invalid serial number"
        if (querySerialNumber == NO_SERIAL_NUMBER)
            ++querySerialNumber;
            
        return result;
    }

    private int trustSerialNumber = 1;
    private synchronized int getNextTrustSerialNumber()
    {
        final int result;
        if (policy.getRequireTrustMessageSerialNumberMatch())
        {
            result = trustSerialNumber;

            ++trustSerialNumber; // increment could overflow: that's ok

            // 0 is a distinguished value meaning "invalid serial number"
            if (trustSerialNumber == NO_SERIAL_NUMBER)
                ++trustSerialNumber;
        }
        else
            result = NO_SERIAL_NUMBER;

        return result;
    }

    /**
     * This is the base class of all naming messages. It contains services that are used by descendant classes. These
     * services are things such as constructors, routines to do run length encoding of various types of data, and routines
     * that implement the trust chain construction operation.
     *
     * When designing the hierarchy of message type several factors were considered:
     *   - origin of message (API/FSM/Peer) determines when/if the <code>olMessage</code> field is valid
     *   - routable vs. non-routable: some naming messages are designed to be sent to peers, others are of local use only.
     *   - contains certificate - some messages contain a certificate, others do not.
     *
     * A rule is: messages are created only during processing of messages (timer task handlers are considered messages for
     * this purpose). In particular code in the methods of various tables does not generate any messages.
     */
    private abstract class NamingMessage
    {
        /**
         * Data stored in the FSM Extension of <code>olMessage</code>. This is stored separately for convenience; storing it
         * avoids having to extract it twice since it is extracted when <code>process()</code> is called on <code>olMessage</code>.
         * This field is <code>null</code> when the <code>NamingMessage</code> object is generated locally by the FSM.
         *
         * @see MessageStoreFSM_Naming#process(OL_Message)
         */
        protected final byte [] controlData;

        /**
         * Constructor used to build NamingMessage with no control data array.
         */
        protected NamingMessage() {
            this (null);
        }

        /**
         * Constructor used to build a NamingMessage with a serialized control data array from an FSM extension.
         */
        protected NamingMessage (final byte [] controlData) {
            this.controlData = controlData;
        }

        /**
         * Descendant classes implement with algorithms for delivering particular message types. These algortims are
         * typically implemented in terms of <code>RoutableNamingMessage.sendBroadcast()</code>,
         * <code>RoutableNamingMessage.sendUnicast()</code>, and application notification events for local delivery.
         *
         * A message should be sent only once. If this method is invoked more than once, subsequent invocations should do nothing (no-op).
         *
         * @see RoutableNamingMessage#sendBroadcast()
         * @see RoutableNamingMessage#sendUnicast()
         * @see hypercast.NotificationHandler
         * @see hypercast.events.NAMING_EVENT
         */
        protected abstract void send();

        /**
         * Descendant classes implement this method with algorithms for handling particular message types. Never called on
         * messages that are generated locally by the FSM.
         */
        public abstract void process();

        /**
         * Encodes a string as a two byte length followed by the bytes of the string.
         */
        protected byte [] runLengthEncodeString (final int prefixBlanks, final String string) {
            return runLengthEncode (prefixBlanks, string.getBytes());
        }

        /**
         * @param data A <code>byte</code> array containing and encoded (serialized) <code>String</code> object.
         * @param index An <code>int</code> array of length 1. Acts as a pointer to the next unprocessed position of parameter
         * <code>data</code>.
         *
         * @throws ArrayIndexOutOfBoundsException
         */
        protected String runLengthDecodeString (final byte [] data, final int [] index)
        {
            final int STRING_LENGTH_BYTES = 2;
            final int stringLength = ByteArrayUtility.toUnsignedInteger (data, index[0], STRING_LENGTH_BYTES);
            index[0] += STRING_LENGTH_BYTES;

            final String result = (stringLength > 0) ? new String (data, index[0], stringLength) : null;
            index[0] += stringLength;
            return result;
        }

        /**
         * Encodes a certificate as a two byte length followed by the bytes of the certificate.
         */
        protected byte [] runLengthEncodeCertificate (final int prefixBlanks, final Certificate certificate) {
            return runLengthEncode (prefixBlanks, CertificateCache.encodeCertificate (certificate));
        }

        /**
         * Decodes a run length encoded certificate. The length is specified by a an unsigned two byte quantity.
         *
         * @throws HyperCastWarningRuntimeException If the certificate cannot be decoded.
         */
        protected Certificate runLengthDecodeCertificate (final byte [] data, final int [] index)
        {
            final int CERTIFICATE_LENGTH_BYTES = 2;
            final int certificateLength = ByteArrayUtility.toUnsignedInteger (data, index[0], CERTIFICATE_LENGTH_BYTES);
            index[0] += CERTIFICATE_LENGTH_BYTES;

            final Certificate certificate = (certificateLength > 0) ? CertificateCache.decodeCertificate (data, index[0], certificateLength) : null;
            index[0] += certificateLength;
            return certificate;
        }

        /**
         * Decodes a private key that is encoded as a two byte length and an array that encodes the private key itself.
         */
        protected PrivateKey runLengthDecodePrivateKey (final byte [] data, final int [] index)
        {
            final int PRIVATE_KEY_LENGTH_BYTES = 2;
            final int privateKeyLength = ByteArrayUtility.toUnsignedInteger (data, index[0], PRIVATE_KEY_LENGTH_BYTES);
            index[0] += PRIVATE_KEY_LENGTH_BYTES;

            final PrivateKey privateKey = (privateKeyLength > 0) ? CertificateCache.decodePrivateKey (data, index[0], privateKeyLength) : null;
            index[0] += privateKeyLength;
            return privateKey;
        }

        /**
         * Run length encodes the array <code>data</code> by prepending a two byte length to the array.
         */
        private byte [] runLengthEncode (final int prefixBlanks, final byte [] data)
        {
            final int bytesLength = data.length;

            final int MAXIMUM_LENGTH = ((1 << 16) - 1);
            if (bytesLength > MAXIMUM_LENGTH)
                throw new HyperCastWarningRuntimeException ("Array for run length encoding exceeds maximum length");

            final int RUN_LENGTH_SIZE_BYTES = 2;

            final byte [] result = new byte [prefixBlanks + RUN_LENGTH_SIZE_BYTES + bytesLength];

            result [prefixBlanks + 0] = (byte) ((bytesLength >>> (8*1)) & 0xFF);
            result [prefixBlanks + 1] = (byte) ((bytesLength >>> (8*0)) & 0xFF);

            System.arraycopy (data, 0, result, prefixBlanks + RUN_LENGTH_SIZE_BYTES, bytesLength);
            
            return result;
        }

        /**
         * @see #initiateTrustChainOperation(int, String, String)
         */
        public boolean initiateTrustChainOperation (final int serialNumber, final String startCertName) {
            return initiateTrustChainOperation (serialNumber, startCertName, null);
        }

        /**
         * Begins the process of assembling a chain of trust from a particular certificate to an installed trust anchor.
         *
         * @param startCertName Name of certificate where trust chain begins
         * @param hintLogicalAddressString string that encodes a location where certificate might be, can be null
         *
         * @return true if a message is actually sent
         */
        public boolean initiateTrustChainOperation (final int serialNumber, final String startCertName, final String hintLogicalAddressString)
        {
            final boolean messageSent;

            if (debugCertificates)
                log.println ("initiateTrustChainOperation (" + serialNumber + ", " + startCertName + ", " + hintLogicalAddressString + ")");

            final CertificateCache.Entry certCacheEntry = policy.getCertificateCaching() ? certificateCache.get (startCertName) : null;
            final String nextCertName = (certCacheEntry == null) ? startCertName : certCacheEntry.getNextNeededIssuerInChain();

            if (nextCertName == null) // Complete chain is already built, no need to send messages
            {
                messageSent = false;
            }
            else
            {
                final boolean nextMessageUnicast = (hintLogicalAddressString != null);
                final boolean newEntry = pendingCertificateRequestTable.put (serialNumber, nextCertName, startCertName, nextMessageUnicast);

                if (newEntry)
                {
                    if (nextMessageUnicast)
                    {
                        final I_LogicalAddress destination = socket.createLogicalAddress (hintLogicalAddressString);
                        new CertificateRequestMessageUnicast (serialNumber, nextCertName, destination).send();
                    }
                    else
                        new CertificateRequestMessageBroadcast (serialNumber, nextCertName).send();

                    messageSent = true;
                }
                else
                    messageSent = false;
            }

            return messageSent;
        }

        /**
         * Exchange (acquire) a single certificate. This is one step in the process of establishing a chain of trust. Called
         * when a certificate is received but another is needed.
         */
        public void continueBuildingTrustChain (final int serialNumber, final String nextCertName)
        {
            if (debugCertificates)
                log.println ("continueBuildingTrustChain (S/N " + serialNumber + ", " + nextCertName + ")");

            new CertificateRequestMessageBroadcast (serialNumber, nextCertName).send();
        }

/*
// This is never called. Such a routine would be useful by interrogating the peer binding cache to acquire a potential location of a certificate.
        public void continueBuildingTrustChain (final String nextCertName, final String hintLogicalAddressString)
        {
            if (debugCertificates)
                log.println ("continueBuildingTrustChain (" + nextCertName + ", " + hintLogicalAddressString + ")");

            final I_LogicalAddress destination = socket.createLogicalAddress (hintLogicalAddressString);
            new CertificateRequestMessageUnicast (nextCertName, destination).send();
        }
*/

        /**
         * Returns the name of the particular type of naming message. Since naming messages are implemented as inner
         * classes, their names have a dollar sign character. The part of the classname after the dollar sign character
         * is the string that is the message type.
         */
        public String toString() {
            return shortNameOfClass (getClass().getName());
        }
    }

    /**
     * This is the base class for all messages that can be routed to peers. One of the goals of this class is to make it
     * possible for the <code>send()</code> method of all types to "do-the-right-thing." That is, some messages are send via
     * unicast, some via broadcast, some are received already having a corresponding OL_Message, other do not have an
     * OL_Message. The methods of this class ensure that all information required to perform a send operation for any
     * particular subtype is generated properly when send() is called.
     */
    private abstract class RoutableNamingMessage
        extends NamingMessage
    {
        /**
         * For a dummy message or message received from peer, this points to the <code>OL_Message</code> object that was
         * received.
         *
         * This is <code>null</code> if generated by local FSM; it will change to a non-<code>null</code> value when it is
         * generated (lazily/on demand) because of a call to <code>send()</code>.
         */
        protected OL_Message olMessage;

        /** <code>null</code> if broadcast or local delivery unicast. */
        protected final I_LogicalAddress destinationAddress;

        /** <code>true</code> iff <code>olMessage != null</code> and message is from application (dummy). */
        protected final boolean generatedLocallyByAPI;

        /**
         * Unique identifier of a message that is sometimes matched by the receiver of the response corresponding to the
         * message. Whether or not matching occurs is based on configuration.
         */
        protected final int serialNumber;

        /**
         * <code>true</code> after either <code>sendBroadcast()</code> or <code>sendUnicast()</code> is called. Ensures that only
         * one send operation occurs per NamingMessage.
         *
         * @see #sendBroadcast()
         * @see #sendUnicast()
         */
        private boolean sent = false;

        // TODO(EXPERIMENTS): remove, only for testing
        public RoutableNamingMessage (final int serialNumber, final boolean generatedLocallyByAPI)
        {
            this.destinationAddress = null;
            this.generatedLocallyByAPI = generatedLocallyByAPI;
            this.olMessage = (OL_Message) socket.createMessage();
            this.serialNumber = serialNumber;
        }

        /**
         * Constructs a message object for a message generated by the FSM without a specific destination address.
         */
        public RoutableNamingMessage() {
            this (NO_SERIAL_NUMBER, null);
        }

        /**
         * Constructs a message object for a message generated by the FSM with a specific destination address.
         *
         * @param destinationAddress The destination of the message, can be null meaning local delivery.
         */
        public RoutableNamingMessage (final int serialNumber, final I_LogicalAddress destinationAddress)
        {
            this.generatedLocallyByAPI = false;
            this.serialNumber = serialNumber;

            this.destinationAddress =
                (destinationAddress != null
                 &&
                 olMessage != null
                 &&
                 olMessage.getDestinationAddress() != null
                 &&
                 getLogicalAddress().equals (olMessage.getDestinationAddress()))
                ? null
                : destinationAddress;
        }

        /**
         * Constructs a message object for a message received from a peer.
         */
        public RoutableNamingMessage (final OL_Message olMessage, final byte [] controlData) {
            this (olMessage, controlData, false);
        }

        /**
         * Constructs a message object for a message received from a peer or generated by an API call.
         */
        public RoutableNamingMessage (final OL_Message olMessage, final byte [] controlData, final boolean generatedLocallyByAPI)
        {
            super (controlData);

            final int serialNumberOffset = generatedLocallyByAPI ? 1 : 0;
            this.serialNumber = ByteArrayUtility.toInteger (controlData, serialNumberOffset);
            if (debugMessages)
                log.println ("RoutableNamingMessage.<init>: serialNumber: " + serialNumber);

            this.olMessage = olMessage;

            this.destinationAddress =
                (olMessage.getDestinationAddress() != null
                 &&
                 getLogicalAddress().equals (olMessage.getDestinationAddress()))
                ? null
                : olMessage.getDestinationAddress();

            this.generatedLocallyByAPI = generatedLocallyByAPI;
        }

        /**
         * Adds the specified type payload information to the FSM extension of the message, creating an FSM extension if
         * necessary. Assumes that there is a single FSM extension per overlay message.
         */
        protected final void createNamingMessageExtension (final byte type, final byte [] particularPayload)
        {
            final byte [] payload = particularPayload;

/*
call of packSerialNumber instead
            final byte [] payload = new byte [SERIAL_NUMBER_LENGTH + particularPayload.length];

            final int serialNumberOffset = 0;

            payload [serialNumberOffset + 0] = (byte) ((serialNumber >>> (8*3)) & 0xFF);
            payload [serialNumberOffset + 1] = (byte) ((serialNumber >>> (8*2)) & 0xFF);
            payload [serialNumberOffset + 2] = (byte) ((serialNumber >>> (8*1)) & 0xFF);
            payload [serialNumberOffset + 3] = (byte) ((serialNumber >>> (8*0)) & 0xFF);

            System.arraycopy (particularPayload, 0, payload, SERIAL_NUMBER_LENGTH, particularPayload.length);
*/

            final FSM_Extension existingFSMExtension = (FSM_Extension) olMessage.getFirstExtensionByType (Extension.FSM);
            if (existingFSMExtension != null)
            {
                // This branch of the "if" handles the situation with query messages which come from the API but are
                // sent to peers. The control data that the API provides is not indended for peers, so the existing fsm
                // extension control data must be replaced with one that has the format expected by peers.

                existingFSMExtension.setType (type);
                existingFSMExtension.setControlData (payload);
            }
            else
            {
                final FSM_Extension newExtension = new FSM_Extension (MessageStore.FSM_NAMING, type, FSM_Extension.NAMING_STREAM_ID, payload);
                olMessage.addExtension (newExtension);
            }
        }

        /**
         * Descendant classes implement this method with algorithms that marshal message information into an overlay
         * message and set overlay message parameters. Typically this means calling
         * <code>createNamingMessageExtension()</code> or setting the <code>hopLimit</code> of the message. This method
         * needs to be called only when a new overlay message is created or an overlay message is received from a
         * NamingAPI call, i.e. this method is not called for messages received from peers - this condition is enforced
         * by <code>ensureMessagePrepared()</code>.
         *
         * When invoked, it can be assumed: <code>olMessage != null</code> due to a prior invocation of
         * <code>ensureMessagePrepared()</code>.
         *
         * This method will be invoked only once per message. Descendant classes should not invoke this method directly to
         * ensure that only one invocation occurs.
         *
         * @see #createNamingMessageExtension(byte,byte[])
         * @see #ensureMessagePrepared()
         */
        protected abstract void prepareMessageForSend();

        /**
         * This method is used by descendant classes to prepend the current serial number to the payload of the naming
         * message FSM extension. For all routable messages, the serial number occupies the first four bytes of the FSM
         * extension payload.
         */
        protected final void packSerialNumber (final byte [] payload)
        {
            final int serialNumberOffset = 0;

            payload [serialNumberOffset + 0] = (byte) ((serialNumber >>> (8*3)) & 0xFF);
            payload [serialNumberOffset + 1] = (byte) ((serialNumber >>> (8*2)) & 0xFF);
            payload [serialNumberOffset + 2] = (byte) ((serialNumber >>> (8*1)) & 0xFF);
            payload [serialNumberOffset + 3] = (byte) ((serialNumber >>> (8*0)) & 0xFF);
        }

        /**
         * Only called from <code>send</code> methods (<code>sendBroadcast()</code> and <code>sendUnicast()</code>)
         *
         * @return <code>true</code> if message was generated locally, i.e. was not received from a peer - useful information in
         * deciding whether to send or forward (the issue is whether or not to set the source address).
         *
         * @see #sendBroadcast()
         * @see #sendUnicast()
         */
        private final boolean ensureMessagePrepared()
        {
            final boolean createdNewOLMessage;

            if (olMessage == null) {
                createdNewOLMessage = true;
                olMessage = (OL_Message) socket.createMessage();
            }
            else
                createdNewOLMessage = false;
                
            final boolean locallyGeneratedMessage = generatedLocallyByAPI || createdNewOLMessage;

            if (locallyGeneratedMessage)
                prepareMessageForSend();

            return locallyGeneratedMessage;
        }

        /**
         * A service provided by this class to its descendants; used to implement <code>send()</code> by descendant classes.
         */
        protected synchronized final void sendBroadcast()
        {
            if (! sent)
            {
                final boolean locallyGeneratedMessage = ensureMessagePrepared();

                if (debugMessages)
                    log.println ("sendBroadcast: " + this);

                // Be careful not to reset source address if message was not generated locally
                if (locallyGeneratedMessage)
                    sendToAll (olMessage);
                else
                    forwardToChildren (olMessage);
            }

            sent = true;
        }

        protected synchronized final void sendUnicast() {
            sendUnicast (destinationAddress);
        }

        protected synchronized final void sendUnicast (final String destinationAddressString) {
            sendUnicast (socket.createLogicalAddress (destinationAddressString));
        }

        /**
         * A service provided by this class to its descendants; used to implement <code>send()</code> by descendant classes.
         */
        protected synchronized final void sendUnicast (final I_LogicalAddress dest)
        {
            if (! sent
                &&
                dest != null)
            {
                final boolean locallyGeneratedMessage = ensureMessagePrepared();

                if (debugMessages)
                    log.println ("sendUnicast: " + this);

                // Be careful not to reset source address if message was not generated locally
                if (locallyGeneratedMessage)
                    sendToNode (olMessage, dest);
                else
                    forwardToParent (olMessage);
            }

            sent = true;
        }

        protected void deliverSatisfiedQueriesLocally (final Map satisfiedQueries)
        {
            final Iterator satisfiedQueriesIterator = satisfiedQueries.entrySet().iterator();
            while (satisfiedQueriesIterator.hasNext())
            {
                final Map.Entry satisfiedQuery = (Map.Entry) satisfiedQueriesIterator.next();
                final Integer serialNumberInteger = (Integer) satisfiedQuery.getKey();
                final Vector satisfyingBindingVector = (Vector) satisfiedQuery.getValue();

                final NamingBinding [] responseBindingsArray = (NamingBinding []) satisfyingBindingVector.toArray (new NamingBinding [0]);
                new LocalQueryResponse (serialNumberInteger.intValue(), responseBindingsArray).send();
            }
        }

        /**
         * Returns a string that contains the source and destination for this naming message.
         */
        public String toString()
        {
            final StringBuffer buffer = new StringBuffer (super.toString());

            final String origin;
            if (olMessage == null)
                origin = "locally by FSM";
            else if (generatedLocallyByAPI)
                origin = "locally by API";
            else
                origin = "by Peer: " + olMessage.getSourceAddress();

            buffer.append (" S/N (" + serialNumber + ")");
            if (olMessage != null)
                buffer.append (" hop limit (" + olMessage.getHopLimit() + ")");
            buffer.append (" (generated " + origin + ")");
            buffer.append (" destination (" + ((destinationAddress == null) ? "this socket" : destinationAddress.toString()) + "): ");

            return buffer.toString();
        }
    }

    /**
     * A message that contains bindings.
     */
    private abstract class BindingsMessage
        extends RoutableNamingMessage
    {
        /**
         * Bindings contained in the bindings message.
         */
        protected /*final*/ NamingBinding [] bindings;

        /**
         * Called to create a binding message by the FSM.
         */
        public BindingsMessage (final int serialNumber, final NamingBinding [] bindings, final I_LogicalAddress destinationAddress) {
            super (serialNumber, destinationAddress);
            this.bindings = bindings;
        }

        /**
         * Called to create a binding message using information received from a peer.
         */
        public BindingsMessage (final OL_Message olMessage, final byte [] controlData, final boolean invalidateBindings)
        {
            super (olMessage, controlData);

            if (debugBindingsMessage)
                log.println ("BindingsMessage: controlData.length: " + controlData.length);

/*
do not decode this information unconditionally - we might not be snooping!
            bindings = NamingBinding.restoreNamingBindings (controlData, SERIAL_NUMBER_LENGTH, invalidateBindings);

            if (debugBindingsMessage && bindings == null) {
                log.println ("BindingsMessage.<init>(): bindings is null");
                Thread.dumpStack();
            }
*/
        }

        /**
         * Returns a string that contains the bindings of this naming message.
         */
        public final String toString()
        {
            final StringBuffer buffer = new StringBuffer (super.toString());

            if (bindings != null)
                for (int i = 0; i < bindings.length; ++i)
                    buffer.append ("\n    " + bindings[i]);
            else
                buffer.append ("\n    bindings not decoded");

            return buffer.toString();
        }

        protected final byte [] bindingsToByteArray()
        {
            if (debugBindingsMessage && bindings == null) {
                log.println ("BindingsMessage.bindingsToByteArray(): bindings is null");
                Thread.dumpStack();
            }
            
            if (debugBindingsMessage)
                log.println ("BindingsMessage.bindingsToByteArray(): bindings.length: " + bindings.length);

            final byte [] result = NamingBinding.bindingsToByteArray (bindings, digitalSigner, certificateCache);
            
            if (debugBindingsMessage && result == null)
                log.println ("BindingsMessage.bindingsToByteArray(): result is null");

            return result;
        }
    }

    /**
     * Parent a message for all message types that propagate binding information. Contains data and algorithms common to
     * binding propagation messages.
     */
    private abstract class PropagateBindingsMessage
        extends BindingsMessage
    {
        /**
         * Used to construct a propagate bindings message that is generated by the naming service FSM.
         */
        public PropagateBindingsMessage (final NamingBinding [] bindings, final I_LogicalAddress destinationAddress) {
            this (NO_SERIAL_NUMBER, bindings, destinationAddress);
        }

        /**
         * Used to construct a propagate bindings message that is generated by the naming service FSM.
         */
        public PropagateBindingsMessage (final int serialNumber, final NamingBinding [] bindings, final I_LogicalAddress destinationAddress) {
            super (serialNumber, bindings, destinationAddress);
        }

        private static final boolean constructInvalidationBindings = false;

        /**
         * Constructor used to reconstruct a binding propagate message received from a peer.
         */
        public PropagateBindingsMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData, constructInvalidationBindings);
        }

        /**
         * <ol>
         *
         * <li>Forward the message along (possibly culling/updating the bindings so that information that is known to be
         * bad/stale is not propagated?)</li>
         *
         * <li>Put all bindings into the binding table - the table will merge new information with existing information
         * (this process could do the culling/updating?)</li>
         *
         * <li>Inspect the pending name and address query tables to see if any new bindings match entries in these
         * tables</li>
         *
         *   <ol>
         *
         *   <li> If there are matching entries, attempt to satisfy queries with new information.</li>
         *
         *   <li> If new information satisfies pending queries, but new information is not trusted, initiate certificate
         *        request (placing binding in the pending certificate table).</li>
         *
         *   </ol>
         * </ol>
         */
        public final void process()
        {
            if (debugPropagateBindingsMessage)
                log.println ("PropagateBindingsMessage.process(): " + this);

            // TODO(EXPERIMENTS)
            if (tracingQueries
                &&
                serialNumber % tracingModulo == 0)
            {
                final I_LogicalAddress previousHop = olMessage == null ? null : olMessage.getPreviousHopAddress();

                if (traceToStderr)
                    System.err.println ("Pull  " + getLogicalAddress() + " " + previousHop + " " + System.currentTimeMillis() + " " + serialNumber);

                synchronized (experimentOutLock)
                {
                    if (experimentOut != null)
                        experimentOut.println ("Pull  " + getLogicalAddress() + " " + previousHop + " " + System.currentTimeMillis() + " " + serialNumber);

                    // + " " + trace_sdf.format (new Date (System.currentTimeMillis())));
                }
            }

            send();

            //
            // There are times (like when latency measurements are being taken) when we do not wish for a Naming Service
            // FSM to use information destined for another Naming Service FSM, i.e. we do not want to snoop pull
            // messages. We want the query to be satisfied only if the pull message is a reply to the query that we sent.
            // Specifying authoritative responses only is not enough since we could snoop an authoritative response.
            //
           
            if (! policy.snoopBindings()
                &&
                destinationAddress != null // (null -> push OR local unicast delivery) -> not a reply to a particular peer, so accepting it is not considered to be snooping
                &&
                ! destinationAddress.equals (getLogicalAddress()))
            {
                if (debugPropagateBindingsMessage)
                    log.println ("PropagateBindingsMessage.process(): not processing message because not snooping: " +
                                 "destinationAddress (" + destinationAddress + ") " +
                                 "logical address (" + getLogicalAddress() + ")");
                return;
            }

            this.bindings = NamingBinding.restoreNamingBindings (controlData, SERIAL_NUMBER_LENGTH, /*invalidateBindings*/ false);

            if (debugBindingsMessage && bindings == null) {
                log.println ("BindingsMessage.<init>(): bindings is null");
                Thread.dumpStack();
            }

            doBindingCaching();

            if (debugPropagateBindingsMessage)
                log.println ("PropagateBindingsMessage.process(): considering " + bindings.length + " bindings");

            final Map satisfiedQueries = new HashMap();

            final Map untrustedLogicalAddressQueries = new HashMap();
            final Map untrustedNameQueries = new HashMap();

            final int serialNumberMatch; // use this quantity to specify which queries may be satisfied by these bindings

            if (! policy.getRequireBindingMessageSerialNumberMatch()) // serial numbers do not have to match, so use NO_SERIAL_NUMBER as a wildcard
                serialNumberMatch = NO_SERIAL_NUMBER;
            else if (serialNumber == NO_SERIAL_NUMBER) // serial numbers must match and this propagate message has no particular serial number (e.g. a push message) -> do not attempt to satisfy anything
                return;
            else
                serialNumberMatch = serialNumber;

            for (int i = 0; i < bindings.length; ++i)
            {
                if (pendingLogicalAddressQueryTable.size() > 0)
                    pendingLogicalAddressQueryTable.attemptToSatisfyQueries (serialNumberMatch, bindings[i], satisfiedQueries, untrustedLogicalAddressQueries);

                if (pendingNameQueryTable.size() > 0)
                    pendingNameQueryTable.attemptToSatisfyQueries (serialNumberMatch, bindings[i], satisfiedQueries, untrustedNameQueries);
            }

            if (debugPropagateBindingsMessage)
                log.println ("PropagateBindingsMessage.process():"                                                 + "\n" +
                             "    satisfiedQueries.size(): "               + satisfiedQueries.size()               + "\n" +
                             "    untrustedLogicalAddressQueries.size(): " + untrustedLogicalAddressQueries.size() + "\n" +
                             "    untrustedNameQueries.size(): "           + untrustedNameQueries.size());

            final Iterator laUntrustedQueriesIterator = untrustedLogicalAddressQueries.entrySet().iterator();
            while (laUntrustedQueriesIterator.hasNext())
            {
                final Map.Entry entry = (Map.Entry) laUntrustedQueriesIterator.next();
                final Vector untrustedLogicalAddressQueryBindingVector = (Vector) entry.getValue();
                for (int i = 0; i < untrustedLogicalAddressQueryBindingVector.size(); ++i)
                {
                    final NamingBinding untrustedBinding = (NamingBinding) untrustedLogicalAddressQueryBindingVector.elementAt(i);

                    final String name = untrustedBinding.getName();
                    final String logicalAddressString = untrustedBinding.getLogicalAddressString();

                    final String startOfTrustChain = untrustedBinding.getSignerName();
                    final String certificateLocationHint = logicalAddressString;

                    final int trustSerialNumber = getNextTrustSerialNumber();
                    final boolean messageSent = initiateTrustChainOperation (trustSerialNumber, startOfTrustChain, certificateLocationHint);
                    if (messageSent)
                        pendingUntrustedQueryTable.putLogicalAddressQuery (trustSerialNumber, serialNumberMatch, startOfTrustChain, name, logicalAddressString);
                }
            }

            final Iterator nameUntrustedQueriesIterator = untrustedNameQueries.entrySet().iterator();
            while (nameUntrustedQueriesIterator.hasNext())
            {
                final Map.Entry entry = (Map.Entry) nameUntrustedQueriesIterator.next();
                final Vector untrustedNameQueryBindingVector = (Vector) entry.getValue();
                for (int i = 0; i < untrustedNameQueryBindingVector.size(); ++i)
                {
                    final NamingBinding untrustedBinding = (NamingBinding) untrustedNameQueryBindingVector.elementAt(i);

                    final String name = untrustedBinding.getName();
                    final String logicalAddressString = untrustedBinding.getLogicalAddressString();

                    final String startOfTrustChain = untrustedBinding.getSignerName();
                    final String certificateLocationHint = logicalAddressString;

                    final int trustSerialNumber = getNextTrustSerialNumber();
                    final boolean messageSent = initiateTrustChainOperation (trustSerialNumber, startOfTrustChain, certificateLocationHint);
                    if (messageSent)
                        pendingUntrustedQueryTable.putNameQuery (trustSerialNumber, serialNumberMatch, startOfTrustChain, logicalAddressString, name);
                }
            }

            deliverSatisfiedQueriesLocally (satisfiedQueries);
        }

        protected abstract void doBindingCaching();
    }

    /**
     * A Pull Message contains bindings that are in response to a Name or Logical Address query. Pull messages are sent
     * unicast from the peer responding to a query to the peer that sent the query. All intermediate peers on the path
     * between the peer that sends the Pull Message and the peer that receives it add the bindings contained in the Pull
     * Messages to their Peer Binding Cache.
     */
    private final class PullBindingsMessage
        extends PropagateBindingsMessage
    {
        /**
         * Constructor used to reconstruct a Pull Message received from a peer.
         */
        public PullBindingsMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData);
        }

        /**
         * Constructor used by FSM to create a Pull Message
         */
        public PullBindingsMessage (final int serialNumber, final NamingBinding [] bindings, final I_LogicalAddress destinationAddress) {
            super (serialNumber, bindings, destinationAddress);
        }

        protected void doBindingCaching()
        {
            if (policy.cachePulledBindings())
            {
                if (debugPropagateBindingsMessage)
                    log.println ("PullBindingsMessage.doBindingCaching(): " + bindings.length + " bindings");

                peerBindingCachePull.put (bindings);
            }
        }

        /**
         * Encode the bindings into a byte array for transmission.
         */
        protected void prepareMessageForSend()
        {
            final byte [] payload = bindingsToByteArray();
            packSerialNumber (payload);
            createNamingMessageExtension (PULL_BINDINGS, payload);
            olMessage.setHopLimit (policy.getQueryRadius()); // pull distance mirrors query distance
        }

        /**
         * Pull messages are unicast.
         */
        protected void send() {
            sendUnicast();
        }
    }

    /**
     * A Push Message contains bindings that are broadcast to peers on a periodic basis or in response to an event such
     * as a logical address change or the installation of a new name for use in bindings. All peers that receive a push
     * message add the bindings contained in the Push Messages to their Peer Binding Cache.
     */
    private final class PushBindingsMessage
        extends PropagateBindingsMessage
    {
        /**
         * Constructor used to reconstruct a Push Message received from a peer.
         */
        public PushBindingsMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData);
        }

        /**
         * Constructor used by FSM to create a Push Message
         */
        public PushBindingsMessage (final NamingBinding [] bindings) {
            super (bindings, null);
        }

        protected void doBindingCaching() {
            peerBindingCachePush.put (bindings);
        }

        /**
         * Encode the bindings into a byte array for transmission. Limit the Push Message to the configured Push radius.
         */
        protected void prepareMessageForSend() {
            createNamingMessageExtension (PUSH_BINDINGS, bindingsToByteArray());
            olMessage.setHopLimit (policy.getPushRadius());
        }

        /**
         * Push messages are broadcast.
         */
        protected void send() {
            sendBroadcast();
        }
    }

    /**
     * Message that causes bindings to be removed from the binding table.
     */
    private final class InvalidateBindingsMessage
        extends BindingsMessage
    {
        /**
         * Constructor used to reconstruct an Invalidate Message received from a peer.
         */
        public InvalidateBindingsMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData, true);
        }

        /**
         * Constructor used by FSM to create an Invalidate Message
         */
        public InvalidateBindingsMessage (final NamingBinding [] bindings) {
            super (NO_SERIAL_NUMBER, bindings, null);
        }

        /**
         * Propagate the invalidate message to peers and remove bindings from local cache.
         */
        public void process() {
            send();
            peerBindingCachePush.remove (bindings);
            peerBindingCachePull.remove (bindings);
        }

        /**
         * Encode the bindings into a byte array for transmission. Limit the Invalidate Message to the configured
         * Invalidate radius.
         */
        protected void prepareMessageForSend() {
            createNamingMessageExtension (INVALIDATE_BINDINGS, bindingsToByteArray());
            olMessage.setHopLimit (policy.getInvalidateRadius());
        }

        /**
         * Invalidate messages are always broadcast.
         */
        protected void send() {
            sendBroadcast();
        }
    }

    private static final int NO_SERIAL_NUMBER = 0;

    /**
     * A message that is sent to search for bindings. The algorithms for logical address and name queries are quite
     * similar, in this implementation the parts that are uncommon are factored into abstract methods that each
     * particular query type implements.
     */
    private abstract class BindingQueryMessage
        extends RoutableNamingMessage
    {
        // local interest only
        protected final boolean requiresTrusted;
        protected final long deadline; // absolute ms // perhaps of local interest only...
        protected final boolean requiresSerialNumberMatch;

        // of remote interest: fixed
        protected final boolean requiresAuthoritative;
        protected final int maxAge; // relative ms

        // of remote interest: variable
        protected int maxResponses;
        protected int minHopCount;
        protected int maxHopCount; // can override queryRadius from configuration

        // TODO(EXPERIMENTS): only for testing
        protected BindingQueryMessage (final boolean requiresTrusted,
                                       final long    deadline,
                                       final boolean requiresSerialNumberMatch,
                                       final boolean requiresAuthoritative,
                                       final int     maxAge,
                                       final int     maxResponses,
                                       final int     minHopCount,
                                       final int     serialNumber)
        {
            super (serialNumber, true); // generatedLocallyByAPI
            this.requiresTrusted = requiresTrusted;
            this.deadline = deadline;
            this.requiresAuthoritative = requiresAuthoritative;
            this.requiresSerialNumberMatch = requiresSerialNumberMatch;
            this.maxAge = maxAge;
            this.maxResponses = maxResponses;
            this.minHopCount = minHopCount;
        }

        /**
         * This is the base class constructor for a query message, either logical address or name query. A query can come
         * from one of two sources: an API call or a peer - queries are not generated by the naming service FSM. Since API
         * calls and peer message are both received as an OL_Message, a single constructor can be used with a flag that
         * indicates the source of the message.
         *
         * This constructor is responsible for parsing the contents of the serialized query naming message. The format is
         * slightly different between the message received from an API versus a message received from a peer.
         *
         * @param olMessage The overlay message received from the API or peer
         * @param controlData A byte array containing the contents of the FSM extension with the naming message data
         * @param generatedLocallyByAPI A flag that indicates the source of the message - true means API, false means peer
         */
        protected BindingQueryMessage (final OL_Message olMessage, final byte [] controlData, final boolean generatedLocallyByAPI)
        {
            super (olMessage, controlData, generatedLocallyByAPI);

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage.<init>: controlData.length: " + controlData.length + ": " + hypercast.HTTP_ServerUtility.toHexString (controlData));

            final int [] index;
            if (generatedLocallyByAPI)
            {
                index = new int [] { MESSAGE_TYPE_LENGTH + SERIAL_NUMBER_LENGTH };
                
                requiresTrusted = (controlData [index [0]] == (byte) 1);
                index[0] += TRUSTED_LENGTH;

                if (debugBindingQueryMessageControlData)
                    log.println ("BindingQueryMessage.<init>: requiresTrusted: " + requiresTrusted);

                deadline = ByteArrayUtility.toLong (controlData, index[0]);
                index[0] += DEADLINE_LENGTH;

                if (debugBindingQueryMessageControlData)
                {
                    final long now = System.currentTimeMillis();
                    log.println ("BindingQueryMessage.<init>: deadline: in " + ((deadline - now) / MILLISECONDS_PER_SECOND) + " secs");
                }

                requiresSerialNumberMatch = (controlData [index [0]] == (byte) 1);
                index[0] += SERIAL_NUMBER_MATCH_LENGTH;

                if (debugBindingQueryMessageControlData)
                    log.println ("BindingQueryMessage.<init>: requiresSerialNumberMatch: " + requiresSerialNumberMatch);
            }
            else
            {
                index = new int [] { SERIAL_NUMBER_LENGTH }; // start at beginning of packet, after the message serial number

                requiresTrusted = false; // request is from peer, only that peer knows if it needs a trusted cert, furthermore only it can decide trust for itself
                deadline = 0; // this value is not used by a remote query
                requiresSerialNumberMatch = false; // this value is not used by a remote query
            }

            requiresAuthoritative = (controlData [index [0]] == (byte) 1);
            index[0] += AUTHORITATIVE_LENGTH;

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage.<init>: requiresAuthoritative: " + requiresAuthoritative);

            maxAge = ByteArrayUtility.toInteger (controlData, index[0]);
            index[0] += MAXIMUM_AGE_LENGTH;

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage.<init>: maxAge: " + (maxAge / MILLISECONDS_PER_SECOND) + " secs old");

            maxResponses = ByteArrayUtility.toInteger (controlData, index[0]);
            index[0] += MAX_RESPONSES_LENGTH;

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage.<init>: maxResponses: " + maxResponses);

            minHopCount = ByteArrayUtility.toShort (controlData, index[0]);
            index[0] += MINIMUM_HOP_COUNT_LENGTH;

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage.<init>: minHopCount: " + minHopCount);

            maxHopCount = ByteArrayUtility.toShort (controlData, index[0]);
            index[0] += MAXIMUM_HOP_COUNT_LENGTH;

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage.<init>: maxHopCount: " + maxHopCount);
        }

        private final static int PARAMETERS_LENGTH = AUTHORITATIVE_LENGTH + MAXIMUM_AGE_LENGTH + MAX_RESPONSES_LENGTH + MINIMUM_HOP_COUNT_LENGTH;

        private void packParameters (final byte [] result, final int offset)
        {
            int index = offset;

            // authoritative response required
            result [index + 0] = (byte) ((requiresAuthoritative ? 1 : 0) & 0xFF);
            index += AUTHORITATIVE_LENGTH;
        
            // maximum allowable age of a response
            result [index + 0] = (byte) ((maxAge >>> (8*3)) & 0xFF);
            result [index + 1] = (byte) ((maxAge >>> (8*2)) & 0xFF);
            result [index + 2] = (byte) ((maxAge >>> (8*1)) & 0xFF);
            result [index + 3] = (byte) ((maxAge >>> (8*0)) & 0xFF);
            index += MAXIMUM_AGE_LENGTH;

            // maxResponses
            result [index + 0] = (byte) ((maxResponses >>> (8*3)) & 0xFF);
            result [index + 1] = (byte) ((maxResponses >>> (8*2)) & 0xFF);
            result [index + 2] = (byte) ((maxResponses >>> (8*1)) & 0xFF);
            result [index + 3] = (byte) ((maxResponses >>> (8*0)) & 0xFF);
            index += MAX_RESPONSES_LENGTH;

            // minHopCount
            result [index + 0] = (byte) ((((minHopCount == 0) ? 0 : minHopCount - 1) >>> (8*1)) & 0xFF);
            result [index + 1] = (byte) ((((minHopCount == 0) ? 0 : minHopCount - 1) >>> (8*0)) & 0xFF);
            index += MINIMUM_HOP_COUNT_LENGTH;
        }

        /** Returns either logical address string or name depending on type. */
        protected abstract String getQueryString();

        /** Returns either LOGICAL_ADDRESS_QUERY or NAME_QUERY depending on type */
        protected abstract byte getQueryType();

        /** Implements type specific (LA or name query) algorithm of extracting relevant local binding */
        protected abstract NamingBinding [] queryLocalBindingTable();

        /** Returns the type specific pending query table, either LA or name */
        protected abstract PendingQueryTable getPendingQueryTable();

        /**
         * Implemented by particular query types to start the process of acquiring necessary information for establishing trust.
         */
        protected abstract void establishTrustChain (final String nextNeededCert, final String name, final String logicalAddressString);

        /**
         * Implemented by particular query types to return relevant bindings from the peer bindings cache.
         */
        protected abstract NamingBinding [] queryPeerBindingsCache();

        /**
              1. if local query, place entry in pending query table (either name or LA)
              2. lookup in LBT and PBC
              3. cull results based on pending query
                  (make list of needed certs to establish trust - note that trust won't help if binding is too old or not auth - these *trump* trust)
                  for example, after lookup in PBC and LBT you will have some bindings, some will be too old, some not auth, some not trusted.
                  these attributes are orthogonal - asssume need trusted: there will be some bindings that satisfy query that are trusted and some that
                  aren't, and some that don't satisfy and aren't trusted - do not ask for certs for the latter group - what good will it do? only ask for certs
                  for those bindings that would sat query and aren't trust - and only do that if the query isn't totally sat with the trusted bindings that you have
              4. update pending query table
              5a. if local query not totally satisfied, propagate query to peers
              5b. if remote query not satisfied or min hop not reached, propagate query to peers
              6. if local query at least partially satisfied, delivery satisfactory binding information to application
              7. if remote query satisfied, respond to peer (note for remote can both propagate and respond)
              8. if satisfiying bindings exist for local query but cannot be trusted currently AND local query not totally sat, initiate exchange for missing certificates
         */
        public void process()
        {
            ++queryMessagesProcessed;

            final I_LogicalAddress querier = generatedLocallyByAPI ? null : olMessage.getSourceAddress();
            final boolean localQuery = (querier == null);

            // TODO(EXPERIMENTS)
            if (tracingQueries
                &&
                serialNumber % tracingModulo == 0)
            {
                final I_LogicalAddress previousHop = localQuery ? null : olMessage.getPreviousHopAddress();

                if (traceToStderr)
                    System.err.println ("Query " + getLogicalAddress() + " " + previousHop + " " + System.currentTimeMillis() + " " + serialNumber);

                synchronized (experimentOutLock)
                {
                    if (experimentOut != null)
                        experimentOut.println ("Query " + getLogicalAddress() + " " + previousHop + " " + System.currentTimeMillis() + " " + serialNumber);

                    // + " " + trace_sdf.format (new Date (System.currentTimeMillis())));
                }
            }

            final PendingQueryTable pendingQueryTable = getPendingQueryTable();
            if (localQuery)
                pendingQueryTable.put (getQueryString(), requiresTrusted, requiresAuthoritative, maxAge, maxResponses, deadline, serialNumber); 

            final NamingBinding [] localBindings = queryLocalBindingTable();
            // Note if query is satisfied by local bindings, the peer cache lookup is unnecessary work - local bindings should always be accepted - right?

            // if query is from a peer and it requires authoritative responses then nothing from our cache will help.
            // For latency experiments with trust I need to cache bindings that are received so that they can be found when
            // cooresponding certificates are received, but I do not want to cache bindings so that they can be found by queries. I
            // want to bypass the peer binding cache on query or "pseudo-cached" bindings.
            final NamingBinding [] peerBindings = ((! localQuery && requiresAuthoritative) || policy.getBypassPeerBindingsCacheOnQuery()) ? new NamingBinding [0] : queryPeerBindingsCache();

            if (peerBindings.length > 0)
                ++cacheHits;

            if (debugQueryMessage)
            {
                log.println ("Query: " + (localQuery ? "local" : "peer") + " query for " + getQueryString());

                log.println ("Query: retrieved " + localBindings.length + " local bindings from table:");
                for (int i = 0; i < localBindings.length; ++i)
                    log.println ("Query: retrieved local binding from table: " + localBindings [i]);

                log.println ("Query: retrieved " + peerBindings.length + " bindings from peer cache:");
                for (int i = 0; i < peerBindings.length; ++i)
                    log.println ("Query: retrieved peer binding from cache: " + peerBindings [i]);
            }

            final Vector resultBindings;

            // Only verify trust at the receiver: responder can propagate bindings that are not (yet) trusted by responder -
            // otherwise periodic push operation and pull operation would be worthless. Basically the system is optimistic
            // and individual naming service FSMs forward information that they do not trust with the assumption that the
            // information is not malicious. The node receiving the information will always verify the trust of the
            // information so malicious info will not make it to the user.

            if (peerBindings.length == 0 && localBindings.length == 0)
            {
                resultBindings = new Vector();
            }
            else if (! localQuery)
            {
                resultBindings = new Vector();
                for (int i = 0; i < localBindings.length + peerBindings.length; ++i)
                {
                    final NamingBinding binding = (i < localBindings.length) ? localBindings [i] : peerBindings [i - localBindings.length];
                    resultBindings.add (binding);
                }
            }
            else // local
            {
                if (debugQueryMessage)
                    log.println ("Query: local query with " + maxResponses + " possible responses remaining");

                final Integer serialNumberInteger = new Integer (serialNumber);

                final Map satisfiedQueries = new HashMap();
                final Map untrustedQueries = new HashMap();

                // Attempt to satisfy the query with each binding that was found in either the local binding table or the
                // peer binding cache. If any binding is found to be untrusted (which means that it is good on all other
                // accounts) add it to a list of untrusted bindings - trust information for these bindings may need to be
                // acquired to satisfy the query.

                int resultsSoFar = 0;
                for (int i = 0; (i < localBindings.length + peerBindings.length) && (resultsSoFar < maxResponses); ++i)
                {
                    final NamingBinding binding = (i < localBindings.length) ? localBindings [i] : peerBindings [i - localBindings.length];

                    if (pendingQueryTable.size() > 0)
                    {
                        pendingQueryTable.attemptToSatisfyQueries (serialNumber, binding, satisfiedQueries, untrustedQueries);
                        final Vector resultsSoFarVector = (Vector) satisfiedQueries.get (serialNumberInteger);
                        resultsSoFar = (resultsSoFarVector == null) ? 0 : resultsSoFarVector.size();
                    }
                }

                final Vector queryBindings = (Vector) satisfiedQueries.get (serialNumberInteger);
                resultBindings = (queryBindings != null) ? queryBindings : new Vector();

                // assert 0 <= satisfiedQueries.size() <= 1. We specified the serial number of only this query, either bindings were found that satisfied it or not
                // assert 0 <= untrustedQueries.size() <= 1. We specified the serial number of only this query, either trusted bindings were found that satisfied it or not

                // If the query is not satisfied, but some untrusted bindings were found, attempt to acquire trust
                // information that will allow these bindings to satisfy the query

                if (resultBindings.size() < maxResponses)
                {
                    final Vector untrustedResultBindings = (Vector) untrustedQueries.get (serialNumberInteger);
                    if (untrustedResultBindings != null)
                    {
                        if (debugQueryMessage && untrustedResultBindings.size() > 0)
                            log.println ("Query: There are " + (maxResponses - resultBindings.size()) +
                                         " possible responses remaining, and " + untrustedResultBindings.size() +
                                         " bindings were found that are not trusted - initiating certificate exchange");

                        // duplicate names/certs in this loop will be ignored in cert request logic
                        for (int i = 0; i < untrustedResultBindings.size(); ++i)
                        {
                            final NamingBinding binding = (NamingBinding) untrustedResultBindings.elementAt (i);

                            final String bindingSignerName = binding.getSignerName();
                            final String bindingLogicalAddressString = binding.getLogicalAddressString();

                            final CertificateCache.Entry entry = policy.getCertificateCaching() ? certificateCache.get (bindingSignerName) : null;

                            // can be null if we have all certs in chain, but the chain does not end with trust anchor.
                            final String nextNeededCert = (entry == null) ? bindingSignerName : entry.getNextNeededIssuerInChain();

                            if (nextNeededCert != null)
                                establishTrustChain (nextNeededCert, bindingSignerName, bindingLogicalAddressString);
                        }
                    }
                }
            }

            // If this is a local query and it remains unsatisfied after inspecting the local binding table and the peer
            // binding cache, send the query to peers. Also, send the query on to peers if it has a minimum hop count
            // specified that has not been fulfilled.

            if (resultBindings.size() < maxResponses
                ||
                --minHopCount > 0)
            {
                if (debugQueryMessage)
                    log.println ("Query: propagating query to peers... " +
                                 " resultBindings.size(): " + resultBindings.size() +
                                 " maxResponses: "          + maxResponses +
                                 " minHopCount: "           + minHopCount);

                send();
            }

            // The results have been computed, deliver them locally or to a peer depending on where the query came from.

            if (resultBindings.size() > 0)
            {
                final NamingBinding [] bindings = (NamingBinding []) resultBindings.toArray (new NamingBinding [0]);

                if (debugQueryMessage)
                    log.println ("Query: returning " + bindings.length + " bindings to " + ((localQuery) ? "application" : querier.toString()));

                if (localQuery)
                    new LocalQueryResponse (serialNumber, bindings).send();
                else
                    new PullBindingsMessage (serialNumber, bindings, querier).send();
            }
        }

        /**
         * Regardless of the origin of this message (either from API or peer) always marshal parameters of query into an
         * array to be sent to the next hop. This is because parameters can change such as minHopCount and maxResponses.
         */
        protected void prepareMessageForSend()
        {
            final byte [] controlDataForPeer = runLengthEncodeString (SERIAL_NUMBER_LENGTH + PARAMETERS_LENGTH, getQueryString());

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage: controlDataForPeer.length: " + controlDataForPeer.length);

            packSerialNumber (controlDataForPeer);
            packParameters (controlDataForPeer, SERIAL_NUMBER_LENGTH);

            if (debugBindingQueryMessageControlData)
                log.println ("BindingQueryMessage: controlDataForPeer.length: " + controlDataForPeer.length + ": " + hypercast.HTTP_ServerUtility.toHexString (controlDataForPeer));

            createNamingMessageExtension (getQueryType(), controlDataForPeer);
            olMessage.setHopLimit (maxHopCount == USE_CONFIGURED_QUERY_RADIUS ? policy.getQueryRadius() : (short) maxHopCount);
        }

        /**
         * Returns a string that contains the parameters of this naming message.
         */
        public String toString()
        {
            final StringBuffer buffer = new StringBuffer (super.toString());
            final long now = System.currentTimeMillis();

            if (generatedLocallyByAPI)
            {
                buffer.append ("\n    requiresTrusted"       + " ("    + requiresTrusted         + ")");
                buffer.append ("\n    deadline"              + " (in " + ((deadline - now) / MILLISECONDS_PER_SECOND) + " secs)");
            }

            buffer.append ("\n    requiresAuthoritative" + " ("    + requiresAuthoritative              + ")");
            buffer.append ("\n    maxAge"                + " ("    + (maxAge / MILLISECONDS_PER_SECOND) + " secs old)");
            buffer.append ("\n    maxResponses"          + " ("    + maxResponses                       + ")");
            buffer.append ("\n    minHopCount"           + " ("    + minHopCount                        + ")");
            buffer.append ("\n    query: ");

            return buffer.toString();
        }
    }

    /**
     * A logical address query message contains a name that is queried for its logical address(es). This type of message
     * can be generated by an API call or it can be received from a peer. Since only a name is supplied this message is
     * propagated via a broadcast operation.
     */
    private final class LogicalAddressQueryMessage
        extends BindingQueryMessage
    {
        /**
         * The name that will be queried for its logical address(es).
         */
        private final String name;

        /**
         * Constructs a logical address query message using the information contained in a message received from an API
         * call or from a peer.
         */
        public LogicalAddressQueryMessage (final OL_Message olMessage, final byte [] controlData, final boolean generatedLocallyByAPI) {
            super (olMessage, controlData, generatedLocallyByAPI);
            final int [] index = new int [] { generatedLocallyByAPI ? LOCAL_QUERY_PREFIX_LENGTH : PEER_QUERY_PREFIX_LENGTH };
            this.name = runLengthDecodeString (controlData, index);
        }

        // TODO(EXPERIMENTS)
        public LogicalAddressQueryMessage (final int serialNumber, final String name, final long deadline)
        {
            super (policy.getTrustedResponsesOnly(),
                   deadline,
                   policy.getRequireBindingMessageSerialNumberMatch(),
                   policy.getAuthoritativeResponsesOnly(),
                   policy.getMaximumResponseBindingAge(),
                   policy.getMaximumResponsesPerQuery(),
                   policy.getMinimumQueryHopCount(),
                   serialNumber);

            this.name = name;
        }

        /**
         * For any given name, there can be only one local binding table entry;
         *
         * @return an array of length one containing the local binding with "name," or a zero length (non-null) array if
         * no such binding exists.
         */
        protected final NamingBinding [] queryLocalBindingTable()
        {
            final NamingBinding binding = localBindingTable.get (name);
            return (binding != null) ? new NamingBinding [] { binding } : new NamingBinding [0];
        }

        /**
         * Interrogates the peer bindings cache by name using the name supplied in the query.
         */
        protected NamingBinding [] queryPeerBindingsCache()
        {
            final NamingBinding [] pull = peerBindingCachePull.getByName (name);
            final NamingBinding [] push = peerBindingCachePush.getByName (name);

            final NamingBinding [] results = new NamingBinding [pull.length + push.length];

            int j = 0;
            for (int i = 0; i < pull.length; ++i)
                results [j++] = pull [i];

            for (int i = 0; i < push.length; ++i)
                results [j++] = push [i];

            return results;
        }

        /**
         * Returns the particular pending query table used by this type of query.
         */
        protected final PendingQueryTable getPendingQueryTable() {
            return pendingLogicalAddressQueryTable;
        }

        /**
         * Called when a trust chain is needed to establish the trust of a binding. An entry for this logical address
         * query is placed in the pending binding table (the table that links the naming binding exchange mechanism with
         * the certificate exchange mechanism).
         *
         * @param nextNeededCert The name of the next missing certificate in the required chain of trust
         * @param name The name portion of the binding for which the chain of trust is being established
         * @param logicalAddressString The logical address portion of the binding for which the chain of trust is being
         * established
         */
        protected final void establishTrustChain (final String nextNeededCert, final String name, final String logicalAddressString)
        {
            final int trustSerialNumber = getNextTrustSerialNumber();
            final boolean messageSent = initiateTrustChainOperation (trustSerialNumber, nextNeededCert, logicalAddressString);
            if (messageSent)
                pendingUntrustedQueryTable.putLogicalAddressQuery (trustSerialNumber, serialNumber, nextNeededCert, name, logicalAddressString);
        }

        /**
         * Returns the string that is used for the query. In the case of a logical address query this is the name.
         */
        protected String getQueryString() {
            return name;
        }

        /**
         * Returns the particular type of query message.
         */
        protected byte getQueryType() {
            return LOGICAL_ADDRESS_QUERY;
        }

        /**
         * Logical address queries are broadcast.
         */
        protected void send() {
            sendBroadcast();
        }

        /**
         * Returns a string that contains the name that is being used for this query.
         */
        public String toString() {
            return new StringBuffer (super.toString()).append (name).toString();
        }
    }

    /**
     * The base class for messages that queries names locally or from peers.
     */
    private abstract class NameQueryMessage
        extends BindingQueryMessage
    {
        /**
         * Construct for naming query messages.
         */
        public NameQueryMessage (final OL_Message olMessage, final byte [] controlData, final boolean generatedLocallyByAPI) {
            super (olMessage, controlData, generatedLocallyByAPI);
        }
    }

    /**
     * This class queries names of remote and local sockets.
     *
     * Note: destination initially null/local but if local table can't satisfy query, destination changes to that of logicalAddressString
     */
    private final class NonLocalNameQueryMessage
        extends NameQueryMessage
    {
        /**
         * The logical address string that is being queried for names bound to it.
         */
        private final String logicalAddressString;

        /**
         * This constructor is called for queries that originate in the local socket (i.e. via API call) as well as
         * those received from peer sockets.
         */
        public NonLocalNameQueryMessage (final OL_Message olMessage, final byte [] controlData, final boolean generatedLocallyByAPI)
        {
            super (olMessage, controlData, generatedLocallyByAPI);

            final int [] index = new int [] { generatedLocallyByAPI ? LOCAL_QUERY_PREFIX_LENGTH : PEER_QUERY_PREFIX_LENGTH };
            this.logicalAddressString = runLengthDecodeString (controlData, index);
        }

        /**
         * Called when a trust chain must be created for a binding that would satisfy this query if it were trusted. The
         * query is noted in the appropriate pending binding table.
         */
        protected final void establishTrustChain (final String nextNeededCert, final String name, final String logicalAddressString)
        {
            final int trustSerialNumber = getNextTrustSerialNumber();
            final boolean messageSent = initiateTrustChainOperation (trustSerialNumber, nextNeededCert, logicalAddressString);
            if (messageSent)
                pendingUntrustedQueryTable.putNameQuery (trustSerialNumber, serialNumber, nextNeededCert, logicalAddressString, name);
        }

        /**
         * Queries the local binding table using the given logical address string.
         */
        protected NamingBinding [] queryLocalBindingTable() {
            final String localLogicalAddressString = getLogicalAddress().toString();
            return logicalAddressString.equals (localLogicalAddressString) ? localBindingTable.getBindings() : new NamingBinding [0];
        }

        /**
         * Queries the peer binding cache using the given logical address string.
         */
        protected NamingBinding [] queryPeerBindingsCache()
        {
            final NamingBinding [] pull = peerBindingCachePull.getByLogicalAddress (logicalAddressString);
            final NamingBinding [] push = peerBindingCachePush.getByLogicalAddress (logicalAddressString);

            final NamingBinding [] results = new NamingBinding [pull.length + push.length];

            int j = 0;
            for (int i = 0; i < pull.length; ++i)
                results [j++] = pull [i];

            for (int i = 0; i < push.length; ++i)
                results [j++] = push [i];

            return results;
        }

        /**
         * Returns the pending query table appropriate to this type of query, i.e. the pending name query table.
         */
        protected PendingQueryTable getPendingQueryTable() {
            return pendingNameQueryTable;
        }

        /**
         * Returns the logical address string used in this name query.
         */
        protected String getQueryString() {
            return logicalAddressString;
        }

        /**
         * Returns the type of query: name
         */
        protected byte getQueryType() {
            return NAME_QUERY;
        }

        /**
         * Do not propagate a name query after it has reached the address of the query - there is no more information
         * to be found in the network - the query has reached the origin of all information in which it is interested.
         *
         * There is a minor problem here. Since the logical address string of the destination is specified here rather than
         * at construction (this class must parse the LA string, by that time the super class constructor has already been
         * invoked) the destination address shows as "this socket" in all debug output.
         */
        protected void send() {
            if (! logicalAddressString.equals (getLogicalAddress().toString()))
                sendUnicast (logicalAddressString);
        }

        /**
         * Returns a string that contains the logical address that is being used for this query.
         */
        public String toString() {
            return super.toString() + logicalAddressString;
        }
    }

    /**
     * Message that handles the case where the local socket wants to know the names that are used in bindings with its
     * logical address.
     */
    private final class LocalNameQueryMessage
        extends NameQueryMessage
    {
        /**
         * The message can only be generated by the local socket so there is only a single constructor.
         */
        public LocalNameQueryMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData, true);

            if (debugQueryMessage)
                log.println ("LocalNameQueryMessage.<init>:" + this);
        }

        // These methods exist so as to factor out pieces of a common query algorithm used by name and logical address query
        // classes. The query algorithm used for local queries does not need this information so the methods are overridden
        // with trival implementations.

        protected final void establishTrustChain (final String nextNeededCert, final String name, final String logicalAddressString) { throw new UnsupportedOperationException(); }
        protected NamingBinding [] queryLocalBindingTable() { throw new UnsupportedOperationException(); }
        protected NamingBinding [] queryPeerBindingsCache() { throw new UnsupportedOperationException(); }
        protected PendingQueryTable getPendingQueryTable() { throw new UnsupportedOperationException(); }
        protected String getQueryString() { throw new UnsupportedOperationException(); }
        protected byte getQueryType() { throw new UnsupportedOperationException(); }

        /**
         * Algorithm for processing a local name query. Simply acquire all entries of the local binding table and return
         * them to the application.
         */
        public final void process()
        {
            final NamingBinding [] localBindings = localBindingTable.getBindings();

            if (debugQueryMessage) {
                log.println ("Local Name Query: retrieved " + localBindings.length + " local bindings from table:");
                for (int i = 0; i < localBindings.length; ++i)
                    log.println ("    " + localBindings [i]);
            }

            new LocalQueryResponse (serialNumber, localBindings).send();
        }

        /**
         * @see BindingQueryMessage#process()
         */
        protected void prepareMessageForSend() {
            // no-op: local message; never sent to peers
        }

        /**
         * @see BindingQueryMessage#process()
         */
        protected void send() {
            // no-op: local message; never sent to peers
        }

        /**
         * Returns a string that shows that this naming message is a query for local names.
         */
        public String toString() {
            return new StringBuffer (super.toString()).append ("<Local Name Query>").toString();
        }
    }

    /**
     * Parent class for all certificate messages.
     */
    private abstract class CertificateMessage
        extends RoutableNamingMessage
    {
        /**
         * Constructor use to reconstruct a certificate message received from a peer.
         */
        public CertificateMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData);
        }

        /**
         * Constructor called by FSM when creating a certificate messages
         */
        public CertificateMessage (final int serialNumber, final I_LogicalAddress destinationAddress) {
            super (serialNumber, destinationAddress);
        }
    }

    /**
     * A message that is sent in response to a certificate request. All sockets that process this message, including
     * intermediate sockets on the path from source to destiniation, add the certificate to their certificate
     * cache. Also, upon reciept of a certificate, each socket determines if the certificate can be used to extend
     * a partially built chain of trust.
     */
    private final class CertificateResponseMessage
        extends CertificateMessage
    {
        /**
         * The certificate that the messages carries.
         */
        private /*final*/ Certificate certificate;

        /**
         * Constructor used to reconstruct a certificate response message that is received from a peer.
         *
         * @throws HyperCastWarningRuntimeException If the certificate cannot be decoded.
         */
        public CertificateResponseMessage (final OL_Message olMessage, final byte [] controlData)
        {
            super (olMessage, controlData);

// Do not decode cert unconditionally! We might not be snooping!
//            final int [] index = new int [] { SERIAL_NUMBER_LENGTH };
//            this.certificate = runLengthDecodeCertificate (controlData, index);
        }

        /**
         * Constructor used to create a certificate response message by the FSM.
         */
        public CertificateResponseMessage (final int serialNumber, final Certificate certificate, final I_LogicalAddress destinationAddress)
        {
            super (serialNumber, destinationAddress);
            this.certificate = certificate;

            if (certificate == null)
                throw new HyperCastFatalRuntimeException ("Logic error in Naming FSM", new NullPointerException ("certificate == null"));
        }

        /**
         * The algorithm for handling a certificate response message. Upon receipt of a certificate we do the following
         * things:
         *
         * <ol>
         *
         * <li>Determine if the certificate is valid for the current date, if not, ignore it.</li>
         *
         * <li>If the certificate is valid, propagate it to peers.</li>
         *
         * <li>Determine if we requested this certificate or if we simply snooped the transmission of a certificate.
         * (pending certificate request table)</li>
         *
         * <li>If it was bound for somebody else and we snooped it AND we weren't waiting for it ourselves, save it to the
         * cert cache and terminate.</li>
         *
         * <li>If we requested this certificate (even if we snooped it), see if it completes a trust chain</li>
         *
         * <li>If it does not complete a trust chain, request the next missing link in the chain and terminate (continue
         * trust exchange operation)</li>
         *
         * <li>If it completes a trust chain, determine the bindings that could not be trusted previously (causing the
         * trust chain to be built), but can be trusted now. (pending binding table)</li>
         *
         * <li>Determine whether or not such bindings are still in demand by any pending queries. (pending name and LA
         * query tables)</li>
         *
         * <li>If newly trusted bindings are still in demand by any pending queries, return bindings to application.</li>
         *
         * </ol>
         */
        public void process()
        {
            if (debugCertificateMessages)
                log.println ("CertificateResponseMessage.process(): " + this);

            if (! policy.snoopCertificates()
                &&
                destinationAddress != null // actually destinationAddress for local delivery will be null
                &&
                ! destinationAddress.equals (getLogicalAddress()))
            {
                if (debugCertificateMessages)
                    log.println ("CertificateResponseMessage.process(): ignoring locally because we're not snooping certs, but propagating to peers");

                send();
            }
            else
            {
                final int [] index = new int [] { SERIAL_NUMBER_LENGTH };
                this.certificate = runLengthDecodeCertificate (controlData, index);

                final boolean validCertificate = CertificateCache.isCertificateDateValid (certificate);

                if (validCertificate)
                {
                    send();

                    if (debugCertificateMessages)
                        log.println ("CertificateResponseMessage.process() snoop info: snoop certs: " + policy.snoopCertificates() +
                                     " destinationAddress: " + destinationAddress + " this logicalAddress: " + getLogicalAddress());
                
                    if (! certificateCache.contains (certificate)
                        ||
                        ! policy.getCertificateCaching())
                    {
                        String certificateName = null;
                        try
                        {
                            final CertificateCache.Entry certificateCacheEntry = certificateCache.put (certificate);
                            certificateName = certificateCacheEntry.getSubjectCommonName();

                            if (debugCertificateMessages)
                                log.println ("CertificateResponseMessage.process() before cert req table update:\n" + pendingCertificateRequestTable + "\n" + pendingUntrustedQueryTable);

                            final Set noLongerPendingCerts = new HashSet();
                            final String nextCertName = pendingCertificateRequestTable.update (serialNumber, certificateName, noLongerPendingCerts);

                            // either nextCertName is not null or noLongerPendingCerts is non-empty - both cannot happen (however, neither can happen)

                            if (debugCertificateMessages)
                                log.println ("CertificateResponseMessage.process() after cert req table update:\n" +
                                             "nextCertName: " + nextCertName + " noLongerPendingCerts.size(): " + noLongerPendingCerts.size() + "\n" +
                                             pendingCertificateRequestTable + "\n" + pendingUntrustedQueryTable);

                            if (nextCertName != null)
                                continueBuildingTrustChain (serialNumber, nextCertName);

                            if (noLongerPendingCerts.size() > 0)
                            {
                                // Map : String (Name) -> (Map : Integer (query serial number) -> (Set : String (LA)))
                                final Map noLongerPendingLogicalAddressQueries = pendingUntrustedQueryTable.getNonPendingLogicalAddressQueries (serialNumber, noLongerPendingCerts);

                                // Map : String (LA) -> (Map : Integer (query serial number) -> (Set : String (Name)))
                                final Map noLongerPendingNameQueries = pendingUntrustedQueryTable.getNonPendingNameQueries (serialNumber, noLongerPendingCerts);

                                if (debugCertificateMessages)
                                    log.println ("CertificateResponseMessage.process():" +
                                                 " noLongerPendingLogicalAddressQueries.size() " + noLongerPendingLogicalAddressQueries.size() +
                                                 " pendingLogicalAddressQueryTable.size() " + pendingLogicalAddressQueryTable.size() +
                                                 " noLongerPendingNameQueries.size() " + noLongerPendingNameQueries.size() +
                                                 " pendingNameQueryTable.size() " + pendingNameQueryTable.size());

                                final Map satisfiedQueries = new HashMap();
                                final Map untrustedQueries = new HashMap();

                                if (pendingLogicalAddressQueryTable.size() > 0
                                    &&
                                    noLongerPendingLogicalAddressQueries != null)
                                {
                                    final Iterator nonPendingLAQueriesIterator = noLongerPendingLogicalAddressQueries.entrySet().iterator();
                                    while (nonPendingLAQueriesIterator.hasNext())
                                    {
                                        final Map.Entry nameToSerialNumberMap = (Map.Entry) nonPendingLAQueriesIterator.next();
                                        final String name = (String) nameToSerialNumberMap.getKey();
                                        final Map serialNumberMap = (Map) nameToSerialNumberMap.getValue();

                                        if (debugCertificateMessages)
                                            log.println ("CertificateResponseMessage.process(): iterating non pending LA queries name: " + name);

                                        final Iterator serialNumberMapIterator = serialNumberMap.entrySet().iterator();
                                        while (serialNumberMapIterator.hasNext())
                                        {
                                            final Map.Entry serialNumberToLASet = (Map.Entry) serialNumberMapIterator.next();
                                            final Integer querySerialNumberInteger = (Integer) serialNumberToLASet.getKey();
                                            final Set laSet = (Set) serialNumberToLASet.getValue();

                                            if (debugCertificateMessages)
                                                log.println ("CertificateResponseMessage.process(): iterating non pending LA queries querySerialNumberInteger: " + querySerialNumberInteger);

                                            final Iterator laSetIterator = laSet.iterator();
                                            while (laSetIterator.hasNext())
                                            {
                                                final String logicalAddressString = (String) laSetIterator.next();

                                                if (debugCertificateMessages)
                                                    log.println ("CertificateResponseMessage.process(): iterating non pending LA queries LA: " + logicalAddressString);

                                                NamingBinding binding = peerBindingCachePull.get (name, logicalAddressString);
                                                if (binding == null) binding = peerBindingCachePush.get (name, logicalAddressString);
                                                if (binding != null)
                                                {
                                                    if (debugCertificateMessages)
                                                        log.println ("CertificateResponseMessage.process(): iterating non pending LA queries binding: " + binding);

                                                    pendingLogicalAddressQueryTable.attemptToSatisfyQueries (querySerialNumberInteger.intValue(), binding, satisfiedQueries, untrustedQueries);
                                                }
                                                else
                                                {
                                                    if (debugCertificateMessages)
                                                        log.println ("CertificateResponseMessage.process(): iterating non pending LA queries binding null");
                                                }
                                            }
                                        }
                                    }
                                }
                        
                                if (pendingNameQueryTable.size() > 0
                                    &&
                                    noLongerPendingNameQueries != null)
                                {
                                    final Iterator nonPendingNameQueriesIterator = noLongerPendingNameQueries.entrySet().iterator();
                                    while (nonPendingNameQueriesIterator.hasNext())
                                    {
                                        final Map.Entry laToSerialNumberMap = (Map.Entry) nonPendingNameQueriesIterator.next();
                                        final String logicalAddressString = (String) laToSerialNumberMap.getKey();
                                        final Map serialNumberMap = (Map) laToSerialNumberMap.getValue();

                                        final Iterator serialNumberMapIterator = serialNumberMap.entrySet().iterator();
                                        while (serialNumberMapIterator.hasNext())
                                        {
                                            final Map.Entry serialNumberToLASet = (Map.Entry) serialNumberMapIterator.next();
                                            final Integer querySerialNumberInteger = (Integer) serialNumberToLASet.getKey();
                                            final Set nameSet = (Set) serialNumberToLASet.getValue();

                                            final Iterator nameSetIterator = nameSet.iterator();
                                            while (nameSetIterator.hasNext())
                                            {
                                                final String name = (String) nameSetIterator.next();
                                                NamingBinding binding = peerBindingCachePull.get (name, logicalAddressString);
                                                if (binding == null) binding = peerBindingCachePush.get (name, logicalAddressString);
                                                if (binding != null)
                                                    pendingNameQueryTable.attemptToSatisfyQueries (querySerialNumberInteger.intValue(), binding, satisfiedQueries, untrustedQueries);
                                            }
                                        }
                                    }
                                }

                                // check that untrustedQueries is empty, otherwise something strange happened?

                                deliverSatisfiedQueriesLocally (satisfiedQueries);
                            }
                            else
                                if (debugCertificateMessages)
                                    log.println ("CertificateResponseMessage: Receipt of certificate \"" + certificateName + "\" does not extend any pending trust chains");
                        }
                        finally
                        {
                            if (! policy.getCertificateCaching())
                                certificateCache.remove (certificateName);
                        }
                    }
                }
                else
                {
                    if (debugCertificateMessages)
                    {
                        final Date notBefore = ((X509Certificate) certificate).getNotBefore();
                        final Date notAfter = ((X509Certificate) certificate).getNotAfter();
                        final Date now = new Date();

                        log.println ("CertificateResponseMessage.process(): Invalid certificate received" +
                                     " (" + notBefore + " - " + notAfter + ") now (" + now + "): ignoring");
                    }
                }
            }
        }

        private String debugShowStuff (final Map secondMapping)
        {
            final StringBuffer buf = new StringBuffer();
            final Iterator secondIterator = secondMapping.entrySet().iterator();
            while (secondIterator.hasNext())
            {
                final Map.Entry entry2 = (Map.Entry) secondIterator.next();
                final String domainElement = (String) entry2.getKey();

                buf.append (domainElement + " -> {");

                final Set rangeElement = (Set) entry2.getValue();
                final Iterator thirdIterator = rangeElement.iterator();
                boolean first = true;
                while (thirdIterator.hasNext())
                {
                    final String s = (String) thirdIterator.next();
                    buf.append ((first ? "" : "; ") + s);
                    first = false;
                }

                buf.append ("}\n");
            }
            return buf.toString();
        }

        /**
         * Encodes the requested certificate.
         */
        protected void prepareMessageForSend()
        {
            if (controlData != null)
                createNamingMessageExtension (CERTIFICATE_RESPONSE, controlData);
            else
            {
                final byte [] payload = runLengthEncodeCertificate (SERIAL_NUMBER_LENGTH, certificate);
                packSerialNumber (payload);
                createNamingMessageExtension (CERTIFICATE_RESPONSE, payload);
            }
        }

        /*
         * Certificate responses are sent via unicast.
         */
        protected void send() {
            sendUnicast();
        }

        /**
         * Returns a string that contains the name of the certificate that is contained in this naming message.
         */
        public String toString() {
            return super.toString() + ((certificate == null) ? "certificate snooping (true): not decoded" : CertificateCache.extractSubjectCommonName (certificate));
        }
    }

    /**
     * Parent class for all certificate request messages. This class contains data and algorithms that are common to all
     * certificate request messages.
     */
    private abstract class CertificateRequestMessage
        extends CertificateMessage
    {
        /**
         * All certificate request messages contain the name of the certificate that is being requested.
         */
        protected final String name;

        /**
         * Constructor that is used to reconstruct a certificate request message received from a peer. The name of the
         * certificate being requested is decoded.
         */
        public CertificateRequestMessage (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData);
            final int [] index = new int [] { SERIAL_NUMBER_LENGTH };
            this.name = runLengthDecodeString (controlData, index);
        }

        /**
         * Constructor that is used to create a certificate request message by the FSM. The certificate name and
         * destination are specified. The destination is null if the message will use broadcast delivery.
         */
        public CertificateRequestMessage (final int serialNumber, final String name, final I_LogicalAddress destinationAddress) {
            super (serialNumber, destinationAddress);
            this.name = name;
        }

        /**
         * The algorithm for handling a certificate request message. Inspect the local certificate cache. If the
         * requested certificate is not found, propagate the request to peers. If the requested certificate is found,
         * send a response message containing the certificate.
         */
        public final void process()
        {
            final CertificateCache.Entry certificateCacheEntry = certificateCache.get (name);

            if (debugCertificateMessages)
                log.println ("CertificateRequestMessage: Certificate Cache lookup of \"" + name + "\": " + ((certificateCacheEntry == null) ? "MISS" : "HIT"));

            if (certificateCacheEntry == null)
                send();
            else
                new CertificateResponseMessage (serialNumber, certificateCacheEntry.getCertificate(), olMessage.getSourceAddress()).send();
        }

        /**
         * Encodes the name of the requested certificate.
         */
        protected final void prepareMessageForSend()
        {
            final byte [] payload = runLengthEncodeString (SERIAL_NUMBER_LENGTH, name);
            packSerialNumber (payload);
            createNamingMessageExtension (CERTIFICATE_REQUEST, payload);
            olMessage.setHopLimit (policy.getQueryRadius());
        }

        /**
         * Returns a string that contains the name that is being used for this certificate query.
         */
        public final String toString() {
            return new StringBuffer (super.toString()).append (name).toString();
        }
    }

    /**
     * A certificate request message that uses a unicast send operation. This type of message is sent when a certificate
     * is needed to establish a chain of trust and the Naming Service has information that might allow it to know the
     * location of the needed certificate.
     */
    private final class CertificateRequestMessageUnicast
        extends CertificateRequestMessage
    {
        /**
         * Constructor used to reconstruct a certificate request unicast message received from a peer.
         */
        public CertificateRequestMessageUnicast (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData);
        }

        /**
         * Constructor used by FSM to generate a certificate request unicast message. A certificate request unicast
         * message needs the name of a certificate to request as well as a destination address for the unicast message.
         */
        public CertificateRequestMessageUnicast (final int serialNumber, final String name, final I_LogicalAddress destinationAddress) {
            super (serialNumber, name, destinationAddress);
        }

        protected void send() {
            sendUnicast();
            setTimer (new CertificateRequestUnicastTimerTask (serialNumber, name), policy.getCertificateRequestTimeout());
            //timer.schedule (new CertificateRequestUnicastTimerTask (serialNumber, name), policy.getCertificateRequestTimeout());
        }
    }

    /**
     * A certificate request message that uses a broadcast send operation. This type of message is sent when a
     * certificate is needed to establish a chain of trust but the Naming Service has no reason to suspect that any
     * particular peer has the needed certificate. The best that the FSM can do is broadcast a request to various peers.
     */
    private final class CertificateRequestMessageBroadcast
        extends CertificateRequestMessage
    {
        /**
         * Constructor used to reconstruct a certificate request broadcast message received from a peer.
         */
        public CertificateRequestMessageBroadcast (final OL_Message olMessage, final byte [] controlData) {
            super (olMessage, controlData);
        }

        /**
         * Constructor used by FSM to generate a certificate request broadcast message. A certificate request broadcast
         * message needs the name of a certificate to request. There is no destination address for a broadcast message.
         */
        public CertificateRequestMessageBroadcast (final int serialNumber, final String name) {
            super (serialNumber, name, null);
        }

        protected void send() {
            sendBroadcast();
            setTimer (new CertificateRequestBroadcastTimerTask (serialNumber, name), policy.getCertificateRequestTimeout());
            //timer.schedule (new CertificateRequestUnicastTimerTask (serialNumber, name), policy.getCertificateRequestTimeout());
        }
    }

    /**
     * Base class of all messages that are send between the API and the FSM.
     */
    private abstract class UserNamingMessage
        extends NamingMessage
    {
        public UserNamingMessage (final byte [] controlData) {
            super (controlData);
        }

        public UserNamingMessage() { }
    }

    /**
     * This is the base class of all messages that are received from the API and are not routable, i.e. these are commands
     * sent by the API to the FSM.
     */
    private abstract class UserNamingMessageFromAPItoFSM
        extends UserNamingMessage
    {
        protected final boolean requiresTrusted;
        protected final long deadline;
        protected final boolean requiresAuthoritative;
        protected final int maxAge;
        protected final int maxResponses;
        protected final int minHopCount;
        protected final int maxHopCount;

        /**
         * Parses the information in the user naming message. There is a single format shared by all user naming messages.
         */
        public UserNamingMessageFromAPItoFSM (final byte [] controlData)
        {
            super (controlData);

            final int [] index;
            
            index = new int [] { MESSAGE_TYPE_LENGTH + SERIAL_NUMBER_LENGTH};

            requiresTrusted = (controlData [index [0]] == (byte) 1);
            index[0] += TRUSTED_LENGTH;

            deadline = ByteArrayUtility.toLong (controlData, index[0]);
            index[0] += DEADLINE_LENGTH;

            requiresAuthoritative = (controlData [index [0]] == (byte) 1);
            index[0] += AUTHORITATIVE_LENGTH;

            maxAge = ByteArrayUtility.toInteger (controlData, index[0]);
            index[0] += MAXIMUM_AGE_LENGTH;

            maxResponses = ByteArrayUtility.toInteger (controlData, index[0]);
            index[0] += MAX_RESPONSES_LENGTH;

            minHopCount = ByteArrayUtility.toShort (controlData, index[0]);
            index[0] += MINIMUM_HOP_COUNT_LENGTH;

            maxHopCount = ByteArrayUtility.toShort (controlData, index[0]);
            index[0] += MAXIMUM_HOP_COUNT_LENGTH;
        }

        /**
         * Returns a string that shows that this message originated with an API call.
         */
        public String toString() {
            return new StringBuffer (super.toString()).append (" (generated locally by API): ").toString();
        }
    }

    /**
     * Base class of messages sent from FSM to API - nonroutable; local only
     */
    private abstract class UserNamingMessageFromFSMtoAPI
        extends UserNamingMessage
    {
        public UserNamingMessageFromFSMtoAPI() { }

        /**
         * Returns a string that shows that this message originated with an API call.
         */
        public String toString() {
            return new StringBuffer (super.toString()).append (" (generated locally by FSM): ").toString();
        }
    }

    /**
     * This class is responsible for delivering query responses, i.e. bindings, to the application. Since NamingBinding
     * objects are designed to be used internally by the Naming Service, the class ImmutableNamingBinding is used to
     * deliver information to the application.
     */
    private final class LocalQueryResponse
        extends UserNamingMessageFromFSMtoAPI
    {
        private final int serialNumber;
        private final ImmutableNamingBinding [] immutableBindings;

        /**
         * Given an array of bindings, construct ImmutableNamingBinding objects to deliver to the application.
         *
         * @param serialNumber The serial number of the query for which this event is a response
         * @param bindings The binding objects that contain the information that will be used to create the immutable bindings given to the application
         */
        public LocalQueryResponse (final int serialNumber, final NamingBinding [] bindings)
        {
            this.serialNumber = serialNumber;

            final Vector b = new Vector();

            for (int i = 0; i < bindings.length; ++i)
                b.add (new ImmutableNamingBinding (bindings[i].isLocal(),
                                                   bindings[i].isTrusted (digitalSigner, certificateCache.get (bindings[i].getName())),
                                                   bindings[i].getName(),
                                                   bindings[i].getLogicalAddressString(),
                                                   bindings[i].getTimestamp(),
                                                   bindings[i].getLogicalAddressChangeCounter(),
                                                   bindings[i].isAuthoritative()));

            this.immutableBindings = (ImmutableNamingBinding []) b.toArray (new ImmutableNamingBinding [0]);
        }

        /**
         * Use the asynchronous event mechanism to deliver information to the application - in a Message Store FSM we
         * have no callback or other means by which to pass information to the application.
         */
        public void send() {
            if (immutableBindings.length > 0)
                notificationHandler.eventOccurred (new NAMING_EVENT (serialNumber, immutableBindings));
        }

        /**
         * There is never a need to call the process method of this class since objects of this class are neither
         * received from a API call nor received from a peer. These objects are designed to be created for sending to an
         * application.
         */
        public void process() {
            throw new UnsupportedOperationException ("A Local Query Response Message is never generated by an application, rather it is sent to an application.");
        }

        /**
         * Returns a string that contains the immutable bindings of this naming message.
         */
        public String toString()
        {
            final StringBuffer buf = new StringBuffer (super.toString());

            buf.append ("Query S/N: " + serialNumber + "\n");

            for (int i = 0; i < immutableBindings.length; ++i)
                buf.append (immutableBindings[i] + "\n");

            return buf.toString();
        }
    }

    /**
     * This message adds a certificate to the certificate cache with the trusted flag set to true.
     */
    private final class InstallTrustedCertificateMessage
        extends UserNamingMessageFromAPItoFSM
    {
        /** The trusted certificate */
        private final Certificate certificate;

        /**
         * @see Naming#packDummyData(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)
         */
        public InstallTrustedCertificateMessage (final byte [] controlData)
        {
            super (controlData);

            final int [] index = new int [] { LOCAL_QUERY_PREFIX_LENGTH + STRING_LENGTH_BYTES };
            this.certificate = runLengthDecodeCertificate (controlData, index);

            if (certificate == null)
                throw new HyperCastWarningRuntimeException ("certificate is null", new NullPointerException());
        }

        protected final void send() {
            throw new UnsupportedOperationException ("A Local Certificate Installation Message cannot be sent to other sockets.");
        }

        /**
         * Adds the certificate to the cache as trusted.
         */
        public void process() {
            certificateCache.installTrustedCertificate (certificate);
        }

        /**
         * Returns a string that contains the name of the certificate that is being installed as a trust anchor.
         */
        public String toString() {
            return new StringBuffer (super.toString()).append (CertificateCache.extractSubjectCommonName (certificate)).toString();
        }
    }

    /**
     * A message used to pass a binding from the API to the FSM.
     */
    private abstract class UserBindingMessage
        extends UserNamingMessageFromAPItoFSM
    {
        /**
         * Never <code>null</code>, might not be specified in message explicitly. Must be either specified in message
         * explicitly or embedded in certificate (can be both, but in that case, must match).
         */
        protected final String name;

        /** Subject common name of certificate needed to verify the digital signature of the binding (digital signature optional) */
        protected final String signerName;

        /** if not null, use this address instead of the local address. */
        protected final String logicalAddressString;

        /** Optionally <code>null</code>. */
        protected final Certificate certificate;

        /** Optionally <code>null</code>. */
        protected final PrivateKey privateKey;

        /**
         * Extracts name, certificate, and private key (certificate and private key are optional)
         *
         * @see Naming#packDummyData(byte, boolean, long, boolean, boolean, int, int, short, short, int, String, I_LogicalAddress, Certificate, PrivateKey)
         */
        public UserBindingMessage (final byte [] controlData)
        {
            super (controlData);

            final int [] index = new int [] { LOCAL_QUERY_PREFIX_LENGTH };
            final String nameInMessage  = runLengthDecodeString (controlData, index);
            this.certificate = (index[0] < controlData.length) ? runLengthDecodeCertificate (controlData, index) : null;

/*
            if (this.certificate != null)
            {
                this.name = CertificateCache.extractSubjectCommonName (certificate);
                if (nameInMessage != null && ! name.equals (nameInMessage))
                    throw new HyperCastWarningRuntimeException ("Given name \"" + nameInMessage + "\" and name embedded in certificate \"" + name + "\" do not match");
                else
                    if (debugUserBindingMessage)
                        log.println ("BindingMessage: nameInMessage: " + nameInMessage + " name: " + name);
            }
            else 
*/

            if (nameInMessage == null && this.certificate == null)
                throw new HyperCastWarningRuntimeException ("Either name or certificate must be specified");

            this.name = (nameInMessage != null) ? nameInMessage : CertificateCache.extractSubjectCommonName (certificate);
            this.signerName = (certificate == null) ? nameInMessage : CertificateCache.extractSubjectCommonName (certificate);

/*
            else
                this.name = nameInMessage;
*/

            this.privateKey = (index[0] < controlData.length) ? runLengthDecodePrivateKey (controlData, index) : null;
            this.logicalAddressString = (index[0] < controlData.length) ? runLengthDecodeString (controlData, index) : null;
        }

        /**
         * This method is inherited from a parent class, but there is nowhere to send this type of message.
         */
        protected final void send() {
            throw new UnsupportedOperationException ("A Local Binding Message cannot be sent to other sockets.");
        }

        /**
         * Returns a string that contains the name as well as the certificate, if present, that are being used to modify
         * a local binding.
         */

        public final String toString()
        {
            final StringBuffer buffer = new StringBuffer (super.toString());
            buffer.append (name);
            if (certificate != null)
                buffer.append (" (have certificate)");
            return buffer.toString();
        }
    }

    /**
     * This class implements the operation of an API call creating a local binding.
     */
    private final class SetBindingMessage
        extends UserBindingMessage
    {
        /**
         * Builds the object by parsing the control data passed from an API.
         */
        public SetBindingMessage (final byte [] controlData) {
            super (controlData);
        }

        /**
         * Creates a local binding.
         */
        public void process()
        {
            final boolean validCertificate = CertificateCache.isCertificateDateValid (certificate);
            final boolean keyMatch = validCertificate ? CertificateCache.certificatePublicKeyMatchesPrivateKey (certificate, privateKey) : false;

            if (certificate == null
                ||
                (validCertificate && keyMatch))
            {
                final boolean createInvalidationBinding = false;
                final NamingBinding binding = new NamingBinding (name, logicalAddressString, signerName);
                localBindingTable.put (binding);

                localName = name; // TODO:(EXPERIMENTS)

                if (certificate != null)
                    certificateCache.put (certificate, privateKey);

                if (policy.doPushOnSetName())
                    pushOperation (name);

                final int pushPeriod = policy.getPushPeriod();
                if (periodicPushTimerTask == null
                    &&
                    pushPeriod > 0
                    && // TODO(EXPERIMENTS): this prohibits the periodic push from beginning immediately - instead it will be started after the network has stabilized in call of startExperiment()
                    experimentRunning)
                {
                    final long delayForStaggering = Math.abs (random.nextLong()) % pushPeriod;
                    periodicPushTimerTask = new PeriodicPushTimerTask (pushPeriod);
                    setTimer (periodicPushTimerTask, delayForStaggering);
                }
            }
            else
                log.println ("Invalid set binding operation detected, have certificate: " + (certificate != null) +
                             ((certificate != null) ? " certificate is valid: " + validCertificate : "") +
                             (validCertificate ? " private key " + (keyMatch ? "matches" : "does not match") + " public key of certificate " : ""));
        }
    }

    /**
     * This class implements the operation of an API call invalidating a local binding.
     */
    private final class UnsetBindingMessage
        extends UserBindingMessage
    {
        /**
         * Builds the object by parsing the control data passed from an API.
         */
        public UnsetBindingMessage (final byte [] controlData) {
            super (controlData);
        }

        /**
         * Creates an invalidation of a local binding.
         */
        public void process()
        {
            final boolean validCertificate = CertificateCache.isCertificateDateValid (certificate);
            final boolean keyMatch = validCertificate ? CertificateCache.certificatePublicKeyMatchesPrivateKey (certificate, privateKey) : false;

            if (certificate == null
                ||
                (validCertificate && keyMatch))
            {
                final boolean createInvalidationBinding = true;
                final NamingBinding binding = new BindingInvalidation (name, getLogicalAddress().toString(), System.currentTimeMillis());
                localBindingTable.remove (binding);

                // Stop periodic push of bindings if there are no bindings to push
                if (localBindingTable.size() == 0)
                    periodicPushTimerTask = null;

                if (policy.doInvalidateOnUnsetName())
                    invalidateOperation (name);
            }
            else
                log.println ("Invalid set binding operation detected, have certificate: " + (certificate != null) +
                             ((certificate != null) ? " certificate is valid: " + validCertificate : "") +
                             (validCertificate ? " private key " + (keyMatch ? "matches" : "does not match") + " public key of certificate " : ""));
        }
    }

    public Element [] getStats (Document doc, XPath xpath) throws HyperCastStatsException { return statsProcessor.getStatsResult (doc, xpath); }
    public Element [] setStats (Document doc, XPath xpath, Element setValue) throws HyperCastStatsException { return statsProcessor.setStatsResult (doc, xpath, setValue); }
    public Element [] getReadSchema (Document doc, XPath xpath) throws HyperCastStatsException { return statsProcessor.getReadSchemaResult (doc, xpath); }
    public Element [] getWriteSchema (Document doc, XPath xpath) throws HyperCastStatsException { return statsProcessor.getWriteSchemaResult (doc, xpath); }

    /** @see I_Stats#getStatsName */
    public String getStatsName()
    {
        return statisticsName;
    }

    /** @see I_Stats#setStatsName */
    public void setStatsName (String name)
    {
        statisticsName = name;
    }

    // TODO(EXPERIMENTS)
    private boolean experimentRunning = false;

    private static java.io.PrintWriter experimentOut;

    // TODO(EXPERIMENTS): remove, only for testing
    public void startExperiment (final boolean push)
    {
        if (! experimentRunning)
        {
            experimentRunning = true;
            openExperimentOut();

            final int socketsInNetwork = policy.getEXPERIMENTsocketsInNetwork();
            final double averageQueryPeriod = (((double) socketsInNetwork) / policy.getEXPERIMENTqueryFrequency()) * 1000.0; // ms

            if (averageQueryPeriod > 0
                &&
                ((socketsInNetwork > 1) || ((localName != null) && localName.equals ("9999"))))
            {
                if (! push)
                {
                    // If more than a single socket is participating in the experiment be sure to stagger the queries of all sockets to avoid a flood of simultaneous initial messages.
                    final long delayForStaggeringQuery = (socketsInNetwork > 1) ? (long) (Math.abs (random.nextLong()) % averageQueryPeriod) : 0;
//                final long absoluteDelay = 3000; // ms; allow time for "experiment start" command to reach all sockets in network

                    randomLAQueryTimerTask = new RandomLAQueryTimerTask (averageQueryPeriod, socketsInNetwork); //, System.currentTimeMillis());

                    // TODO:? schedule the first event to happen at some time modulo 10 seconds? (sync clocks)
//                setTimer (randomLAQueryTimerTask, absoluteDelay + delayForStaggeringQuery);
                    timer.schedule (new MetaRandomLAQueryTimerTask (randomLAQueryTimerTask), /*absoluteDelay +*/ delayForStaggeringQuery);
                }
                else
                {
                    experimentPushes = 1;

                    final int pushPeriod = policy.getPushPeriod();
                    if (periodicPushTimerTask == null
                        &&
                        pushPeriod > 0)
                    {
                        final long delayForStaggeringPush = Math.abs (random.nextLong()) % pushPeriod;
                        periodicPushTimerTask = new PeriodicPushTimerTask (pushPeriod);
                        setTimer (periodicPushTimerTask, delayForStaggeringPush);
                    }
                }
            }

//            preLoadClasses();
        }
    }

    private RandomLAQueryTimerTask randomLAQueryTimerTask;

    private void stopExperiment (final boolean push)
    {
        if (experimentRunning)
        {
            if (! push)
            {
                randomLAQueryTimerTask = null;

                closeExperimentOut();
                experimentRunning = false;
            }
            else
            {
                periodicPushTimerTask = null;
                experimentRunning = false;
            }
        }
    }

    private void pauseExperiment (/*final String pauseString*/)
    {
        if (experimentRunning
            &&
            randomLAQueryTimerTask != null)
        {
            log.println ("pausing experiment at " + System.currentTimeMillis());
            randomLAQueryTimerTask.pause();
/*
            final String pauseDurationString = pauseString.substring ("pause".length());
            try {
                final int pauseDuration.parseInt (pauseDurationString); // seconds
                randomLAQueryTimerTask.pause (pauseDuration * 1000 * 1000); // convert to microseconds
                log.println ("pausing experiment for " + pauseDuration + " seconds");
            } catch (final NumberFormatException nfe) {
                // ignored
            }
*/
        }
    }

    private void unpauseExperiment()
    {
        if (experimentRunning
            &&
            randomLAQueryTimerTask != null)
        {
            log.println ("unpausing experiment at " + System.currentTimeMillis());
            randomLAQueryTimerTask.unpause();
        }
    }

    private static Object experimentOutLock = new Object();

    private static void openExperimentOut()
    {
        synchronized (experimentOutLock)
        {
            if (experimentOut == null)
            {
                try
                {
                    final String hostname = java.net.InetAddress.getLocalHost().getHostName();
                    final String experimentOutFilename = "./" + hostname + "-" + Math.abs (random.nextInt()) + "-experiment.txt";
                    try
                    {
                        final boolean autoFlush = false;
                        //final int flushAmount = 8192;
                        final int flushAmount = 8192 * 4;
                        //final int flushAmount = (1 << 20);
                        experimentOut = new java.io.PrintWriter (new java.io.BufferedOutputStream (new java.io.FileOutputStream (experimentOutFilename), flushAmount), autoFlush);
                    }
                    catch (final java.io.FileNotFoundException fnfe)
                    {
                        System.err.println ("Could not open file for experiment data output: \"" + experimentOutFilename + "\"");
                        System.exit (1);
                    }
                }
                catch (final java.net.UnknownHostException uhe)
                {
                    System.err.println ("Could not get host name of local computer for experiment output file");
                    System.exit (1);
                }
            }
        }
    }

    private static void closeExperimentOut()
    {
        synchronized (experimentOutLock)
        {
            if (experimentOut != null)
            {
                //experimentOut.println ("Closing");
                experimentOut.flush();
                experimentOut.close();
                experimentOut = null;
            }
        }
    }
}

/**
 * Class that implements the meta operations for all objects of this type of FSM. In the case of Naming there is only a
 * single FSM per socket so these operations actually describe the behavior of only one FSM.
 */
final class NamingMetaOperations
    implements FiniteStateMachineMetaOperations
{
    /**
     * The naming FSM will do the forwarding for all naming messages.
     */
    public boolean messageStoreWillForwardMessage (final OL_Message message) {
        return true;
    }

    /**
     * Only unicast and multicast delivery modes are used in naming messages.
     */
    public boolean isValidMessage (final OL_Message message)
    {
        switch (message.getDeliveryMode())
        {
            case OL_Message.DELIVERY_MODE_INVALID: /* fall-through */
            case OL_Message.DELIVERY_MODE_MULTICAST: /* fall-through */
            case OL_Message.DELIVERY_MODE_UNICAST:
                return true;

            default:
                return false;
        }
    }

    /**
     * Maybe we can satisfy this request before it travels the whole
     * way to the specified logical address. Of course, we could have
     * stale information. Also, snoop passing bindings and certificates.
     */
    public boolean processIntermediateUnicastMessage (final OL_Message message)
    {
        final FSM_Extension extension = (FSM_Extension) message.getFirstExtensionByType (Extension.FSM);

        if (extension == null)
            throw new HyperCastFatalRuntimeException ("Message found that must have FSM extension, but it does not have an FSM extension...");

        return true;
    }

    /*
     * No need to notify the application when naming service messages are sent.
     */
    public NOTIFICATION_EVENT notifyOnSend (final OL_Message message) {
        return null;
    }
}

interface I_LRUObject
{
    public LRUTimeoutQueue.Entry getTimeoutQueueEntry();
    public void setTimeoutQueueEntry (final LRUTimeoutQueue.Entry entry);
}

/**
 * Java's LinkedList class does not permit references to the internal node objects of the list. I would like for each object
 * that is timed out in an LRU manner to be able to find its TimeoutQueue entry in O(1) time.
 *
 * This queue is designed to store objects in a list with the least recently used object at the head of the list. Each object
 * that is in the list is expected to store a pointer/reference to its list entry. This list entry allows the object to find
 * its place in the list in constant time, it also allows the object to place itself at the end of the list when it is
 * used. Each time an object is considered to be used it must invoke the append() method on its list entry object to place
 * itself at the end of the list.
 *
 * This implementation does not contain a timer, rather it relies on an external timing event that should call the
 * remove() method of this list repeatedly until remove() returns null. A null return from the remove() method indicates that
 * none of the entries in the list have timed out yet.
 */
final class LRUTimeoutQueue
{
    private static final boolean debugLRUTimeoutQueue = false;

    private String name;
    public void addName (final String name) { this.name = name; }

    /** Circular, doubly linked, need to know head and tail: tail is head.prev (which might be head for list of length 1) */
    private Entry head = null;

    /** Measured in milliseconds, the amount of time after which an LRU entry will be removed from the queue. */
    private long timeoutTime;

    /**
     * A distinguished value used in calling the remove method that forces the remove method to remove and return the head
     * of the list if the list is not empty.
     */
    private static final int FORCE_REMOVE = -1;

    /**
     * Initializes a queue with the given timeout time for unused entries.
     *
     * @param timeoutTime The time in milliseconds that an unused entry will remain in the queue until it is considered to be
     * timed out (calling remove() repeatedly will eventually return it).
     */
    public LRUTimeoutQueue (final long timeoutTime) {
        this.timeoutTime = timeoutTime;
    }

    /**
     * Add to the back of the queue
     */
    public synchronized void append (final I_LRUObject lruObject)
    {
        final long now = System.currentTimeMillis();

        Entry entry = lruObject.getTimeoutQueueEntry();

        if (debugLRUTimeoutQueue)
            System.err.println (name + " append 0 " + entry + "\n" + this);

        if (entry != null)
        {
            if (debugLRUTimeoutQueue)
                System.err.println (name + " append 0.0");

            entry.deadline = now + timeoutTime;

            if (head == null)
                throw new HyperCastWarningRuntimeException ("Cannot splice entry from zero length queue");
        
            //  check: is this entry in the queue?
            if (entry.prev.next != entry
                ||
                entry.next.prev != entry)
                throw new HyperCastWarningRuntimeException ("Entry is not element of queue");        

            // Append = Remove the entry from its current postion + add to end
            removeEntry (entry);
        }
        else
        {
            if (debugLRUTimeoutQueue)
                System.err.println (name + " append 0.1");

            final long deadline = now + timeoutTime;
            entry = createEntry (lruObject, deadline);
        }

        if (debugLRUTimeoutQueue)
            System.err.println (name + " append 1 " + this);

        // append to the end of the queue
        if (head == null)
        {
            if (debugLRUTimeoutQueue)
                System.err.println (name + " append 1.0");

            head = entry;
            head.next = head.prev = head;
        }
        else
        {
            if (debugLRUTimeoutQueue)
                System.err.println (name + " append 1.1");

            head.prev.next = entry;
            entry.next = head;
            entry.prev = head.prev;
            head.prev = entry;
        }

        if (debugLRUTimeoutQueue)
            System.err.println (name + " append 2 " + this);
    }

    /**
     * remove entry with timeout time before parameter "before" if any
     * exist (if one exists it will be the head of the queue)
     *
     * Returns null if no such entry exists
     *
     * The idea is to repeatedly call this until null is returned.
     *
     * @param before An absolute time (relative to the epoch) used to determine which entries are considered to be timed out.
     */
    public synchronized I_LRUObject remove (final long before)
    {
        if (debugLRUTimeoutQueue)
            System.err.println (name + " remove long 1 " + this);

        final I_LRUObject result;
        if (head != null
            &&
            ((before == FORCE_REMOVE) || (head.deadline < before)))
        {
            final Entry removed = head;
            result = head.lruObject;

            removeEntry (removed);
            removed.free();
        }
        else
            result = null;

        if (debugLRUTimeoutQueue)
            System.err.println (name + " remove long 2 " + this);

        return result;
    }

    /**
     * Unconditionally remove an entry from the timeout queue.
     *
     * Returns null if the queue is empty.
     */
    public synchronized I_LRUObject removeLRU() {
        return remove (FORCE_REMOVE);
    }

    public synchronized void remove (final I_LRUObject lruObject)
    {
        if (debugLRUTimeoutQueue)
            System.err.println (name + " remove object 1 " + this);

        final Entry entry = lruObject.getTimeoutQueueEntry();

        if (entry != null)
        {
            removeEntry (entry);
            entry.free();
        }

        if (debugLRUTimeoutQueue)
            System.err.println (name + " remove object 2 " + this);
    }

    private void removeEntry (final Entry entry)
    {
        if (entry == head)
            head = (head.next == head) ? null : head.next;

        entry.prev.next = entry.next;
        entry.next.prev = entry.prev;
    }

    // not circular, singly linked (next only), don't care where tail is
    private Entry freeList = null;

    private Entry createEntry (final I_LRUObject lruObject, final long deadline)
    {
        final Entry result;
        if (freeList != null)
        {
            result = freeList;
            freeList = result.next;
            result.reset (lruObject, deadline);
        }
        else
            result = new Entry (lruObject, deadline);

        return result;
    }

    public String toString()
    {
        if (head == null)
            return "EMPTY LIST";
        else
        {
            final StringBuffer buf = new StringBuffer();
            final Entry end = head.prev;
            Entry current = head;
            do
            {
                buf.append (current + " prev: " + current.prev + " next: " + current.next + "\n");
                if (current == end)
                    break;
                current = current.next;
            } while (current != null);

            return buf.toString();
        }
    }

    public final class Entry
    {
        public final long DEADLINE_NOT_SET = -1;

        protected I_LRUObject lruObject;
        protected long deadline; // absolute ms
        protected Entry next;
        protected Entry prev;

        public String toString()
        {
            final NamingBinding b = (NamingBinding) lruObject;
            return (b == null) ? "***NULL***" : "<" + b.getName() + ", " + b.getLogicalAddressString() + ", " + (deadline % 100) + ">";
        }

        protected Entry (final I_LRUObject lruObject, final long deadline) {
            reset (lruObject, deadline);
        }

        protected void reset (final I_LRUObject lruObject, final long deadline)
        {
            lruObject.setTimeoutQueueEntry (this);
            this.lruObject = lruObject;
            this.deadline = deadline;
            this.next = this.prev = null;
        }

        protected void free()
        {
            lruObject.setTimeoutQueueEntry (null);
            lruObject = null;
            deadline = DEADLINE_NOT_SET;
            prev = null;
            next = freeList;
            //freeList = this;
        }
    }
}

/**
 * Think of this cache as a mapping: <code>String</code> (certificate subject common name) -> <code>Certificate</code>.
 *
 * This class is written in a general way so that it can support various types of certificates, however there are places in
 * the code where <code>X509Certificate</code> specific operations occur. Extending this class to handle other type of
 * certificates should be relatively painless.
 *
 * Assumption is that certificate are never forgotten. They are written immediately to backing store. They can be retrieved
 * from backing store at any time. If free memory is low, the JVM will automatically reclaim the memory of unused
 * certificates by inspecting the <code>SoftReferences</code> through which they are linked to the
 * <code>CertificateCache</code>.
 *
 * Note that Private Keys are <i>never</i> written to backing store.
 *
 * "Validation" means all checks: date, signature, revocation, etc.
 * "Valid" means that the current time is after the cert activation time, but before the cert expiration time.
 * "Verified" means that it has been proven that the signature of the cert was signed by the named issuer - verified does not imply trusted.
 * "Trusted" means that a chain of verified certs has been found to a trust anchor - trusted implies verified only for non-trust anchors.
 *
 * A trust anchor need not be self signed.
 *
 * Certificate Revocation Lists are not supported.
 */
final class CertificateCache
{
    private static final boolean debug = false;
    private static final boolean debugValidate = false;
    private static final boolean debugBackingStore = false;

    private HashMap cache = new HashMap(); // String (subject common name) -> Entry

    private final boolean writeCertificatesToNonVolatileBackingStore;

    /**
     * Creates a certificate cache that saves all certificates to backing store.
     */
    public CertificateCache() {
        this (true);
    }

    /**
     * Creates a certificate cache.
     * @param writeCertificatesToNonVolatileBackingStore if <code>true</code> all certificates are saved to backing store and
     * <code>SoftReference</code> objects are used to point to in memory copies of certificates.
     */
    public CertificateCache (final boolean writeCertificatesToNonVolatileBackingStore) {
        this.writeCertificatesToNonVolatileBackingStore = writeCertificatesToNonVolatileBackingStore;
    }

    /**
     * Returns the entry corresponding the specified subject common name.
     */
    public Entry get (final String subjectCommonName) {
        return (Entry) cache.get (subjectCommonName);
    }

    /**
     * Returns true if the specified certificate is in the cache.
     */
    public boolean contains (final Certificate certificate) {
        final String subjectCommonName = extractSubjectCommonName (certificate);
        return cache.get (subjectCommonName) != null;
    }

    /**
     * Adds a certificate to the certificate cache.
     *
     * @see #put(Certificate,PrivateKey,boolean)
     */
    public Entry put (final Certificate certificate) {
        return put (certificate, null, false);
    }

    /**
     * Adds a certificate to the certificate cache with a matching private key. A certificate with a matching private
     * key can be used as a local name.
     *
     * @see #put(Certificate,PrivateKey,boolean)
     */
    public Entry put (final Certificate certificate, final PrivateKey privateKey) {
        return put (certificate, privateKey, false);
    }

    /**
     * Removes an entry from the cache (corresponding disk file is not deleted).
     */
    public Entry remove (final String subjectCommonName) {
        return (Entry) cache.remove (subjectCommonName);
    }

    /**
     * Adds a certificate to the certificate cache. The certificate is marked as trusted.
     *
     * @see #put(Certificate,PrivateKey,boolean)
     */
    public Entry installTrustedCertificate (final Certificate certificate) {
        return put (certificate, null, true);
    }

    /**
     * Adds a certificate to the certificate cache with a matching private key. A certificate with a matching private
     * key can be used as a local name. The certificate is marked as trusted.
     *
     * @see #put(Certificate,PrivateKey,boolean)
     */
    public Entry installTrustedCertificate (final Certificate certificate, final PrivateKey privateKey) {
        return put (certificate, privateKey, true);
    }

    /**
     * Places an entry into the certificate cache.
     *
     * @param certificate The certificate that is added to the cache
     * @param privateKey Private keys are specified for certificates that are used for local names
     * @param trusted true if this certificate is being installed as a trust anchor
     *
     * @return The CertificateCache Entry for the certificate and private key information - never null.
     */
    private Entry put (final Certificate certificate, final PrivateKey privateKey, final boolean trusted)
    {
        if (certificate == null)
            throw new HyperCastWarningRuntimeException ("CertificateCache.put(Certificate,PrivateKey,boolean)", new NullPointerException ("certificate == null"));

        if (debug)
            System.err.println ("CertificateCache.put: "                                                   +
                                "certificate" + " (" + (certificate == null ? "" : "not ") + "null" + ") " +
                                "privateKey"  + " (" + (privateKey  == null ? "" : "not ") + "null" + ") " +
                                "trusted"     + " (" + trusted                                      + ")");

        final Entry result;

        final String subjectCommonName = extractSubjectCommonName (certificate);
        final Entry existingEntry = (Entry) cache.get (subjectCommonName);

        if (existingEntry != null)
        {
            if (privateKey != null)
                existingEntry.privateKey = privateKey;

            if (trusted && ! existingEntry.trusted)
                existingEntry.trusted = true;

            result = existingEntry;
        }
        else
        {
            final Entry newEntry = new Entry (certificate, privateKey, trusted);
            cache.put (newEntry.getSubjectCommonName(), newEntry);
            result = newEntry;
        }
        return result;
    }

    /**
     * Creates a string with one line for each entry of the certficate cache.
     */
    public String toString()
    {
        final StringBuffer buf = new StringBuffer ("Certificate Cache:\n\n");
        final Iterator iterator = cache.entrySet().iterator();
        while (iterator.hasNext())
        {
            final Entry entry = (Entry) ((Map.Entry) iterator.next()).getValue();
            buf.append (entry + "\n");
        }
        return buf.toString();
    }

    /**
     * A handle for a certificate in the certificate cache. Basically instances of this class contain enough information
     * to retrieve a certificate from memory or backing store.
     */
    public final class Entry
    {
        /**
         * The reference to the managed certificate. The Certificate data structure pointed to by this reference can be
         * reclaimed by the garbage collector if available memory becomes low.
         */
        private SoftReference certificateSoftReference;

        /** The common name portion of the X.500 subject name. */
        private final String subjectCommonName;

        /** The common name portion of the X.500 issuer name. */
        private final String issuerCommonName;

        /**
         * The private key that matches the public key in the certificate of this entry. Non-null only if the certificate is
         * used as a local name. This value can be updated if a private key is provided after the corresponding certificate is
         * initially placed into the cache.
         */
        private PrivateKey privateKey;

        /** The expiration time of the certificate. */
        private final long expirationTime;

        /** true -> EITHER defined to be trusted because installed as trusted OR a verified path to a trusted certificate has been established */
        private boolean trusted;

        /** true -> confirmed that issuer private key signed this certificate */
        private boolean signatureVerified = false;

        /** true -> validation process has proven that this certificate cannot be valid, e.g. signature is incorrect, expiration date has passed, etc. */
        private boolean invalidated = false;

        /**
         * @param certificate The certificate that this entry will manage
         * @param privateKey Private keys are specified for certificates that are used for local names
         * @param trusted true if this certificate is being installed as a trust anchor
         */
        public Entry (final Certificate certificate, final PrivateKey privateKey, final boolean trusted)
        {
            this.certificateSoftReference = new SoftReference (certificate);

            this.privateKey = privateKey;
            this.trusted = trusted;

            this.subjectCommonName = extractSubjectCommonName (certificate);
            this.issuerCommonName = extractIssuerCommonName (certificate);

            final Date notAfter = ((X509Certificate) certificate).getNotAfter();
            calendar.setTime (notAfter);
            this.expirationTime = calendar.getTimeInMillis();

            saveCertificateToBackingStore (certificate);
        }

        /**
         * Saves a certificate to backing store so that it can be purged from memory if memory is needed for other
         * objects. This method ensures that the certificate cache directory is exists before attempting to write the
         * certificate.
         */
        private void saveCertificateToBackingStore (final Certificate certificate)
        {
            if (writeCertificatesToNonVolatileBackingStore)
            {
                // ./.certcache/00/ etc. Use first two chars of subject common name. Crude database; avoids having directory of tens of thousands of entries.
                final String certificateCacheDirectoryString = certificateCacheFilename + File.separator + subjectCommonName.substring (0,2);
                final File certificateCacheDirectory = new File (certificateCacheDirectoryString);

                final String certificateFilename = certificateCacheDirectoryString + File.separator + subjectCommonName + ".der";

                if (! certificateCacheDirectory.isDirectory())
                {
                    final boolean success = certificateCacheDirectory.mkdirs();
                    if (! success)
                        throw new HyperCastWarningRuntimeException ("Could not create directory \"" + certificateCacheDirectoryString + "\" " +
                                                                    "when attempting to write certificate \"" + subjectCommonName + "\" to backing store: " + certificateFilename);
                }

                if (debugBackingStore)
                    System.err.println ("Writing certificate \"" + subjectCommonName + "\" to backing store: " + certificateFilename);

                writeX509Certificate (certificateFilename, (X509Certificate) certificate);
            }
            else
            {
                synchronized (this)
                {
                    if (staticRAMcache == null)
                        staticRAMcache = new HashSet();

                    staticRAMcache.add (certificate);
                }
            }
        }

        /**
         * Reads a certificate from disk.
         */
        private Certificate readCertificateFromBackingStore()
        {
            if (writeCertificatesToNonVolatileBackingStore)
            {
                // ./.certcache/00/ etc. Use first two chars of subject common name
                final String certificateCacheDirectoryString = certificateCacheFilename + File.separator + subjectCommonName.substring (0,2);

                final String certificateFilename = certificateCacheDirectoryString + File.separator + subjectCommonName + ".der";

                if (debugBackingStore)
                    System.err.println ("Reading certificate \"" + subjectCommonName + "\" from backing store: " + certificateFilename);

                return readX509Certificate (certificateFilename);
            }
            else
            {
                throw new HyperCastFatalRuntimeException ("readCertificateFromBackingStore should never be called when using a RAM disk.");
            }
        }

        /**
         * Returns the subject name of the next certificate needed to establish a chain of trust from this
         * certificate. Returns null if a non-self signed certificate is installed as trusted or if an untrusted
         * self-signed certificate is encountered.
         */
        public String getNextNeededIssuerInChain() {
            return getNextNeededIssuerInChain ("");
        }

        /**
         * @see #getNextNeededIssuerInChain()
         *
         * @param indent textual indentation used for debugging messages
         */
        private String getNextNeededIssuerInChain (final String indent)
        {
            final String result;

            if (trusted)
            {
                if (debug)
                    System.err.println ("getNextNeededIssuerInChain: \"" + subjectCommonName + "\" is trusted, returning null");

                result = null;
            }
            else
            {
                final Entry issuerEntry = get (issuerCommonName);

                if (debug)
                    System.err.println ("getNextNeededIssuerInChain: \"" +
                                        subjectCommonName + "\" issued by \"" + issuerCommonName + "\" - " +
                                        ((issuerEntry == null) ? "do not " : "") + "have entry for issuer, subject currently " +
                                        (trusted ? "" : "not") + " trusted");

                if (issuerCommonName.equals (subjectCommonName)) // This is the case where we have a non-trusted self signed cert
                    result = null;
                else
                    result = (issuerEntry == null) ? issuerCommonName : issuerEntry.getNextNeededIssuerInChain (indent + " ");

                if (debug)
                    System.err.println ("getNextNeededIssuerInChain: \"" +
                                        subjectCommonName + "\" issued by \"" + issuerCommonName + "\" returning \"" + result + "\"");
            }

            return result;
        }

        /*
         * Front end to validate(), records time and initiates prefixed used in debugging recursion.
         */
        private void validate() {
            final long now = System.currentTimeMillis();
            validate (now, "");
        }

        /*
         * Validates a certificate by creating a chain of trust from the current certificate to a trust anchor. This
         * method is recursive. It stops when a self-signed cert is encounted, a trusted cert is encountered, the issuer
         * certificate is missing, or the maximum recursion limit is reached (this is an artificial number that is
         * present to prevent a program crash).
         */
        private void validate (final long now, final String indent)
        {
            if (indent.length() > 64)
            {
                System.err.println (indent + "Certificate Validation Recursion Limit Reached  - giving up! " + this);
                return;
            }

            if (debug || debugValidate)
                System.err.println (indent + "CertificateCache.Entry.validate(): " + this);

            if (now > expirationTime) {
                invalidated = true;
            }

            if (invalidated)
            {
                if (debug || debugValidate)
                    System.err.println (indent + "CertificateCache.Entry.validate(): invalidated arm");

                return;
            }
            else if (trusted)
            {
                if (debug || debugValidate)
                    System.err.println (indent + "CertificateCache.Entry.validate(): subject: \"" + subjectCommonName + "\" is marked as trusted");

                if (signatureVerified)
                    return;
                else if (issuerCommonName.equals (subjectCommonName))
                {
                    signatureVerified = verify (getCertificate());
                    if (! signatureVerified)
                    {
                        if (debug || debugValidate)
                            System.err.println (indent + "CertificateCache.Entry.validate(): subject: \"" + subjectCommonName + "\" self signed cert cannot authenticate itself! subject INVALIDATED!");

                        trusted = false;
                        invalidated = true;
                    }
                }
            }
            else if (signatureVerified)
            {
                if (debug || debugValidate)
                    System.err.println (indent + "CertificateCache.Entry.validate(): signature verified arm: " + issuerCommonName + " " + subjectCommonName);

                final Entry issuer = (Entry) cache.get (issuerCommonName);

                if (issuerCommonName.equals (subjectCommonName))
                {
                    if (debug || debugValidate)
                        System.err.println (indent + "CertificateCache.Entry.validate(): subject: \"" + subjectCommonName + "\" issuer: \"" + issuerCommonName + "\" THIS CASE SHOULD NEVER HAPPEN");
                }

                issuer.validate (now, indent + "  "); // must have issuer cert or sig won't be verified
                if (issuer.trusted)
                {
                    if (debug || debugValidate)
                        System.err.println (indent + "CertificateCache.Entry.validate(): \"" + subjectCommonName + "\" issuer: \"" + issuerCommonName + "\" subject has verified sig and issuer is trusted -> subject considered TRUSTED!");

                    trusted = true;
                }
                else if (issuer.invalidated)
                {
                    if (debug || debugValidate)
                        System.err.println (indent + "CertificateCache.Entry.validate(): \"" + subjectCommonName + "\" issuer: \"" + issuerCommonName + "\" ISSUER INVALIDATED, subject also INVALIDATED!");

                    // although our signature was verified with our issuer cert, it has been determined that our issuer cert is not valid
                    invalidated = true;
                    signatureVerified = false;
                }
            }
            else
            {
                if (debug || debugValidate)
                    System.err.println (indent + "CertificateCache.Entry.validate(): subject: \"" + subjectCommonName +
                                        "\" not yet trusted, not yet verified, not invalidated. Issuer is \"" + issuerCommonName + "\"");

                final Entry issuer = (Entry) cache.get (issuerCommonName);

                if (issuer == null)
                {
                    if (debug || debugValidate)
                        System.err.println (indent + "CertificateCache.Entry.validate(): subject: \"" + subjectCommonName + "\" - do not have certificate for issuer: \"" + issuerCommonName + "\"");
                }
                else
                {
                    if (issuerCommonName.equals (subjectCommonName))
                    {
                        // untrusted self-signed cert
                        return;
                    }

                    issuer.validate (now, indent + "  ");
                
                    if (! issuer.invalidated)
                    {
                        signatureVerified = verify (issuer.getCertificate());

                        if ((debug || debugValidate) && ! signatureVerified)
                            System.err.println (indent + "CertificateCache.Entry.validate(): \"" + subjectCommonName + "\" - issuer: \"" + issuerCommonName + "\" INVALID SIGNATURE!");

                        if (issuer.trusted)
                        {
                            if (signatureVerified)
                            {
                                if (debug || debugValidate)
                                    System.err.println (indent + "CertificateCache.Entry.validate(): \"" + subjectCommonName + "\" issuer: \"" + issuerCommonName + "\" subject has verified sig and issuer is trusted -> subject considered TRUSTED!");

                                trusted = true;
                            }
                            else if (! signatureVerified)
                            {
                                if (debug || debugValidate)
                                    System.err.println (indent + "CertificateCache.Entry.validate(): \"" + subjectCommonName + "\" issuer: \"" + issuerCommonName + "\"; issuer is trusted but cannot verify subject signature -> subject is INVALIDATED!");

                                invalidated = true;
                            }
                        }
                    }
                }
            }
        }

        /**
         * Verifies that the signature of this certificate was signed by the issuer named in the certificate. This
         * verification does not necessarily establish trust - only a chain of trust to a trust anchor can do that.
         */
        private boolean verify (final Certificate issuerCertificate)
        {
            boolean keyException = false;
            try
            {
                getCertificate().verify (issuerCertificate.getPublicKey());
            }
            catch (final NoSuchAlgorithmException nsae) { throw new HyperCastWarningRuntimeException (nsae); }
            catch (final NoSuchProviderException  nspe) { throw new HyperCastWarningRuntimeException (nspe); }
            catch (final SignatureException       se)   { throw new HyperCastWarningRuntimeException (se);   }
            catch (final CertificateException     ce)   { throw new HyperCastWarningRuntimeException (ce);   }
            catch (final InvalidKeyException      ike)
            {
                keyException = true;
            }

            return ! keyException;
        }

        /**
         * Returns the trusted status of this certificate.
         */
        public boolean isTrusted()
        {
            if (invalidated)
                return false;

            if (! trusted)
                validate();

            return trusted;
        }

        /**
         * Returns the trusted status of this certificate.
         *
         * @param boundCommonName A name of a certificate
         */

        /**
         * Returns the common name of the subject of the certificate represented by this entry. No disk activity is
         * needed; this string is always available in memory.
         */
        public String getSubjectCommonName() {
            return subjectCommonName;
        }

        /**
         * Returns the common name of the issuer of the certificate represented by this entry. No disk activity is
         * needed; this string is always available in memory.
         */
        public String getIssuerCommonName() {
            return issuerCommonName;
        }

        /**
         * Returns the certificate represented by this entry, reading it from disk if necessary.
         */
        public Certificate getCertificate()
        {
            Certificate certificate = (Certificate) certificateSoftReference.get();

            if (certificate == null)
            {
                certificate = readCertificateFromBackingStore();
                if (certificate != null)
                    certificateSoftReference = new SoftReference (certificate);
            }

            return certificate;
        }

        /**
         * Returns a private key for a certificate that is being used for a local name; null otherwise. Private keys are
         * never written to disk.
         */
        public PrivateKey getPrivateKey() {
            return privateKey;
        }

        /**
         * Returns the public key of the certificate represented by this entry, reading it from disk if necessary.
         */
        public PublicKey getPublicKey() {
            return getCertificate().getPublicKey();
        }

        /**
         * Encodes the values of this object into a String.
         */
        public String toString() {
            return
                "Subject "            + "(" + subjectCommonName                   + ") " +
                "Issuer "             + "(" + issuerCommonName                    + ") " +
                "PrivateKey "         + "(" + (privateKey != null ? "Yes" : "No") + ") " +
                "Trusted "            + "(" + (trusted            ? "Yes" : "No") + ") " +
                "Verified "           + "(" + (signatureVerified  ? "Yes" : "No") + ")";
//                "Invalidated "        + "(" + (invalidated        ? "Yes" : "No") + ")";
        }
    }

    /** Used to manipulate certificate expiration dates */
    private static GregorianCalendar calendar = new GregorianCalendar();

    /** Directory where certificates are cached */
    private static final String certificateCacheFilename = "." + File.separator + ".certcache";

    /**
     * This is an implementation of a volatile, in-memory backing store rather than a non-volatile, on disk backing
     * store. The <code>CertificateCache</code> is designed so that it never forgets a certificate. It was originally
     * designed to save each certificate to disk. If disk I/O is not desired, e.g because it's too slow during experiments,
     * this structure will contain hard references to certificates in memory so that the garbage collector will not reclaim
     * them.
     */
    private static Set staticRAMcache;

    /*
     * Some of these values are hardcoded. These hardcoded values appear elsewhere in the HyperCast code, see
     * SecInfoExchange_Message.java and KeyVault.java.
     */

    /** The name of the class that provides crypto algorithm implementations. */
    private static final String securityProviderClassName = MessageStoreFSM_Naming.securityProviderClassName;

    /** The short name of the provider that has crypto algorithm implementations. */
    private static final String securityProviderShortName = MessageStoreFSM_Naming.securityProviderShortName;

    private final static String certificateFormat = "X.509";
    private final static String asymmetricEncryptionAlgorithm = "RSA";

    private static CertificateFactory certificateFactory;
    private static KeyFactory keyFactory;

    private static boolean initStaticCryptoFactories = false;

    /*
     * This static block initializes structures that are used in various static "utility" methods of this class.
     */
    public synchronized static void initStaticCryptoFactories()
    {
        if (initStaticCryptoFactories)
            return;

        initStaticCryptoFactories = true;

        ensureSecurityProviderInstalled (securityProviderShortName, securityProviderClassName);

        try
        {
            certificateFactory = CertificateFactory.getInstance (certificateFormat, securityProviderShortName);
            keyFactory = KeyFactory.getInstance (asymmetricEncryptionAlgorithm, securityProviderShortName);
        }
        catch (final CertificateException ce) {
            throw new HyperCastFatalRuntimeException ("Could not instanciate \"" + certificateFormat + "\" certificate factory: ", ce);
        }
        catch (final NoSuchAlgorithmException nsae) {
            throw new HyperCastWarningRuntimeException ("Could not instanciate \"" + asymmetricEncryptionAlgorithm + "\" key factory: ", nsae);
        }
        catch (final NoSuchProviderException nspe) {
            throw new HyperCastFatalRuntimeException ("Could not instanciate \"" + certificateFormat + "\" certificate factory, or \"" +
                                                      asymmetricEncryptionAlgorithm + "\" key factory because provider " + securityProviderShortName + " is unknown", nspe);
        }
    }

    /**
     * Installs a security provider by name using reflection to avoid any compile time bindings. General purpose utility
     * placed here because this class "does crypto stuff."
     */
    public static void ensureSecurityProviderInstalled (final String shortName, final String className)
    {
        if (Security.getProvider (shortName) == null)
        {
            final String errorString = "NamingFSM: Adding " + shortName + " provider \"" + className + "\" is not supported - you must include a separate jar file in the classpath for the security provider";
            try
            {
                final Class providerClass = Class.forName (className);
                final Constructor providerConstructor = providerClass.getConstructor (new Class [0]);
                final Provider provider = (Provider) providerConstructor.newInstance (new Object [0]);

                // According to research on the web this seems to be the de facto way to install a provider.
                //Security.insertProviderAt (provider, 2);

                // This is how existing HyperCast code installs a provider...
                Security.addProvider (provider);
            } catch (final ClassNotFoundException cnfe) {
                throw new HyperCastConfigException (errorString, cnfe);
            }
            catch (final NoSuchMethodException nsme) {
                throw new HyperCastConfigException (errorString, nsme);
            }
            catch (final InstantiationException ie) {
                throw new HyperCastConfigException (errorString, ie);
            }
            catch (final IllegalAccessException iae) {
                throw new HyperCastConfigException (errorString, iae);
            }
            catch (final InvocationTargetException ite) {
                throw new HyperCastConfigException (errorString, ite);
            }
        }
    }

    /**
     * Reads a certificate from a disk file.
     *
     * @param certificateFilename Name of file where certificate is stored
     * @return An X.509 v3 certificate initialized from the contents of certificateFilename
     * @throws HyperCastWarningRuntimeException If there is any problem while reading the file
     */
    public static X509Certificate readX509Certificate (final String certificateFilename)
    {
        FileInputStream fis = null;
        DataInputStream dis = null;
        ByteArrayInputStream bais = null;

        try {
            fis = new FileInputStream (certificateFilename);
            dis = new DataInputStream (fis);

            byte [] certificateData = new byte [dis.available()];
            dis.readFully (certificateData);

            if (debug)
                System.err.println ("Certificate Data Size: " + certificateData.length);

            bais = new ByteArrayInputStream (certificateData);
            final Certificate certificate = certificateFactory.generateCertificate (bais);

            return (X509Certificate) certificate;
        }
        catch (final FileNotFoundException fnfe) {
            throw new HyperCastWarningRuntimeException ("Could not read certificate from file \"" + certificateFilename + "\"", fnfe);
        }
        catch (final IOException ioe) {
            throw new HyperCastWarningRuntimeException ("Could not read certificate from file \"" + certificateFilename + "\"", ioe);
        }
        catch (final CertificateException ce) {
            throw new HyperCastWarningRuntimeException ("Could not read certificate from file \"" + certificateFilename + "\"", ce);
        }
        finally {
            try { bais.close(); } catch (final IOException ignored) { }
            try { dis.close(); } catch (final IOException ignored) { }
            try { fis.close(); } catch (final IOException ignored) { }
        }
    }

    /**
     * Writes a certificate to a disk file.
     *
     * @param certificateFilename Name of file where certificate is stored
     * @param certificate Certificate object to store in file
     * @throws HyperCastWarningRuntimeException If there is any problem while writing the file
     */
    public static void writeX509Certificate (final String certificateFilename, final X509Certificate certificate)
    {
        FileOutputStream fos = null;

        try {
            fos = new FileOutputStream (certificateFilename);
            fos.write (certificate.getEncoded());
        }
        catch (final FileNotFoundException fnfe) {
            throw new HyperCastWarningRuntimeException ("Could not write certificate \"" + certificate + "\" to file \"" + certificateFilename + "\"", fnfe);
        }
        catch (final IOException ioe) {
            throw new HyperCastWarningRuntimeException ("Could not write certificate \"" + certificate + "\" to file \"" + certificateFilename + "\"", ioe);
        }
        catch (final CertificateException ce) {
            throw new HyperCastWarningRuntimeException ("Could not write certificate \"" + certificate + "\" to file \"" + certificateFilename + "\"", ce);
        }
        finally {
            try { fos.close(); } catch (final IOException ignored) { }
        }
    }

    /**
     * Decodes a certificate from a byte array to a Java certificate object.
     *
     * @param encodedCertificate Byte array that contains data for a certificate object
     * @param offset The first byte of the certificate object data
     * @param length The number of bytes of certificate object data
     * @throws HyperCastWarningRuntimeException If the given byte array cannot be decoded into a certificate object
     */
    public static Certificate decodeCertificate (final byte [] encodedCertificate, final int offset, final int length)
    {
        initStaticCryptoFactories();

        ByteArrayInputStream encodedCertificateStream = null;

        try {
            encodedCertificateStream = new ByteArrayInputStream (encodedCertificate, offset, length);
            final Certificate certificate = certificateFactory.generateCertificate (encodedCertificateStream);

            return certificate;
        }
        catch (final CertificateException ce) {
            throw new HyperCastWarningRuntimeException ("Could not decode certificate", ce);
        }
        finally {
            try { encodedCertificateStream.close(); } catch (final IOException ignored) { }
        }
    }

    /**
     * Encodes a Java certificate object to a byte array.
     *
     * @param certificate certificate object to be encoded into a byte array
     * @throws HyperCastWarningRuntimeException If the given certificate object cannot be encoded into a  byte array
     */
    public static byte [] encodeCertificate (final Certificate certificate)
    {
        try {
            return certificate.getEncoded();
        }
        catch (final CertificateEncodingException cee) {
            throw new HyperCastWarningRuntimeException ("Could not encode certificate.", cee);
        }
    }

    /**
     * Decodes a private key from a byte array to a Java private key object.
     *
     * @param encodedPrivateKey Byte array that contains data for a private key object
     * @param offset The first byte of the private key object data
     * @param length The number of bytes of private key object data
     * @throws HyperCastWarningRuntimeException If the given byte array cannot be decoded into a private key object
     */
    public static PrivateKey decodePrivateKey (final byte [] encodedPrivateKey, final int offset, final int length)
    {
        initStaticCryptoFactories();

        try {
            final byte [] privateKeyData = new byte [length];
            System.arraycopy (encodedPrivateKey, offset, privateKeyData, 0, length);

            final PKCS8EncodedKeySpec privateKeySpec = new PKCS8EncodedKeySpec (privateKeyData);
            final PrivateKey privateKey = keyFactory.generatePrivate (privateKeySpec);

            return privateKey;
        }
        catch (final InvalidKeySpecException ikse) {
            throw new HyperCastWarningRuntimeException ("Could not decode private key", ikse);
        }
    }

    /**
     * The Java API does not supply a method that gets the Common Name so it must be extracted from full subject
     * name. Note that only X509Certificate are supported.
     *
     * @param certificate Certificate object that contains a common name
     * @param signer If true, the common name of the issuer of this certificate will be returned, if false, the common
     * name of the subject will be returned
     */
    private static String extractCommonName (final Certificate certificate, final boolean signer)
    {
        final String commonName;

        if (certificate == null)
            throw new HyperCastWarningRuntimeException ("certificate is null", new NullPointerException());

        if (certificate instanceof X509Certificate)
        {
            final X509Certificate x509Certificate = (X509Certificate) certificate;

            // Alternate way to get the name, results in a different format - decoding code below with "firstIndex" and
            // "secondIndex" does not work with this alternate format.
            //
            //final String fullName = signer ? x509Certificate.getIssuerDN().getName() : x509Certificate.getSubjectDN().getName();

            final X500Principal x500Principal = signer ? x509Certificate.getIssuerX500Principal() : x509Certificate.getSubjectX500Principal();
            final String fullName = x500Principal.getName();

            // Extract the common name substring
            final int firstIndex = fullName.indexOf (",CN=") + ",CN=".length();
            final int secondIndex = fullName.indexOf (",OU");

            commonName = fullName.substring (firstIndex, secondIndex);
        }
        else
            throw new HyperCastWarningRuntimeException ("Only X.509 v3 certificates are currently supported");

        return commonName;
    }

    /**
     * Returns the common name of the issuer of the certificate
     *
     * @param certificate Certificate object that contains an issuer common name
     */
    public static String extractIssuerCommonName (final Certificate certificate) {
        return extractCommonName (certificate, true);
    }

    /**
     * Returns the common name of the subject of the certificate
     *
     * @param certificate Certificate object that contains a subject common name
     */
    public static String extractSubjectCommonName (final Certificate certificate) {
        return extractCommonName (certificate, false);
    }

    /**
     * A certificate is considered to be invalid if it has expired or it is not yet valid. In particular the naming
     * system discards certificates that are scheduled to become valid in the future. Only X509Certificate are fully
     * supported, all other types are considered to be valid without a check.
     *
     * @param certificate Certificate object that will be checked for validity
     */
    public static boolean isCertificateDateValid (final Certificate certificate)
    {
        if (certificate instanceof X509Certificate) {
            try
            {
                ((X509Certificate) certificate).checkValidity();
            }
            catch (final CertificateExpiredException     cee)   { return false; }
            catch (final CertificateNotYetValidException cnyve) { return false; }

            return true;
        }
        else
            return true;
    }

    /*
     * Returns true if the public key in the certificate matches the specified private key.
     */
    public static boolean certificatePublicKeyMatchesPrivateKey (final Certificate certificate, final PrivateKey privateKey)
    {
        if (certificate == null && privateKey == null) return true;

        try
        {
            final Signature signature = Signature.getInstance ("SHA1withRSA", "BC");
            final String testString = "Test for public key, private key match";

            signature.initSign (privateKey);
            signature.update (testString.getBytes());
            final byte [] digitalSignature = signature.sign();

            final PublicKey publicKey = certificate.getPublicKey();
            signature.initVerify (publicKey);
            signature.update (testString.getBytes());
            final boolean match = signature.verify (digitalSignature);

            return match;
        }
        catch (final SignatureException se) {
            throw new HyperCastWarningRuntimeException ("public/private key match", se);
        } catch (final InvalidKeyException ike) {
            throw new HyperCastWarningRuntimeException ("public/private key match", ike);
        } catch (final NoSuchAlgorithmException nsae) {
            throw new HyperCastWarningRuntimeException ("public/private key match", nsae);
        } catch (final NoSuchProviderException nspe) {
            throw new HyperCastWarningRuntimeException ("public/private key match", nspe);
        }
    }
}
